/**
   @file gl_color.h
   @brief 色成分に対するクラスヘッダ

   @author tadah_fussy
   @date 2014/11/29 コメントを Doxygen 用に変更 
   @date 2018/02/24 ファイル名変更 ( color.h → gl_color.h )
   @date 2018/02/26 全関数をヘッダファイルで実装する形に変更 gl_color.cpp は廃止
   @date 2018/03/08 overlay 関数の追加 ( アルファチャンネルを考慮した重ね合わせに対応 )
   @date 2018/04/17 in, out, add 関数の追加 ( アルファチャンネル演算の追加 )
                    RGB_Op::getParam, RGB_Op::setParam 関数の追加
   @date 2021/02/25 ワーニングの出力に対応
   @date 2021/03/26 名前空間の見直し
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_COLOR_20210402_H
#define GRAPHIC_LIBRARY_COLOR_20210402_H

#include <cassert>

#include <limits>
#include <functional>
#include <stdexcept>

#include <common/math_library.h>

namespace tadah_fussy
{
  namespace gl
  {
    /**
       @brief RGB成分用パレット構造体
    **/
    class RGB
    {
      /*
        RGB_Operator : rgb オブジェクトの operator@ 関数 func を src を引数として呼び出す
        演算子の多重定義の実装に利用する
      */
      template< class T, class MemFunc >
      static RGB RGB_Operator( RGB rgb, T src, MemFunc func )
      {
        return( ( ( &rgb )->*func )( src ) );
      }

    public:

      // 型宣言
      using color_type = uint32_t;  ///< 色成分の型(AARRGGBB)
      using primary_type = uint8_t; ///< RGB各成分の型
      using size_type = size_t;     ///< 添字の型

      // 定数宣言
      static const primary_type MAX = std::numeric_limits< primary_type >::max();  ///< RGB各成分の最大値
      static const color_type MASK = std::numeric_limits< primary_type >::max();   ///< RGB各成分のマスク
      static const primary_type BIT = std::numeric_limits< primary_type >::digits; ///< RGB各成分のビット数

      // 添字定義用定数
      static const size_type RED = 0;       ///< 赤成分の添字
      static const size_type GREEN = 1;     ///< 緑成分の添字
      static const size_type BLUE = 2;      ///< 青成分の添字
      static const size_type ALPHA = 3;     ///< アルファチャンネル成分の添字
      static const size_type RGB_SIZE = 3;  ///< (アルファチャンネルを除いた)RGB成分のサイズ
      static const size_type RGBA_SIZE = 4; ///< RGBA成分のサイズ

      // 公開メンバ変数
      primary_type r; ///< 赤成分
      primary_type g; ///< 緑成分
      primary_type b; ///< 青成分
      primary_type a; ///< アルファチャンネル

      /* 色コードから各成分を取得する関数 */

      /// @brief 色コードから赤成分を抽出する
      ///
      /// 赤成分は、color_type の下位より 16 から 23 ビットまでとする。
      ///
      /// ... xxxxxxxx OOOOOOOO xxxxxxxx xxxxxxxx
      ///
      /// @param color 対象の色コード(AARRGGBB)
      /// @return 抽出した赤成分
      static primary_type getRed( color_type color ) { return( ( color >> 16 ) & MASK ); }

      /// @brief 色コードから緑成分を抽出する
      ///
      /// 緑成分は、color_type の下位より 8 から 15 ビットまでとする。
      ///
      /// ... xxxxxxxx xxxxxxxx OOOOOOOO xxxxxxxx
      ///
      /// @param color 対象の色コード(AARRGGBB)
      /// @return 抽出した緑成分
      static primary_type getGreen( color_type color ) { return( ( color >> 8 ) & MASK ); }

      /// @brief 色コードから青成分を抽出する
      ///
      /// 青成分は、color_type の下位より 0 から 7 ビットまでとする。
      ///
      /// ... xxxxxxxx xxxxxxxx xxxxxxxx OOOOOOOO
      ///
      /// @param color 対象の色コード(AARRGGBB)
      /// @return 抽出した青成分
      static primary_type getBlue( color_type color ) { return( color & MASK ); }

      /// @brief 色コードからアルファチャンネルを抽出する
      ///
      /// アルファチャンネルは、color_type の下位より 24 から 31 ビットまでとする。
      ///
      /// ... OOOOOOOO xxxxxxxx xxxxxxxx xxxxxxxx
      ///
      /// @param color 対象の色コード(AARRGGBB)
      /// @return 抽出したアルファチャンネル
      static primary_type getAlpha( color_type color ) { return( ( color >> 24 ) & MASK ); }

      /* コンストラクタ */

      /// @brief RGB各成分からの構築
      ///
      /// @note アルファチャンネルを指定しない場合、RGB::a は RGB::MAX で初期化される
      ///
      /// @param red 赤成分
      /// @param green 緑成分
      /// @param blue 青成分
      /// @param alpha アルファチャンネル
      RGB( primary_type red, primary_type green, primary_type blue, primary_type alpha = MAX )
        : r( red ), g( green ), b( blue ), a( alpha ) {}

      /// @brief 色コードからの構築
      ///
      /// @param c 対象の色コード(AARRGGBB)
      explicit RGB( color_type c = 0 )
      {
        r = getRed( c );
        g = getGreen( c );
        b = getBlue( c );
        a = getAlpha( c );
      }

      /* 各成分の取得 */

      /// @brief RGB各成分とアルファチャンネルを引数側に渡す
      ///
      /// @param red 赤成分を渡す変数へのポインタ
      /// @param green 緑成分を渡す変数へのポインタ
      /// @param blue 青成分を渡す変数へのポインタ
      /// @param alpha アルファチャンネルを渡す変数へのポインタ
      void sendToParam( primary_type* red, primary_type* green, primary_type* blue, primary_type* alpha ) const
      {
        if ( red != nullptr ) *red = r;
        if ( green != nullptr ) *green = g;
        if ( blue != nullptr ) *blue = b;
        if ( alpha != nullptr ) *alpha = a;
      }

      /// @brief RGB各成分を引数側に渡す
      ///
      /// @param red 取得した赤成分
      /// @param green 取得した緑成分
      /// @param blue 取得した青成分
      void sendToParam( primary_type* red, primary_type* green, primary_type* blue ) const
      {
        if ( red != nullptr ) *red = r;
        if ( green != nullptr ) *green = g;
        if ( blue != nullptr ) *blue = b;
      }

      /// @brief 引数から渡された RGBA 各成分をセットする
      ///
      /// @param red セットする赤成分
      /// @param green セットする緑成分
      /// @param blue セットする青成分
      /// @param alpha セットするアルファチャンネル
      void receiveFromParam( primary_type red, primary_type green, primary_type blue, primary_type alpha = RGB::MAX )
      { r = red; g = green; b = blue; a = alpha; }

      /// @brief 添字を利用して各成分への参照を返す
      ///
      /// @note 未定義の size_type が指定された場合 std::out_of_range 例外を投げる。
      ///
      /// @param index 対象の成分への添字
      /// @return 対象の成分への参照
      primary_type& operator[]( size_type index )
      {
        switch ( index ) {
        case RED :
          return( r );
        case GREEN :
          return( g );
        case BLUE :
          return( b );
        case ALPHA :
          return( a );
        default :
          throw std::out_of_range( "Out of range at RGB::operator[]" );
        }
      }

      /// @brief 添字を利用して各成分の値を返す
      ///
      /// @note 未定義の size_type が指定された場合 std::out_of_range 例外を投げる。
      ///
      /// @param index 対象の成分への添字
      /// @return 対象の成分
      primary_type operator[]( size_type index ) const
      {
        switch ( index ) {
        case RED :
          return( r );
        case GREEN :
          return( g );
        case BLUE :
          return( b );
        case ALPHA :
          return( a );
        default :
          throw std::out_of_range( "Out of range at RGB::operator[]" );
        }
      }

      /// @brief 色を重ね合わせる ( 乗算合成 )
      ///
      /// アルファチャンネルを考慮して色の重ね合わせを行う。
      ///
      /// 成分を ( 0, 1 ) の範囲に正規化した時、bg に fg を重ねあわせた時の計算式は以下の通り。
      ///
      /// alpha = fg.alpha + bg.alpha ( 1 - fg.alpha )
      ///
      /// rgb = [ fg.rgb・fg.alpha + bg.rgb・bg.alpha ( 1 - fg.alpha ) ] / alpha
      ///
      /// 参考リンク : https://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AB%E3%83%95%E3%82%A1%E3%83%96%E3%83%AC%E3%83%B3%E3%83%89
      ///
      /// @param fg 上側の(重ねあわせる)色
      /// @return 自分自身への参照
      RGB& overlay( RGB fg )
      {
        // 1 で正規化した値
        double fa = static_cast< double >( fg.a ) / MAX;
        double ba = static_cast< double >( a ) / MAX;
        double oa = fa + ba * ( 1 - fa );

        a = Check_RGBPrimary< double >( std::round( oa * MAX ) );
        r = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.r * fa + r * ba * ( 1.0 - fa ) ) / oa ) );
        g = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.g * fa + g * ba * ( 1.0 - fa ) ) / oa ) );
        b = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.b * fa + b * ba * ( 1.0 - fa ) ) / oa ) );

        return( *this );
      }

      /// @brief アルファチャンネルの計算 ( in )
      ///
      /// 成分を ( 0, 1 ) の範囲に正規化した時、bg に fg を in 演算した時の計算式は以下の通り。
      ///
      /// alpha = fg.alpha・bg.alpha
      ///
      /// rgb = fg.rgb
      ///
      /// 参考リンク : https://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AB%E3%83%95%E3%82%A1%E3%83%81%E3%83%A3%E3%83%B3%E3%83%8D%E3%83%AB
      ///
      /// @param fg 上側の(重ねあわせる)色
      /// @return 自分自身への参照
      RGB& in( RGB fg )
      {
        // 1 で正規化した値
        double fa = static_cast< double >( fg.a ) / MAX;
        double ba = static_cast< double >( a ) / MAX;
        double oa = fa * ba;

        a = Check_RGBPrimary< double >( std::round( oa * MAX ) );
        r = fg.r;
        g = fg.g;
        b = fg.b;

        return( *this );
      }

      /// @brief アルファチャンネルの計算 ( out )
      ///
      /// 成分を ( 0, 1 ) の範囲に正規化した時、bg に fg を out 演算した時の計算式は以下の通り。
      ///
      /// alpha = fg.alpha ( 1 - bg.alpha )
      ///
      /// rgb = fg.rgb
      ///
      /// 参考リンク : https://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AB%E3%83%95%E3%82%A1%E3%83%81%E3%83%A3%E3%83%B3%E3%83%8D%E3%83%AB
      ///
      /// @param fg 上側の(重ねあわせる)色
      /// @return 自分自身への参照
      RGB& out( RGB fg )
      {
        // 1 で正規化した値
        double fa = static_cast< double >( fg.a ) / MAX;
        double ba = static_cast< double >( a ) / MAX;
        double oa = fa * ( 1.0 - ba );

        a = Check_RGBPrimary< double >( std::round( oa * MAX ) );
        r = fg.r;
        g = fg.g;
        b = fg.b;

        return( *this );
      }

      /// @brief 加算合成 ( add )
      ///
      /// 成分を ( 0, 1 ) の範囲に正規化した時、bg に fg を加算合成した時の計算式は以下の通り。
      ///
      /// alpha = fg.alpha + bg.alpha
      ///
      /// rgb = ( fg.rgb・fg.alpha + bg.rgb・bg.alpha ) / alpha
      ///
      /// 参考リンク : https://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AB%E3%83%95%E3%82%A1%E3%83%81%E3%83%A3%E3%83%B3%E3%83%8D%E3%83%AB
      ///
      /// @param fg 上側の(重ねあわせる)色
      /// @return 自分自身への参照
      RGB& add( RGB fg )
      {
        // 1 で正規化した値
        double fa = static_cast< double >( fg.a ) / MAX;
        double ba = static_cast< double >( a ) / MAX;
        double oa = std::min( fa + ba, 1.0 );

        a = Check_RGBPrimary< double >( std::round( oa * MAX ) );
        r = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.r * fa + r * ba ) / oa ) );
        g = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.g * fa + g * ba ) / oa ) );
        b = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.b * fa + b * ba ) / oa ) );

        return( *this );
      }

      /// @brief 排他的論理和 ( xor )
      ///
      /// 成分を ( 0, 1 ) の範囲に正規化した時、bg に fg を xor 演算した時の計算式は以下の通り。
      ///
      /// alpha = fg.alpha ( 1 - bg.alpha ) + bg.alpha ( 1 - fg.alpha )
      ///
      /// rgb = [ fg.rgb・fg.alpha ( 1 - bg.alpha ) + bg.rgb・bg.alpha ( 1 - fg.alpha ) ] / alpha
      ///
      /// ビット演算子の排他的論理和とは異なることに注意。
      ///
      /// 参考リンク : https://ja.wikipedia.org/wiki/%E3%82%A2%E3%83%AB%E3%83%95%E3%82%A1%E3%83%81%E3%83%A3%E3%83%B3%E3%83%8D%E3%83%AB
      ///
      /// @param fg 上側の(重ねあわせる)色
      /// @return 自分自身への参照
      RGB& alphaXor( RGB fg )
      {
        // 1 で正規化した値
        double fa = static_cast< double >( fg.a ) / MAX;
        double ba = static_cast< double >( a ) / MAX;
        double oa = fa * ( 1.0 - ba ) + ba * ( 1.0 - fa );

        a = Check_RGBPrimary< double >( std::round( oa * MAX ) );
        r = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.r * fa * ( 1.0 - ba ) + r * ba * ( 1.0 - fa ) ) / oa ) );
        g = ( oa == 0 ) ? 0 :
          Check_RGBPrimary< double >
          ( std::round
            ( ( fg.g * fa * ( 1.0 - ba ) + g * ba * ( 1.0 - fa ) ) / oa ) );
        b = ( oa == 0 ) ? 0
          : Check_RGBPrimary< double >
          ( std::round
            ( ( fg.b * fa * ( 1.0 - ba ) + b * ba * ( 1.0 - fa ) ) / oa ) );

        return( *this );
      }

      /*
        演算子の多重定義
      */

      /// @brief ビット反転演算子(~)への多重定義
      ///
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @return 演算結果
      RGB operator~() const
      { return( RGB( ~r & MASK, ~g & MASK, ~b & MASK, a ) ); }

      /// @brief 一致判定(==)への多重定義
      ///
      /// @note アルファチャンネルは比較されない。
      ///
      /// @param p 判定対象の RGB 成分
      /// @return 判定結果 ( 一致なら true )
      bool operator==( RGB p ) const
      { return( r == p.r && g == p.g && b == p.b ); }

      /// @brief 不一致判定(!=)への多重定義
      ///
      /// @note アルファチャンネルは比較されない。
      ///
      /// @param p 判定対象の RGB 成分
      /// @return 判定結果 ( 不一致なら true )
      bool operator!=( RGB p ) const
      { return( r != p.r || g != p.g || b != p.b ); }

      /* 四則演算(和・差) */

      /// @brief RGB 成分を加算する
      ///
      /// @note 各成分に対し、加算結果がオーバーフローした場合は最大値に丸められる。
      /// @note アルファチャンネルは加算対象とならない。
      ///
      /// @param rgb 加算する値
      /// @return 自分自身への参照
      RGB& operator+=( RGB rgb )
      {
        r = RGB::Check_RGBPrimary< int16_t >( static_cast< int16_t >( r ) + static_cast< int16_t >( rgb.r ) );
        g = RGB::Check_RGBPrimary< int16_t >( static_cast< int16_t >( g ) + static_cast< int16_t >( rgb.g ) );
        b = RGB::Check_RGBPrimary< int16_t >( static_cast< int16_t >( b ) + static_cast< int16_t >( rgb.b ) );

        return( *this );
      }

      /// @brief RGB に同じ成分を加算する
      ///
      /// @note 各成分に対し、加算結果がオーバーフローした場合は最大値に丸められる。
      /// @note アルファチャンネルは加算対象とならない。
      ///
      /// @param value 加算する値
      /// @return 自分自身への参照
      RGB& operator+=( primary_type value )
      {
        r = Check_RGBPrimary< int16_t >( static_cast< int16_t >( r ) + static_cast< int16_t >( value ) );
        g = Check_RGBPrimary< int16_t >( static_cast< int16_t >( g ) + static_cast< int16_t >( value ) );
        b = Check_RGBPrimary< int16_t >( static_cast< int16_t >( b ) + static_cast< int16_t >( value ) );

        return( *this );
      }

      /// @brief RGB 成分を減算する
      ///
      /// @note 各成分に対し、減算結果が負数になった場合はゼロに丸められる。
      /// @note アルファチャンネルは減算対象とならない。
      ///
      /// @param rgb 減算する値
      /// @return 自分自身への参照
      RGB& operator-=( RGB rgb )
      {
        r = Check_RGBPrimary< int16_t >( static_cast< int16_t >( r ) - static_cast< int16_t >( rgb.r ) );
        g = Check_RGBPrimary< int16_t >( static_cast< int16_t >( g ) - static_cast< int16_t >( rgb.g ) );
        b = Check_RGBPrimary< int16_t >( static_cast< int16_t >( b ) - static_cast< int16_t >( rgb.b ) );

        return( *this );
      }

      /// @brief RGB から同じ成分を減算する
      ///
      /// @note 各成分に対し、減算結果が負数になった場合はゼロに丸められる。
      /// @note アルファチャンネルは減算対象とならない。
      ///
      /// @param value 減算する値
      /// @return 自分自身への参照
      RGB& operator-=( primary_type value )
      {
        r = Check_RGBPrimary< int >( static_cast< int16_t >( r ) - static_cast< int16_t >( value ) );
        g = Check_RGBPrimary< int >( static_cast< int16_t >( g ) - static_cast< int16_t >( value ) );
        b = Check_RGBPrimary< int >( static_cast< int16_t >( b ) - static_cast< int16_t >( value ) );

        return( *this );
      }

      /* ビット演算子 */

      /// @brief RGB成分との論理積(AND)をとる
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param mask 論理積(AMD)用マスク
      /// @return 自分自身への参照
      RGB& operator&=( RGB mask )
      {
        r &= mask.r;
        g &= mask.g;
        b &= mask.b;

        return( *this );
      }

      /// @brief 同じ成分との論理積(AND)をとる
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param mask 論理積(AMD)用マスク
      /// @return 自分自身への参照
      RGB& operator&=( primary_type mask )
      {
        r &= mask;
        g &= mask;
        b &= mask;

        return( *this );
      }

      /// @brief RGB成分との論理和(OR)をとる
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param mask 論理和(OR)用マスク
      /// @return 自分自身への参照
      RGB& operator|=( RGB mask )
      {
        r |= mask.r;
        g |= mask.g;
        b |= mask.b;

        return( *this );
      }

      /// @brief 同じ成分との論理和(OR)をとる
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param mask 論理和(OR)用マスク
      /// @return 自分自身への参照
      RGB& operator|=( primary_type mask )
      {
        r |= mask;
        g |= mask;
        b |= mask;

        return( *this );
      }

      /// @brief RGB成分との排他的論理和(XOR)をとる
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param mask 排他的論理和(XOR)用マスク
      /// @return 自分自身への参照
      RGB& operator^=( RGB mask )
      {
        r ^= mask.r;
        g ^= mask.g;
        b ^= mask.b;

        return( *this );
      }

      /// @brief 同じ成分との排他的論理和(XOR)をとる
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param mask 排他的論理和(XOR)用マスク
      /// @return 自分自身への参照
      RGB& operator^=( primary_type mask )
      {
        r ^= mask;
        g ^= mask;
        b ^= mask;

        return( *this );
      }

      /* ビットシフト */

      /// @brief RGB各成分のビットを右シフトする
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param shift シフト回数
      /// @return 自分自身への参照
      RGB& operator>>=( uint32_t shift )
      {
        r >>= shift;
        g >>= shift;
        b >>= shift;

        return( *this );
      }

      /// @brief RGB各成分のビットを左シフトする
      ///
      /// @note アルファチャンネルは演算対象とならない。
      ///
      /// @param shift シフト回数
      /// @return 自分自身への参照
      RGB& operator<<=( uint32_t shift )
      {
        r <<= shift;
        g <<= shift;
        b <<= shift;

        return( *this );
      }

      /* 二つの要素の演算結果を返す */

      /// @brief RGB成分どうしの和を返す
      ///
      /// @note 各成分に対し、加算結果がオーバーフローした場合は最大値に丸められる
      /// @note アルファチャンネルは左辺の値がそのまま反映される。
      ///
      /// @param src 加算する値
      /// @return 計算結果
      RGB operator+( RGB src ) const
      {
        using MemFunc = RGB& ( RGB::* )( RGB );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator+= ) ) );
      }

      /// @brief RGB各成分に同じ成分を加算した結果を返す
      ///
      /// @note 各成分に対し、加算結果がオーバーフローした場合は最大値に丸められる
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator+( primary_type src ) const
      {
        using MemFunc = RGB& ( RGB::* )( primary_type );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator+= ) ) );
      }

      /// @brief RGB成分どうしの差を返す
      ///
      /// @note 各成分に対し、減算結果が負数になった場合はゼロに丸められる
      /// @note アルファチャンネルは左辺の値がそのまま反映される。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator-( RGB src ) const
      {
        using MemFunc = RGB& ( RGB::* )( RGB );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator-= ) ) );
      }

      /// @brief RGB各成分に同じ成分を減算した結果を返す
      ///
      /// @note 各成分に対し、減算結果が負数になった場合はゼロに丸められる
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator-( primary_type src ) const
      {
        using MemFunc = RGB& ( RGB::* )( primary_type );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator-= ) ) );
      }

      /// @brief RGB各成分どうしの論理積(AND)を計算して返す
      ///
      /// @note アルファチャンネルは左辺の値がそのまま反映される。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator&( RGB src ) const
      {
        using MemFunc = RGB& ( RGB::* )( RGB );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator&= ) ) );
      }

      /// @brief RGB各成分に対して同じ成分で論理積(AND)を計算して返す
      ///
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator&( primary_type src ) const
      {
        using MemFunc = RGB& ( RGB::* )( primary_type );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator&= ) ) );
      }

      /// @brief RGB各成分どうしの論理和(OR)を計算して返す
      ///
      /// @note アルファチャンネルは左辺の値がそのまま反映される。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator|( RGB src ) const
      {
        using MemFunc = RGB& ( RGB::* )( RGB );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator|= ) ) );
      }

      /// @brief RGB各成分に対して同じ成分で論理和(OR)を計算して返す
      ///
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator|( primary_type src ) const
      {
        using MemFunc = RGB& ( RGB::* )( primary_type );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator|= ) ) );
      }

      /// @brief RGB各成分どうしの排他的論理和(XOR)を計算して返す
      ///
      /// @note アルファチャンネルは左辺の値がそのまま反映される。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator^( RGB src ) const
      {
        using MemFunc = RGB& ( RGB::* )( RGB );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator^= ) ) );
      }

      /// @brief RGB各成分に対して同じ成分で排他的論理和(XOR)を計算して返す
      ///
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param src 演算対象
      /// @return 計算結果
      RGB operator^( primary_type src ) const
      {
        using MemFunc = RGB& ( RGB::* )( primary_type );

        return( RGB_Operator( *this, src, static_cast< MemFunc >( &RGB::operator^= ) ) );
      }

      /// @brief RGB各成分に対して右ビットシフトした結果を返す
      ///
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param shift シフト回数
      /// @return 計算結果
      RGB operator>>( uint32_t shift ) const
      { return( RGB_Operator( *this, shift, &RGB::operator>>= ) ); }

      /// @brief RGB各成分に対して左ビットシフトした結果を返す
      ///
      /// @note アルファチャンネルは影響を受けない。
      ///
      /// @param shift シフト回数
      /// @return 計算結果
      RGB operator<<( uint32_t shift ) const
      { return( RGB_Operator( *this, shift, &RGB::operator<<= ) ); }

      /// @brief RGB::primary_typeの値の範囲内にあるかを判定し、そうでなければ範囲内に収める
      ///
      /// @param p 対象の値
      /// @return 修正後の値
      template< class T > static RGB::primary_type Check_RGBPrimary( T p )
      {
        if ( p < 0 ) p = 0;
        if ( p > RGB::MAX ) p = RGB::MAX;

        return( static_cast< RGB::primary_type >( p ) );
      }
    };

    /*****************************************************
      RGB_Op : RGB 成分への処理用関数オブジェクト群
    *****************************************************/

    /**
       @brief RGB 成分への処理用関数インターフェース

       RGB 成分に対して処理を行う関数オブジェクトを呼び出す場合に利用するインターフェース。
    **/
    struct RGB_Op_IF
    {
      /// @brief RGB 成分への処理 (純粋仮想関数)
      ///
      /// @param rgb RGB 成分
      /// @return 実装に依存
      virtual bool operator()( RGB* rgb ) const = 0;

      /// @brief 仮想デストラクタ (何もしない)
      virtual ~RGB_Op_IF() {}
    };

    /**
       @brief RGB 成分への処理用関数

       RGB 成分に対するメンバ関数を呼び出すことができる
    **/
    template< class T > class RGB_Op : public RGB_Op_IF
    {
    public:

      using MemFun = RGB& ( RGB::* )( T ); ///< RGB 成分に対するメンバ関数の型

    private:

      MemFun mf_; // RGB 成分に対するメンバ関数
      T t_;   // mf_ に渡す引数

    public:

      /* コンストラクタ */

      /// @brief RGB 成分に対するメンバ関数とその引数の値を指定して構築
      ///
      /// 例えば、以下のコードは白色 (FF,FF,FF) をグレー (0F,0F,0F) に変える
      /// @code
      /// RGB_Op< RGB::primary_type > op( &RGB::operator&=, 0xF );
      /// RGB rgb( 0xFF, 0xFF, 0xFF );
      /// op( &rgb );
      /// @endcode
      ///
      /// @param mf RGB 成分に対するメンバ関数へのポインタ
      /// @param t f に渡す引数
      RGB_Op( MemFun mf, const T& t )
        : mf_( mf ), t_( t ) {}

      /// @brief RGB 成分に対するメンバ関数をセットする
      ///
      /// @param mf 新たなメンバ関数へのポインタ
      void setMemFun( MemFun mf )
      { mf_ = mf; }

      /// @brief 引数の値を返す
      ///
      /// @return 引数の値
      T getArg() const
      { return( t_ ); }

      /// @brief 引数をセットする
      ///
      /// @param t 新たな引数の値
      void setArg( T t )
      { t_ = t; }

      /// @brief コンストラクタで指定したメンバ関数 f を呼び出す
      ///
      /// mf が未定義だった場合は assert を実行する。
      ///
      /// @param rgb 処理対象の RGB 成分へのポインタ
      /// @return 常に true
      virtual bool operator()( RGB* rgb ) const
      {
        assert( mf_ != nullptr );

        ( rgb->*mf_ )( t_ );

        return( true );
      }
    };

    /**
       @brief 色のマスキング用クラス
    **/
    template< class T > class MaskColor : private RGB_Op< T >
    {
    public:

      /* コンストラクタ */

      /// @brief マスクパターンを指定して構築
      ///
      /// @param mask マスクパターン
      /// @param andMask 論理積をとるか (true)、論理和をとるか (false)
      MaskColor( const T& mask, bool andMask )
        : RGB_Op< T >( ( ( andMask ) ?
                         static_cast< typename RGB_Op< T >::MemFun >( &RGB::operator&= ) :
                         static_cast< typename RGB_Op< T >::MemFun >( &RGB::operator|= ) ),
                       mask ) {}

      /// @brief マスクパターンの値を返す
      ///
      /// @return マスクパターンの値
      T getMask() const
      { return( RGB_Op< T >::getArg() ); }

      /// @brief マスクパターンをセットする
      ///
      /// @param mask 新たなマスクパターンの値
      void setMask( T mask )
      { RGB_Op< T >::setArg( mask ); }

      /// @brief 色のマスキングを実行する
      ///
      /// @param rgb マスキング処理対象の RGB 成分へのポインタ
      /// @return 常に true
      bool operator()( RGB* rgb ) const
      { return( RGB_Op< T >::operator()( rgb ) ); }
    };

    /**
       @brief 色成分変換用ルーチン
    **/
    namespace color
    {
      /* RGB → YUV 成分変換 */

      /// @brief RGB各成分から Y 成分を求める
      ///
      /// @note Y 成分を求める計算式は Y = 0.299R + 0.587G + 0.114B
      ///
      /// @param rgb 対象の RGB 成分
      /// @return 求めた Y 成分
      inline double RGBtoY( const RGB& rgb )
      {
        return(   ( 0.299 * static_cast< double >( rgb.r ) )
                  + ( 0.587 * static_cast< double >( rgb.g ) )
                  + ( 0.114 * static_cast< double >( rgb.b ) ) );
      }

      /// @brief RGB各成分から U 成分を求める
      ///
      /// @note U 成分を求める計算式は U = -0.1687R - 0.3313G + 0.5B
      ///
      /// @param rgb 対象の RGB 成分
      /// @return 求めた U 成分
      inline double RGBtoU( const RGB& rgb )
      {
        return( - ( 0.1687 * static_cast< double >( rgb.r ) )
                - ( 0.3313 * static_cast< double >( rgb.g ) )
                + ( 0.5    * static_cast< double >( rgb.b ) ) );
      }

      /// @brief RGB各成分から V 成分を求める
      ///
      /// @note V 成分を求める計算式は V = 0.5R - 0.4187G - 0.0813B
      ///
      /// @param rgb 対象の RGB 成分
      /// @return 求めた V 成分
      inline double RGBtoV( const RGB& rgb )
      {
        return(   ( 0.5    * static_cast< double >( rgb.r ) )
                  - ( 0.4187 * static_cast< double >( rgb.g ) )
                  - ( 0.0813 * static_cast< double >( rgb.b ) ) );
      }

      /* RGB → IQ 成分変換 */

      /// @brief RGB各成分から I 成分を求める
      ///
      /// @note I 成分を求める計算式は I = 0.596R - 0.274G - 0.322B
      ///
      /// @param rgb 対象の RGB 成分
      /// @return 求めた I 成分
      inline double RGBtoI( const RGB& rgb )
      {
        return( ( 0.596 * static_cast< double >( rgb.r ) ) -
                ( 0.274 * static_cast< double >( rgb.g ) ) -
                ( 0.322 * static_cast< double >( rgb.b ) ) );
      }

      /// @brief RGB各成分から Q 成分を求める
      ///
      /// @note Q 成分を求める計算式は Q = 0.211R - 0.522G + 0.311B
      ///
      /// @param rgb 対象の RGB 成分
      /// @return 求めた Q 成分
      inline double RGBtoQ( const RGB& rgb )
      {
        return( ( 0.211 * static_cast< double >( rgb.r ) ) -
                ( 0.522 * static_cast< double >( rgb.g ) ) +
                ( 0.311 * static_cast< double >( rgb.b ) ) );
      }

      /* YUV → RGB 成分変換 */

      /// @brief YUV(YCbCr) 各成分から R 成分を求める
      ///
      /// @note R 成分を求める計算式は R = Y + 1.402V
      ///
      /// @param y,u,v 対象の YUV 成分
      /// @return 求めた R 成分
      inline RGB::primary_type YUVtoR( double y, double u, double v )
      {
        return( RGB::Check_RGBPrimary< double >( y + 1.402 * v + 0.5 ) );
      }
      /// @brief YUV(YCbCr) 各成分から G 成分を求める
      ///
      /// @note G 成分を求める計算式は G = Y - 0.34414U - 0.71414V
      ///
      /// @param y,u,v 対象の YUV 成分
      /// @return 求めた G 成分
      inline RGB::primary_type YUVtoG( double y, double u, double v )
      {
        return( RGB::Check_RGBPrimary< double >( y - 0.34414 * u - 0.71414 * v + 0.5 ) );
      }
      /// @brief YUV(YCbCr) 各成分から B 成分を求める
      ///
      /// @note B 成分を求める計算式は G = Y + 1.772U
      ///
      /// @param y,u,v 対象の YUV 成分
      /// @return 求めた B 成分
      inline RGB::primary_type YUVtoB( double y, double u, double v )
      {
        return( RGB::Check_RGBPrimary< double >( y + 1.772 * u + 0.5 ) );
      }

    } // namespace color

  } // namespace gl

} // namespace tadah_fussy

#endif
