#include "gl_curve.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace shape
    {
      namespace curve
      {

        using std::vector;

        /*
          BezierCurve::calc : バーンスタイン多項式の値を計算する

          p : 点列の座標成分(多項式の係数)
          t : 関数に代入する値

          戻り値 : 計算結果
        */
        double BezierCurve::calc( const vector< double >& p, double t )
        {
          size_t n = size(); // 点列のサイズ
          double d = 0;      // 計算結果

          for ( size_t k = 0 ; k < n ; ++k )
            d += func_( k, t ) * p[k];

          return( d );
        }

        /*
          DeCasteljau::calc : De Casteljau のアルゴリズムを使って曲線の X/Y 座標を求める

          p : 点列の座標成分(多項式の係数)
          t : 関数に代入する値

          戻り値 : 計算結果
        */
        double DeCasteljau::calc( const vector< double >& p, double t )
        {
          if ( size() <= 1 ) return( NAN );

          // 点列から最初の内点を求めて buff_ に格納する
          for ( size_t s = 1 ; s < p.size() ; ++s )
            buff_[s - 1] = t * p[s - 1] + ( 1 - t ) * p[s];

          // 再帰的に内点を計算
          for ( size_t e = buff_.size() ; e > 1 ; --e )
            for ( size_t s = 1 ; s < e ; ++s )
              buff_[s - 1] = t * buff_[s - 1] + ( 1 - t ) * buff_[s];

          return( buff_[0] );
        }

        /*
          CalcBSplineBasisFunc : B-スプライン基底関数の計算

          knot : ノット列
          j : ノット列の開始番号
          p : 次数
          t : 計算対象の独立変数

          ノット列は昇順である必要があるが、そのチェックは行わない

          戻り値 : 計算結果
        */
        double CalcBSplineBasisFunc( const vector< double >& knot, size_t j, size_t p, double t )
        {
          if ( knot.size() == 0 ) return( NAN );

          // ノット列のデータ長が充分でない場合は nan を返す
          size_t m = knot.size() - 1;
          if ( m < j + p + 1 ) return( NAN );

          // 正値をとる範囲外ならゼロを返す
          if ( ( t < knot[j] ) || ( t > knot[j + p + 1] ) ) return( 0 );
          // p = 0 かつ knot[j] <= t <= knot[j + p + 1] なら 1 を返す
          if ( p == 0 ) return( 1 );
          // p = 1 の場合、三角の頂点の値は特別扱い
          if ( p == 1 && t == knot[j + 1]  ) return( 1 );

          // 漸化式の計算
          double d1 = ( knot[j + p] == knot[j] ) ? 0 :
            ( t - knot[j] ) * CalcBSplineBasisFunc( knot, j, p - 1, t ) / ( knot[j + p] - knot[j] );
          double d2 = ( knot[j + p + 1] == knot[j + 1] ) ? 0 :
            ( knot[j + p + 1] - t ) * CalcBSplineBasisFunc( knot, j + 1, p - 1, t ) / ( knot[j + p + 1] - knot[j + 1] );

          return( d1 + d2 );
        }

        /*
          CalcBSplineBasisDerivedFunc : B-スプライン基底関数の導関数の計算

          knot : ノット列
          j : ノット列の開始番号
          p : 次数
          t : 計算対象の独立変数

          ノット列は昇順である必要があるが、そのチェックは行わない
          ノット列のデータ長が充分でない場合は nan を返す

          戻り値 : 計算結果
        */
        double CalcBSplineBasisDerivedFunc( const vector< double >& knot, size_t j, size_t p, double t )
        {
          if ( knot.size() == 0 ) return( NAN );
          size_t m = knot.size() - 1; // ノット列の添字の最大値
          if ( m < j + p + 1 ) return( NAN );

          if ( p == 0 ) return( 0 );
          if ( ( t < knot[j] ) || ( t > knot[j + p + 1] ) ) return( 0 );

          // 漸化式のいずれの項も計算不能であれば、nan を返す
          if ( knot[j + p] == knot[j] && knot[j + p + 1] == knot[j + 1] ) return( NAN );

          double d1 = ( knot[j + p] == knot[j] ) ? 0 :
            p * CalcBSplineBasisFunc( knot, j, p - 1, t ) / ( knot[j + p] - knot[j] );
          double d2 = ( knot[j + p + 1] == knot[j + 1] ) ? 0 :
            p * CalcBSplineBasisFunc( knot, j + 1, p - 1, t ) / ( knot[j + p + 1] - knot[j + 1] );

          return( d1 - d2 );
        }

        /*
          BSplineCurve::calc : 曲線の X/Y 座標値を計算する

          pnt : 点列の X/Y 座標
          t : 計算対象の独立変数

          返り値 : 計算結果
        */
        double BSplineCurve::calc( const vector< double >& pnt, double t )
        {
          double ans = 0; // 計算結果
          size_t p = order(); // 次数

          for ( size_t j = 0 ; j < size() ; ++j ) {
            double d = pnt[j] * CalcBSplineBasisFunc( knot_, j, p, t );
            if ( ! std::isnan( d ) )
              ans += d;
          }

          return( ans );
        }

        /*
          DeBoorCox::calc : 曲線の X/Y 座標値を計算する

          pnt : 点列の X/Y 座標
          t : 計算対象の独立変数

          返り値 : 計算結果
        */
        double DeBoorCox::calc( const vector< double >& pnt, double t )
        {
          if ( size() <= 1 ) return( NAN );
          size_t p = order(); // 次数

          // t を含むノット列の領域 [ knot_[s-1], knot_[s] ] の添字 s を求める
          size_t s = 0;
          for ( ; s < knot_.size() ; ++s )
            if ( t < knot_[s] ) break;
          if ( s <= p || s > size() )
            return( NAN );
          s -= p;

          // p = 0 のときは点列の座標値そのものを返す
          if ( p == 0 ) return( pnt[s - 1] );

          // 点列から最初の内点を求めて buff_ に格納する
          vector< double >::iterator it = buff_.begin();
          for ( size_t j = s ; j < p + s ; ++j )
            *it++ = ( ( knot_[j + p] - t ) * pnt[j - 1] + ( t - knot_[j] ) * pnt[j] )
              / ( knot_[j + p] - knot_[j] );

          // 再帰的に内点を計算
          for ( size_t e = p ; e > 1 ; --e ) {
            size_t inc = p - e + 1; // 前側のノットの添字に対する増分
            it = buff_.begin();
            for ( size_t j = s ; j < s + e - 1 ; ++j ) {
              *it = ( ( knot_[j + p] - t ) * *it + ( t - knot_[j + inc] ) * *( it + 1 ) )
                / ( knot_[j + p] - knot_[j + inc] );
              ++it;
            }
          }

          return( buff_[0] );
        }

        /*
          NurbsCurve::calc : 曲線の X/Y 座標値を計算する

          pnt : 点列の X/Y 座標
          t : 計算対象の独立変数

          返り値 : 計算結果
        */
        double NurbsCurve::calc( const vector< double >& pnt, double t )
        {
          size_t p = order(); // 次数

          vector< double > bs( size() ); // B-Spline 基底関数の計算結果(重み値を積算)
          double wSum = 0; // bs の合計
          for ( size_t j = 0 ; j < size() ; ++j ) {
            bs[j] = w_[j] * CalcBSplineBasisFunc( knot_, j, p, t );
            wSum += bs[j];
          }

          double ans = 0; // 計算結果
          for ( size_t j = 0 ; j < size() ; ++j ) {
            double d = pnt[j] * bs[j] / wSum;
            if ( ! std::isnan( d ) )
              ans += d;
          }

          return( ans );
        }

      } // namespace curve

    } // namespace shape

  } // namespace gl

} // namespace tadah_fussy
