/**
   @file gl_curve.h
   @brief 曲線描画用クラス定義ヘッダ

   @author tadah_fussy
   @date 2018/02/26 新規作成
   @date 2021/02/25 ワーニングの出力に対応
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_CURVE_20210402_H
#define GRAPHIC_LIBRARY_CURVE_20210402_H

#include <vector>
#include <common/coord.h>
#include <common/math_library.h>

namespace tadah_fussy
{
  namespace gl
  {
    namespace shape
    {
      namespace curve
      {
      /**
         @brief パラメトリック方程式用基底クラス

         媒介変数 t を用いて ( x, y ) の関数を表した式。例えば、

         x = r・cos( t ), y = r・sin( t )

         で円の方程式を表す。
      **/
      struct ParametricEquation
      {
        /// @brief x(t) の値を求める
        ///
        /// @param t 媒介変数
        /// @return x の値
        virtual double x( double t ) = 0;
        /// @brief y(t) の値を求める
        ///
        /// @param t 媒介変数
        /// @return y の値
        virtual double y( double t ) = 0;

        /// @brief 仮想デストラクタ(何もしない)
        virtual ~ParametricEquation() {}
      };

      /**
         @brief Ferguson/Coons曲線

         3 次式 x(t) = at^3 + bt^2 + ct + d に 2 点 x(0), x(1) とその導関数 x'(0), x'(1) を代入して
         係数を連立方程式で解くと以下のようになる。

         <pre>
         a = 2[ x(0) - x(1) ] + [ x'(0) + x'(1) ]
         b = -3[ x(0) - x(1) ] - [ 2x'(0) + x'(1) ]
         c = x'(0)
         d = x(0)
         </pre>

         y についても同様の式が得られる。この 3 次式を解いて ( x(t), y(t) ) を描画することで曲線を得る。
      **/
      class FergusonCoons_Spline : public ParametricEquation
      {
        Coord< double > a[4]; // 係数(高次の項から順に並ぶ)

      public:

        /// @brief 曲線の両端の点とその導関数の値を指定して構築
        ///
        /// @param p0,p1 曲線の両端の点の座標
        /// @param dp0,dp1 p0,p1 での微分係数
        template< class T, class D >
        FergusonCoons_Spline( const Coord< T >& p0, const Coord< T >& p1,
                              const Coord< D >& dp0, const Coord< D >& dp1 )
        {
          a[0].x = 2 * ( p0.x - p1.x ) + ( dp0.x + dp1.x );
          a[1].x = -3 * ( p0.x - p1.x ) - ( 2 * dp0.x + dp1.x );
          a[2].x = dp0.x;
          a[3].x = p0.x;

          a[0].y = 2 * ( p0.y - p1.y ) + ( dp0.y + dp1.y );
          a[1].y = -3 * ( p0.y - p1.y ) - ( 2 * dp0.y + dp1.y );
          a[2].y = dp0.y;
          a[3].y = p0.y;
        }

        // x,y  の値を求める
        virtual double x( double t )
        { return( a[0].x * pow( t, 3 ) + a[1].x * pow( t, 2 ) + a[2].x * pow( t, 1 ) + a[3].x ); }
        virtual double y( double t )
        { return( a[0].y * pow( t, 3 ) + a[1].y * pow( t, 2 ) + a[2].y * pow( t, 1 ) + a[3].y ); }
      };

      /**
         @brief Catmull-Rom 曲線

         FergusonCoons/Spline 曲線において導関数を以下の式で固定したものを Catmull-Rom 曲線という。

         x_i'(0) = d * ( x_i+1 - x_i-1 ) / 2

         但し、d は補正係数とする。
      **/
      class CatmullRom_Spline : public FergusonCoons_Spline
      {
        /*
          diff : 微分係数の計算

          const Coord< double >& p0, p1 : 計算対象の前後の点の座標
          double d : 補正係数

          戻り値 : 計算した微分係数
        */
        template< class T >
        static Coord< double > diff( const Coord< T >& p0, const Coord< T >& p1, double d )
        { return( Coord< double >( d * static_cast< double >( p1.x - p0.x ) / 2.0, d * static_cast< double >( p1.y - p0.y ) / 2.0 ) ); }

      public:

        /// @brief 対象の 2 点の前後も含めた 4 点を指定して構築
        ///
        /// 点 p の導関数の値は、前後の点 pa, pb と補正係数 d を使って以下の式で計算する
        ///
        /// d( pb - pa ) / 2
        ///
        /// @param before p0 の前の点の座標
        /// @param p0, p1 曲線の両端の点の座標
        /// @param after p1 の後の点の座標
        /// @param d 補正係数
        template< class T >
        CatmullRom_Spline( const Coord< T >& before, const Coord< T >& p0, const Coord< T >& p1, const Coord< T >& after, double d = 1 )
          : FergusonCoons_Spline( p0, p1, diff( before, p1, d ), diff( p0, after, d ) ) {}
      };

      /**
         @brief ベジェ曲線描画用基底クラス

         BezierCurve、DeCasteljau の基底となり、ベジェ曲線の描画に利用される。
      **/
      class BezierCurveBase : public ParametricEquation
      {
        size_t size_;             // 点列のサイズ
        std::vector< double > x_; // 点列の x 座標
        std::vector< double > y_; // 点列の y 座標

        // 曲線の X/Y 座標値を計算する(純粋仮想関数)
        //
        // 具体的な処理は派生クラスで実装を行う。
        //
        // p : 点列の座標成分(多項式の係数)
        // t : 関数に代入する値
        virtual double calc( const std::vector< double >& p, double t ) = 0;

      public:

        /*
          BezierCurveBase コンストラクタ
        */

        /// @brief 点列の数を指定して構築
        ///
        /// @param n 点列の数
        BezierCurveBase( size_t n )
          : size_( n ), x_( size() ), y_( size() ) {}

        /// @brief 配列の範囲を指定して構築
        ///
        /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
        ///
        /// @param in0,in1 配列の範囲 (in1 は末尾の次を表すため含まれない)
        template< class In >
        BezierCurveBase( In in0, In in1 )
          : size_( std::distance( in0, in1 ) ), x_( size() ), y_( size() )
        { set( in0, in1 ); }

        /// @brief 配列全体を指定して構築
        ///
        /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
        ///
        /// @param p 対象の点列
        template< class Vec >
        BezierCurveBase( const Vec& p )
          : size_( p.size() ), x_( size() ), y_( size() )
        { set( p.begin(), p.end() ); }

        /*
          set : 点列の座標をセットする
        */

        /// @brief 指定された位置に点列の座標をセットする
        ///
        /// @param k セットする位置
        /// @param p 対象の点の座標
        template< class T >
        void set( size_t k, const Coord< T >& p )
        {
          if ( k < size() ) {
            x_[k] = p.x;
            y_[k] = p.y;
          }
        }

        /// @brief 指定した範囲の配列の要素を、点列の指定位置から順にセットする
        ///
        /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
        ///
        /// セットできる点列の範囲が指定された範囲 [ in0, in1 ) より小さい場合、
        /// 指定された範囲の末尾側の要素は無視される
        ///
        /// @param k 点列にセットする位置
        /// @param in0,in1 配列の範囲 (in1 は末尾の次を表すため含まれない)
        template< class In >
        void set( size_t k, In in0, In in1 )
        {
          while ( in0 != in1 ) {
            if ( k >= size_ ) break;
            set( k, *in0 );
            ++in0;
            ++k;
          }
        }

        /// @brief 指定した範囲の配列の要素を点列の先頭から順にセットする
        ///
        /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
        ///
        /// セットできる点列の範囲が指定された範囲 [ in0, in1 ) より小さい場合、
        /// 指定された範囲の末尾側の要素は無視される
        ///
        /// @param in0,in1 配列の範囲 (in1 は末尾の次を表すため含まれない)
        template< class In >
        void set( In in0, In in1 )
        { set( 0, in0, in1 ); }

        /// @brief 点列のサイズを返す
        ///
        /// @return 点列のサイズ
        size_t size() const
        { return( size_ ); }

        // x, y の値を求める
        virtual double x( double t )
        { return( calc( x_, t ) ); }
        virtual double y( double t )
        { return( calc( y_, t ) ); }
      };

        /**
           @brief バーンスタイン多項式を使ったベジェ曲線
        **/
        class BezierCurve : public BezierCurveBase
        {
          math_lib::polynomial::Bernstein< double > func_; // Bernstein多項式

          // calc : バーンスタイン多項式の値を計算する
          virtual double calc( const std::vector< double >& p, double t );

        public:

          /*
            BezierCurve コンストラクタ

            BezierCurveBaseコンストラクタをそのまま利用
          */

          /// @brief 点列の数を指定して構築
          ///
          /// @param n 点列の数
          BezierCurve( size_t n )
            : BezierCurveBase( n ), func_( ( size() == 0 ) ? 0 : size() - 1 ) {}

          /// @brief 配列の範囲を点列として指定して構築
          ///
          /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
          ///
          /// @param in0,in1 配列の範囲 (in1 は末尾の次を表すため含まれない)
          template< class In >
          BezierCurve( In in0, In in1 )
            : BezierCurveBase( in0, in1 ), func_( ( size() == 0 ) ? 0 : size() - 1 ) {}

          /// @brief 配列全体を点列として指定して構築
          ///
          /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
          ///
          /// @param p 対象の点列
          template< class Vec >
          BezierCurve( const Vec& p )
            : BezierCurveBase( p ), func_( ( size() == 0 ) ? 0 : size() - 1 ) {}
        };

        /**
           @brief De Casteljau のアルゴリズムを使ったベジェ曲線
        **/
        class DeCasteljau : public BezierCurveBase
        {
          std::vector< double > buff_; // 計算用のバッファ

          // calc : De Casteljau のアルゴリズムを使って曲線の X/Y 座標を求める
          virtual double calc( const std::vector< double >& p, double t );

        public:

          /// @brief 点列の数を指定して構築
          ///
          /// @param n 点列の数
          DeCasteljau( size_t n )
            : BezierCurveBase( n ), buff_( ( size() == 0 ) ? 0 : size() - 1 ) {}

          /// @brief 配列の範囲を指定して構築
          ///
          /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
          ///
          /// @param in0,in1 配列の範囲 (in1 は末尾の次を表すため含まれない)
          template< class In >
          DeCasteljau( In in0, In in1 )
            : BezierCurveBase( in0, in1 ), buff_( ( size() == 0 ) ? 0 : size() - 1 ) {}

          /// @brief 配列全体を指定して構築
          ///
          /// 引数に指定した配列は Coord< T > 型の要素を持つことを想定している。
          ///
          /// @param p 対象の点列
          template< class Vec >
          DeCasteljau( const Vec& p )
            : BezierCurveBase( p ), buff_( ( size() == 0 ) ? 0 : size() - 1 ) {}
        };

        /**
           @brief B-スプライン曲線描画用基底クラス

           非公開な純粋仮想関数 calc を内部で定義しているため、単独での利用はできない。
           なお、calc は曲線の座標値計算に利用される。
        **/
        class BSplineCurveBase : public ParametricEquation
        {
          size_t p_; // 次数
          std::vector< double > x_; // 点列の x 座標
          std::vector< double > y_; // 点列の y 座標

        protected:

          std::vector< double > knot_; ///< ノット列

        private:

          // calc : 曲線の X/Y 座標値を計算する(純粋仮想関数)
          virtual double calc( const std::vector< double >& p, double t ) = 0;

          // set : メンバ変数の初期化
          template< class In1, class In2 >
          void set( In1 knot0, In1 knot1, In2 p0, In2 p1, size_t p )
          {
            p_ = p;

            // 点列の大きさ
            size_t pntSize = std::distance( p0, p1 );
            // ノット列の大きさ
            size_t knotSize = std::distance( knot0, knot1 );

            // ノット列の初期化
            // 用意したノット列の要素が pntSize + p_ + 1 より少なければ、末尾側は最後の要素で補う
            // 逆に多ければ末尾側の要素は無視する
            if ( knotSize < pntSize + p_ + 1 ) {
              knot_.assign( knot0, knot1 );
              for ( size_t i = knotSize ; i < pntSize + p_ + 1 ; ++i )
                knot_.push_back( knot_.back() );
            } else {
              knot_.assign( knot0, knot0 + pntSize + p_ + 1 );
            }

            // 点列を X/Y に分解して初期化
            while ( p0 != p1 ) {
              x_.push_back( p0->x );
              y_.push_back( p0->y );
              ++p0;
            }
          }

        public:

          /*
            BSplineCurveBase コンストラクタ
          */

          /// @brief 配列の範囲を指定して構築
          ///
          /// @param knot0,knot1 ノット列の範囲 (knot1 は末尾の次を表すため含まれない)
          /// @param p0,p1 点列の範囲 (p1 は末尾の次を表すため含まれない)
          /// @param p 次数
          template< class In1, class In2 >
          BSplineCurveBase( In1 knot0, In1 knot1, In2 p0, In2 p1, size_t p )
          { set( knot0, knot1, p0, p1, p ); }

          /// @brief 配列全体を指定して構築
          ///
          /// @param knot ノット列
          /// @param pnt 点列
          /// @param p 次数
          template< class Vec1, class Vec2 >
          BSplineCurveBase( const Vec1& knot, const Vec2& pnt, size_t p )
          { set( knot.begin(), knot.end(), pnt.begin(), pnt.end(), p ); }

          /// @brief 点列のサイズを返す
          ///
          /// @return 点列のサイズ
          size_t size() const
          { return( x_.size() ); }

          /// @brief 次数を返す
          ///
          /// @return 次数
          size_t order() const
          { return( p_ ); }

          // x, y の値を求める
          virtual double x( double t )
          { return( calc( x_, t ) ); }
          virtual double y( double t )
          { return( calc( y_, t ) ); }
        };

        /**
           @brief B-スプライン曲線描画用クラス
        **/
        class BSplineCurve : public BSplineCurveBase
        {
          // calc : 曲線の X/Y 座標値を計算する
          virtual double calc( const std::vector< double >& p, double t );

        public:

          /*
            BSplineCurve コンストラクタ

            BSplineCurveBase をそのまま利用
          */

          /// @brief 配列の範囲を指定して構築
          ///
          /// @param knot0,knot1 ノット列の範囲 (knot1 は末尾の次を表すため含まれない)
          /// @param p0,p1 点列の範囲 (p1 は末尾の次を表すため含まれない)
          /// @param p 次数
          template< class In1, class In2 >
          BSplineCurve( In1 knot0, In1 knot1, In2 p0, In2 p1, size_t p )
            : BSplineCurveBase( knot0, knot1, p0, p1, p ) {}

          /// @brief 配列全体を指定して構築
          ///
          /// @param knot ノット列
          /// @param pnt 点列
          /// @param p 次数
          template< class Vec1, class Vec2 >
          BSplineCurve( const Vec1& knot, const Vec2& pnt, size_t p )
            : BSplineCurveBase( knot, pnt, p ) {}
        };

        /**
           @brief De Boor-Cox のアルゴリズムを使った B-スプライン曲線
        **/
        class DeBoorCox : public BSplineCurveBase
        {
          std::vector< double > buff_; // 内点の計算結果を保持するバッファ

          // calc : 曲線の X/Y 座標値を計算する
          virtual double calc( const std::vector< double >& p, double t );

        public:

          /*
            DeBoorCox コンストラクタ

            BSplineCurveBase をそのまま利用
          */

          /// @brief 配列の範囲を指定して構築
          ///
          /// @param knot0,knot1 ノット列の範囲 (knot1 は末尾の次を表すため含まれない)
          /// @param p0,p1 点列の範囲 (p1 は末尾の次を表すため含まれない)
          /// @param p 次数
          template< class In1, class In2 >
          DeBoorCox( In1 knot0, In1 knot1, In2 p0, In2 p1, size_t p )
            : BSplineCurveBase( knot0, knot1, p0, p1, p ),
              buff_( p ) {}

          /// @brief 配列全体を指定して構築
          ///
          /// @param knot ノット列
          /// @param pnt 点列
          /// @param p 次数
          template< class Vec1, class Vec2 >
          DeBoorCox( const Vec1& knot, const Vec2& pnt, size_t p )
            : BSplineCurveBase( knot, pnt, p ),
              buff_( p ) {}
        };

        /**
           @brief NURBS 曲線描画用クラス
        **/
        class NurbsCurve : public BSplineCurveBase
        {
          std::vector< double > w_; // 重み値

          // calc : 曲線の X/Y 座標値を計算する
          virtual double calc( const std::vector< double >& p, double t );

          // setWeight : 重み値の取り込み
          template< class In >
          void setWeight( In w0, In w1 )
          {
            w_.assign( size(), 1 ); // 重み値のデフォルトは 1 とする

            size_t sz = std::distance( w0, w1 );

            // データ数が超過している場合、末尾は捨てる
            if ( sz > size() ) {
              w1 = w0;
              std::advance( w1, size() );
            }
            std::copy( w0, w1, w_.begin() );          // データは範囲内
          }

        public:

          /*
            NurbsCurve コンストラクタ
          */

          /// @brief 配列の範囲を指定して構築
          ///
          /// @param knot0,knot1 ノット列の範囲 (knot1 は末尾の次を表すため含まれない)
          /// @param p0,p1 点列の範囲 (p1 は末尾の次を表すため含まれない)
          /// @param w0,w1 重み値の範囲 (w1 は末尾の次を表すため含まれない)
          /// @param p 次数
          template< class In1, class In2, class In3 >
          NurbsCurve( In1 knot0, In1 knot1,
                      In2 p0, In2 p1,
                      In3 w0, In3 w1,
                      size_t p )
            : BSplineCurveBase( knot0, knot1, p0, p1, p )
          { setWeight( w0, w1 ); }

          /// @brief 配列全体を指定して構築
          ///
          /// @param knot ノット列
          /// @param pnt 点列
          /// @param w 重み値の列
          /// @param p 次数
          template< class Vec1, class Vec2, class Vec3 >
          NurbsCurve( const Vec1& knot, const Vec2& pnt, const Vec3& w, size_t p )
            : BSplineCurveBase( knot, pnt, p )
          { setWeight( w.begin(), w.end() ); }
        };

      } // namespace curve

    } // namespace shape

  } // namespace gl

} // namespace tadah_fussy

#endif
