#include "gl_drawing_method.h"
#include "gl_curve.h"

#include <queue>
#include <set>

namespace tadah_fussy
{
  namespace gl
  {
    using std::vector;
    using std::round;
    using coord_lib::Gradient;
    using coord_lib::Intercept;
    using coord_lib::Diff;
    using coord_lib::Sign;
    using coord_lib::Distance;
    using canvas::DrawingArea_IF;
    using pen::GPixelOp;

    namespace shape
    {
      /**************************************************************
          GLine : 線分描画用関数
      **************************************************************/
      namespace line
      {
        namespace
        {
          /*
            Edge : 端点分類コード

            線分の端点が描画エリアの外側にあれば、それぞれのビット位置を立てる
          */
          enum Edge {
            LEFT = 1,
            RIGHT = 2,
            TOP = 4,
            BOTTOM = 8,
          };

          /*
            GetEdgeCode : 端点分類コードを求める
          */
          int GetEdgeCode( Coord<int> c, Coord<int> drawSize )
          {
            int code = 0;
            if ( c.x < 0 ) code += LEFT;
            if ( c.x >= drawSize.x ) code += RIGHT;
            if ( c.y < 0 ) code += TOP;
            if ( c.y >= drawSize.y ) code += BOTTOM;

            return( code );
          }

          /*
            CalcPI_X : 線分 c0-c1 の x = clipX での交点を求める

            * PI は Point of Intersection の略
            */
          bool CalcPI_X( Coord<int> c0, Coord<int> c1, int clipX,
                         Coord<int> drawSize, Coord<int>* c )
          {
            if ( c1.x == c0.x )
              return( c0.x < drawSize.x );

            // 切片を求める
            double cy = coord_lib::Intercept< double >( c0, c1, false, clipX );

            // 交点がエリア外の場合はFalseを返す
            if ( cy < 0 || cy >= static_cast<double>( drawSize.y ) )
              return( false );

            c->x = clipX;
            c->y = static_cast<int>( cy );

            return( true );
          }

          /*
            CalcPI_Y : 線分 c0-c1 の y = clipY での交点を求める

            * PI は Point of Intersection の略
            */
          bool CalcPI_Y( Coord<int> c0, Coord<int> c1, int clipY,
                         Coord<int> drawSize, Coord<int>* c )
          {
            if ( c1.y == c0.y )
              return( c0.y < drawSize.y );

            double cx = coord_lib::Intercept< double >( c0, c1, true, clipY );

            // エリア外の場合はFalseを返す
            if ( cx < 0 || cx >= static_cast<double>( drawSize.x ) )
              return( false );

            c->x = static_cast<int>( cx );
            c->y = clipY;

            return( true );
          }

          /*
            CalcPI_Main : 端点の片側をクリッピングするためのサブルーチン

            code で表される領域にある c0-c1 上の端点を drawSize の大きさでクリッピングする
            結果は c に代入する
          */
          bool CalcPI_Main( int code, Coord<int> c0, Coord<int> c1, Coord<int> drawSize, Coord<int>* c )
          {
            // ウィンドウの左端より外側にある
            if ( ( code & LEFT ) != 0 )
              if ( CalcPI_X( c0, c1, 0, drawSize, c ) )
                return( true );

            // ウィンドウの右端より外側にある
            if ( ( code & RIGHT ) != 0 )
              if ( CalcPI_X( c0, c1, drawSize.x, drawSize, c ) )
                return( true );

            // ウィンドウの上端より外側にある
            if ( ( code & TOP ) != 0)
              if ( CalcPI_Y( c0, c1, 0, drawSize, c ) )
                return( true );

            // ウィンドウの下端より外側にある
            if ( ( code & BOTTOM ) != 0 )
              if ( CalcPI_Y( c0, c1, drawSize.y, drawSize, c ) )
                return( true );

            return( false );  // クリッピングされなかった場合、線分は完全に不可視
          }

        } // namespace

        /*
          Clipping : 線分 c0-c1 を drawSize の大きさのエリアでクリッピングする

          戻り値 :
          >0 ... クリッピングされた
          =0 ... クリッピングの必要なし
          <0 ... 線分は完全に不可視 ( または c0, c1 に未定義のものがある )
        */
        int Clipping( Coord<int> drawSize, Coord<int>* c0, Coord<int>* c1 )
        {
          if ( c0 == 0 || c1 == 0 ) return( -1 );

          int code0 = GetEdgeCode( *c0, drawSize ); // 始点の端点分類コードを求める
          int code1 = GetEdgeCode( *c1, drawSize ); // 終点の端点分類コードを求める

          // 端点分類コードがどちらも0000ならばクリッピングは必要なし
          if ( ( code0 == 0 ) && ( code1 == 0 ) ) return( 0 );

          // 始点・終点の端点分類コードの論理積が非０ならば線分は完全に不可視
          if ( ( code0 & code1 ) != 0 ) return( -1 );

          Coord<int> buff0( *c0 );
          Coord<int> buff1( *c1 );

          // 始点のクリッピング
          if( code0 != 0 )
            if ( ! CalcPI_Main( code0, buff0, buff1, drawSize, c0 ) )
              return( -1 );

          // 終点のクリッピング
          if ( code1 != 0 )
            if ( ! CalcPI_Main( code1, buff0, buff1, drawSize, c1 ) )
              return( -1 );

          return( 1 );
        }

        /*
          Draw : Bresenham線分発生アルゴリズムを用いて線分 c0-c1 を描画する
        */
        void Draw( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c0, Coord<int> c1, bool needClipping, bool connect4 )
        {
          Coord<int> d; // 二点間の距離
          Coord<int8_t> s; // 二点の方向

          Coord<int> drawSize = draw.size(); // 描画領域の大きさ

          // クリッピング処理
          if ( needClipping )
            Clipping( drawSize, &c0, &c1 );

          // c0, c1 の X, Y ぞれぞれについて差分と向きを求める
          coord_lib::Diff( c0, c1, &d );
          coord_lib::Sign( c0, c1, &s );

          // 傾きが1以下の場合
          if( d.x >= d.y ) {
            int error = -d.x; // 誤差項
            for( int i = 0 ; i <= d.x ; i++ ) {
              pixelOp( draw, c0 );
              c0.x += s.x;
              error += 2 * d.y;
              if ( error >= 0 ) {
                if ( connect4 ) pixelOp( draw, c0 ); // connect4=true なら 4 連結で描画
                c0.y += s.y;
                error -= 2 * d.x;
              }
            }
            // 傾きが1より大きい場合
          } else {
            int error = -d.y; // 誤差項
            for( int i = 0 ; i <= d.y ; i++ ) {
              pixelOp( draw, c0 );
              c0.y += s.y;
              error += 2 * d.x;
              if ( error >= 0 ) {
                if ( connect4 ) pixelOp( draw, c0 ); // connect4=true なら 4 連結で描画
                c0.x += s.x;
                error -= 2 * d.y;
              }
            }
          }
        }

        /*
          DrawMultiple : 複数のラインを描画する
        */
        void DrawMultiple( DrawingArea_IF& draw, GPixelOp& pixelOp, const vector< Coord<int> >& vertexList )
        {
          typedef std::pair< Coord< int >, Coord< int > > Edge; // 各ラインの端点を表す型

          Coord<int> drawSize = draw.size(); // 描画領域のサイズ
          vector<Edge> edgeList;             // ラインの端点用配列

          for ( unsigned int i = 1 ; i < vertexList.size() ; ++i ) {
            Edge edge( vertexList[i - 1], vertexList[i % vertexList.size()] );
            Clipping( drawSize, &( edge.first ), &( edge.second ) );
            edgeList.push_back( edge );
          }

          Coord<int> d; // 二点間の距離
          double a = 0; // 傾き
          for ( unsigned int i = 0 ; i < edgeList.size() ; ++i ) {
            Coord<int> c0 = edgeList[i].first;
            Coord<int> c1 = edgeList[i].second;
            Diff( c0, c1, &d );
            if ( d.x > d.y ) {
              if ( c0.x > c1.x ) std::swap( c0, c1 );
              a = coord_lib::Gradient< double >( c0, c1, false );
            } else {
              if ( c0.y > c1.y ) std::swap( c0, c1 );
              a = coord_lib::Gradient< double >( c0, c1, true );
            }
            if ( d.x > d.y ) {
              double y = c0.y;
              if ( edgeList[( i + edgeList.size() - 1 ) % edgeList.size()].second == c0 ) {
                ++( c0.x );
                y += a;
              }
              for ( ; c0.x < c1.x ; ++( c0.x ) ) {
                pixelOp( draw, Coord<int>( c0.x, round( y ) ) );
                y += a;
              }
              if ( edgeList[( i + edgeList.size() - 1 ) % edgeList.size()].second != c1 )
                pixelOp( draw, Coord<int>( c1.x, round( y ) ) );
            } else {
              double x = c0.x;
              if ( edgeList[( i + edgeList.size() - 1 ) % edgeList.size()].second == c0 ) {
                ++( c0.y );
                x += a;
              }
              for ( ; c0.y < c1.y ; ++( c0.y ) ) {
                pixelOp( draw, Coord<int>( round( x ), c0.y ) );
                x += a;
              }
              if ( edgeList[( i + edgeList.size() - 1 ) % edgeList.size()].second != c1 )
                pixelOp( draw, Coord<int>( round( x ), c1.y ) );
            }
          }
        }

        /*
          DrawH : c0 を起点として、X 座標が ex である点を終点とする水平線分を描画する
        */
        void DrawH( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c0, int ex )
        {
          if ( c0.x > ex ) std::swap( c0.x, ex );

          for ( ; c0.x <= ex ; ++( c0.x ) )
            pixelOp( draw, c0 );
        }

        /*
          DrawV : c0 を起点として、Y 座標が ey である点を終点とする垂直線分を描画する
        */
        void DrawV( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c0, int ey )
        {
          if ( c0.y > ey ) std::swap( c0.y, ey );

          for ( ; c0.y <= ey ; ++( c0.y ) )
            pixelOp( draw, c0 );
        }

      } // namespace line

      /*****************************************************
         GBox : 矩形描画用関数
      *****************************************************/
      namespace box
      {
        using line::DrawH;
        using line::DrawV;

        /*
          Stroke : c0, c1 を頂点とする矩形を描画する
        */
        void Stroke( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c0, Coord<int> c1 )
        {
          Coord<int> drawSize = draw.size(); // 描画領域のサイズ

          // 二点の位置を調整(c0を左上、c1を右下に)
          if ( c0.x > c1.x ) std::swap( c0.x, c1.x );
          if ( c0.y > c1.y ) std::swap( c0.y, c1.y );

          // 領域範囲外なら何もせずに終了
          if ( c0.x >= drawSize.x ) return;
          if ( c0.y >= drawSize.y ) return;
          if ( c1.x < 0 ) return;
          if ( c1.y < 0 ) return;

          // クリッピング(4本の線分に対して描画するかも併せて決める)
          bool l = true, u = true, r = true, d = true;
          if ( c0.x < 0 ) {
            l = false;
            c0.x = 0;
          }
          if ( c1.x >= drawSize.x ) {
            r = false;
            c1.x = drawSize.x - 1;
          }
          if ( c0.y < 0 ) {
            u = false;
            c0.y = 0;
          }
          if ( c1.y >= drawSize.y ) {
            d = false;
            c1.y = drawSize.y - 1;
          }

          if ( c0.y == c1.y ) {
            DrawH( draw, pixelOp, c0, c1.x );
          } else if ( c0.x == c1.x ) {
            DrawV( draw, pixelOp, c0, c1.y );
          } else {
            if ( u ) DrawH( draw, pixelOp, c0, c1.x );
            if ( d ) DrawH( draw, pixelOp, Coord< int >( c0.x, c1.y ), c1.x );
            int uy = ( u ) ? c0.y + 1 : c0.y;
            int dy = ( d ) ? c1.y - 1 : c1.y;
            if ( r ) DrawV( draw, pixelOp, Coord< int >( c1.x, uy ), dy );
            if ( l ) DrawV( draw, pixelOp, Coord< int >( c0.x, uy ), dy );
          }
        }

        /*
          Fill : c0, c1 を頂点とする矩形の塗りつぶし
        */
        void Fill( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c0, Coord<int> c1 )
        {
          Coord<int> drawSize = draw.size(); // 描画領域のサイズ

          // 二点の位置を調整(c0を左上、c1を右下に)
          if ( c0.x > c1.x ) std::swap( c0.x, c1.x );
          if ( c0.y > c1.y ) std::swap( c0.y, c1.y );

          // 領域範囲外なら何もせずに終了
          if ( c0.x >= drawSize.x ) return;
          if ( c0.y >= drawSize.y ) return;
          if ( c1.x < 0 ) return;
          if ( c1.y < 0 ) return;

          // クリッピング
          if ( c0.x < 0 ) c0.x = 0;
          if ( c1.x >= drawSize.x ) c1.x = drawSize.x - 1;
          if ( c0.y < 0 ) c0.y = 0;
          if ( c1.y >= drawSize.y ) c1.y = drawSize.y - 1;

          for ( int y = c0.y ; y <= c1.y ; ++y )
            DrawH( draw, pixelOp, Coord<int>( c0.x, y ), c1.x );
        }

      } // namespace box

      /**************************************************************
        GEllipse : 楕円描画関数
      **************************************************************/

      namespace ellipse
      {
        /*
          GetParamByRectangle : 二点 c1, c2 を対角線とする長方形に接する楕円のパラメータを求める
          (楕円のパラメータは無限に存在するため、楕円の中心から対角線の端点への相対座標(u,v)に対し
          k = Coord( v, u ), r = uv( u + v ) を代表値とする)
        */
        bool GetParamByRectangle( Coord< int > c1, Coord< int > c2, Coord< int >* o, int* r, Coord<int>* k )
        {
          if ( c1.x == c2.x || c1.y == c2.y ) return( false );

          // c1, c2 の中点 (楕円の中心)
          Coord< double > ori( static_cast< double >( c1.x + c2.x ) / 2,
                               static_cast< double >( c1.y + c2.y ) / 2 );
          // 楕円の中心から端点までの相対座標 ( u, v )
          Coord< double > p( static_cast< double >( std::max( c1.x, c2.x ) ) - ori.x,
                           static_cast< double >( std::max( c1.y, c2.y ) ) - ori.y );

          if ( o != 0 ) {
            o->x = round( ori.x );
            o->y = round( ori.y );
          }

          if ( r != 0 )
            *r = round( sqrt( ( p.x + p.y ) * p.x * p.y ) );

          if ( k != 0 ) {
            k->x = round( p.y );
            k->y = round( p.x );
          }

          return( true );
        }

        /*
          StrokeEllipse : 楕円描画関数

          a( x - o.x )^2 + b( y - o.y )^2 = r^2を描画する
        */
        void Stroke( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > k )
        {
          // 半径 r と係数 k はどちらも正数が前提条件
          if ( r <= 0 ) return;
          if ( k.x <= 0 || k.y <= 0 ) return;

          Coord< int > c0( round( static_cast<double>( r ) / sqrt( k.x ) ), 0 ); // 描画開始座標
          Coord< int > drawSize = draw.size(); // 描画領域の大きさ

          double d = sqrt( k.x ) * static_cast<double>( r );
          int f = round( -2.0 * d ) +     k.x + 2 * k.y;
          int h = round( -4.0 * d ) + 2 * k.x     + k.y;

          while ( c0.x >= 0 ) {
            //if ( CheckYRange( o.y, c0.y, drawSize.y ) ) {
            if ( o.y + c0.y >= 0 && o.y - c0.y < drawSize.y ) {
              pixelOp( draw, Coord<int>( o.x + c0.x, o.y + c0.y ) );
              if ( c0.x != 0 )
                pixelOp( draw, Coord<int>( o.x - c0.x, o.y + c0.y ) );
              if ( c0.y != 0 ) {
                pixelOp( draw, Coord<int>( o.x + c0.x, o.y - c0.y ) );
                if ( c0.x != 0 )
                  pixelOp( draw, Coord<int>( o.x - c0.x, o.y - c0.y ) );
              }
            }
            if ( f >= 0 ) {
              --( c0.x );
              f -= 4 * k.x * c0.x;
              h -= 4 * k.x * c0.x - 2 * k.x;
            }
            if ( h < 0 ) {
              ++( c0.y );
              f += 4 * k.y * c0.y + 2 * k.y;
              h += 4 * k.y * c0.y;
            }
          }
        }

        /*
          Fill : 楕円の塗りつぶし関数

          a( x - o.x )^2 + b( y - o.y )^2 = r^2 の内部を塗りつぶす
        */
        void Fill( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> o, int r, Coord<int> k )
        {
          // 半径 r と係数 k はどちらも正数が前提条件
          if ( r <= 0 ) return;
          if ( k.x <= 0 || k.y <= 0 ) return;

          Coord<int> c0( round( static_cast<double>( r ) / sqrt( k.x ) ), 0 ); // 描画開始座標
          Coord<int> drawSize = draw.size(); // 描画領域の大きさ

          double d = sqrt( k.x ) * static_cast<double>( r );
          int f = round( -2.0 * d ) +     k.x + 2 * k.y;
          int h = round( -4.0 * d ) + 2 * k.x     + k.y;
          bool drawn = false; // 対象の Y 座標がすでに描画済みか

          while ( c0.x >= 0 ) {
            //if ( CheckYRange( o.y, c0.y, drawSize.y ) && ( ! drawn ) ) {
            if ( o.y + c0.y >= 0 && o.y - c0.y < drawSize.y && ( ! drawn ) ) {
              int x = o.x - c0.x;
              int maxX = o.x + c0.x;
              if ( x < drawSize.x && maxX >= 0 ) {
                if ( x < 0 ) x = 0;
                if ( maxX >= drawSize.x ) maxX = drawSize.x - 1;
                for ( ; x <= maxX ; ++x ) {
                  if ( o.y - c0.y >= 0 ) pixelOp( draw, Coord<int>( x, o.y - c0.y ) );
                  if ( o.y + c0.y >= 0 && o.y + c0.y < drawSize.y && c0.y != 0 )
                    pixelOp( draw, Coord<int>( x, o.y + c0.y ) );
                }
              }
            }
            if ( f >= 0 ) {
              --( c0.x );
              f -= 4 * k.x * c0.x;
              h -= 4 * k.x * c0.x - 2 * k.x;
              drawn = true;
            }
            if ( h < 0 ) {
              ++( c0.y );
              f += 4 * k.y * c0.y + 2 * k.y;
              h += 4 * k.y * c0.y;
              drawn = false;
            }
          }
        }

        typedef void (*DrawEllipseFuncType)( DrawingArea_IF&, GPixelOp&, Coord<int>, int, Coord<int> );

        /*
          Draw_2p : 二点 c1, c2 を対角線とする長方形に接する楕円を描画する(Stroke/Fill用)
        */
        void Draw_2p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2,
                      DrawEllipseFuncType f )
        {
          Coord<int> o; // 楕円の中心
          int r;        // 楕円の半径
          Coord<int> k; // 楕円の係数

          if ( ! GetParamByRectangle( c1, c2, &o, &r, &k ) ) return;

          (*f)( draw, pixelOp, o, r, k );
        }

        /*
          Stroke_2p : 二点 c1, c2 を対角線とする長方形に接する楕円を描画する
        */
        void Stroke_2p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2 )
        { Draw_2p( draw, pixelOp, c1, c2, Stroke ); }

        /*
          Fill_2p : 二点 c1, c2 を対角線とする長方形に接する楕円を塗りつぶし描画する
        */
        void Fill_2p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2 )
        { Draw_2p( draw, pixelOp, c1, c2, Fill ); }

      } // namespace ellipse

      /**************************************************************
           GCircle : 真円描画用関数
      **************************************************************/

      namespace circle
      {
        /*
          CalcMidPoint : 三点 c1,c2,c3 間の距離の中で最大のものを抽出してその中点を返す
          (三点が同一線分上にあった場合の中心と半径を求める)
          d には距離の半分を代入する
        */
        Coord<int> CalcMidPoint( Coord<int> c1, Coord<int> c2, Coord<int> c3, int* d )
        {
          double d12 = coord_lib::Distance< double >( c1, c2 );
          double d23 = coord_lib::Distance< double >( c2, c3 );
          double d31 = coord_lib::Distance< double >( c3, c1 );

          // "+ 1"は四捨五入のため
          if ( d12 > d23 && d12 > d31 ) {
            if ( d != 0 ) *d = static_cast<int>( ( d12 + 1 ) / 2 );
            return( Coord<int>( ( c1.x + c2.x + 1 ) / 2, ( c1.y + c2.y + 1 ) / 2 ) );
          } else if ( d23 > d31 && d23 > d12 ) {
            if ( d != 0 ) *d = static_cast<int>( ( d23 + 1 ) / 2 );
            return( Coord<int>( ( c2.x + c3.x + 1 ) / 2, ( c2.y + c3.y + 1 ) / 2 ) );
          } else {
            if ( d != 0 ) *d = static_cast<int>( ( d31 + 1 ) / 2 );
            return( Coord<int>( ( c3.x + c1.x + 1 ) / 2, ( c3.y + c1.y + 1 ) / 2 ) );
          }
        }

        /*
          CalcPerpendicularBisector : 二点 c1, c2 を結ぶ線分の垂直二等分線に対する交点 m と傾き a を求める
        */
        bool CalcPerpendicularBisector( const Coord<int>& c1, const Coord<int>& c2, Coord<double>* m, double* a )
        {
          // 水平線分なら false を返す
          if ( c1.y == c2.y ) return( false );

          if ( m != 0 )
            *m = Coord<double>( static_cast<double>( c1.x + c2.x ) / 2, static_cast<double>( c1.y + c2.y ) / 2 );
          if ( a != 0 )
            *a = static_cast<double>( c1.x - c2.x ) / static_cast<double>( c2.y - c1.y );

          return( true );
        }

        /*
          CalcOrigin : 三角形 c1-c2-c3 の外心(戻り値)と半径 r を求める
        */
        Coord<int> CalcOrigin( Coord<int> c1, Coord<int> c2, Coord<int> c3, int* r )
        {
          Coord<double> m1, m2; // 二点間の中点
          double a1, a2;        // 二点間の傾き

          if ( c1.y == c2.y && c1.y == c3.y )
            return( CalcMidPoint( c1, c2, c3, r ) ); // 水平線上にある場合は求められない

          if ( c1.y == c2.y ) {
            CalcPerpendicularBisector( c2, c3, &m1, &a1 );
            CalcPerpendicularBisector( c3, c1, &m2, &a2 );
          } else if ( c2.y == c3.y ) {
            CalcPerpendicularBisector( c1, c2, &m1, &a1 );
            CalcPerpendicularBisector( c3, c1, &m2, &a2 );
          } else {
            CalcPerpendicularBisector( c1, c2, &m1, &a1 );
            CalcPerpendicularBisector( c2, c3, &m2, &a2 );
          }

          if ( a1 == a2 )
            return( CalcMidPoint( c1, c2, c3, r ) ); // 直線上に並んでいる場合は求められない

          double x = ( a1 * m1.x - a2 * m2.x - m1.y + m2.y );
          double y = ( a1 * a2 * ( m1.x - m2.x ) - a2 * m1.y + a1 * m2.y );
          if ( r != 0 )
            *r = static_cast<int>(
                                  sqrt( pow( ( ( a1 - a2 ) * static_cast<double>( c1.x ) - x ) / ( a1 - a2 ), 2 ) +
                                        pow( ( ( a1 - a2 ) * static_cast<double>( c1.y ) - y ) / ( a1 - a2 ), 2 ) ) + 0.5 );

          x /= a1 - a2;
          y /= a1 - a2;

          return( Coord<int>( round( x ), round( y ) ) );
        }

        typedef void (*DrawCircleFuncType)( DrawingArea_IF&, GPixelOp&, Coord<int>, int );

        /*
          DrawCircle_3p : 円周上の三点 c1, c2, c3 から中心と半径を決定し、関数 f で描画する
        */
        void Draw_3p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2, Coord<int> c3,
                      DrawCircleFuncType f )
        {
          int r; // 半径
          Coord<int> o = CalcOrigin( c1, c2, c3, &r ); // 円の中心

          (*f)( draw, pixelOp, o, r );
        }

        /*
          StrokeCircle_3p : 円周上の三点 c1, c2, c3 から中心と半径を決定して描画する
          FillCircle_3p : 円周上の三点 c1, c2, c3 から中心と半径を決定して塗りつぶす
        */
        void Stroke_3p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2, Coord<int> c3 )
        { Draw_3p( draw, pixelOp, c1, c2, c3, Stroke ); }
        void Fill_3p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2, Coord<int> c3 )
        { Draw_3p( draw, pixelOp, c1, c2, c3, Fill ); }

        /*
          DrawCircle_3p : 二点 c1, c2 を直径の両端として円を決定して関数 f で描画する
        */
        void Draw_2p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2,
                      DrawCircleFuncType f )
        {
          Coord<int> o( ( c1.x + c2.x + 1 ) / 2, ( c1.y + c2.y + 1 ) / 2 ); // "+ 1"は四捨五入のため
          int r = round( coord_lib::Distance< double >( o, c1 ) );

          (*f)( draw, pixelOp, o, r );
        }

        /*
          StrokeCircle_2p : 二点 c1, c2 を直径の両端として円を決定して描画する
          FillCircle_2p : 二点 c1. c2 を直径の両端として円を決定して塗りつぶす
        */
        void Stroke_2p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2 )
        { Draw_2p( draw, pixelOp, c1, c2, Stroke ); }
        void Fill_2p( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2 )
        { Draw_2p( draw, pixelOp, c1, c2, Fill ); }

      } // namespace circle

      /**************************************************************
          GArc : 円弧描画用関数
      **************************************************************/

      namespace arc
      {
        /*
          GetAngleFromCoord : 原点 o から c へのベクトルに対し、X 軸からの角度 ( 0 ～ 2π ) を求める
        */
        double GetAngleFromCoord( Coord< int > c, Coord< int > o )
        {
          c -= o; // 中心からの相対座標

          // 傾きから角度を求める ( atan は [ -π/2, π/2 ] の範囲の値を返す )
          double angle = ( c.x != 0 ) ?
            std::atan( static_cast< double >( c.y ) / static_cast< double >( c.x ) ) :
            ( c.y >= 0 ) ? math_lib::Pi< double >() / 2.0 : -math_lib::Pi< double >() / 2.0;

          // x < 0 の場合は [ π/2, 3π/2 ] の範囲に変換する
          if ( c.x < 0 ) angle += math_lib::Pi< double >();
          // 角度が負 ( [ -π/2, 0 ) ) の場合は [ 3π/2, 2π ) の範囲に変換する
          if ( angle < 0 ) angle += math_lib::Pi< double >() * 2;

          return( angle );
        }

        /*
          GetCoordFromAngle : 原点を o とする半径 r の円周上にある、
          X 軸からの角度が t ( 0 ～ 2π ) にある点の座標を求める
        */
        Coord< int > GetCoordFromAngle( Coord< int > o, int r, Coord< int > k, double t )
        {
          const double t2 = math_lib::Pi< double >() / 2.0;       // 直角(PI/2)
          const double t3 = math_lib::Pi< double >();             // PI
          const double t4 = math_lib::Pi< double >() * 3.0 / 4.0; // 3*PI/4

          assert( r > 0 && k.x > 0 && k.y > 0 );

          t = fmod( t, math_lib::Pi< double >() * 2.0 );

          if ( t == t2 ) // Y軸上正方向
            return( Coord< int >( o.x, round( static_cast< double >( r ) / std::sqrt( static_cast< double >( k.y ) ) ) + o.y ) );
          if ( t == t4 ) // Y軸上負方向
            return( Coord< int >( o.x, round( static_cast< double >( r ) / std::sqrt( static_cast< double >( k.y ) ) ) * -1 + o.y ) );

          // t から求められる傾きの二乗
          double powM = std::pow( std::tan( t ), 2 );
          Coord< int > c(
                         round( static_cast< double >( r ) /
                                         sqrt( static_cast< double >( k.x ) + static_cast< double >( k.y ) * powM ) ),
                         round( static_cast< double >( r ) * sqrt( powM ) /
                                         sqrt( static_cast< double >( k.x ) + static_cast< double >( k.y ) * powM ) )
                         );

          if ( t >= t2 && t < t4 ) c.x *= -1;
          if ( t >= t3 ) c.y *= -1;

          c += o;

          return( c );
        }

        /*
          GClipByAngle : 指定した角度内だけを点描画する

          扇形の塗りつぶしを行う場合 GCircle::FillCircle と組み合わせることで実現することができる。
          例えば、以下のコードは中心を ( 100, 100 ) とした中心角 30 度の扇形を draw 上で描画する。

          GPSet gpset( RGB( RGB::MAX, RGB::MAX, RGB::MAX ) );
          GClipByAngle clipByAngle( gpset, Coord< int >( 100, 100 ), 0, 2 * MathLib::Pi< double >() / 12.0 );
          GCircle::Fill( draw, clipByAngle, Coord< int >( 100, 100 ), 50 );
        */
        class GClipByAngle : public GPixelOp
        {
          GPixelOp& pixelOp_; // ピクセル操作用オブジェクトへの参照

          Coord< int > origin_; // 中心点
          double sArg_; // 開始角度
          double eArg_; // 終了角度

        public:

          /* 公開メンバ変数 */
          bool isOut; // 範囲外を描画するなら true

          /* コンストラクタ */

          /*
            中心点と開始・終了角度を指定して構築

            pixelOp : 点描画に使う関数オブジェクト
            o : 中心点の座標
            sa,ea : 開始・終了角度
            isOut : 範囲外を描画するならtrue
          */
          GClipByAngle( GPixelOp& pixelOp, Coord< int > o, double sa, double ea, bool isOut );

          /*
            中心点と、範囲を決めるその他の二点を指定して構築

            pixelOp : 点描画に使う関数オブジェクト
            o : 中心点の座標
            c1,c2 : 描画範囲の座標
            isOut : 範囲外を描画するならtrue
          */
          GClipByAngle( GPixelOp& pixelOp, Coord< int > o, Coord< int > c1, Coord< int > c2, bool isOut );

          /*
            指定した角度内(または角度外)だけを点描画する

            draw : ピクセル操作対象の DrawingArea_IF
            c : ピクセル操作するピクセルの座標
          */
          virtual bool operator()( DrawingArea_IF& draw, Coord< int > c );
        };

        /*
          GClipByAngle::GClipByAngle コンストラクタ
        */
        GClipByAngle::GClipByAngle( GPixelOp& pixelOp, Coord< int > o, double sa, double ea, bool isOut )
          : pixelOp_( pixelOp ), origin_( o ), isOut( isOut )
        {
          const double PI2 = math_lib::Pi< double >() * 2.0;

          sArg_ = fmod( sa, PI2 );
          if ( sArg_ < 0 ) sArg_ += PI2;
          eArg_ = fmod( ea, PI2 );
          if ( eArg_ < 0 ) eArg_ += PI2;

          //std::cout << sArg_ << "-" << eArg_ << std::endl;
        }

        /*
          GClipByAngle::GClipByAngle コンストラクタ
        */
        GClipByAngle::GClipByAngle( GPixelOp& pixelOp, Coord< int > o, Coord< int > c1, Coord< int > c2, bool isOut )
          : pixelOp_( pixelOp ), origin_( o ), isOut( isOut )
        {
          double a1 = GetAngleFromCoord( c1, origin_ );
          double a2 = GetAngleFromCoord( c2, origin_ );

          //if ( c1.x > c2.x ) {
          //  sArg_ = a2;
          //  eArg_ = a1;
          //} else {
          sArg_ = a1;
          eArg_ = a2;
          //}
        }

        /*
          GClipByAngle::operator() : 描画処理
        */
        bool GClipByAngle::operator()( DrawingArea_IF& draw, Coord< int > c )
        {
          double angle = GetAngleFromCoord( c, origin_ );

          if ( sArg_ > eArg_ ) {
            if ( ( (   isOut ) && ( angle <  sArg_ && angle >  eArg_ ) ) ||
                 ( ( ! isOut ) && ( angle >= sArg_ || angle <= eArg_ ) ) )
              return( pixelOp_( draw, c ) );
          } else {
            if ( ( (   isOut ) && ( angle <  sArg_ || angle >  eArg_ ) ) ||
                 ( ( ! isOut ) && ( angle >= sArg_ && angle <= eArg_ ) ) )
              return( pixelOp_( draw, c ) );
          }

          return( false );
        }

        /*
          Stroke / Fill : 円弧描画と塗りつぶし

          draw : 描画領域
          pixelOp : 描画用関数オブジェクト
          o : 円弧の中心
          r : 円弧の半径
          sa,ea : 円弧の描画開始角度・終了角度
          isOut : 開始から終了までの範囲外を描画するなら true
        */
        void Stroke( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > o, int r, double sa, double ea, bool isOut )
        {
          GClipByAngle gcba( pixelOp, o, sa, ea, isOut );
          circle::Stroke( draw, gcba, o, r );
        }
        void Fill( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > o, int r, double sa, double ea, bool isOut )
        {
          GClipByAngle gcba( pixelOp, o, sa, ea, isOut );
          circle::Fill( draw, gcba, o, r );
        }

        /*
          Stroke / Fill : 円弧描画と塗りつぶし

          draw : 描画領域
          pixelOp : 描画用関数オブジェクト
          o : 円弧の中心
          r : 円弧の半径
          c1,c2 : 円弧の描画範囲を決める点の座標
          isOut : 開始から終了までの範囲外を描画するなら true
        */
        void Stroke( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > c1, Coord< int > c2, bool isOut )
        {
          GClipByAngle gcba( pixelOp, o, c1, c2, isOut );
          circle::Stroke( draw, gcba, o, r );
        }
        void Fill( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > c1, Coord< int > c2, bool isOut )
        {
          GClipByAngle gcba( pixelOp, o, c1, c2, isOut );
          circle::Fill( draw, gcba, o, r );
        }

      } // namespace arc

    } // namespace shape

    namespace draw
    {
      /**************************************************************
         GPaint : ペイント用関数
      **************************************************************/

      namespace paint
      {
        typedef std::pair<int,int> LineX;
        typedef std::map< int, vector<LineX> > MapDrawn; // 描画済領域を保持するコンテナの型

        /* バッファのデータ構成 */
        struct Buffer {
          int lx; // 領域左端のX座標
          int rx; // 領域右端のX座標
          int y;  // 領域のY座標
          int oy; // 親ラインのY座標
        };

        typedef std::queue<Buffer> QueueBuff; // バッファ用 Queue の型

        /*
          GCheck : 指定した RGB コードと一致するか比較する

          GPaint::Paint にこのインスタンスを適用すると、領域色の塗りつぶしができる。
          GPaint::AreaPaint で利用。
        */
        class GCheck : public GPixelOp
        {
        public:

          RGB src; // 指定した RGB コード

          /* コンストラクタ */

          // 比較対象となる RGB コードを指定して構築
          explicit GCheck( RGB rgb )
            : src( rgb ) {}

          // RGB コード比較処理
          virtual bool operator()( DrawingArea_IF& draw, Coord< int > c )
          {
            RGB dest;

            if ( draw.point( c, &dest ) )
              return( src == dest );
            else
              return( false );
          }
        };

        /*
          GNotCheck : 指定した RGB コードと異なるか比較する

          GPaint::Paint にこのインスタンスを適用すると、境界色の塗りつぶしができる。
          GPaint::BorderPaint で利用。
        */
        class GNotCheck : public GCheck
        {
        public:

          /* コンストラクタ */

          // 比較対象として指定する RGB コードを指定して構築
          explicit GNotCheck( RGB rgb )
            : GCheck( rgb ) {}

          // RGB コード比較処理
          virtual bool operator()( DrawingArea_IF& draw, Coord< int > c )
          {
            RGB dest;

            if ( draw.point( c, &dest ) )
              return( src != dest );
            else
              return( false );
          }
        };

        /*
          Drawn : mapDrawn をチェックして、座標 ( x, y ) がすでに描画済みなら true を返す
        */
        bool Drawn( const MapDrawn& mapDrawn, int x, int y )
        {
          MapDrawn::const_iterator mapCit = mapDrawn.find( y );
          if ( mapCit == mapDrawn.end() )
            return( false );

          vector<LineX>::const_iterator vecCit = ( mapCit->second ).begin();
          while ( vecCit != ( mapCit->second ).end() ) {
            if ( x >= vecCit->first && x <= vecCit->second ) return( true );
            ++vecCit;
          }

          return( false );
        }

        /*
          ScanLine : 線分 ( lx, y )-( rx, y ) からシードを探索してバッファ queueBuff に登録する
          探索には check 関数オブジェクトを使用する
          oy は y 方向にたどった時の親ラインの Y 座標を表す。
        */
        void ScanLine( DrawingArea_IF& draw, int lx, int rx, int y, int oy, GCheck& check, QueueBuff* queueBuff )
        {
          Buffer buff;

          while ( lx <= rx ) {
            /* 非領域色を飛ばす */
            for ( ; lx <= rx ; ++lx ) {
              if ( check( draw, Coord<int>( lx, y ) ) )
                break;
            }
            if ( lx > rx ) break;
            buff.lx = lx;

            /* 領域色を飛ばす */
            for ( ; lx <= rx ; ++lx ) {
              if ( ! check( draw, Coord<int>( lx, y ) ) )
                break;
            }
            buff.rx = lx - 1;
            buff.y = y;
            buff.oy = oy;
            queueBuff->push( buff );
          }
        }

        /*
          Paint : c を起点にシードフィルアルゴリズムを使って領域を塗りつぶす
        */
        void Paint( DrawingArea_IF& draw, GPixelOp& pixelOp, GCheck& check, Coord<int> c )
        {
          QueueBuff queueBuff; // シード候補を登録するqueue
          MapDrawn mapDrawn;   // 描画済みの領域

          Buffer buff;
          Coord<int> drawSize = draw.size();

          if ( c.x < 0 || c.x >= drawSize.x || c.y < 0 || c.y >= drawSize.y ) return;

          buff.lx = buff.rx = c.x;
          buff.y = buff.oy = c.y;
          queueBuff.push( buff );

          do {
            buff = queueBuff.front();
            queueBuff.pop();

            /* すでに描画済みならスキップ */
            if ( Drawn( mapDrawn, buff.lx, buff.y ) ) continue;

            int lxsav = buff.lx - 1;
            int rxsav = buff.rx + 1;

            /* 右方向の境界を探す */
            while ( buff.rx < drawSize.x - 1 ) {
              if ( ! check( draw, Coord<int>( buff.rx + 1, buff.y ) ) )
                break;
              ++( buff.rx );
            }
            /* 左方向の境界を探す */
            while ( buff.lx > 0 ) {
              if ( ! check( draw, Coord<int>( buff.lx - 1, buff.y ) ) )
                break;
              --( buff.lx );
            }
            /* lx-rxの線分を描画 */
            shape::line::DrawH( draw, pixelOp, Coord<int>( buff.lx, buff.y ), buff.rx );
            //for ( int x = buff.lx; x <= buff.rx; x++ )
            //  pixelOp( draw, Coord<int>( x, buff.y ) );

            ( mapDrawn[buff.y] ).push_back( LineX( buff.lx, buff.rx ) );

            /* 真上のスキャンラインを走査する */
            int uy = buff.y - 1;
            if ( uy >= 0 ) {
              if ( uy == buff.oy ) {
                if ( lxsav >= 0 ) ScanLine( draw, buff.lx, lxsav, uy, uy + 1, check, &queueBuff );
                if ( rxsav < drawSize.x ) ScanLine( draw, rxsav, buff.rx, uy, uy + 1, check, &queueBuff );
              } else {
                ScanLine( draw, buff.lx, buff.rx, uy, uy + 1, check, &queueBuff );
              }
            }

            /* 真下のスキャンラインを走査する */
            int dy = buff.y + 1;
            if ( dy < drawSize.y ) {
              if ( dy == buff.oy ) {
                if ( lxsav >= 0 ) ScanLine( draw, buff.lx, lxsav, dy, dy - 1, check, &queueBuff );
                if ( rxsav < drawSize.x ) ScanLine( draw, rxsav, buff.rx, dy, dy - 1, check, &queueBuff );
              } else {
                ScanLine( draw, buff.lx, buff.rx, dy, dy - 1, check, &queueBuff );
              }
            }

          } while ( queueBuff.size() > 0 );
        }

        /*
          AreaPaint : 塗りつぶし開始座標上の色を領域色として塗りつぶす
        */
        void AreaPaint( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > c )
        {
          RGB rgb;

          if ( ! draw.point( c, &rgb ) ) return;
          GCheck check( rgb );
          Paint( draw, pixelOp, check, c );
        }

        /*
          BorderPaint : 境界色による塗りつぶし処理
        */
        void BorderPaint( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord< int > c, RGB rgb )
        {
          GNotCheck notCheck( rgb );
          Paint( draw, pixelOp, notCheck, c );
        }

      } // namespace paint

    } // namespace draw

    namespace shape
    {
      /**************************************************************
         GPolygon : ポリゴン描画関数
      **************************************************************/

      namespace polygon
      {
        // vector< Coord<int> > 型
        typedef vector< Coord<int> > VecCoord;

        // vector< Coord<int> > 用イテレータの型
        typedef VecCoord::iterator itCoord;
        typedef VecCoord::const_iterator citCoord;

        /* 辺の定義 */
        struct Edge {
          double x;        // 交点のX座標(初期値は始点のX座標)
          int    y0;       // 始点のY座標
          int    y1;       // 終点のY座標(必ずy0<y1)
          bool   downward; // 始点から終点の方向(下向きのときtrue)
          double a;        // 傾き dX/dY
        };

        // 辺定義ベクタの型
        typedef vector<Edge> VecEdge;

        // vector<Edge> 用イテレータの型
        typedef VecEdge::iterator itEdge;
        typedef VecEdge::const_iterator citEdge;

        // X 座標によるソート用比較関数オブジェクト
        class CmpByX {
        public:
          int operator()( citEdge e0, citEdge e1 ) const
          { return( e0->x < e1->x ); }
        };

        /*
          LineClip : 線分 ip1-ip2 を Y 座標 drawMin_y と drawMax_y の間でクリッピングする
          ip2 は必ず描画領域の外側になることを想定している
        */
        Coord<int> LineClip( Coord<int> ip1, Coord<int> ip2, int drawMin_y, int drawMax_y )
        {
          // 描画領域と端点との Y 方向の距離を求める
          int dy = ( ip2.y < drawMin_y ) ? drawMin_y - ip1.y : drawMax_y - ip1.y;
          // X 方向の距離に変換した上で、描画領域の端と線分の交点の X 座標を求める
          int x = ( ip2.x - ip1.x ) * dy / ( ip2.y - ip1.y ) + ip1.x;
          // 描画領域の端と線分の交点の Y 座標
          int y = ( ip2.y < drawMin_y ) ? drawMin_y : drawMax_y;

          return( Coord<int>( x, y ) );
        }

        /*
          Clipping : vertex を Y 座標 drawMin_y と drawMax_y の間でクリッピングして
          clippedVertex に再登録する
        */
        void Clipping( const VecCoord& vertex, int drawMin_y, int drawMax_y, VecCoord* clippedVertex )
        {
          clippedVertex->clear();
          unsigned int vertexCnt = vertex.size(); // 頂点の数

          for ( unsigned int i = 1 ; i <= vertexCnt ; ++i ) {
            const Coord<int>& c0 = vertex[i-1];           // 始点
            const Coord<int>& c1 = vertex[i % vertexCnt]; // 終点
            if ( c0 == c1 ) continue;
            // 始点がエリア外
            if ( ( c0.y < drawMin_y ) || ( c0.y > drawMax_y ) ) {
              // 終点はエリア内
              if ( ( c1.y >= drawMin_y ) && ( c1.y <= drawMax_y ) ) {
                clippedVertex->push_back( LineClip( c1, c0, drawMin_y, drawMax_y ) );
                // 終点もエリア外(クリッピングエリアの上下境界をまたぐ)
              } else if ( ( ( c0.y < drawMin_y ) && ( c1.y > drawMax_y ) ) ||
                          ( ( c1.y < drawMin_y ) && ( c0.y > drawMax_y ) ) ) {
                clippedVertex->push_back( LineClip( c1, c0, drawMin_y, drawMax_y ) );
                clippedVertex->push_back( LineClip( c0, c1, drawMin_y, drawMax_y ) );
              }
              // 始点がエリア内
            } else {
              clippedVertex->push_back( c0 );
              // 終点がエリア外ならクリッピングして頂点を追加
              if ( ( c1.y < drawMin_y ) || ( c1.y > drawMax_y ) )
                clippedVertex->push_back( LineClip( c0, c1, drawMin_y, drawMax_y ) );
            }
          }
        }

        /*
          GenerateEdgeList : 頂点のリスト vertex から辺リスト edgeList を生成する
        */
        void GenerateEdgeList( const VecCoord& vertex, VecEdge* edgeList )
        {
          edgeList->clear();
          if ( vertex.empty() ) return;
          unsigned int vertexCnt = vertex.size(); // 頂点の個数

          for ( unsigned int i = 0 ; i < vertexCnt ; ++i ) {
            const Coord<int>& c0 = vertex[i]; // 始点
            Coord<int> c1 = vertex[( i + 1 ) % vertexCnt]; // 終点
            if ( c0.y == c1.y ) continue;

            const Coord<int>& c2 = vertex[( i + 2 ) % vertexCnt]; // 終点の次の点

            Edge edge; // 辺データ用バッファ

            edge.downward = c0.y < c1.y;
            edge.x = ( edge.downward ) ? c0.x : c1.x;
            edge.a = static_cast<double>( c1.x - c0.x ) / static_cast<double>( c1.y - c0.y );

            // 終点が左右向きの場合、終点のY座標をひとつずらす
            if ( ( c1.y - c0.y ) * ( c1.y - c2.y ) < 0 ) {
              if ( edge.downward ) {
                --( c1.y ); // 終点が辺の下側なら上側にずらす
              } else {
                ++( c1.y ); // 終点が辺の上側なら下側にずらす
                edge.x += edge.a; // X座標の初期値を補正する
              }
            }

            edge.y0 = ( edge.downward ) ? c0.y : c1.y;
            edge.y1 = ( edge.downward ) ? c1.y : c0.y;
            edgeList->push_back( edge );
          }
        }

        /*
          InsertionSort : first と last の間のデータを挿入ソートで並び替える
          比較関数に less を用いる
        */
        template<class Ran, class Less> void InsertionSort( Ran first, Ran last, Less less )
        {
          typedef typename Ran::value_type value_type;

          for ( Ran e = first + 1 ; e < last ; ++e ) {
            value_type v = *e;     // 未ソート列の先頭の要素を取得
            bool inserted = false; // v が挿入されたか
            for ( Ran s = e - 1 ; s >= first ; --s ) {
              if ( less( v, *s ) ) {
                *(s + 1) = *s;
              } else {
                *(s + 1) = v;
                inserted = true;
                break;
              }
            }
            // v が未挿入の場合は先頭に要素を代入
            if ( ! inserted ) *first = v;
          }
        }
        template void InsertionSort( vector<itEdge>::iterator first, vector<itEdge>::iterator last, CmpByX cmpByX );
        template void InsertionSort( vector<int>::iterator first, vector<int>::iterator last, std::less<int> less );

        /*
          FillLine : ( x0, y ) - ( x1, y ) までの水平線分の描画
        */
        void FillLine( DrawingArea_IF& draw, GPixelOp& pixelOp, int x0, int x1, int y )
        {
          // X座標のクリッピング
          if ( x1 < 0 || x0 > draw.size().x - 1 ) return;
          if ( x0 < 0 ) x0 = 0;
          if ( x1 > draw.size().x - 1 ) x1 = draw.size().x - 1;

          line::DrawH( draw, pixelOp, Coord<int>( x0, y ), x1 );
        }

        /*
          Poly_CN_DrawLine : クリッピング・エリア外の交点をチェックしながら Y 座標 y のラインを描画する
          アクティブな辺リストの開始位置を s、終了の次の位置を e とする
          描画領域判定は Crossing Number (Even-Odd) Test を利用
        */
        void Poly_CN_DrawLine( DrawingArea_IF& draw, GPixelOp& pixelOp, vector<itEdge>::iterator s, vector<itEdge>::iterator e, int y )
        {
          for ( vector<itEdge>::iterator i = s + 1 ; i < e ; i += 2 ) {
            int x0 = round( ( *( i - 1 ) )->x ); // 左端の点の X 座標
            int x1 = round( ( *i )->x );         // 右端の点の X 座標

            // X座標の更新
            ( *( i - 1 ) )->x += ( *( i - 1 ) )->a;
            ( *i )->x += ( *i )->a;

            FillLine( draw, pixelOp, x0, x1, y );
          }
        }

        /*
          Poly_WN_DrawLine : クリッピング・エリア外の交点をチェックしながら Y 座標 y のラインを描画する
          アクティブな辺リストの開始位置を s、終了の次の位置を e とする
          描画領域判定は Winding Number を利用
        */
        void Poly_WN_DrawLine( DrawingArea_IF& draw, GPixelOp& pixelOp, vector<itEdge>::iterator s, vector<itEdge>::iterator e, int y )
        {
          for ( vector<itEdge>::iterator i = s ; i < e ; ) {
            vector<itEdge>::iterator e0 = i; // 描画領域の左端
            unsigned int wn = ( ( *e0 )->downward ) ? 1 : -1;
            for ( ++i ; i < e ; ++i ) {
              wn += ( ( *i )->downward ) ? 1 : -1;
              if ( wn == 0 ) break;
            }
            if ( i == e ) break; // 終端位置が見つからない場合はデータがおかしい

            vector<itEdge>::iterator e1 = i++; // 描画領域の右端
            int x0 = round( ( *e0 )->x ); // 左端の点の X 座標
            int x1 = round( ( *e1 )->x ); // 右端の点の X 座標

            // X座標の更新
            for ( vector<itEdge>::iterator j = e0 ; j <= e1 ; ++ j )
              ( *j )->x += ( *j )->a;

            FillLine( draw, pixelOp, x0, x1, y );
          }
        }

        /*
          領域判定方法の型
        */
        typedef void (*IOTestType)( DrawingArea_IF&, GPixelOp&, vector<itEdge>::iterator, vector<itEdge>::iterator, int );

        /*
          PolyFill_Group : 多角形描画 ( edgeList を「アクティブ」「非アクティブ」の群に分ける方法)
          func はスキャンラインの描画用関数
        */
        void PolyFill_Group( DrawingArea_IF& draw, GPixelOp& pixelOp, VecEdge& edgeList, IOTestType func )
        {
          // edgeList の反復子を保持するベクタ
          vector<VecEdge::iterator> edgeItList( edgeList.size() );
          for ( unsigned int i = 0 ; i < edgeList.size() ; ++i )
            edgeItList[i] = edgeList.begin() + i;

          // 「アクティブ」な辺の末尾の次の位置
          unsigned int activeIndexEnd = edgeItList.size();
          // 「アクティブ」な辺の開始位置(最初は「アクティブ」な辺はない)
          // activeIndexStart は「非アクティブ」な辺の末尾の次の位置も表す
          unsigned int activeIndexStart = activeIndexEnd;

          for ( int y = 0 ; y < draw.size().y ; ++y ) {
            // yと始点が一致した「未アクティブ」な辺を「アクティブ」にする
            for ( unsigned int i = activeIndexStart ; i > 0 ; --i ) {
              if ( edgeItList[i - 1]->y0 == y ) {
                itEdge edgeIt = edgeItList[i - 1];
                edgeItList[i - 1] = edgeItList[--activeIndexStart];
                edgeItList[activeIndexStart] = edgeIt;
              }
            }

            // yより点が小さくなった「アクティブ」な辺を削除する
            for ( unsigned int i = activeIndexEnd ; i > activeIndexStart ; --i )
              if ( edgeItList[i - 1]->y1 < y )
                edgeItList[i - 1] = edgeItList[--activeIndexEnd];

            // 交点のソート
            //InsertionSort( edgeItList.begin() + activeIndexStart, edgeItList.begin() + activeIndexEnd, CmpByX() );
            std::sort( edgeItList.begin() + activeIndexStart, edgeItList.begin() + activeIndexEnd, CmpByX() );

            // クリッピング・エリア外の交点をチェックしながらライン描画
            (*func)( draw, pixelOp, edgeItList.begin() + activeIndexStart, edgeItList.begin() + activeIndexEnd, y );
          }
        }

        /*
          PolyFill_Flag : 多角形描画(アクティブ・非アクティブの判断をフラグで行う)
        */
        void PolyFill_Flag( DrawingArea_IF& draw, GPixelOp& pixelOp, VecEdge& edgeList )
        {
          unsigned int edgeSize = edgeList.size(); // 辺リストのサイズ
          vector<int> vec_x( edgeSize ); // X 座標のリスト

          for ( int y = 0 ; y < draw.size().y ; ++y ) {
            vector< int >::iterator ep = vec_x.begin(); // 抽出した X 座標の末尾(開始位置で初期化)
            for ( unsigned int i = 0 ; i < edgeSize ; ++i ) {
              // アクティブな辺の X 座標を抽出
              if ( edgeList[i].y0 <= y && y <= edgeList[i].y1 ) {
                *ep++ = edgeList[i].x;
                edgeList[i].x += edgeList[i].a; // X座標の更新
              }
            }

            // 交点のソート
            //InsertionSort( vec_x.begin(), ep, std::less<int>() );
            std::sort( vec_x.begin(), ep );

            // クリッピング・エリア外の交点をチェックしながらライン描画
            for ( vector< int >::iterator sp = vec_x.begin() + 1 ;
                  sp < ep ; sp += 2 ) {
              int x0 = *( sp - 1 ); // 左端の点のX座標
              int x1 = *sp;         // 右端の点のX座標

              // X座標のクリッピング
              if ( x1 < 0 || x0 > draw.size().x - 1 ) continue;
              if ( x0 < 0 ) x0 = 0;
              if ( x1 > draw.size().x - 1 ) x1 = draw.size().x - 1;

              // 直線の描画
              for ( int x = x0 ; x <= x1 ; x++ )
                pixelOp( draw, Coord<int>( x, y ) );
            }
          }
        }

        /*
          GPolygon::PolyFill_MultiSet : 多角形描画(multisetを利用した場合)
        */
        void PolyFill_MultiSet( DrawingArea_IF& draw, GPixelOp& pixelOp, VecEdge& edgeList )
        {
          unsigned int edgeSize = edgeList.size(); // 辺リストのサイズ

          for ( int y = 0 ; y < draw.size().y ; ++y ) {
            std::multiset< int > multiset_x; // X 座標のリスト
            for ( unsigned int i = 0 ; i < edgeSize ; ++i ) {
              // アクティブな辺の X 座標を抽出
              if ( edgeList[i].y0 <= y && y <= edgeList[i].y1 ) {
                multiset_x.insert( edgeList[i].x );
                edgeList[i].x += edgeList[i].a; // X座標の更新
              }
            }

            // クリッピング・エリア外の交点をチェックしながらライン描画
            std::multiset< int >::iterator sp = multiset_x.begin();
            while ( sp != multiset_x.end() ) {
              int x0 = *sp++; // 左端の点のX座標
              if ( sp == multiset_x.end() ) break;
              int x1 = *sp++; // 右端の点のX座標

              // X座標のクリッピング
              if ( x1 < 0 || x0 > draw.size().x - 1 ) continue;
              if ( x0 < 0 ) x0 = 0;
              if ( x1 > draw.size().x - 1 ) x1 = draw.size().x - 1;

              // 直線の描画
              for ( int x = x0 ; x <= x1 ; x++ )
                pixelOp( draw, Coord<int>( x, y ) );
            }
          }
        }

        /*
          GPolygon::Fill : ポリゴン描画 メインルーチン
          vertex にある頂点を持つポリゴンを描画する
        */
        void Fill( DrawingArea_IF& draw, GPixelOp& pixelOp, const VecCoord& vertex, METHOD method )
        {
          VecCoord clippedVertex;
          vector<Edge> edgeList;

          Clipping( vertex, 0, draw.size().y - 1, &clippedVertex );
          GenerateEdgeList( clippedVertex, &edgeList );
          //PolyFill_MultiSet( draw, pixelOp, edgeList );
          //polyFill_Flag( edgeList, pixelOp );
          if ( method == CROSSING_NUMBER )
            PolyFill_Group( draw, pixelOp, edgeList, Poly_CN_DrawLine );
          else if ( method == WINDING_NUMBER )
            PolyFill_Group( draw, pixelOp, edgeList, Poly_WN_DrawLine );
        }

        /*
          TriFill_XDraw : 三角形上のスキャンラインを Y 座標 sy から ey まで描画する
          l, r は要素 first に X 座標の初期値、要素 second に増分を持つ pair で、
          処理後には次に描画する X 座標が示される
        */
        void TriFill_XDraw( DrawingArea_IF& draw, GPixelOp& pixelOp, int sy, int ey,
                            std::pair<double, double>* l, std::pair<double,double>* r )
        {
          for ( ; sy < ey ; ++sy ) {
            int sx = round( l->first ); // 描画開始 X 座標
            int ex = round( r->first ); // 描画終了 X 座標

            // X 座標のクリッピング
            if ( sx < 0 ) sx = 0;
            if ( ex >= draw.size().x ) ex = draw.size().x - 1;

            // スキャンライン描画
            for ( ; sx <= ex ; ++sx )
              pixelOp( draw, Coord<int>( sx, sy ) );

            // X 座標の更新
            l->first += l->second;
            r->first += r->second;
          }
        }

        /*
          TriFill_Main : top, middle, bottom の三つの頂点からなる三角形を描画する
          top, middle, bottom は Y 座標が昇順に並んでいるものとする
        */
        void TriFill_Main( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> top, Coord<int> middle, Coord<int> bottom )
        {
          // 上側の頂点からの描画開始 X 座標(頂角が描画領域外の場合、異なる座標になる)
          double top_mid_x = top.x; // 上側 - 中央
          double top_btm_x = top.x; // 上側 - 下側

          // 上側に水平な辺がある場合は中央の頂点で初期化する
          if ( top.y == middle.y )
            top_mid_x = middle.x;

          int sy = top.y;    // 描画開始 Y 座標
          int my = middle.y; // 中央の頂点の Y 座標
          int ey = bottom.y; // 描画終了 Y 座標

          // クリッピング

          // 上側の頂点が領域外の場合
          if ( top.y < 0 ) {
            sy = 0;
            // 上側から中央への辺をクリッピング
            if ( middle.y >= 0 ) {
              if ( top.y != middle.y )
                top_mid_x = static_cast<double>( middle.x - top.x ) *
                  static_cast<double>( middle.y ) / static_cast<double>( top.y - middle.y ) +
                  static_cast<double>( middle.x );
            } else {
              my = 0;
              if ( middle.y != bottom.y )
                top_mid_x = static_cast<double>( bottom.x - middle.x ) *
                  static_cast<double>( bottom.y ) / static_cast<double>( middle.y - bottom.y ) +
                  static_cast<double>( bottom.x );
            }
            // 上側から下側への辺をクリッピング
            if ( top.y != bottom.y )
              top_btm_x = static_cast<double>( bottom.x - top.x ) *
                static_cast<double>( bottom.y ) / static_cast<double>( top.y - bottom.y ) +
                static_cast<double>( bottom.x );
          }

          // 下側の頂点が領域外の場合は描画終了 Y 座標を描画領域内にする
          if ( bottom.y >= draw.size().y )
            ey = draw.size().y - 1;

          // 二辺の描画開始座標とその増分
          double top_mid_a = ( middle.y != top.y ) ?
            static_cast<double>( middle.x - top.x ) / static_cast<double>( middle.y - top.y ) : 0;       // 上側 - 中央
          double mid_btm_a = ( middle.y != bottom.y ) ?
            static_cast<double>( middle.x - bottom.x ) / static_cast<double>( middle.y - bottom.y ) : 0; // 中央 - 下側
          double top_btm_a = ( top.y != bottom.y ) ?
            static_cast<double>( top.x - bottom.x ) / static_cast<double>( top.y - bottom.y ) : 0;       // 上側 - 下側

          // 描画開始 X 座標とその増分の pair
          std::pair<double, double> top_mid( top_mid_x, top_mid_a ); // 上側 - 中央
          std::pair<double, double> top_btm( top_btm_x, top_btm_a ); // 上側 - 下側

          // 中央の頂点が右向きか左向きかを判定して、各辺が左側・右側ののいずれかを決定する
          // 中央の頂点を通る水平線が、上側・下側を通る直線と交わる点の X 座標
          int splitLine_x = ( top.y != bottom.y ) ?
            ( top.x - bottom.x ) * ( middle.y - top.y ) / ( top.y - bottom.y ) + top.x :
            bottom.x; // 中央・下側の Y 座標が等しい場合、下側の X 座標
          std::pair<double, double>* l = ( middle.x < splitLine_x ) ? &top_mid : &top_btm; // 左側
          std::pair<double, double>* r = ( middle.x < splitLine_x ) ? &top_btm : &top_mid; // 右側

          // 描画開始
          // 中央の頂点を通る水平線を境界に、上下に分けて描画する
          TriFill_XDraw( draw, pixelOp, sy, my, l, r );
          top_mid.second = mid_btm_a;
          TriFill_XDraw( draw, pixelOp, my, ey + 1, l, r );
        }

        /*
          Triangle : c1, c2, c3 を頂点とする三角形描画 呼び出しルーチン
        */
        void Triangle( DrawingArea_IF& draw, GPixelOp& pixelOp, Coord<int> c1, Coord<int> c2, Coord<int> c3 )
        {
          // Y 座標で昇順にソート
          if ( c1.y > c2.y ) std::swap( c1, c2 );
          if ( c1.y > c3.y ) std::swap( c1, c3 );
          if ( c2.y > c3.y ) std::swap( c2, c3 );

          // ポリゴンが描画領域外なら処理しない
          if ( c1.y >= draw.size().y ) return;
          if ( c3.y < 0 ) return;

          // 描画ルーチン メインへ
          TriFill_Main( draw, pixelOp, c1, c2, c3 );
        }

      } // namespace polygon

      /**************************************************************
          GCurve : 曲線描画関数
      **************************************************************/

      namespace curve
      {
        /*
          LessCoord< T > : 座標用の比較関数
        */
        template< class T > struct LessCoord
        {
          bool operator()( const Coord< T >& src, const Coord< T >& dest ) const
          {
            if ( src.y != dest.y )
              return( src.y < dest.y );
            else
              return( src.x < dest.x );
          }
        };

        class SetCandidate : public GPixelOp
        {
          // 求めた座標を保持する map
          //LessCoord< int > less;
          std::map< Coord< int >, bool, LessCoord< int > > drawCoord;

        public:

          typedef std::map< Coord< int >, bool > map;

          SetCandidate()
            : drawCoord( LessCoord< int >() ) {}

          bool operator()( DrawingArea_IF& draw, Coord< int > c )
          {
            if ( c.x >= 0 && c.x < draw.size().x && c.y >= 0 && c.y < draw.size().y ) {
              drawCoord[c] = true;
              return( true );
            } else {
              return( false );
            }
          }

          map::const_iterator begin() const
          { return( drawCoord.begin() ); }

          map::iterator begin()
          { return( drawCoord.begin() ); }

          map::const_iterator end() const
          { return( drawCoord.end() ); }

          map::iterator end()
          { return( drawCoord.end() ); }

          map::const_iterator find( Coord< int > c ) const
          { return( drawCoord.find( c ) ); }

          map::iterator find( Coord< int > c )
          { return( drawCoord.find( c ) ); }
        };

        /*
          CalcInnerPoint : パラメトリック曲線の内点を計算する

          t0, t1 : 定義域
          first : t0 に対する座標
          second : t1 に対する座標
          f : パラメトリック方程式
          drawCoord : 描画対象の座標を登録する set
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void CalcInnerPoint( DrawingArea_IF& draw, double t0, double t1, const Coord< int >& first, const Coord< int >& second,
                             ParametricEquation& f, SetCandidate& drawCoord, double rate )
        {
          // t0 と t1 の差の最小値
          const double THRESHOLD = 1E-9;

          // THRESHOLDより小さければ処理はしない
          if ( fabs( t0 - t1 ) < THRESHOLD ) return;

          // 内点に対する t の値
          double inner = ( t0 < t1 ) ?
            ( 1 - rate ) * t0 + rate * t1 :
            rate * t0 + ( 1 - rate ) * t1;

          // 内点の座標
          Coord< int > c( round( f.x( inner ) ), round( f.y( inner ) ) );
          drawCoord( draw, c );

          // 再帰的に内点を計算
          if ( abs( first.x - c.x ) > 0 || abs( first.y - c.y ) > 0 )
            CalcInnerPoint( draw, t0, inner, first, c, f, drawCoord, rate );
          if ( abs( second.x - c.x ) > 0 || abs( second.y - c.y ) > 0 )
            CalcInnerPoint( draw, inner, t1, c, second, f, drawCoord, rate );
        }

        /*
          DrawPramaetricCurve : パラメトリック曲線描画関数

          内点を再帰的に描画していく形式

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          f : パラメトリック方程式
          range : f の定義域
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void DrawParametricCurve( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                  ParametricEquation& f, const std::pair< double, double >& range, double rate )
        {
          assert( rate > 0 && rate < 1 );

          // 求めた座標を保持するバッファ
          SetCandidate drawCoord;

          // 両端の点の座標
          Coord< int > first( round( f.x( range.first ) ), round( f.y( range.first ) ) );
          Coord< int > second( round( f.x( range.second ) ), round( f.y( range.second ) ) );

          // 両端の点を登録
          drawCoord( draw, first );
          drawCoord( draw, second );

          // 再起処理開始
          CalcInnerPoint( draw, range.first, range.second, first, second, f, drawCoord, rate );

          // 四連結の削除
          for ( SetCandidate::map::iterator it = drawCoord.begin() ;
                it != drawCoord.end() ; ++it ) {
            SetCandidate::map::iterator l = drawCoord.find( Coord< int >( ( it->first ).x - 1, ( it->first ).y ) );
            SetCandidate::map::iterator u = drawCoord.find( Coord< int >( ( it->first ).x, ( it->first ).y - 1 ) );
            SetCandidate::map::iterator r = drawCoord.find( Coord< int >( ( it->first ).x + 1, ( it->first ).y ) );
            SetCandidate::map::iterator d = drawCoord.find( Coord< int >( ( it->first ).x, ( it->first ).y + 1 ) );
            if ( l != drawCoord.end() && u != drawCoord.end() && l->second && u->second ) {
              it->second = false;
              continue;
            }
            if ( r != drawCoord.end() && u != drawCoord.end() && r->second && u->second ) {
              it->second = false;
              continue;
            }
            if ( l != drawCoord.end() && d != drawCoord.end() && l->second && d->second ) {
              it->second = false;
              continue;
            }
            if ( r != drawCoord.end() && d != drawCoord.end() && r->second && d->second ) {
              it->second = false;
              continue;
            }
          }

          // 描画処理
          for ( SetCandidate::map::const_iterator it = drawCoord.begin() ;
                it != drawCoord.end() ; ++it ) {
            if ( it->second )
              pixelOp( draw, it->first );
          }
        }

        /*
          DrawPramaetricCurve : パラメトリック曲線描画関数

          ステップ幅固定で描画し、隙間は線分で補間する。
          開一様のスプライン曲線は途中で曲線が途切れるためこれは使用できない。

          * 未使用の関数

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          f : パラメトリック方程式
          range : f の定義域
        */
        void DrawParametricCurve_ConnectByLine( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                                ParametricEquation& f, const std::pair< double, double >& range )
        {
          // 求めた座標を保持するバッファ
          SetCandidate drawCoord;

          double delta = ( range.second - range.first ) / 10000;
          Coord< int > b( round( f.x( range.first ) ), round( f.y( range.first ) ) );
          for ( double d = range.first ; d < range.second ; d += delta ) {
            Coord< int > c( round( f.x( d ) ), round( f.y( d ) ) );
            line::Draw( draw, drawCoord, b, c );
            b = c;
          }

          // 四連結の削除
          for ( SetCandidate::map::iterator it = drawCoord.begin() ;
                it != drawCoord.end() ; ++it ) {
            SetCandidate::map::iterator l = drawCoord.find( Coord< int >( ( it->first ).x - 1, ( it->first ).y ) );
            SetCandidate::map::iterator u = drawCoord.find( Coord< int >( ( it->first ).x, ( it->first ).y - 1 ) );
            SetCandidate::map::iterator r = drawCoord.find( Coord< int >( ( it->first ).x + 1, ( it->first ).y ) );
            SetCandidate::map::iterator d = drawCoord.find( Coord< int >( ( it->first ).x, ( it->first ).y + 1 ) );
            if ( l != drawCoord.end() && u != drawCoord.end() && l->second && u->second ) {
              it->second = false;
              continue;
            }
            if ( r != drawCoord.end() && u != drawCoord.end() && r->second && u->second ) {
              it->second = false;
              continue;
            }
            if ( l != drawCoord.end() && d != drawCoord.end() && l->second && d->second ) {
              it->second = false;
              continue;
            }
            if ( r != drawCoord.end() && d != drawCoord.end() && r->second && d->second ) {
              it->second = false;
              continue;
            }
          }

          // 描画処理
          for ( SetCandidate::map::const_iterator it = drawCoord.begin() ;
                it != drawCoord.end() ; ++it ) {
            if ( it->second )
              pixelOp( draw, it->first );
          }
        }

        /*
          DrawPramaetricCurve : パラメトリック曲線描画関数

          ステップ幅を決めて描画していく形式

          * 未使用の関数

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          f : パラメトリック方程式
          range : f の定義域
        */
        void DrawParametricCurve_ByStep( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                         ParametricEquation& f, const std::pair< double, double >& range )
        {
          // t0 と t1 の差の最小値
          //const double THRESHOLD = 1E-16;

          // 求めた座標を保持するバッファ
          SetCandidate drawCoord;

          double delta = ( range.second - range.first ) / 10000;
          Coord< int > b( round( f.x( range.first ) ), round( f.y( range.first ) ) );
          drawCoord( draw, b );
          for ( double d = range.first ; d < range.second ; ) {
            Coord< int > c( round( f.x( d ) ), round( f.y( d ) ) );
            if ( b == c ) {
              delta *= 2;
              d += delta;
            } else if ( std::abs( c.x - b.x ) > 1 || std::abs( c.y - b.y ) > 1 ) {
              //if ( delta >= THRESHOLD ) {
              delta /= 2;
              d -= delta;
              //} else {
              //GLine::Draw( draw, drawCoord, b, c );
              //b = c;
              //}
            } else {
              drawCoord( draw, c );
              b = c;
            }
          }

          // 四連結の削除
          for ( SetCandidate::map::iterator it = drawCoord.begin() ;
                it != drawCoord.end() ; ++it ) {
            SetCandidate::map::iterator l = drawCoord.find( Coord< int >( ( it->first ).x - 1, ( it->first ).y ) );
            SetCandidate::map::iterator u = drawCoord.find( Coord< int >( ( it->first ).x, ( it->first ).y - 1 ) );
            SetCandidate::map::iterator r = drawCoord.find( Coord< int >( ( it->first ).x + 1, ( it->first ).y ) );
            SetCandidate::map::iterator d = drawCoord.find( Coord< int >( ( it->first ).x, ( it->first ).y + 1 ) );
            if ( l != drawCoord.end() && u != drawCoord.end() && l->second && u->second ) {
              it->second = false;
              continue;
            }
            if ( r != drawCoord.end() && u != drawCoord.end() && r->second && u->second ) {
              it->second = false;
              continue;
            }
            if ( l != drawCoord.end() && d != drawCoord.end() && l->second && d->second ) {
              it->second = false;
              continue;
            }
            if ( r != drawCoord.end() && d != drawCoord.end() && r->second && d->second ) {
              it->second = false;
              continue;
            }
          }

          // 描画処理
          for ( SetCandidate::map::const_iterator it = drawCoord.begin() ;
                it != drawCoord.end() ; ++it ) {
            if ( it->second )
              pixelOp( draw, it->first );
          }
        }

        /*
          StrokeFergusonCoonsSpline : Ferguson/Coons曲線の描画

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          dp : 各点に対する微分係数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void StrokeFergusonCoonsSpline( DrawingArea_IF& draw, GPixelOp& pixelOp, const vector< Coord< int > >& p,
                                        const vector< Coord< double > >& dp, double rate )
        {
          size_t cnt = std::min( p.size(), dp.size() );
          if ( cnt == 0 ) return;

          for ( size_t i = 0 ; i < cnt - 1 ; ++i ) {
            FergusonCoons_Spline spline( p[i], p[i + 1], dp[i], dp[i + 1] );
            DrawParametricCurve( draw, pixelOp, spline, std::pair< double, double >( 0, 1 ) , rate );
          }
        }

        /*
          StrokeCatmullRomSpline : Catmull-Rom曲線の描画

          p の最初と最後の要素は描画対象の点には含まれない(微分係数の計算のみに使用)

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          d : 微分係数計算時の補正係数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void StrokeCatmullRomSpline( DrawingArea_IF& draw, GPixelOp& pixelOp, const vector< Coord< int > >& p,
                                     double d, double rate )
        {
          size_t cnt = p.size();
          if ( cnt < 4 ) return;

          for ( size_t i = 0 ; i <= cnt - 4 ; ++i ) {
            CatmullRom_Spline spline( p[i], p[i + 1], p[i + 2], p[i + 3], d );
            DrawParametricCurve( draw, pixelOp, spline, std::pair< double, double >( 0, 1 ) , rate );
          }
        }

        /*
          StrokeBezierSpline_Base : ベジェ曲線によるスプライン曲線描画

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          n : 次数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        template< class BEZIER >
        void StrokeBezierSpline_Base( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                      const vector< Coord< int > >& p, size_t n, double rate )
        {
          // n が 1 以下の場合は描画しない
          if ( n < 2 ) return;

          BEZIER bezier( n + 1 ); // n 次ベジェ曲線(点列数は n + 1)
          Coord< double > edgeP( p.front().x, p.front().y ); // スプラインの端点

          // 中間点の開始反復子
          vector< Coord< int > >::const_iterator middleP = p.begin() + 1;
          // 次のスプライン開始の反復子
          vector< Coord< int > >::const_iterator nextP =
            ( p.size() > n ) ? p.begin() + n : p.end();

          for (;;) {
            // nextP が末尾を超えている場合は点列の数は n より小さい
            if ( nextP == p.end() ) {
              vector< Coord< int > > tail( middleP, nextP );
              tail.insert( tail.begin(), Coord< int >( edgeP.x, edgeP.y ) );
              BEZIER bezier( tail );
              DrawParametricCurve( draw, pixelOp, bezier, std::pair< double, double >( 0, 1 ) , rate );
              break;
            }

            // 点列のセット
            bezier.set( 0, edgeP );
            bezier.set( 1, middleP, nextP );

            // 最終点は、第 n - 1 点と第 n 点の中点とする。
            // 但し、第 n 点が点列の末尾なら第 n 点をそのまま使用する
            edgeP = ( nextP + 1 == p.end() ) ?
              Coord< double >( nextP->x, nextP->y ) :
              Coord< double >( ( ( nextP - 1 )->x + nextP->x ) / 2,
                               ( ( nextP - 1 )->y + nextP->y ) / 2 );
            bezier.set( n, edgeP );

            DrawParametricCurve( draw, pixelOp, bezier, std::pair< double, double >( 0, 1 ) , rate );

            // 反復子の更新
            middleP = nextP;
            nextP = ( p.end() - ( n - 1 ) <= nextP ) ? p.end() : nextP + n - 1;
          }
        }
        void StrokeBezierSpline( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                 const vector< Coord< int > >& p, size_t n, double rate )
        { StrokeBezierSpline_Base< BezierCurve >( draw, pixelOp, p, n, rate ); }

        void StrokeDeCasteljauSpline( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                      const vector< Coord< int > >& p, size_t n, double rate )
        { StrokeBezierSpline_Base< DeCasteljau >( draw, pixelOp, p, n, rate ); }


        /*
          StrokeBSpline_Base : B-スプライン曲線描画

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          knot : ノット列
          n : 次数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        template< class BSPLINE >
        void StrokeBSpline_Base( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                 const vector< Coord< int > >& p, const vector< double >& knot,
                                 size_t n, double rate )
        {
          // ノット列は点列のサイズ + 次数 + 1 以上でなければならない
          assert( knot.size() >= p.size() + n + 1 );

          BSPLINE bSpline( knot, p, n );
          DrawParametricCurve( draw, pixelOp, bSpline, std::pair< double, double >( knot[n], knot[p.size()] ) , rate );
        }
        void StrokeBSpline( DrawingArea_IF& draw, GPixelOp& pixelOp,
                            const vector< Coord< int > >& p, const vector< double >& knot,
                            size_t n, double rate )
        { StrokeBSpline_Base< BSplineCurve >( draw, pixelOp, p, knot, n, rate ); }

        void StrokeDeBoorCox( DrawingArea_IF& draw, GPixelOp& pixelOp,
                              const vector< Coord< int > >& p, const vector< double >& knot,
                              size_t n, double rate )
        { StrokeBSpline_Base< DeBoorCox >( draw, pixelOp, p, knot, n, rate ); }

        /*
          CreateUniformKnot : 一様なノット列の作成

          knot : ノット列を登録する配列へのポインタ
          pSize : 点列の数
          n : 次数
        */
        void CreateUniformKnot( vector< double >* knot, size_t pSize, size_t n )
        {
          knot->clear();
          for ( size_t i = 0 ; i < pSize + n + 1 ; ++i )
            knot->push_back( i );
        }

        /*
          StrokeUniformBSpline : 一様 B-スプライン曲線描画

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          n : 次数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void StrokeUniformBSpline( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                   const vector< Coord< int > >& p, size_t n, double rate )
        {
          vector< double > knot;
          CreateUniformKnot( &knot, p.size(), n );
          StrokeBSpline( draw, pixelOp, p, knot, n, rate );
        }

        void StrokeUniformDeBoorCox( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                     const vector< Coord< int > >& p, size_t n, double rate )
        {
          vector< double > knot;
          CreateUniformKnot( &knot, p.size(), n );
          StrokeDeBoorCox( draw, pixelOp, p, knot, n, rate );
        }

        /*
          CreateOpenUniformKnot : 開一様なノット列の作成

          knot : ノット列を登録する配列へのポインタ
          pSize : 点列の数
          n : 次数
        */
        void CreateOpenUniformKnot( vector< double >* knot, size_t pSize, size_t n )
        {
          knot->clear();
          for ( size_t i = 0 ; i < pSize + n + 1 ; ++i ) {
            if ( i <= n )
              knot->push_back( 0 );
            else if ( i >= pSize )
              knot->push_back( 2 );
            else
              knot->push_back( 1 );
          }
        }

        /*
          StrokeOpenUniformBSpline : 開一様 B-スプライン曲線描画

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          n : 次数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void StrokeOpenUniformBSpline( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                       const vector< Coord< int > >& p, size_t n, double rate )
        {
          vector< double > knot;
          CreateOpenUniformKnot( &knot, p.size(), n );
          StrokeBSpline( draw, pixelOp, p, knot, n, rate );
        }

        void StrokeOpenUniformDeBoorCox( DrawingArea_IF& draw, GPixelOp& pixelOp,
                                         const vector< Coord< int > >& p, size_t n, double rate )
        {
          vector< double > knot;
          CreateOpenUniformKnot( &knot, p.size(), n );
          StrokeDeBoorCox( draw, pixelOp, p, knot, n, rate );
        }

        /*
          StrokeNurbs : Nurbs 曲線描画

          draw : 描画オブジェクト
          pixelOp : 描画用関数
          p : 描画する曲線上の点の座標
          knot : ノット列
          w : 重み
          n : 次数
          rate : 再帰的に内点を描画するときの内点の位置(比率)
        */
        void StrokeNurbs( DrawingArea_IF& draw, GPixelOp& pixelOp,
                          const vector< Coord< int > >& p, const vector< double >& knot, const vector< double >& w,
                          size_t n, double rate )
        {
          // ノット列は点列のサイズ + 次数 + 1 以上でなければならない
          assert( knot.size() >= p.size() + n + 1 );

          NurbsCurve nurbs( knot, p, w, n );
          DrawParametricCurve( draw, pixelOp, nurbs, std::pair< double, double >( knot[n], knot[p.size()] ) , rate );
        }

      } // namespace curve

    } // namespace shape

    namespace pen
    {
      /*
        中心からの距離に応じて透明度を変化させてドットを描画する
      */
      class GPSetByDistance : public GRGBOp
      {
        Coord< int > center_; // 円形ブラシの中心座標
        int maxRadius_;       // 円形ブラシの最大半径
        int intensity_;       // 最大強度

      public:

        /*
          コンストラクタ

          op : 描画用関数オブジェクト
          rgb : 描画色(アルファチャンネルは描画位置に応じて変更される)
          center : 中心位置
          maxRadius : 円形ブラシの半径
          intensity : ブラシの強度
        */
        GPSetByDistance( RGB& ( RGB::*op )( RGB ), RGB rgb, Coord< int > center, int maxRadius, int intensity )
          : GRGBOp( op, rgb ), center_( center ), maxRadius_( maxRadius ), intensity_( intensity ) {}

        /*
          描画領域 draw の位置 c に、中心位置 center_ からの距離に応じてアルファチャンネルを変化させて描画を行う
        */
        virtual bool operator()( DrawingArea_IF& draw, Coord< int > c )
        {
          int rSq = ( c.x - center_.x ) * ( c.x - center_.x ) + ( c.y - center_.y ) * ( c.y - center_.y );
          int maxRSq = maxRadius_ * maxRadius_;
          if ( rSq > maxRSq ) return( true );
          if ( maxRSq == 0 ) return( true );

          RGB rgb = getRGB();
          rgb.a = RGB::Check_RGBPrimary( std::sqrt( static_cast< double >( maxRSq - rSq ) ) * intensity_ / std::sqrt( static_cast< double >( maxRSq ) ) );
          setRGB( rgb );

          return( GRGBOp::operator()( draw, c ) );
        }
      };

      /*
        描画領域 draw の位置 c を中心に円形ブラシを描画する
      */
      bool GCircleBrush::operator()( DrawingArea_IF& draw, Coord< int > c )
      {
        GPSetByDistance pset( mf_, rgb_, c, r_, intensity_ );

        shape::ellipse::Fill( draw, pset, c, r_, Coord< int >( 1, 1 ) );

        return( true );
      }

    } // namespace pen

  } // namespace gl

} // namespace tadah_fussy
