/**
   @file gl_drawing_method.h
   @brief 描画ルーチン定義用ヘッダ

   @author tadah_fussy
   @date 2018/02/25 DrawingMehod.h から分離
   @date 2021/02/25 ワーニング出力に対応
   @date 2021/03/29 名前空間の見直し
   @date 2021/03/30 tadah_fussy::gl::shape::polygon::Triangleのバグ修正など
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_DRAWING_METHOD_20210402_H
#define GRAPHIC_LIBRARY_DRAWING_METHOD_20210402_H

#include "gl_g_pixel_op.h"
#include <vector>

namespace tadah_fussy
{
  namespace gl
  {
    namespace shape
    {
      /**
         @brief 矩形描画用関数
      **/
      namespace box
      {
        /// @brief 矩形描画関数
        ///
        /// c0, c1 は矩形の対角の座標を表す。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c0,c1 矩形の対角の座標(二点指定)
        void Stroke( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, Coord< int > c1 );

        /// @brief 矩形の塗りつぶし関数
        ///
        /// c0, c1 は矩形の対角の座標を表す。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c0,c1 矩形の対角の座標(二点指定)
        void Fill( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, Coord< int > c1 );

        /// @brief 矩形の描画/塗りつぶし関数
        ///
        /// 引数 filled に応じて StrokeBox と FillBox を切り替えて呼び出す
        ///
        /// c0, c1 は矩形の対角の座標を表す。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c0,c1 矩形の対角の座標(二点指定)
        /// @param filled true で塗りつぶし (FillBox)、false で描画 (StrokeBox) を呼び出す
        inline void Draw( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const Coord< int >& c0, const Coord< int >& c1, bool filled = false )
        {
          if ( filled )
            Fill( draw, pixelOp, c0, c1 );
          else
            Stroke( draw, pixelOp, c0, c1 );
        }

      } // namespace box

      /**
         @brief 線分描画用関数
      **/
      namespace line
      {
        /// @brief 線分描画用クリッピング処理
        ///
        /// @param drawSize 描画エリアの大きさ
        /// @param c0,c1 始点・終点の座標へのポインタ(クリッピング後の座標に変換される)
        /// @retval >0 ... クリッピングされた
        /// @retval =0 ... クリッピングの必要なし
        /// @retval <0 ... 線分は完全に不可視
        int Clipping( Coord< int > drawSize, Coord< int >* c0, Coord< int >* c1 );

        /// @brief 線分描画関数
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c0,c1 始点・終点座標
        /// @param needClipping trueの場合はクリッピングを行う
        /// @param connect4 trueの場合は 4 連結で描画する
        void Draw( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, Coord< int > c1, bool needClipping = true, bool connect4 = false );

        /// @brief 複数のラインを描画する
        ///
        /// 線分は、添字を一つずつたどりながら描画するため、描画結果は各線分が連結された形になる。
        /// 但し、配列の末尾と開始は連結しない。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param vertexList 各頂点の座標
        void DrawMultiple( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const std::vector< Coord< int > >& vertexList );

        /// @brief 水平線分を描画する
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c0 描画開始点の座標
        /// @param ex 描画終了点の X 座標
        void DrawH( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, int ex );

        /// @brief 垂直線分を描画する
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c0 描画開始点の座標
        /// @param ey 描画終了点の Y 座標
        void DrawV( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, int ey );

      } // namespace line

      /**
         @brief 楕円描画用関数
      **/
      namespace ellipse
      {
        /// @brief 二点を対角線とする長方形の頂点を通る楕円のパラメータを求める
        ///
        /// 楕円のパラメータは無限に存在するため、楕円の中心から対角線の端点への相対座標(u,v)に対し
        /// k = Coord( v, u ), r = uv( u + v ) を代表値とする。
        ///
        /// @param c1,c2 対角線の座標
        /// @param o 求めた中心座標を代入する変数へのポインタ
        /// @param r 求めた半径を代入する変数へのポインタ
        /// @param k 求めた係数を代入する変数へのポインタ ( a( x - o.x )^2 + b( y - o.y )^2 = r^2 に対する ( a, b ) )
        /// @return 長方形が線分となる場合は false を返す
        bool GetParamByRectangle( Coord< int > c1, Coord< int > c2, Coord< int >* o, int* r, Coord< int >* k );

        /// @brief 楕円描画関数
        ///
        /// 楕円 a( x - o.x )^2 + b( y - o.y )^2 = r^2 を描画する
        ///
        /// 半径 r、係数 k の中にゼロ以下の値が含まれる場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 楕円の中心
        /// @param r 半径
        /// @param k 係数
        void Stroke( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > k );

        /// @brief 楕円塗りつぶし関数
        ///
        /// 楕円 a( x - o.x )^2 + b( y - o.y )^2 = r^2 を塗りつぶす
        ///
        /// 半径 r、係数 k の中にゼロ以下の値が含まれる場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 楕円の中心
        /// @param r 半径
        /// @param k 係数
        void Fill( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > k );

        /// @brief 楕円描画/塗りつぶし関数
        ///
        /// 引数 filled に応じて GEllipse::Stroke と GEllipse::Fill を切り替えて呼び出す
        ///
        /// 半径 r、係数 k の中にゼロ以下の値が含まれる場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 楕円の中心
        /// @param r 半径
        /// @param k 係数
        /// @param filled 塗りつぶしをするか
        inline void Draw( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > k, bool filled = false )
        {
          if ( filled )
            Fill( draw, pixelOp, o, r, k );
          else
            Stroke( draw, pixelOp, o, r, k );
        }

        /// @brief 二点を対角線とする長方形の頂点を通る楕円を描画する
        ///
        /// 長方形が線分となる場合 ( 二点の X / Y 座標のいずれかが等しくなった場合 ) は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2 対角線の両端点
        void Stroke_2p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2 );
        /// @brief 二点を対角線とする長方形の頂点を通る楕円を塗りつぶす
        ///
        /// 長方形が線分となる場合 ( 二点の X / Y 座標のいずれかが等しくなった場合 ) は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2 対角線の両端点
        void Fill_2p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2 );

        /// @brief 二点を対角線とする長方形の頂点を通る楕円を描画または塗りつぶす
        ///
        /// 引数 filled に応じて GEllipse::Stroke_2p と GEllipse::Fill_2p を切り替えて呼び出す。
        ///
        /// 長方形が線分となる場合 ( 二点の X / Y 座標のいずれかが等しくなった場合 ) は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2 対角線の両端点
        /// @param filled 塗りつぶしをするか
        inline void Draw_2p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const Coord< int >& c1, const Coord< int >& c2, bool filled = false )
        {
          if ( filled )
            Fill_2p( draw, pixelOp, c1, c2 );
          else
            Stroke_2p( draw, pixelOp, c1, c2 );
        }

      } // namespace ellipse

      /**
         @brief 真円描画用関数
      **/
      namespace circle
      {
        /// @brief 三角形の外心と外接円の半径を求める
        ///
        /// @param c1,c2,c3 三角形の座標
        /// @param r 求めた半径を代入する変数へのポインタ ( NULL の場合は無視される )
        /// @return 中心座標 (求められない場合は c1〜c3 でできる線分の中点を返す)
        Coord< int > CalcOrigin( Coord< int > c1, Coord< int > c2, Coord< int > c3, int* r );
        /// @brief 三角形の外心のみを求める ( 外接円の半径は求めない )
        ///
        /// @param c1,c2,c3 三角形の座標
        /// @return 中心座標 (求められない場合は c1〜c3 でできる線分の中点を返す)
        inline Coord< int > CalcOrigin( Coord< int > c1, Coord< int > c2, Coord< int > c3 )
        {
          return( CalcOrigin( c1, c2, c3, 0 ) );
        }

        /// @brief 真円描画関数
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円の中心
        /// @param r 半径
        inline void Stroke( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r )
        { ellipse::Stroke( draw, pixelOp, o, r, Coord< int >( 1, 1 ) ); }
        /// @brief 真円塗りつぶし関数
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円の中心
        /// @param r 半径
        inline void Fill( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r )
        { ellipse::Fill( draw, pixelOp, o, r, Coord< int >( 1, 1 ) ); }
        /// @brief 真円描画/塗りつぶし関数
        ///
        /// 引数 filled に応じて GCircle::Stroke と GCircle::Fill を切り替えて呼び出す
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円の中心
        /// @param r 半径
        /// @param filled 塗りつぶしをするか
        inline void Draw( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, bool filled = false )
        { ellipse::Draw( draw, pixelOp, o, r, Coord< int >( 1, 1 ), filled ); }

        /// @brief 円周上の三点から中心と半径を決定して描画する
        ///
        /// 求めた半径がゼロの場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2,c3 円周上を通る三点
        void Stroke_3p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2, Coord< int > c3 );
        /// @brief 円周上の三点から中心と半径を決定して塗りつぶす
        ///
        /// 求めた半径がゼロの場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2,c3 円周上を通る三点
        void Fill_3p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2, Coord< int > c3 );
        /// @brief 円周上の三点から中心と半径を決定して描画する/塗りつぶす
        ///
        /// 引数 filled に応じて GCircle::Stroke_3p と GCircle::Fill_3p を切り替えて呼び出す。
        ///
        /// 求めた半径がゼロの場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2,c3 円周上を通る三点
        /// @param filled 塗りつぶしをするか
        inline void Draw_3p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2, Coord< int > c3, bool filled = false )
        {
          if ( filled )
            Fill_3p( draw, pixelOp, c1, c2, c3 );
          else
            Stroke_3p( draw, pixelOp, c1, c2, c3 );
        }

        /// @brief 二点を直径の両端として円を決定して描画する
        ///
        /// 二点が等しい場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2 円周上を通る二点
        void Stroke_2p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2 );
        /// @brief 二点を直径の両端として円を決定して塗りつぶす
        ///
        /// 二点が等しい場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2 円周上を通る二点
        void Fill_2p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2 );
        /// @brief 二点を直径の両端として円を決定して描画する/塗りつぶす
        ///
        /// 引数 filled に応じて GCircle::Stroke_2p と GCircle::Fill_2p を切り替えて呼び出す。
        ///
        /// 二点が等しい場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2 円周上を通る二点
        /// @param filled 塗りつぶしをするか
        inline void Draw_2p( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const Coord< int >& c1, const Coord< int >& c2, bool filled = false )
        {
          if ( filled )
            Fill_2p( draw, pixelOp, c1, c2 );
          else
            Stroke_2p( draw, pixelOp, c1, c2 );
        }

      } // namespace circle

      /**
         @brief 円弧描画用関数
      **/
      namespace arc
      {
        /// @brief 円弧描画関数
        ///
        /// 描画する範囲の角度は sa から ea まで時計回りで表される。どちらの角度も 0 から 2π の間の値に変換されることに注意。
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円弧の中心
        /// @param r 円弧の半径
        /// @param sa,ea 円弧の開始・終了角度
        /// @param isOut 範囲内を描画するなら false
        void Stroke( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, double sa, double ea, bool isOut = false );

        /// @brief 円弧塗りつぶし関数
        ///
        /// 描画する範囲の角度は sa から ea まで時計回りで表される。どちらの角度も 0 から 2π の間の値に変換されることに注意。
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円弧の中心
        /// @param r 円弧の半径
        /// @param sa,ea 円弧の開始・終了角度
        /// @param isOut 範囲内を描画するなら false
        void Fill( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, double sa, double ea, bool isOut = false );

        /// @brief 円弧描画/塗りつぶし関数
        ///
        /// 描画する範囲の角度は sa から ea まで時計回りで表される。どちらの角度も 0 から 2π の間の値に変換されることに注意。
        ///
        /// 引数 filled に応じて GArc::Stroke と GArc::Fill を切り替えて呼び出す。
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円弧の中心
        /// @param r 円弧の半径
        /// @param sa,ea 円弧の開始・終了角度
        /// @param isOut 範囲内を描画するなら false
        /// @param filled 塗りつぶしをするなら true
        inline void Draw( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, double sa, double ea, bool isOut = false, bool filled = false )
        {
          if ( filled )
            Fill( draw, pixelOp, o, r, sa, ea, isOut );
          else
            Stroke( draw, pixelOp, o, r, sa, ea, isOut );
        }

        /// @brief 円弧描画関数
        ///
        /// 描画する範囲の角度は点 c1, c2 を元に計算され、c1 から c2 まで時計回りで表される。
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円弧の中心
        /// @param r 円弧の半径
        /// @param c1,c2 円弧の開始・終了角度を示す点
        /// @param isOut 範囲内を描画するなら false
        void Stroke( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > c1, Coord< int > c2, bool isOut = false );

        /// @brief 円弧塗りつぶし関数
        ///
        /// 描画する範囲の角度は点 c1, c2 を元に計算され、c1 から c2 まで時計回りで表される。
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円弧の中心
        /// @param r 円弧の半径
        /// @param c1,c2 円弧の開始・終了角度を示す点
        /// @param isOut 範囲内を描画するなら false
        void Fill( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > c1, Coord< int > c2, bool isOut = false );

        /// @brief 円弧描画/塗りつぶし関数
        ///
        /// 描画する範囲の角度は点 c1, c2 を元に計算され、c1 から c2 まで時計回りで表される。
        ///
        /// 引数 filled に応じて GArc::Stroke と GArc::Fill を切り替えて呼び出す。
        ///
        /// 半径 r がゼロ以下の場合は処理をしない ( エラーにはならないことに注意 )。
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param o 円弧の中心
        /// @param r 円弧の半径
        /// @param c1,c2 円弧の開始・終了角度を示す点
        /// @param isOut 範囲内を描画するなら false
        /// @param filled 塗りつぶしをするなら true
        inline void Draw( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, int r, Coord< int > c1, Coord< int > c2, bool isOut = false, bool filled = false )
        {
          if ( filled )
            Fill( draw, pixelOp, o, r, c1, c2, isOut );
          else
            Stroke( draw, pixelOp, o, r, c1, c2, isOut );
        }

      } // namespace arc

    } // namespace shape

    namespace draw
    {
      /**
         @brief ペイント用関数
      **/
      namespace paint
      {
        /// @brief 領域色による塗りつぶし処理
        ///
        /// 塗りつぶし開始座標上の色を領域色として塗りつぶす
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c 塗りつぶし開始座標
        void AreaPaint( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c );

        /// @brief 境界色による塗りつぶし処理
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c 塗りつぶし開始座標
        /// @param rgb 境界色
        void BorderPaint( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c, RGB rgb );
      }

    } // namespace draw

    namespace shape
    {
      /**
         @brief スキャン・ライン・コンバージョンによるポリゴン描画用クラス
      **/
      namespace polygon
      {
        /// @brief 描画領域の判定方法
        ///
        /// ポリゴンの辺が交差している場合の描画領域判定として、
        /// 偶奇判定 (Crossing Number Algorithm) と
        /// 巻き数判定 (Winding Number Algorithm)
        /// のいずれかを指定することができる
        enum METHOD {
                     CROSSING_NUMBER, ///< Crossing Number Algorithm
                     WINDING_NUMBER,  ///< Winding Number Algorithm
        };

        /// @brief ポリゴン描画関数
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp、頂点 vertex  が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param vertex 描画するポリゴンの頂点
        /// @param method 描画領域の判定方法
        void Fill( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const std::vector< Coord< int > >& vertex, METHOD method = CROSSING_NUMBER );
        /// @brief 三角形の塗りつぶし関数
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作領域
        /// @param pixelOp ピクセル操作関数オブジェクト
        /// @param c1,c2,c3 三角形の頂点
        void Triangle( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c1, Coord< int > c2, Coord< int > c3 );
      }

      /**
         @brief 曲線描画用関数
      **/
      namespace curve
      {
        /// @brief Ferguson/Coons曲線を描画する
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p、微分係数列 dp が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画領域
        /// @param pixelOp 描画処理関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param dp 各点に対する微分係数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeFergusonCoonsSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const std::vector< Coord< int > >& p,
                                        const std::vector< Coord< double > >& dp, double rate );

        /// @brief Catmull-Rom曲線を描画する
        ///
        /// p の最初と最後の要素は描画対象の点には含まれない(微分係数の計算のみに使用)
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画領域
        /// @param pixelOp 描画処理関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param d 微分係数計算時の補正係数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeCatmullRomSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const std::vector< Coord< int > >& p,
                                     double d, double rate );

        /// @brief 任意の次数のベジェ曲線によるスプライン曲線描画
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeBezierSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& p, size_t n, double rate );

        /// @brief 任意の次数のベジェ曲線によるスプライン曲線描画 ( De Casteljau のアルゴリズムを使用 )
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeDeCasteljauSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                      const std::vector< Coord< int > >& p, size_t n, double rate );

        /// @brief 任意のノット列を使った非一様 B-スプライン曲線描画
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p、ノット列 knot が未定義の場合は assert を実行する。
        ///
        /// ノット列 knot は点列 p のサイズ + 次数 + 1 以上のサイズでなければならない。この条件を満たしていない場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param knot ノット列
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeBSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                            const std::vector< Coord< int > >& p, const std::vector< double >& knot,
                            size_t n, double rate );

        /// @brief 任意のノット列を使った非一様 B-スプライン曲線描画 ( De Boor Cox のアルゴリズムを使用)
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// ノット列 knot は点列 p のサイズ + 次数 + 1 以上のサイズでなければならない。この条件を満たしていない場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param knot ノット列
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeDeBoorCox( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                              const std::vector< Coord< int > >& p, const std::vector< double >& knot,
                              size_t n, double rate );

        /// @brief 一様 B-スプライン曲線描画
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeUniformBSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                   const std::vector< Coord< int > >& p, size_t n, double rate );

        /// @brief 一様 B-スプライン曲線描画 ( De Boor Cox のアルゴリズムを使用)
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeUniformDeBoorCox( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                     const std::vector< Coord< int > >& p, size_t n, double rate );

        /// @brief 開一様 B-スプライン曲線描画
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeOpenUniformBSpline( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                       const std::vector< Coord< int > >& p, size_t n, double rate );

        /// @brief 開一様 B-スプライン曲線描画 ( De Boor Cox のアルゴリズムを使用)
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p が未定義の場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeOpenUniformDeBoorCox( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                         const std::vector< Coord< int > >& p, size_t n, double rate );

        /// @brief Nurbs 曲線描画
        ///
        /// ピクセル操作領域 draw、ピクセル操作関数オブジェクト pixelOp が未定義の場合は assert を実行する。
        ///
        /// 点列 p、ノット列 knot、重み値 w が未定義の場合は assert を実行する。
        ///
        /// ノット列 knot は点列 p のサイズ + 次数 + 1 以上のサイズでなければならない。この条件を満たしていない場合は assert を実行する。
        ///
        /// 内点の位置 rate は 1 より小さい正数である必要がある。そうでない場合は assert を実行する。
        ///
        /// @param draw 描画オブジェクト
        /// @param pixelOp 描画用関数オブジェクト
        /// @param p 描画する曲線上の点の座標
        /// @param knot ノット列
        /// @param w 重み値
        /// @param n 次数
        /// @param rate 再帰的に内点を描画するときの内点の位置(比率)
        void StrokeNurbs( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                          const std::vector< Coord< int > >& p, const std::vector< double >& knot, const std::vector< double >& w,
                          size_t n, double rate );

      } // namespace curve

    } // namespace shape

    namespace pen
    {
      /**
         @brief 楕円形のペン

         指定した座標 ( c.x, c.y ) を中心に半径 r、係数 ( k.x, k.y ) の楕円 k.x( x - c.x )^2 + k.y( y - c.y )^2 = r^2 の形の点を描画する
      **/
      class GEllipsePen : public GPixelOp
      {
        GPixelOp& pixelOp_; // ドット描画
        int r_;             // 半径
        Coord< int > k_;    // 係数

      public:

        /// @brief ドット描画用の関数オブジェクトと楕円のパラメータを指定して構築
        ///
        /// 関数オブジェクト pixelOp が未定義の場合は assert を実行する
        ///
        /// @param pixelOp ドット描画用の関数オブジェクト
        /// @param r 楕円の半径
        /// @param k 楕円の係数
        GEllipsePen( GPixelOp& pixelOp, int r, Coord< int > k )
          : pixelOp_( pixelOp )
        {
          setR( r );
          setK( k );
        }

        /// @brief 半径をセットする
        ///
        /// セットする半径 r が負数の場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param r セットする半径
        void setR( int r )
        { r_ = error_lib::NNNum( r ); }

        /// @brief セットされた半径を取得する
        ///
        /// @return セットされた半径
        int getR() const
        { return( r_ ); }

        /// @brief 係数をセットする
        ///
        /// セットする係数 k に負数が含まれる場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param k セットする係数
        void setK( Coord< int > k )
        { k_ = Coord< int >( error_lib::NNNum( k.x ), error_lib::NNNum( k.y ) ); }

        /// @brief セットされた係数を取得する
        ///
        /// @return セットされた係数
        Coord< int > getK() const
        { return( k_ ); }

        /// @brief 指定した座標を中心とした楕円形を描画する
        ///
        /// 描画対象 draw が未定義なら assert を実行する
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c 操作するピクセルの座標
        /// @return 常に true
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c )
        {
          shape::ellipse::Fill( draw, pixelOp_, c, r_, k_ );
          return( true );
        }
      };

      /**
         @brief 円形のブラシ

         指定した座標 ( c.x, c.y ) を中心に半径 r の円 ( x - c.x )^2 + ( y - c.y )^2 = r^2 の形の点を描画する。
         中心から離れるほど透明度が増加する。
      **/
      class GCircleBrush : public GPixelDraw
      {
        RGB& ( RGB::*mf_ )( RGB ); // 描画用関数オブジェクト
        RGB rgb_;                  // 描画色(アルファ・チャンネルは描画位置に応じて変更されることに注意)
        int r_;                    // 円形ブラシの半径
        int intensity_;            // アルファ・チャンネルの最大強度

      public:

        /// @brief 描画時の RGB メンバ関数、描画色、ブラシの半径・強度を指定して構築
        ///
        /// ブラシの半径 r、強度 intensity が負数なら 0 で初期化される。
        ///
        /// @param mf 描画時の RGB メンバ関数へのポインタ
        /// @param rgb 描画色
        /// @param r ブラシの半径
        /// @param intensity ブラシの強度
        GCircleBrush( RGB& ( RGB::*mf )( RGB ), RGB rgb, int r, int intensity )
          : mf_( mf ), rgb_( rgb ), r_( 0 ), intensity_( 0 )
        {
          if ( r > 0 ) setR( r );
          if ( intensity > 0 ) setIntensity( intensity );
        }

        /// @brief 描画時の RGB メンバ関数をセットする
        ///
        /// @param mf 描画時の RGB メンバ関数へのポインタ
        void setMemFun( RGB& ( RGB::*mf )( RGB ) )
        { mf_ = mf; }

        /// @brief 描画時の RGB コードをセットする
        ///
        /// @param rgb 描画時の RGB コード
        virtual void setRGB( RGB rgb )
        { rgb_ = rgb; }

        /// @brief 描画時の RGB コードを返す
        ///
        /// アルファチャンネルは常に RGB::MAX として返される
        ///
        /// @return 描画時の RGB コード
        virtual RGB getRGB() const
        { return( RGB( rgb_.r, rgb_.g, rgb_.b, RGB::MAX ) ); }

        /// @brief 半径をセットする
        ///
        /// ブラシの半径 r が負数なら例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param r セットする半径
        void setR( int r )
        { r_ = error_lib::NNNum( r ); }

        /// @brief セットされた半径を取得する
        ///
        /// @return セットされた半径
        int getR() const
        { return( r_ ); }

        /// @brief ブラシの強度をセットする
        ///
        /// ブラシの強度 intensity が負数なら例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param intensity ブラシの強度
        void setIntensity( int intensity )
        { intensity_ = error_lib::NNNum( intensity ); }

        /// @brief セットされたブラシの強度を返す
        ///
        /// @return セットされたブラシの強度
        int getIntensity() const
        { return( intensity_ ); }

        /// @brief 指定した座標を中心とした円形を描画する
        ///
        /// 中心から離れるほど透明度は増加し、端のぼやけた円が描画される。
        ///
        /// 描画対象 draw が未定義の場合は assert を実行する
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c 操作するピクセルの座標
        /// @return draw が未定義なら false を返す
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c );
      };

    } // namespace pen

  } // namespace gl

} // namespace tadah_fussy

#endif
