#include <vector>

#include "gl_g_interpolation.h"

namespace tadah_fussy
{
  namespace gl
  {
    using std::vector;
    using canvas::DrawingArea_IF;

    namespace interpolation
    {

      /**************************************************************
        Bilinear : 線形補間法(Bilinear Interpolation)
      **************************************************************/

      /*
        Bilinear::operator() : p の位置に対する色成分を線形補間法(Bilinear Interpolation)で補間して返す
      */
      bool Bilinear::operator()( const DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const
      {
        assert( rgb != 0 );

        Coord< int > c( p.x, p.y ); // pを含む格子の左上の座標(格子の左上なのでRoundで丸めてはいけない)
        RGB q11, q21, q12, q22;

        // 各格子の色コードを取得
        // p を含む格子が画像の範囲外だった場合は何もせずに false を返す
        if ( ! draw.point( c, &q11 ) ) return( false );
        // その他の格子が画像の範囲外だった場合は c 上のピクセルをコピーする
        if ( ! draw.point( Coord< int >( c.x + 1, c.y ), &q21 ) )
          q21 = q11;
        if ( ! draw.point( Coord< int >( c.x, c.y + 1 ), &q12 ) )
          q12 = q11;
        if ( ! draw.point( Coord< int >( c.x + 1, c.y + 1 ), &q22 ) )
          q22 = q11;

        // 格子の各頂点から p までの距離から合成比率を求める
        double rd = ( static_cast< double >( c.x + 1 ) - p.x ) * ( static_cast< double >( c.y + 1 ) - p.y );
        double ld = ( p.x - static_cast< double >( c.x ) ) * ( static_cast< double >( c.y + 1 ) - p.y );
        double ru = ( static_cast< double >( c.x + 1 ) - p.x ) * ( p.y - static_cast< double >( c.y ) );
        double lu = ( p.x - static_cast< double >( c.x ) ) * ( p.y - static_cast< double >( c.y ) );

        // 線形補間処理
        rgb->r = RGB::Check_RGBPrimary< double >
          ( std::round( rd * static_cast< double >( q11.r ) + ld * static_cast< double >( q21.r ) +
                            ru * static_cast< double >( q12.r ) + lu * static_cast< double >( q22.r ) ) );
        rgb->g = RGB::Check_RGBPrimary< double >
          ( std::round( rd * static_cast< double >( q11.g ) + ld * static_cast< double >( q21.g ) +
                            ru * static_cast< double >( q12.g ) + lu * static_cast< double >( q22.g ) ) );
        rgb->b = RGB::Check_RGBPrimary< double >
          ( std::round( rd * static_cast< double >( q11.b ) + ld * static_cast< double >( q21.b ) +
                            ru * static_cast< double >( q12.b ) + lu * static_cast< double >( q22.b ) ) );
        rgb->a = RGB::Check_RGBPrimary< double >
          ( std::round( rd * static_cast< double >( q11.a ) + ld * static_cast< double >( q21.a ) +
                            ru * static_cast< double >( q12.a ) + lu * static_cast< double >( q22.a ) ) );

        return( true );
      }

      /***************************************************************************************
        BicubicSpline : 三次スプライン補間法(Bicubic Spline Interpolation)
      ****************************************************************************************/

      // 係数行列の逆行列
      const double BicubicSpline::COEF_MAT[16][16] = {
                                                       {  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
                                                       {  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
                                                       { -3,  3,  0,  0, -2, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
                                                       {  2, -2,  0,  0,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
                                                       {  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0 },
                                                       {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0 },
                                                       {  0,  0,  0,  0,  0,  0,  0,  0,  3, -3,  0,  0, -2, -1,  0,  0 },
                                                       {  0,  0,  0,  0,  0,  0,  0,  0, -2,  2,  0,  0,  1,  1,  0,  0 },
                                                       { -3,  0,  3,  0,  0,  0,  0,  0,  0, -2, -1,  0,  0,  0,  0,  0 },
                                                       {  0,  0,  0,  0, -3,  0,  3,  0,  0,  0,  0,  0, -2,  0, -1,  0 },
                                                       {  9, -9, -9,  9,  6,  3, -6, -3, -6,  6,  3, -3,  4,  2,  2,  1 },
                                                       { -6,  6,  6, -6, -3, -3,  3,  3,  4, -4, -2,  2, -2, -2, -1, -1 },
                                                       {  2,  0, -2,  0,  0,  0,  0,  0,  0,  1,  1,  0,  0,  0,  0,  0 },
                                                       {  0,  0,  0,  0,  2,  0, -2,  0,  0,  0,  0,  0,  1,  0,  1,  0 },
                                                       { -6,  6,  6, -6, -4, -2,  4,  2,  3, -3, -3,  3, -2, -1, -2, -1 },
                                                       {  4, -4, -4,  4,  2,  2, -2, -2, -2,  2,  2, -2,  1,  1,  1,  1 },
      };

      /*
        BicubicSpline::operator() : p の位置に対する色成分を三次スプライン補間法(Bicubic Spline Interpolation)で補間して返す
      */
      bool BicubicSpline::operator()( const DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const
      {
        assert( rgb != 0 );

        Coord< int > c( p.x, p.y ); // pを含む格子の左上の座標(16点の中の左から二番目、上から二番目)
        Coord< double > un( p.x - static_cast< double >( c.x ),
                            p.y - static_cast< double >( c.y ) ); // cを原点としたpの座標
        Coord< double > sq( un.x * un.x, un.y * un.y ); // unの二乗
        Coord< double > cb( sq.x * un.x, sq.y * un.y ); // unの三乗
        RGB col[4][4];   // 16点の格子の色コード
        double b[16][4]; // 連立方程式の右辺
        double a[16][4]; // 連立方程式の未知数

        // 各格子の色コードを取得
        // p を含む格子が画像の範囲外だった場合は何もせずに false を返す
        if ( ! draw.point( Coord< int >( c.x, c.y ), &( col[1][1] ) ) )
          return( false );

        // その他の格子が画像の範囲外だった場合は p を含む格子上の画素から取得する
        for ( int y = -1 ; y < 3 ; ++y ) {
          for ( int x = -1 ; x < 3 ; ++x ) {
            if ( x == 0 && y == 0 ) continue;
            if ( ! draw.point( Coord< int >( c.x + x, c.y + y ), &( col[x+1][y+1] ) ) )
              col[x+1][y+1] = col[1][1];
          }
        }

        // 連立方程式の右辺を計算
        double sqrt8 = sqrt( 8 );
        for ( int y = 0 ; y < 2 ; ++y ) {
          for ( int x = 0 ; x < 2 ; ++x ) {
            // 各点の色コード
            const RGB& center = col[x+1][y+1];
            // 各点におけるx方向の微分係数
            const RGB& right = col[x+2][y+1];
            const RGB& left = col[x][y+1];
            // 各点におけるy方向の微分係数
            const RGB& up = col[x+1][y];
            const RGB& down = col[x+1][y+2];
            // 各点における混合微分係数
            const RGB& rd = col[x+2][y+2];
            const RGB& lu = col[x][y];
            for ( RGB::size_type ci = 0 ; ci < RGB::RGBA_SIZE ; ++ci ) {
              b[y*2+x][ci] = static_cast< double >( center[ci] );
              b[y*2+x+4][ci] = ( static_cast< double >( right[ci] ) -
                                 static_cast< double >( left[ci] ) ) / 2;
              b[y*2+x+8][ci] = ( static_cast< double >( down[ci] ) -
                                 static_cast< double >( up[ci] ) ) / 2;
              b[y*2+x+12][ci] = ( static_cast< double >( rd[ci] ) -
                                  static_cast< double >( lu[ci] ) ) / sqrt8;
            }
          }
        }

        // 連立方程式を解く(係数行列の逆行列との積)
        for ( int i = 0 ; i < 16 ; ++i ) {
          a[i][0] = a[i][1] = a[i][2] = a[i][3] = 0;
          for ( int j = 0 ; j < 16 ; ++j )
            for ( RGB::size_type k = 0 ; k < RGB::RGBA_SIZE ; ++k )
              a[i][k] += COEF_MAT[i][j] * b[j][k];
        }

        // 求めた解を三次式の係数として pでの値を計算
        for ( RGB::size_type ci = 0 ; ci < RGB::RGBA_SIZE ; ++ci ) {
          (*rgb)[ci] = RGB::Check_RGBPrimary< double >
            ( std::round( a[0][ci] + a[1][ci] * un.x + a[2][ci] * sq.x + a[3][ci] * cb.x +
                              a[4][ci] * un.y + a[5][ci] * un.x * un.y + a[6][ci] * sq.x * un.y + a[7][ci] * cb.x * un.y +
                              a[8][ci] * sq.y + a[9][ci] * un.x * sq.y + a[10][ci] * sq.x * sq.y + a[11][ci] * cb.x * sq.y +
                              a[12][ci] * cb.y + a[13][ci] * un.x * cb.y + a[14][ci] * sq.x * cb.y + a[15][ci] * cb.x * cb.y ) );
        }

        return( true );
      }

    } // namespace interpolation

    namespace convolution_kernel
    {
      /****************************************************
        Bicubic : 三次畳み込み補間関数
      *****************************************************/

      /*
        Bicubic::operator() : 三次畳み込み補間関数の t ± kh での値を求めて phai に代入する
      */
      void Bicubic::operator()( double t, vector< double >* phai ) const
      {
        // phai が必要なサイズを満たしていない場合は assert
        assert( phai->size() >= static_cast< vector< double >::size_type >( size() ) * 2 );

        double t2 = t * t;
        double t_1 = t - 1;

        // -a・t^2・(t - 1) = -at^3 + at^2
        (*phai)[0] = -a_ * t2 * t_1;
        // t[ ( -a - 2 )t^2 + ( 2a + 3 )t - a ] = -( a + 2 )t^3 + ( 2a + 3 )t^2 - at
        (*phai)[1] = t * ( ( -a_ - 2 ) * t2 + ( 2 * a_ + 3 ) * t - a_ );
        // ( a + 2 )t^2t - ( a + 3 )t^2 + 1 = ( a + 2 )t^3 - ( a + 3 )t^2 + 1
        (*phai)[2] = ( a_ + 2 ) * t2 * t - ( a_ + 3 ) * t2 + 1;
        // at( t - 1 )^2 = at^3 - 2at^2 + at
        (*phai)[3] = a_ * t * t_1 * t_1;
      }

      /*************************************************
        Lanczos : Lanczos補間関数
      **************************************************/

      /*
        Lanczos::operator() : Lanczos補間関数の t ± kh での値を求めて phai に代入する
      */
      void Lanczos::operator()( double t, vector< double >* phai ) const
      {
        // ループ内で sin を使わない方法
        // しかし、ResizePut で境界付近に縞模様が生じるため利用を断念
        /*
          double a = size();             // a = N
          double x = M_PI * ( t - a );   // x = π( t - N )
          double sin_x = a * std::sin( std::fmod( x, 2.0 * M_PI ) );   // sin_x = N・sin( x )
          double sin_x_a = std::sin( std::fmod( x / a, 2.0 * M_PI ) ); // sin_x_a = sin( x / N )
          double cos_x_a = std::cos( std::fmod( x / a, 2.0 * M_PI ) ); // cos_x_a = cos( x / N )
          for ( vector< double >::iterator it = phai->begin() ; it != phai->end() ; ++it ) {
          *it = ( x == 0 ) ?
          1 :
          ( sin_x * sin_x_a / ( x * x ) ); // N・sin( x )sin( x / N ) / x^2
          if ( *it > 1 ) *it = 1;
          x += M_PI;
          sin_x *= -1;
          double buff = sin_x_a;
          sin_x_a = sin_x_a * cosPi_n_ + cos_x_a * sinPi_n_; // sin( x / N + π / N ) の計算
          cos_x_a = cos_x_a * cosPi_n_ - buff * sinPi_n_;    // cos( x / N + π / N ) の計算
          }
        */
        ///*
        double a = size();
        double x = math_lib::Pi< double >() * ( t - a );
        for ( unsigned int i = 0 ; i < phai->size() ; ++i ) {
          (*phai)[i] = ( x == 0 ) ?
            1 :
            ( a * std::sin( x ) * std::sin( x / a ) / ( x * x ) );
          x += math_lib::Pi< double >();
        }
        //*/
      }

      /************************************************************
        Mitchell : Mitchell-Netravali補間関数
      *************************************************************/

      /*
        Mitchell コンストラクタ

        double b : パラメータ B
        double c : パラメータ C
      */
      Mitchell::Mitchell( double b, double c )
        : b_( b ), c_( c )
      {
        if ( b_ < 0 ) b_ = 0;
        if ( b_ > 1 ) b_ = 1;
        if ( c_ < 0 ) c_ = 0;
        if ( c_ > 1 ) c_ = 1;

        // 定数項をあらかじめ求める
        a00_ = 12 - 9 * b_ - 6 * c_;
        a01_ = -18 + 12 * b_ + 6 * c_;
        a02_ = 6 - 2 * b_;
        a10_ = -b_ - 6 * c_;
        a11_ = 6 * b_ + 30 * c_;
        a12_ = -12 * b_ - 48 * c_;
        a13_ = 8 * b_ + 24 * c_;
      }

      /*
        Mitchell::operator() : Mitchell-Netravali補間関数の t ± kh での値を求めて phai に代入する
      */
      void Mitchell::operator()( double t, vector< double >* phai ) const
      {
        double x = t - 2;
        for ( vector< double >::size_type i = 0 ; i < phai->size() ; ++i ) {
          double absX = fabs( x );
          if ( absX < 1 )
            (*phai)[i] = ( ( a00_ * absX + a01_ ) * x * x + a02_ ) / 6;
          else if ( absX < 2 )
            (*phai)[i] = ( ( ( a10_ * absX + a11_ ) * absX + a12_ ) * absX + a13_ ) / 6;
          else
            (*phai)[i] = 0;
          x += 1;
        }
      }

    } // namespace convolution_kernel

    namespace interpolation
    {
      /****************************************************
        Convolution : 畳み込みによる補間法
      *****************************************************/

      /*
        Convolution::convolution : 画素列 col の t の位置の色コードを畳み込みによる補間法で求める
      */
      RGB Convolution::convolution( const vector< RGB >& col, const vector< double >& phai ) const
      {
        unsigned int n = col.size(); // ピクセル幅

        // 畳み込み積分を行う
        double ans[] = { 0, 0, 0, 0 };
        for ( unsigned int i = n ; i > 0 ; --i ) {
          ans[0] += static_cast< double >( ( col[n-i] ).r ) * phai[i-1];
          ans[1] += static_cast< double >( ( col[n-i] ).g ) * phai[i-1];
          ans[2] += static_cast< double >( ( col[n-i] ).b ) * phai[i-1];
          ans[3] += static_cast< double >( ( col[n-i] ).a ) * phai[i-1];
        }

        return( RGB( RGB::Check_RGBPrimary< double >( std::round( ans[0] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[1] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[2] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[3] ) )
                     ) );
      }

      /*
        Convolution::operator() : 畳み込み補間により座標 p の色を求める
      */
      bool Convolution::operator()( const DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const
      {
        assert( rgb != 0 );

        int n = kernel_.size(); // 補間関数のカーネルサイズ(半分の値)
        assert( n > 0 );

        //Coord< int > c( std::round( p.x ), std::round( p.y ) );     // pを含む格子の左上の座標
        Coord< int > c( p.x, p.y );     // pを含む格子の左上の座標
        vector< RGB > col( n * 2 );     // X方向の色コード
        vector< RGB > res( n * 2 );     // Y方向の色コード
        vector< double > phai( n * 2 ); // 補間関数による計算結果

        // p を含む格子が画像の範囲外だった場合は何もせずに false を返す
        // このとき同時に rgb へ RGB 成分を取得する
        if ( ! draw.point( Coord< int >( c.x, c.y ), rgb ) )
          return( false );

        // 補間関数による値の計算
        kernel_( p.x - static_cast< double >( c.x ), &phai );

        for ( int y = 1 - n ; y <= n ; ++y ) {
          // 各格子の色コードを取得
          // 取得に失敗した場合は p を含む格子状の点を採用する
          for ( int x = 1 - n ; x <= n ; ++x ) {
            if ( ! draw.point( Coord< int >( c.x + x, c.y + y ), &( col[x+n-1] ) ) )
              col[x+n-1] = *rgb;
          }
          // X方向の畳み込み積分
          res[y+n-1] = convolution( col, phai );
        }

        // Y方向に畳み込み積分を行なって結果を返す
        kernel_( p.y - static_cast< double >( c.y ), &phai );
        *rgb = convolution( res, phai );

        return( true );
      }

      /*************************************************
        Polynomial : 多項式による補間法
      *************************************************/

      /*
        Polynomial::operator() : 多項式による補間を使い、座標 p の色を求める
      */
      bool Polynomial::operator()( const DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const
      {
        assert( rgb != 0 );

        Coord< int > c( p.x, p.y ); // pを含む格子の左上の座標
        vector< RGB > col( n_ );
        vector< RGB > res( n_ );

        // p を含む格子が画像の範囲外だった場合は何もせずに false を返す
        // このとき同時に rgb へ RGB 成分を取得する
        if ( ! draw.point( Coord< int >( c.x, c.y ), rgb ) )
          return( false );

        // 色コードの取得
        // 取得に失敗した場合は p を含む格子状の点を採用する
        int iy = 0;
        for ( int y = -( ( n_ - 1 ) / 2 ) ; y <= n_ / 2 ; ++y ) {
          int ix = 0;
          for ( int x = -( ( n_ - 1 ) / 2 ) ; x <= n_ / 2 ; ++x ) {
            if ( ! draw.point( Coord< int >( c.x + x, c.y + y ), &( col[ix] ) ) )
              col[ix] = *rgb;
            ++ix;
          }
          res[iy] = (*pf_)( col, p.x - static_cast< double >( c.x ) + static_cast< double >( n_ - 1 ) / 2 );
          ++iy;
        }

        *rgb = (*pf_)( res, p.y - static_cast< double >( c.y ) + static_cast< double >( n_ - 1 ) / 2 );

        return( true );
      }

      /******************************************************
       LagrangePolynomial : ラグランジュ多項式による補間法
      *******************************************************/

      /*
        LagrangePolynomial : Lagrange多項式補間により画素列 col の位置 t の色を求める
      */
      RGB LagrangePolynomial( const vector< RGB >& col, double t )
      {
        double ans[] = { 0, 0, 0, 0 };
        int n = col.size();

        for ( int j = 0 ; j < n ; ++j ) {
          double d = 1;
          for ( int i = 0 ; i < n ; ++i ) {
            if ( i == j ) continue;
            d *= ( t - static_cast< double >( i ) ) / static_cast< double >( j - i );
          }
          ans[0] += static_cast< double >( ( col[j] ).r ) * d;
          ans[1] += static_cast< double >( ( col[j] ).g ) * d;
          ans[2] += static_cast< double >( ( col[j] ).b ) * d;
          ans[3] += static_cast< double >( ( col[j] ).a ) * d;
        }

        return( RGB( RGB::Check_RGBPrimary< double >( std::round( ans[0] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[1] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[2] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[3] ) )
                     ) );
      }

        /******************************************************
          NewtonPolynomial : ニュートン多項式による補間法
        *******************************************************/

      /*
        NewtonPolynomial : Newton多項式補間により画素列 col の位置 t の色を求める
      */
      RGB NewtonPolynomial( const vector< RGB >& col, double t )
      {
        double ans[] = {
                        static_cast< double >( ( col[0] ).r ),
                        static_cast< double >( ( col[0] ).g ),
                        static_cast< double >( ( col[0] ).b ),
                        static_cast< double >( ( col[0] ).a ),
        };
        int n = col.size();

        // 差分商の初期化
        vector< double > diffR( n, 0 );
        vector< double > diffG( n, 0 );
        vector< double > diffB( n, 0 );
        vector< double > diffA( n, 0 );
        for ( int i = 0 ; i < n ; ++i ) {
          diffR[i] = ( col[i] ).r;
          diffG[i] = ( col[i] ).g;
          diffB[i] = ( col[i] ).b;
          diffA[i] = ( col[i] ).a;
        }

        double d = t;
        for ( int i = 1 ; i < n ; ++i ) {
          // 差分商の計算
          for ( int j = 0 ; j < n - i ; ++j ) {
            diffR[j] = ( diffR[j+1] - diffR[j] ) / static_cast< double >( i );
            diffG[j] = ( diffG[j+1] - diffG[j] ) / static_cast< double >( i );
            diffB[j] = ( diffB[j+1] - diffB[j] ) / static_cast< double >( i );
            diffA[j] = ( diffA[j+1] - diffA[j] ) / static_cast< double >( i );
          }
          ans[0] += diffR[0] * d;
          ans[1] += diffG[0] * d;
          ans[2] += diffB[0] * d;
          ans[3] += diffA[0] * d;
          d *= t - static_cast< double >( i );
        }

        return( RGB( RGB::Check_RGBPrimary< double >( std::round( ans[0] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[1] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[2] ) ),
                     RGB::Check_RGBPrimary< double >( std::round( ans[3] ) )
                     ) );
      }

    } // namespace interpolation

  } // namespace gl

} // namespace tadah_fussy
