/**
   @file gl_g_interpolation.h
   @brief サンプル補間定義用ヘッダ

   @author tadah_fussy
   @date 2018/02/25 DrawingMehod.h から分離
   @date 2018/03/08 アルファ・チャンネルへの対応
   @date 2021/02/25 ワーニングの出力に対応
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_INTERPOLATION_20210402_H
#define GRAPHIC_LIBRARY_INTERPOLATION_20210402_H

#include <common/math_library.h>

#include "gl_color.h"
#include "gl_drawing_area_if.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace interpolation
    {
      /**
         @brief サンプル補間用基底クラス
      **/
      class Base
      {
      public:

        /// @brief サンプル補間処理用関数
        ///
        /// @param draw 描画領域
        /// @param p 補間対象のピクセル座標
        /// @param rgb 補間した RGB 成分を返す変数へのポインタ
        /// @return draw のメンバ関数 canvas::DrawingArea_IF::point の返り値をそのまま返す ( 通常、RGB 成分が取得できたら true を返す )
        virtual bool operator()( const canvas::DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const = 0;

        /// @brief 仮想デストラクタ (何もしない)
        virtual ~Base() {}
      };

      /**
         @brief 最近傍法 (Nearest Neighbor) によるサンプル補間
      **/
      class NearestNeighbor : public Base
      {
      public:

        /// @brief 最近傍法によるサンプル補間処理
        ///
        /// 描画領域 draw、補間結果へのポインタ rgb が未定義の場合は assert を実行する
        ///
        /// @param draw 描画領域
        /// @param p 補間対象のピクセル座標
        /// @param rgb 補間した RGB 成分を返す変数へのポインタ
        /// @return draw のメンバ関数 canvas::DrawingArea_IF::point の返り値をそのまま返す ( 通常、RGB 成分が取得できたら true を返す )
        virtual bool operator()( const canvas::DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const
        {
          assert( rgb != 0 );

          return( draw.point( Coord< int32_t >( std::round( p.x ), std::round( p.y ) ), rgb ) );
        }
      };

      /**
         @brief 線形補間法 (Bilinear Interpolation) によるサンプル補間
      **/
      class Bilinear : public Base
      {
      public:

        /// @brief 線形補間法によるサンプル補間処理
        ///
        /// 描画領域 draw、補間結果へのポインタ rgb が未定義の場合は assert を実行する
        ///
        /// @param draw 描画領域
        /// @param p 補間対象のピクセル座標
        /// @param rgb 補間した RGB 成分を返す変数へのポインタ
        /// @return 座標 p を含むピクセルに対する draw のメンバ関数 canvas::DrawingArea_IF::point の返り値をそのまま返す ( 通常、RGB 成分が取得できたら true を返す )
        virtual bool operator()( const canvas::DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const;
      };

      /**
         @brief 三次スプライン補間法(Bicubic Spline Interpolation) によるサンプル補間

         4 x 4 の点からなる矩形に対して三次の多項式を当てはめる。各係数は、画素の値とその微分係数から求められる。
      **/
      class BicubicSpline : public Base
      {
        // 係数行列の逆行列
        static const double COEF_MAT[16][16];

      public:

        /// @brief 三次スプライン補間法によるサンプル補間処理
        ///
        /// 描画領域 draw、補間結果へのポインタ rgb が未定義の場合は assert を実行する
        ///
        /// @param draw 描画領域
        /// @param p 補間対象のピクセル座標
        /// @param rgb 補間した RGB 成分を返す変数へのポインタ
        /// @return draw のメンバ関数 canvas::DrawingArea_IF::point の返り値をそのまま返す ( 通常、RGB 成分が取得できたら true を返す )
        virtual bool operator()( const canvas::DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const;
      };

    } // namespace interpolation

    namespace convolution_kernel
    {
      /**
         @brief 畳み込み積分用補間関数基底クラス

         任意の補間関数を使い、位置 t を原点にゼロ以外の値を持つ点の値を計算する。
         計算する値は、t - N から t + N ( N は整数 ) までの位置であり、N は補間関数の種類により異なる。
      **/
      class Base
      {
      public:

        /// @brief 補間関数の値を計算する
        ///
        /// @param phai 求める値を保持する配列への参照
        /// @param t 変数 t
        virtual void operator()( double t, std::vector< double >* phai ) const = 0;

        /// @brief ゼロ以外の値を持つピクセル幅を返す
        ///
        /// 実際には中心からの距離(半分の値)を返す
        ///
        /// @return ゼロ以外の値を持つピクセルの幅(実際には中心からの距離)
        virtual int size() const = 0;

        // 仮想デストラクタ
        virtual ~Base() {}
      };

      /**
         @brief 三次畳み込み補間関数

         下式で表される三次畳み込み補間関数を計算するための関数オブジェクト。
         値を持つ範囲は t - 2 から t + 2 までとなる。

         <pre>
         F(t) = ( a + 2 )|t|^3 - ( a + 3 )|t|^2 + 1 [ 0 < |t| < 1 ]
         = a|t|^3 - 5a|t|^2 + 8a|t| - 4a       [ 1 < |t| < 2 ]
         = 0                                   [ |t| > 2 ]
         </pre>
      **/
      class Bicubic : public Base
      {
        double a_; // 三次式の係数を決定するパラメータ

      public:

        /* コンストラクタ */

        /// @brief 三次畳み込み補間関数用コンストラクタ
        ///
        /// @param a 三次式の係数を決定するパラメータ
        Bicubic( double a = -0.5 )
          : a_( a ) {}

        /// @brief 三次畳み込み補間関数の t - 1 から t + 2 までの値を求める
        ///
        /// 求めた値を保持する配列 phai は、あらかじめサイズを 4 ( size() の 2 倍 )にしておく必要がある。そうでない場合は assert を実行する。
        ///
        /// @param phai 求めた値を保持する配列への参照
        /// @param t 変数 t
        virtual void operator()( double t, std::vector< double >* phai ) const;

        /// @brief ゼロ以外の値を持つピクセルの幅を返す
        ///
        /// 値を持つ範囲は - 2 <= t <= 2 なので 2 固定
        ///
        /// @return ゼロ以外の値を持つピクセルの幅
        virtual int size() const { return( 2 ); }
      };

      /**
         @brief Lanczos 補間関数

         下式で表されるランツォシュ ( Lanczos ) 補間関数を計算するための関数オブジェクト。
         値を持つ範囲は t - a から t + a までとなる。

         <pre>
         L(t) = sinc(t)sinc(t/a) [ |t| <= a ]
         = 0                [ |t| > a ]
         </pre>
      **/
      class Lanczos : public Base
      {
        int32_t n_; // ゼロでない値を持つ範囲
        //sin( πx / n ) の計算に利用する変数
        //double sinPi_n_; // sin( π / n )
        //double cosPi_n_; // cos( π / n )

      public:

        /* コンストラクタ */

        /// @brief 関数値がゼロでない範囲を指定して構築
        ///
        /// @param n 関数値がゼロでない定義域 ( -n <= t <= n の形で表される )
        Lanczos( int32_t n = 2 )
          : n_( ( n <= 0 ) ? 1 : n )
        {
          //sinPi_n_ = std::sin( MathLib::Pi< double >() / static_cast< double >( n_ ) );
          //cosPi_n_ = std::cos( MathLib::Pi< double >() / static_cast< double >( n_ ) );
        }

        /// @brief Lanczos補間関数の t ± n での値を求める
        ///
        /// 求められる範囲は phai のサイズに依存する ( 最低 2n のサイズは必要 )
        ///
        /// @param phai 求めた値を保持する配列への参照
        /// @param t 変数 t
        virtual void operator()( double t, std::vector< double >* phai ) const;

        /// @brief ゼロ以外の値を持つピクセルの幅を返す
        ///
        /// 値を持つ範囲は -n <= t <= n なので n 固定
        ///
        /// @return ゼロ以外の値を持つピクセルの幅
        virtual int size() const { return( n_ ); }
      };

      /**
         @brief Mitchell-Netravali 補間関数

         下式で表されるミッチェル ( Mitchell-Netravali ) 補間関数を計算するための関数オブジェクト。
         値を持つ範囲は t - 2 から t + 2 までとなる。

         <pre>
         k(t) = (1/6)( ( 12 - 9B - 6C )|t|3 + ( -18 + 12B + 6C )|t|2 + ( 6 - 2B ) )            [ |t| < 1 ]
         = (1/6)( ( -B - 6C )|t|3 + ( 6B + 30C )|t|2 + ( -12B - 48C )|t| + ( 8B + 24C ) ) [ 1 <= |t| < 2 ]
         = 0                                                                              [ |t| >= 2 ]
         </pre>
      **/
      class Mitchell : public Base
      {
        double b_; // パラメータ B
        double c_; // パラメータ C

        // 計算に使う係数
        double a00_;
        double a01_;
        double a02_;
        double a10_;
        double a11_;
        double a12_;
        double a13_;

      public:

        /* コンストラクタ */

        /// @brief 関数の二つのパラメータを指定して構築
        ///
        /// @param b パラメータ B
        /// @param c パラメータ C
        Mitchell( double b = 1 / 3, double c = 1 / 3 );

        /// @brief Mitchell-Netravali 補間関数の t ± n での値を求める
        ///
        /// 求められる範囲は phai のサイズに依存する ( 最低サイズ 4 は必要 )
        ///
        /// @param phai 求めた値を保持する配列への参照
        /// @param t 変数 t
        virtual void operator()( double t, std::vector< double >* phai ) const;

        /// @brief ゼロ以外の値を持つピクセルの幅を返す
        ///
        /// 値を持つ範囲は - 2 <= t <= 2 なので 2 固定
        ///
        /// @return ゼロ以外の値を持つピクセルの幅
        virtual int size() const { return( 2 ); }
      };

    } // namespace convolution_kernel

    namespace interpolation
    {
      /**
         @brief 補間関数を利用した畳み込み積分による補間法

         ConvolutionKernel_IF を補間関数として畳み込み積分を使って位置 p の色コードを計算する。
      **/
      class Convolution : public Base
      {
        const convolution_kernel::Base& kernel_; // 利用する補間関数

        /*
          convolution : 畳み込みによる補間法メインルーチン

          const std::vector<RGB>& col : 色コード
          const vector< double >& phai : 補間関数により求めた重み値

          戻り値 : 求めた色コード
        */
        RGB convolution( const std::vector< RGB >& col, const std::vector< double >& phai ) const;

      public:

        /* コンストラクタ */

        /// @brief 補間関数を指定して構築
        ///
        /// @param kernel 利用する補間関数
        explicit Convolution( const convolution_kernel::Base& kernel )
          : kernel_( kernel ) {}

        /// @brief 畳み込み積分によるサンプル補間処理
        ///
        /// 描画領域 draw、補間関数 kernel、補間結果へのポインタ rgb が未定義の場合は assert を実行する
        ///
        /// @param draw 描画領域
        /// @param p 補間対象のピクセル座標
        /// @param rgb 求めた色コードを返す変数へのポインタ
        /// @return draw のメンバ関数 canvas::DrawingArea_IF::point の返り値をそのまま返す ( 通常、RGB 成分が取得できたら true を返す )
        virtual bool operator()( const canvas::DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const;
      };

      /**
         @brief 多項式による補間法

         指定した範囲のピクセルの色コードを通る多項式を求めて補間処理を行う。

         interpolation::LagrangePolynomial と interpolation::NewtonPolynomial の利用を想定している。
         次のコードで、Lagrange 多項式を利用した補間をしながら、パターン pattern を描画領域 draw に c0 - c1 の大きさで拡大・縮小貼り付けすることができる。 

         @code
         interpolation::Polynomial inter( 5, interpolation::LagrangePolynomial );
         draw::ResizePut( pattern, draw, c0, c1, false, false, false, inter );
         @endcode
      **/
      class Polynomial : public Base
      {
      public:

        /// @brief 多項式関数の型
        using PolynomialFunc = RGB ( * )( const std::vector< RGB >& col, double t );

      private:

        int32_t n_;             // 多項式関数の係数を求めるための変数の範囲
        PolynomialFunc pf_; // 利用する多項式関数

      public:

        /* コンストラクタ */

        /// @brief 利用する多項式を指定して構築
        ///
        /// 変数の範囲 n は 2 以上でなければならない。2 より小さい場合は assert を実行する。
        ///
        /// @param n 多項式関数の係数を求めるための変数の範囲
        /// @param pf 利用する補間関数
        Polynomial( uint32_t n, PolynomialFunc pf )
          : n_( n ), pf_( pf ) { assert( n_ >= 2 ); }

        /// @brief 多項式によるサンプル補間処理
        ///
        /// 描画領域 draw、補間関数 pf、補間結果へのポインタ rgb が未定義の場合は assert を実行する
        ///
        /// @param draw 描画領域
        /// @param p 補間対象のピクセル座標
        /// @param rgb 求めた色コードを返す変数へのポインタ
        /// @return draw のメンバ関数 canvas::DrawingArea_IF::point の返り値をそのまま返す ( 通常、RGB 成分が取得できたら true を返す )
        virtual bool operator()( const canvas::DrawingArea_IF& draw, Coord< double > p, RGB* rgb ) const;
      };

      /**
         @brief サンプル補間法関数
      **/
      /// @brief Lagrange 多項式で補間した色を求める
      ///
      /// @param col サンプル点の色コード
      /// @param t 色を求める位置
      /// @return 求めた色コード
      RGB LagrangePolynomial( const std::vector< RGB >& col, double t );

      /// @brief Newton 多項式で補間した色を求める
      ///
      /// @param col サンプル点の色コード
      /// @param t 色を求める位置
      /// @return 求めた色コード
      RGB NewtonPolynomial( const std::vector< RGB >& col, double t );

    } // namespace interpolation

  } // namespace gl

} // namespace tadah_fussy

#endif
