#include "gl_g_pattern.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace canvas
    {

      using error_lib::NNNum;

      /**************************************************************
  GPattern : 配列による描画領域の表現
      **************************************************************/

      /*
        GPattern::GPattern コンストラクタ

        サイズ sz のみをしてして構築
        各要素は RGB 成分のデフォルト値で初期化される
      */
      GPattern::GPattern( Coord< int > sz )
        : patSize_( Coord< int >( NNNum( sz.x ), NNNum( sz.y ) ) )
      {
        if ( patSize_.x > 0 && patSize_.y > 0 )
          pattern_.assign( patSize_.y, patSize_.x );
      }

      /*
        GPattern::GPattern コンストラクタ

        対象の描画オブジェクト draw と取得する範囲 s - e を指定して構築
      */
      void GPattern::copy( const DrawingArea_IF& draw, Coord< int > s, Coord< int > e )
      {
        // 座標を左上・右下に揃える
        if ( s.x > e.x ) std::swap( s.x, e.x );
        if ( s.y > e.y ) std::swap( s.y, e.y );

        // パターンサイズの取得とパターンの初期化
        patSize_.x = e.x - s.x + 1;
        patSize_.y = e.y - s.y + 1;
        pattern_.assign( patSize_.y, patSize_.x );

        /*
          オフセットと描画の終端の取得

          s=-4     e=5
          <----++++++>       pattern.size=10
          |pOffset=4
          <++++++----->  draw.size=11
          |dOffset=0
          <--6->

          10
          s=4   |  e=13
          <+++++++---> pattern.size=10
          |pOffset=0
          <----+++++++>    draw.size=11
          |dOffset=4
          <--7-->

          5
          s=-4     |    e=10
          <----++++++-----> pattern.size=15
          |pOffset=4
          <++++++>      draw.size=6
          |dOffset=0
          <--6->

          s=4   e=10
          <+++++++>      pattern.size=7
          |pOffset=0
          <----+++++++-----> draw.size=16
          |dOffset=4
          <--7-->
        */
        Coord< int > dOffset( 0, 0 ); // draw 側の取得開始位置
        Coord< int > pOffset( 0, 0 ); // パターン側の描画開始位置
        if ( e.x < 0 || s.x >= draw.size().x || e.y < 0 || s.y >= draw.size().y ) return;
        if ( s.x < 0 )
          pOffset.x = -s.x;
        else
          dOffset.x = s.x;
        if ( s.y < 0 )
          pOffset.y = -s.y;
        else
          dOffset.y = s.y;
        if ( e.x >= draw.size().x )
          e.x = draw.size().x - 1;
        if ( e.y >= draw.size().y )
          e.y = draw.size().y - 1;

        // パターンの取り込み
        RGB rgb;
        for ( Coord< int > d( dOffset ), p( pOffset ) ; d.y <= e.y ; ++( d.y ), ++( p.y ) ) {
          for ( d.x = dOffset.x , p.x = pOffset.x ; d.x <= e.x ; ++( d.x ) , ++ ( p.x ) ) {
            draw.point( d, &( pattern_[p.y][p.x] ) );
          }
        }
      }

      /*
        GPattern::point : 座標 c 上の RGB コードを rgb に取得する
      */
      bool GPattern::point( Coord< int > c, RGB* rgb ) const
      {
        assert( rgb != 0 );

        if ( size().x == 0 || size().y == 0 ) return( false );

        bool ans = true; // 戻り値

        // 指定した座標がパターン外部の場合、鏡像を取得する。
        // 例えば x = -10 のときは x = 10 が取得対象となる。
        // 但し、鏡像もはみ出る場合は端のピクセルを対象とする。
        // 上記の例の場合、サイズが 5 であれば x = 4 となる。
        if ( c.x < 0 || c.x >= size().x || c.y < 0 || c.y >= size().y ) {
          ans = false;
          if ( c.x < 0 ) {
            c.x = -( c.x );
            if ( c.x >= size().x ) c.x = size().x - 1;
          } else if ( c.x >= size().x ) {
            c.x = 2 * ( size().x - 1 ) - c.x;
            if ( c.x < 0 ) c.x = 0;
          }
          if ( c.y < 0 ) {
            c.y = -( c.y );
            if ( c.y >= size().y ) c.y = size().y - 1;
          } else if ( c.y >= size().y ) {
            c.y = 2 * ( size().y - 1 ) - c.y;
            if ( c.y < 0 ) c.y = 0;
          }
        }

        *rgb = pattern_[c.y][c.x];

        return( ans );
      }

      /*
        GPattern::pset : 座標 c 上を RGB コード rgb で描画する
      */
      bool GPattern::pset( Coord< int > c, RGB rgb )
      {
        if ( c.x < 0 || c.x >= size().x || c.y < 0 || c.y >=size().y )
          return( false );

        pattern_[c.y][c.x] = rgb;

        return( true );
      }

      /*
        GPattern::resize : 画像サイズを sz に変更する

        サイズを大きくする場合は、拡大されたエリアは RGB のデフォルト値で初期化される
      */
      void GPattern::resize( Coord< int > sz )
      {
        NNNum( sz.x );
        NNNum( sz.y );

        if ( size().x == sz.x && size().y == sz.y ) return;

        pattern_.resize( sz.y, sz.x );
        patSize_ = sz;
      }

    } // namespace canvas

  } // namespace gl

} // namespace tadah_fussy
