/**
   @file gl_g_pattern.h
   @brief 描画領域クラス定義用ヘッダ

   @author tadah_fussy
   @date 2018/02/24 gl_drawing_area_if.h から分離
   @date 2021/02/25 ワーニングの出力に対応
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_G_PATTERN_20210402_H
#define GRAPHIC_LIBRARY_G_PATTERN_20210402_H

#include <common/matrix.h>
#include <common/error_library.h>

#include "gl_drawing_area_if.h"
#include "gl_color.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace canvas
    {
      /**
         @brief 配列による描画領域の表現

         配列内に画像データを保持する。ディスプレイ上に画像を表示することはこのクラスだけではできない。
      **/
      class GPattern : public DrawingArea_IF
      {
        Matrix< RGB > pattern_; // 矩形パターン
        Coord< int > patSize_;  // 矩形パターンのサイズ

      public:

        /* コンストラクタ */

        /// @brief デフォルトコンストラクタ
        ///
        /// 描画領域の大きさはゼロで初期化される
        GPattern() : patSize_( 0, 0 ) {}

        /// @brief 描画領域の大きさを指定して構築
        ///
        /// パターンの大きさ sz に負数が含まれていた場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param sz パターンの大きさ
        explicit GPattern( Coord< int > sz );

        /// @brief 他の描画領域 draw からコピーして構築
        ///
        /// 描画領域 draw が未定義だった場合は assert を実行する。
        /// コピーする範囲 s, e が draw の領域外だった場合は、領域内になるようにクリッピングしてコピーする ( サイズも変化する )。
        ///
        /// @param draw パターンを取得する描画オブジェクト
        /// @param s,e パターンの範囲
        GPattern( const DrawingArea_IF& draw, Coord< int > s, Coord< int > e )
          : patSize_( Coord< int >( 0, 0 ) )
        { copy( draw, s, e ); }

        /* パターンのコピー */

        /// @brief 他の描画領域 d からコピーする
        ///
        /// もし、描画領域 d が存在しなければ何もせずに終了する。
        /// コピーする範囲 s, e が d の領域外だった場合は、領域内になるようにクリッピングしてコピーする ( サイズも変化する )
        ///
        /// 描画領域から取得した RGB コードは透明度を含めてそのままコピーされる。重ねあわせ等は行わないことに注意。
        ///
        /// @param draw パターンを取得する描画オブジェクト
        /// @param s,e パターンの範囲
        void copy( const DrawingArea_IF& draw, Coord< int > s, Coord< int > e );

        /* パターン内の色コードの取得と設定(override) */

        /// @brief 指定した座標の RGB コードを取得する
        ///
        /// 描画領域の x, y 方向のサイズのいずれかがゼロだった場合は RGB コードは取得せずに false を返す。
        ///
        /// 指定した座標がパターン外部の場合は鏡像を取得する。例えば x = -10 のときは x = 10 が取得対象となる。
        /// 但し、鏡像もはみ出る場合は端のピクセルを対象とする。先の例の場合、サイズが 5 であれば x = 4 となる。
        /// なお、指定した座標がパターン外部の場合は戻り値が false になる。
        ///
        /// 変数 rgb が未定義だった場合は assert を実行する。
        ///
        /// @param c RGB コードを取得する位置
        /// @param rgb 取得した RGB コードを返す変数へのポインタ
        ///
        /// @return 指定した座標がパターンの範囲内なら true を返す
        virtual bool point( Coord< int > c, RGB* rgb ) const;

        /// @brief 指定した座標のピクセルに RGB コードをセットする
        ///
        /// RGB コードは透明度を含めてそのままセットされる。重ねあわせ等は行わないことに注意。
        ///
        /// @param c RGB コードをセットする座標値
        /// @param rgb セットする RGB コード
        ///
        /// @return ピクセルに RGB コードをセットできたら true を返す
        virtual bool pset( Coord< int > c, RGB rgb );

        /* パターン内のサイズの取得と設定(override) */

        virtual Coord< int > size() const { return( patSize_ ); }

        /// @brief 画像サイズを変更する
        ///
        /// サイズを大きくする場合は、拡大されたエリアは RGB のデフォルト値で初期化される。
        ///
        /// パターンの大きさ sz に負数が含まれていた場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param sz 変更後のサイズ
        virtual void resize( Coord< int > sz );
      };

    } // namespace canvas

  } // namespace gl

} // namespace tadah_fussy
#endif
