#include "gl_g_pattern_op.h"
#include <set>

namespace tadah_fussy
{
  namespace gl
  {
    namespace draw
    {
      using std::vector;
      using std::round;
      using coord_lib::Diff;
      using coord_lib::Sign;

      /**************************************************************
          GPatternOp : 矩形パターンの処理関数群
      **************************************************************/

      namespace
      {
        /*
          iterator_base : ループ制御用反復子基底クラス
        */
        class iterator_base
        {
        protected:

          int current_; // 現在値
          int limit_;   // 終了値

        public:

          /*
            set : 反復子の始点 i1 と終点 i2 をセットする(純粋仮想関数)
          */
          virtual void set( int i1, int i2 ) = 0;

          /*
            next : 次の値へ移動する(純粋仮想関数)
          */
          virtual void next() = 0;

          /*
            valid : 現在値が終了値の範囲内なら true を返す(純粋仮想関数)
          */
          virtual bool valid() const = 0;

          /*
            value : 現在値を返す
          */
          virtual int value() const { return( current_ ); }
        };

        /*
          iterator : 通常ループ制御用反復子クラス
        */
        class iterator : public iterator_base
        {
        public:

          /*
            set : 反復子の始点 i1 と終点 i2 をセットする
          */
          virtual void set( int i1, int i2 )
          {
            if ( i1 < i2 ) {
              current_ = i1;
              limit_ = i2;
            } else {
              current_ = i2;
              limit_ = i1;
            }
          }

          /*
            next : 次の値へ移動する
          */
          virtual void next() { ++current_; }

          /*
            valid : 現在値が終了値の範囲内なら true を返す
          */
          virtual bool valid() const { return( current_ <= limit_ ); }
        };

        /*
          reverse_iterator : 反転ループ制御用反復子クラス
        */
        class reverse_iterator : public iterator_base
        {
        public:

          /*
            set : 反復子の始点 i1 と終点 i2 をセットする
          */
          virtual void set( int i1, int i2 )
          {
            if ( i1 > i2 ) {
              current_ = i1;
              limit_ = i2;
            } else {
              current_ = i2;
              limit_ = i1;
            }
          }

          /*
            next : 次の値へ移動する
          */
          virtual void next() { --current_; }

          /*
            valid : 現在値が終了値の範囲内なら true を返す
          */
          virtual bool valid() const { return( current_ >= limit_ ); }
        };

      } // namespace

      /*
        Put : パターン pattern を描画領域 draw 上の点 s を始点として描画する。

        x,y は X,Y それぞれの描画方向を表す。
        swapXY が true なら X,Y を入れ替える。
      */
      void Put( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw, Coord< int >& s,
                iterator_base& x, iterator_base& y, bool reverseX, bool reverseY, bool swapXY )
      {
        if ( &pattern == &draw ) return;

        // パターンと描画領域の大きさを取得
        Coord< int > pSize = pattern.size();
        if ( pSize.x == 0 || pSize.y == 0 ) return;
        Coord< int > dSize = draw.size();
        if ( dSize.x == 0 || dSize.y == 0 ) return;

        // パターンサイズを読み込んで右下の座標を決定
        Coord< int > e( s.x + pSize.x - 1, s.y + pSize.y - 1 );

        // X,Y を入れ替える場合、始点と大きさをあらかじめ入れ替えておく必要がある
        if ( swapXY ) {
          Coord< int > rSize( e.x - s.x, e.y - s.y );
          std::swap( s.x, s.y );
          e.x = s.x + rSize.x;
          e.y = s.y + rSize.y;
          std::swap( dSize.x, dSize.y );
        }

        // 描画開始位置はクリッピングエリア範囲内か
        if ( s.x >= dSize.x || s.y >= dSize.y ) return;

        // パターンはクリッピングエリア範囲内か
        if ( e.x < 0 || e.y < 0 ) return;

        // パターンの読み込み開始位置
        Coord< int > p( 0, 0 );

        // クリッピング処理
        if ( s.x < 0 ) {
          if ( ! reverseX )
            p.x = -s.x;
          s.x = 0;
        }
        if ( s.y < 0 ) {
          if ( ! reverseY )
            p.y = -s.y;
          s.y = 0;
        }
        if ( e.x >= dSize.x ) {
          if ( reverseX )
            p.x += e.x - dSize.x;
          e.x = dSize.x - 1;
        }
        if ( e.y >= dSize.y ) {
          if ( reverseY )
            p.y += e.y - dSize.y;
          e.y = dSize.y - 1;
        }

        // パターンの描画
        RGB rgb;
        for ( y.set( s.y, e.y ) ; y.valid() ; y.next() ) {
          int px = p.x;
          for ( x.set( s.x, e.x ) ; x.valid() ; x.next() ) {
            if ( ! pattern.point( Coord< int >( px++, p.y ), &rgb ) )
              continue;
            pixelDraw.setRGB( rgb );
            if ( swapXY )
              pixelDraw( draw, Coord< int >( y.value(), x.value() ) );
            else
              pixelDraw( draw, Coord< int >( x.value(), y.value() ) );
          }
          ++( p.y );
        }
      }

      /*
        Put : パターン pattern を描画領域 draw 上の点 s を始点として描画する。

        reverseX が true なら X 方向に反転する
        reverseY が true なら Y 方向に反転する
        swapXY が true なら X,Y を入れ替える
      */
      void Put( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw, Coord< int > s,
                bool reverseX, bool reverseY, bool swapXY )
      {
        if ( ! reverseX ) {
          iterator x;
          if ( ! reverseY ) {
            iterator y;
            Put( pattern, draw, pixelDraw, s, x, y, reverseX, reverseY, swapXY );
          } else {
            reverse_iterator y;
            Put( pattern, draw, pixelDraw, s, x, y, reverseX, reverseY, swapXY );
          }
        } else {
          reverse_iterator x;
          if ( ! reverseY ) {
            iterator y;
            Put( pattern, draw, pixelDraw, s, x, y, reverseX, reverseY, swapXY );
          } else {
            reverse_iterator y;
            Put( pattern, draw, pixelDraw, s, x, y, reverseX, reverseY, swapXY );
          }
        }
      }

      /*
        ResizePut : パターン pattern を描画領域 draw 上の範囲 s - e に拡大・縮小して描画する。

        x,y は X,Y それぞれの描画方向を表す。
        swapXY が true なら X,Y を入れ替える。
        ipF を使い補間処理を行う。
      */
      void ResizePut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                      Coord< int >& s, Coord< int >& e,
                      iterator_base& x, iterator_base& y,
                      bool reverseX, bool reverseY, bool swapXY,
                      const interpolation::Base& ipF )
      {
        if ( &pattern == &draw ) return;

        // パターンと描画領域の大きさを取得
        Coord< int > pSize = pattern.size();
        if ( pSize.x == 0 || pSize.y == 0 ) return;
        Coord< int > dSize = draw.size();
        if ( dSize.x == 0 || dSize.y == 0 ) return;

        // 左上・右下の大小関係が逆の場合は入れ替える
        if ( s.x > e.x ) std::swap( s.x, e.x );
        if ( s.y > e.y ) std::swap( s.y, e.y );

        // X,Y を入れ替える場合、始点と大きさをあらかじめ入れ替えておく必要がある
        if ( swapXY ) {
          Coord< int > rSize( e.x - s.x, e.y - s.y );
          std::swap( s.x, s.y );
          e.x = s.x + rSize.x;
          e.y = s.y + rSize.y;
          std::swap( dSize.x, dSize.y );
        }

        // 描画範囲内か
        if ( s.x >= dSize.x || s.y >= dSize.y ) return;
        if ( e.x < 0 || e.y < 0 ) return;

        // パターンの読み込み位置
        Coord< double > p( 0, 0 );
        // 画像 1ピクセルあたりのパターンのピクセルサイズ
        Coord< double > d;
        d.x = static_cast< double >( pSize.x ) / static_cast< double >( e.x - s.x + 1 );
        d.y = static_cast< double >( pSize.y ) / static_cast< double >( e.y - s.y + 1 );

        /*
          クリッピング処理
        */
        if ( s.x < 0 ) {
          if ( ! reverseX )
            p.x -= static_cast< double >( s.x ) * d.x;
          s.x = 0;
        }
        if ( s.y < 0 ) {
          if ( ! reverseY )
            p.y -= static_cast< double >( s.y ) * d.y;
          s.y = 0;
        }
        if ( e.x >= dSize.x ) {
          if ( reverseX )
            p.x += static_cast< double >( e.x - dSize.x ) * d.x;
          e.x = dSize.x - 1;
        }
        if ( e.y >= dSize.y ) {
          if ( reverseY )
            p.y += static_cast< double >( e.y - dSize.y ) * d.y;
          e.y = dSize.y - 1;
        }

        // 描画処理
        RGB rgb;
        for ( y.set( s.y, e.y ) ; y.valid() ; y.next() ) {
          double px = p.x;
          for ( x.set( s.x, e.x ) ; x.valid() ; x.next() ) {
            if ( ipF( pattern, Coord< double >( px, p.y ), &rgb ) ) {
              pixelDraw.setRGB( rgb );
              if ( swapXY )
                pixelDraw( draw, Coord< int >( y.value(), x.value() ) );
              else
                pixelDraw( draw, Coord< int >( x.value(), y.value() ) );
            }
            px += d.x;
          }
          p.y += d.y;
        }
      }

      /*
        ResizePut : パターン pattern を描画領域 draw 上の範囲 s - e に拡大・縮小して描画する。

        reverseX が true なら X 方向に反転する
        reverseY が true なら Y 方向に反転する
        swapXY が true なら X,Y を入れ替える
        ipF を使い補間処理を行う。
      */
      void ResizePut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                      Coord< int > s, Coord< int > e,
                      bool reverseX, bool reverseY, bool swapXY,
                      const interpolation::Base& ipF )
      {
        if ( ! reverseX ) {
          iterator x;
          if ( ! reverseY ) {
            iterator y;
            ResizePut( pattern, draw, pixelDraw, s, e, x, y, reverseX, reverseY, swapXY, ipF );
          } else {
            reverse_iterator y;
            ResizePut( pattern, draw, pixelDraw, s, e, x, y, reverseX, reverseY, swapXY, ipF );
          }
        } else {
          reverse_iterator x;
          if ( ! reverseY ) {
            iterator y;
            ResizePut( pattern, draw, pixelDraw, s, e, x, y, reverseX, reverseY, swapXY, ipF );
          } else {
            reverse_iterator y;
            ResizePut( pattern, draw, pixelDraw, s, e, x, y, reverseX, reverseY, swapXY, ipF );
          }
        }
      }

      namespace
      {
        /*
          自由変形用パラメータ
        */
        struct TransputParam {
          double len;              // 描画領域上の線分の長さ(x,y成分の絶対値のうち大きい側)
          double patternAdd;       // パターンの次の座標への増分
          Coord< double > drawAdd; // 描画領域の増分

          /*
            TransputParam コンストラクタ : 線分 s - e に対するパラメータを計算する(自由変形処理専用)

            patSize はパターンの X または Y 成分のサイズを表す。
          */
          TransputParam( const Coord< double >& s, const Coord< double >& e, int patSize )
          {
            Coord< double > diff; // 差分
            Coord< int8_t > sign; // 向き

	    coord_lib::Diff( s, e, &diff );
	    coord_lib::Sign( s, e, &sign );

            len = std::max( diff.x, diff.y ); // diff.x,diff.yの絶対値のうち大きい方を登録
            patternAdd = static_cast< double >( patSize - 1 );
            drawAdd = Coord< double >( diff.x * sign.x,
                                       diff.y * sign.y );
          }
        };

        /*
          TransPut_SetCache : キャッシュ cache に座標 d が登録されているかをチェックし、されていなければ新たに登録する(自由変形処理専用)

          キャッシュに登録されていた場合は false を返す
        */
        bool TransPut_SetCache( const Coord< int >& d, std::map< int, std::set< int > >* cache )
        {
          std::map< int, std::set< int > >::iterator cacheIt = cache->find( d.x );
          if ( cacheIt != cache->end() ) {
            if ( ( cacheIt->second ).find( d.y ) != ( cacheIt->second ).end() )
              return( false );
            ( cacheIt->second ).insert( d.y );
          } else {
            cache->insert( std::pair< int, std::set< int > >( d.x, std::set< int >() ) );
          }

          return( true );
        }

        /*
          TransPut_DrawLine : パターン pattern の点 p から始まる線分の色コードを length 分読み込み draw へ描画する(自由変形処理専用)

          line は TransPut_CalcBresenhamParam で計算したパラメータを表す。
        */
        void TransPut_DrawLine( const canvas::DrawingArea_IF& pattern, Coord< double > p,
                                canvas::DrawingArea_IF& draw, Coord< double > c, pen::GPixelDraw& pixelDraw,
                                int length, TransputParam& line, const interpolation::Base& ipF, std::map< int, std::set< int > >* cache )
        {
          RGB rgb;

          for ( int i = 0 ; i <= length ; i++ ) {
            // 描画直前に座標値を半分に
            Coord< int > d( std::round( c.x / 2 ), std::round( c.y / 2 ) );
            if ( d.x >= 0 && d.y >= 0 && d.x < ( draw.size() ).x && d.y < ( draw.size() ).y ) {
              if ( cache == 0 || TransPut_SetCache( d, cache ) ) {
                if ( ipF( pattern, p, &rgb ) ) {
                  pixelDraw.setRGB( rgb );
                  pixelDraw( draw, d );
                }
              }
            }
            c += line.drawAdd;
            p.x += line.patternAdd;
          }
        }

      } // namespace

      /*
        TransPut : パターン pattern の左上・右上・右下・左下各座標がそれぞれ lu, ru, rd, ld になるように自由変形して draw に描画する
      */
      void TransPut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                     Coord< int > lu, Coord< int > ru, Coord< int > rd, Coord< int > ld,
                     const interpolation::Base& ipF, bool priorFirst )
      {
        if ( &pattern == &draw ) return;

        // 描画で生じる隙間を埋めるため擬似的に 2 倍の解像度とする(描画直前で値を半分にする)
        Coord< double > lu2( lu.x * 2, lu.y * 2 );
        Coord< double > ru2( ru.x * 2, ru.y * 2 );
        Coord< double > rd2( rd.x * 2, rd.y * 2 );
        Coord< double > ld2( ld.x * 2, ld.y * 2 );

        Coord< double > p( 0, 0 ); // パターン読み取り位置
        Coord< int > patSize = pattern.size(); // 矩形の大きさ
        if ( patSize.x == 0 || patSize.y == 0 ) return;

        // 左右の辺用パラメータ
        TransputParam left( lu2, ld2, patSize.y );
        TransputParam right( ru2, rd2, patSize.y );
        // 左右の辺の最大値
        int len = std::max( left.len, right.len );

        left.patternAdd /= len;
        left.drawAdd.x /= len;
        left.drawAdd.y /= len;
        right.patternAdd /= len;
        right.drawAdd.x /= len;
        right.drawAdd.y /= len;

        std::map< int, std::set< int > > cache;
        for ( int i = 0 ; i <= len ; i++ ) {
          // 左右の点を結んだ線分
          TransputParam line( lu2, ru2, patSize.x );
          line.patternAdd /= line.len;
          line.drawAdd.x /= line.len;
          line.drawAdd.y /= line.len;
          TransPut_DrawLine( pattern, p, draw, lu2, pixelDraw, line.len, line, ipF, ( priorFirst ) ? &cache : 0 );

          TransputParam& longerParam = ( left.len >= right.len ) ? left : right;
          TransputParam& shorterParam = ( left.len < right.len ) ? left : right;
          Coord< double >& longerCoord = ( left.len >= right.len ) ? lu2 : ru2;
          Coord< double >& shorterCoord = ( left.len < right.len ) ? lu2 : ru2;

          //std::cout << longerParam.drawAdd.x << "," << longerParam.drawAdd.y << ">";
          //std::cout << shorterParam.drawAdd.x << "," << shorterParam.drawAdd.y << std::endl;
          longerCoord += longerParam.drawAdd;
          shorterCoord += shorterParam.drawAdd;
          p.y += longerParam.patternAdd;
        }
      }

      namespace
      {
        /*
          RotPut_GetClippingArea : 描画対象が含まれる矩形の座標を取得する

          d はパターンの中心から端までの距離、o は描画中心位置、r は回転角度、scale は拡大・縮小倍率をそれぞれ表す。
          矩形の左端を lu, 右端を rd に取得する。
        */
        void RotPut_GetClippingArea( Coord< double > d, Coord< int > o, double r, double scale, Coord< int >* lu, Coord< int >* rd )
        {
          double cosr = cos( r );
          double sinr = sin( r );
          Coord< double > p0( ( static_cast< double >( -d.x ) * cosr - static_cast< double >( -d.y ) * sinr ) / scale + o.x,
                              ( static_cast< double >( -d.x ) * sinr + static_cast< double >( -d.y ) * cosr ) / scale + o.y );
          Coord< double > p1( ( static_cast< double >( d.x ) * cosr - static_cast< double >( -d.y ) * sinr ) / scale + o.x,
                              ( static_cast< double >( d.x ) * sinr + static_cast< double >( -d.y ) * cosr ) / scale + o.y );
          Coord< double > p2( ( static_cast< double >( d.x ) * cosr - static_cast< double >( d.y ) * sinr ) / scale + o.x,
                              ( static_cast< double >( d.x ) * sinr + static_cast< double >( d.y ) * cosr ) / scale + o.y );
          Coord< double > p3( ( static_cast< double >( -d.x ) * cosr - static_cast< double >( d.y ) * sinr ) / scale + o.x,
                              ( static_cast< double >( -d.x ) * sinr + static_cast< double >( d.y ) * cosr ) / scale + o.y );
          lu->x = std::round( std::min( p0.x, std::min( p1.x, std::min( p2.x, p3.x ) ) ) );
          lu->y = std::round( std::min( p0.y, std::min( p1.y, std::min( p2.y, p3.y ) ) ) );
          rd->x = std::round( std::max( p0.x, std::max( p1.x, std::max( p2.x, p3.x ) ) ) );
          rd->y = std::round( std::max( p0.y, std::max( p1.y, std::max( p2.y, p3.y ) ) ) );
        }

      }

      /*
        RotPut : パターン pattern を角度 r だけ回転して draw 上に描画(Bresenhamを使わないバージョン)

        o は draw 上の pattern の中心位置を表し、scale は拡大・縮小率を表す。
        ipF を利用して補間処理を行う。
      */
      void RotPut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                   Coord< int > o, double r, const interpolation::Base& ipF, double scale )
      {
        if ( &pattern == &draw ) return;

        Coord< int > pSize( pattern.size() ); // パターンの大きさ
        Coord< int > dSize( draw.size() ); // パターンの大きさ

        if ( scale <= 0 ) return;
        if ( pSize.x == 0 || pSize.y == 0 ) return;
        if ( dSize.x == 0 || dSize.y == 0 ) return;

        Coord< double > d( static_cast< double >( pSize.x ) / 2, static_cast< double >( pSize.y ) / 2 ); // パターンの中心から両端までの距離
        Coord< int > lu, rd;
        RotPut_GetClippingArea( d, o, r, scale, &lu, &rd );
        //Coord< int > s( Round( static_cast< double >( o.x ) - d.x ), Round( static_cast< double >( o.y ) - d.y ) ); // 描画開始位置
        if ( lu.x >= dSize.x || lu.y >= dSize.y ) return;

        //Coord< int > max( s.x + pattern.size().x, s.y + pattern.size().y ); // 描画終了位置
        if ( rd.x < 0 || rd.y < 0 ) return;

        if ( lu.x < 0 ) lu.x = 0;
        if ( lu.y < 0 ) lu.x = 0;
        if ( rd.x >= dSize.x ) rd.x = dSize.x - 1;
        if ( rd.y >= dSize.y ) rd.y = dSize.y - 1;

        Coord< double > p;
        double cosr = cos( -r );
        double sinr = sin( -r );
        RGB rgb;
        for ( ; lu.y <= rd.y ; ++( lu.y ) ) {
          Coord< int > c( lu );
          for ( ; c.x <= rd.x ; ++( c.x ) ) {
            p.x = ( static_cast< double >( c.x - o.x ) * cosr - static_cast< double >( c.y - o.y ) * sinr ) / scale + d.x;
            p.y = ( static_cast< double >( c.x - o.x ) * sinr + static_cast< double >( c.y - o.y ) * cosr ) / scale + d.y;
            if ( ipF( pattern, p, &rgb ) ) {
              //if ( p.x >= 0 && p.x < pSize.x && p.y >= 0 && p.y < pSize.y ) {
              pixelDraw.setRGB( rgb );
              pixelDraw( draw, c );
            }
          }
        }
      }

    } // namespace draw

  } // namespace gl

} // namespace tadah_fussy
