/**
   @file gl_g_pattern_op.h
   @brief パターン取得・描画関数用ヘッダ

   @author tadah_fussy
   @date 2018/02/26 DrawingMethod.h から分離
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_G_PATTERN_OP_20210402_H
#define GRAPHIC_LIBRARY_G_PATTERN_OP_20210402_H

#include <vector>

#include <common/coord.h>
#include "gl_drawing_area_if.h"
#include "gl_g_pixel_op.h"
#include "gl_drawing_method.h"
#include "gl_g_interpolation.h"

namespace tadah_fussy
{
  namespace gl
  {
    /**
       @brief 描画機能用名前空間
    **/
    namespace draw
    {
      /* 通常処理 */

      /// @brief パターンの描画
      ///
      /// reverseX, reverseY, swapXY はパターンの向きを変えたり反転させる場合に利用できる。
      /// 全て false なら通常の描画になる。その他、よく利用する例は
      /// @code
      /// Put( pattern, draw, pixelDraw, s, true, false, false ); // 水平方向の鏡像
      /// Put( pattern, draw, pixelDraw, s, false, true, false ); // 垂直方向の鏡像
      /// Put( pattern, draw, pixelDraw, s, true, true, false );  // 180度回転
      /// Put( pattern, draw, pixelDraw, s, true, false, true );  // 反時計回りに90度回転
      /// Put( pattern, draw, pixelDraw, s, false, true, true );  // 時計回りに90度回転
      /// @endcode
      /// など。swapXY は、描画する座標を単純に ( x, y ) から ( y, x ) にするだけなので、これだけでは回転処理にならないことに注意。
      /// また、swapXY が true の場合、reverseX は垂直方向、reverseY は水平方向に適用される。
      ///
      /// パターン pattern、描画領域 draw のいずれかが未定義の場合は assert を実行する。
      ///
      /// パターン pattern と描画領域 draw が等しい場合、処理は行わないことに注意。
      ///
      /// @param pattern 描画するパターン
      /// @param draw パターンを描画する領域
      /// @param pixelDraw ピクセル描画用関数オブジェクト
      /// @param s パターンの描画開始位置
      /// @param reverseX X 方向の描画を反転させるか
      /// @param reverseY Y 方向の描画を反転させるか
      /// @param swapXY X,Y の描画を反転させるか
      void Put( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw,
                pen::GPixelDraw& pixelDraw, Coord< int > s,
                bool reverseX, bool reverseY, bool swapXY );

      /* 拡大・縮小処理(Bresenhamを使わないバージョン) */

      /// @brief パターンの拡大・縮小描画
      ///
      /// reverseX, reverseY, swapXY はパターンの向きを変えたり反転させる場合に利用できる。
      /// 全て false なら通常の描画になる。その他、よく利用する例は
      /// @code
      /// ResizePut( pattern, draw, pixelDraw, s, e, true, false, false, ipF ); // 水平方向の鏡像
      /// ResizePut( pattern, draw, pixelDraw, s, e, false, true, false, ipF ); // 垂直方向の鏡像
      /// ResizePut( pattern, draw, pixelDraw, s, e, true, true, false, ipF );  // 180度回転
      /// ResizePut( pattern, draw, pixelDraw, s, e, true, false, true, ipF );  // 反時計回りに90度回転
      /// ResizePut( pattern, draw, pixelDraw, s, e, false, true, true, ipF );  // 時計回りに90度回転
      /// @endcode
      /// など。swapXY は、描画する座標を単純に ( x, y ) から ( y, x ) にするだけなので、これだけでは回転処理にならないことに注意。
      /// また、swapXY が true の場合、reverseX は垂直方向、reverseY は水平方向に適用される。
      ///
      /// ipF はサンプル補間用の関数オブジェクト GInterpolation_IF を指定するために利用する。例えば
      /// @code
      /// GInterpolation_NN nn;
      /// ResizePut( pattern, draw, pixelDraw, s, e, false, false, false, nn ); // NN 法による補間
      /// @endcode
      /// によって「最近傍法 (Nearest Neighbour)」によるサンプル補間が行われる。
      ///
      /// パターン pattern、描画領域 draw のいずれかが未定義の場合は assert を実行する。
      ///
      /// パターン pattern と描画領域 draw が等しい場合、処理は行わないことに注意。
      ///
      /// @param pattern 描画するパターン
      /// @param draw パターンを描画する領域
      /// @param pixelDraw ピクセル描画用関数オブジェクト
      /// @param s,e パターンの描画開始・終了位置
      /// @param reverseX X 方向の描画を反転させるか
      /// @param reverseY Y 方向の描画を反転させるか
      /// @param swapXY X,Y の描画を反転させるか
      /// @param ipF サンプル補間用関数オブジェクト
      void ResizePut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                      Coord< int > s, Coord< int > e,
                      bool reverseX, bool reverseY, bool swapXY,
                      const interpolation::Base& ipF );

      /// @brief パターンの自由変形
      ///
      /// 処理の性格上、一つのピクセルを複数回描画する場合がある。priorFirst を true にすると最初に描画されたときの色が優先される。
      /// 描画はパターンの上側から順に行われるので、重なった部分は priorFirst が true なら先に描画される上側の方が、false なら
      /// 後で描画される下側の方が残ることになる。
      /// priorFirst を true にした場合、内部で座標をキャッシュするため処理速度は落ちることになるが、点描画のコストが高い場合は
      /// 重複した描画がなくなる分、逆に早く描画できるようになる。
      ///
      /// ipF はサンプル補間用の関数オブジェクト GInterpolation_IF を指定するために利用する。例えば
      /// @code
      /// GInterpolation_NN nn;
      /// TransPut( pattern, draw, pixelDraw, lu, ru, rd, ld, nn, false ); // NN 法による補間
      /// @endcode
      /// によって「最近傍法 (Nearest Neighbour)」によるサンプル補間が行われる。
      ///
      /// パターン pattern、描画領域 draw のいずれかが未定義の場合は assert を実行する。
      ///
      /// パターン pattern と描画領域 draw が等しい場合、処理は行わないことに注意。
      ///
      /// @param pattern 描画するパターン
      /// @param draw 描画領域
      /// @param pixelDraw ピクセル描画用関数オブジェクト
      /// @param lu,ru,rd,ld 変形後の各頂点(パターンの左上・右上・右下・左下にそれぞれ対応)
      /// @param ipF サンプル補間用関数オブジェクト
      /// @param priorFirst 先に描画されたときの色を優先するなら true
      void TransPut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                     Coord< int > lu, Coord< int > ru, Coord< int > rd, Coord< int > ld,
                     const interpolation::Base& ipF, bool priorFirst );

      /// @brief パターンの回転描画
      ///
      /// ipF はサンプル補間用の関数オブジェクト GInterpolation_IF を指定するために利用する。例えば
      /// @code
      /// GInterpolation_NN nn;
      /// RotPut( pattern, draw, pixelDraw, o, r, nn ); // NN 法による補間
      /// @endcode
      /// によって「最近傍法 (Nearest Neighbour)」によるサンプル補間が行われる。
      ///
      /// パターン pattern、描画領域 draw のいずれかが未定義の場合は assert を実行する。
      ///
      /// パターン pattern と描画領域 draw が等しい場合、処理は行わないことに注意。
      ///
      /// @param pattern 描画するパターン
      /// @param draw 描画領域
      /// @param pixelDraw ピクセル描画用関数オブジェクト
      /// @param o パターンの中心位置
      /// @param r 回転角度
      /// @param ipF サンプル補間用関数オブジェクト
      /// @param scale 拡大・縮小率
      void RotPut( const canvas::DrawingArea_IF& pattern, canvas::DrawingArea_IF& draw, pen::GPixelDraw& pixelDraw,
                   Coord< int > o, double r, const interpolation::Base& ipF, double scale = 1 );

    } // namespace draw

  } // namespace gl

} // namespace tadah_fussy

#endif
