/**
   @file gl_g_pixel_op.h
   @brief 点描画クラス定義用ヘッダ

   @author tadah_fussy
   @date 2018/02/25 DrawingMethod.h から分離
   @date 2021/02/25 ワーニングの出力に対応
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_G_PIXEL_OP_20210402_H
#define GRAPHIC_LIBRARY_G_PIXEL_OP_20210402_H

#include <cassert>

#include "gl_g_pattern.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace pen
    {
      /**
         @brief ピクセル操作用基底(抽象)クラス
      **/
      class GPixelOp
      {
      public:

        /// @brief ピクセル操作用関数
        ///
        /// @param draw 操作対象の canvas::DrawingArea_IF
        /// @param c 操作するピクセルの座標
        ///
        /// @return 派生クラスの実装に依存
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c ) = 0;

        /// @brief 仮想デストラクタ (何もしない)
        virtual ~GPixelOp() {}
      };

      /**
         @brief ピクセル描画用基底(抽象)クラス

         色コードのセットと、セットした色を使った描画を目的とする関数オブジェクトの基底クラス。
      **/
      class GPixelDraw : public GPixelOp
      {
      public:

        /// @brief RGB 成分のセット
        ///
        /// @param rgb セットする RGB 成分
        virtual void setRGB( RGB rgb ) = 0;

        /// @brief セットされた RGB 成分の取得
        ///
        /// @return セットされた RGB 成分
        virtual RGB getRGB() const = 0;

        /// @brief 色コードのセット
        ///
        /// @param col セットする色コード(AARRGGBB)
        void setColor( RGB::color_type col )
        { setRGB( RGB( col ) ); }
      };

      /**
         @brief 点描画クラス

         描画領域 canvas::DrawingArea_IF の任意の位置に RGB コードをセットする。
         指定した座標にあった RGB コードは、アルファチャンネルを含めて完全に新たなコードに置き換えられる
         (下地の色は考慮されない)。
      **/
      class GPSet : public GPixelDraw
      {
        RGB rgb_; // 描画する RGB コード

      public:

        /* コンストラクタ */

        /// @brief デフォルト・コンストラクタ
        ///
        /// メンバ変数 rgb は RGB のデフォルト値で初期化する
        GPSet() : rgb_() {}

        /// @brief 色コードを指定して構築
        ///
        /// @param col 色コード(AARRGGBB)
        explicit GPSet( RGB::color_type col )
          : rgb_( col ) {}

        /// @brief RGB 成分を指定して構築
        ///
        /// @param rgb RGBコード
        explicit GPSet( RGB rgb )
          : rgb_( rgb ) {}

        /// @brief RGB 成分のセット
        ///
        /// @param rgb セットする RGB 成分
        virtual void setRGB( RGB rgb )
        { rgb_ = rgb; }

        /// @brief 指定された RGB 成分の取得
        ///
        /// @return 指定された RGB 成分
        virtual RGB getRGB() const
        { return( rgb_ ); }

        /// @brief 指定した座標のピクセルに RGB コードをセットする
        ///
        /// draw が未定義の場合は assert を実行する。
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c RGB コードをセットするピクセルの座標
        /// @return 描画対象 draw のメンバ関数 canvas::DrawingArea_IF::pset の戻り値
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c )
        {
          return( draw.pset( c, rgb_ ) );
        }
      };

      /**
         @brief ピクセルの RGB 成分を取得するためのクラス

         描画領域 canvas::DrawingArea_IF の任意の位置の RGB コードを取得する。
         取得した RGB コードは、登録された点描画用インスタンスにメンバ関数 setRGB を
         通してセットされる。対象の点描画用インスタンスは、setPixcelDraw を使って
         複数登録することができる。
      **/
      class GPGet : public GPixelOp
      {
        std::vector< GPixelDraw* > pixelDrawList_; // 取得した RGB コードをセットする点描画クラス

      public:

        /* コンストラクタ */

        /// @brief デフォルト・コンストラクタ
        GPGet() {}

        /// @brief 点描画クラスを登録する
        ///
        /// 点描画インスタンス pixelDraw が未定義の場合は assert を実行する。
        ///
        /// @param pixelDraw 登録対象の点描画インスタンスへのポインタ
        void setPixelDraw( GPixelDraw* pixelDraw )
        {
          assert( pixelDraw != 0 );

          pixelDrawList_.push_back( pixelDraw );
        }

        /// @brief 指定した座標のピクセルの RGB コードを取得する
        ///
        /// draw が未定義の場合は assert を実行する。
        ///
        /// @param draw 取得対象の canvas::DrawingArea_IF
        /// @param c RGB コードを取得するピクセルの座標
        /// @return 取得対象 draw のメンバ関数 canvas::DrawingArea_IF::point の戻り値
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c )
        {
          RGB rgb;
          if ( ! draw.point( c, &rgb ) )
            return( false );

          for ( std::vector< GPixelDraw* >::iterator it = pixelDrawList_.begin() ;
                it != pixelDrawList_.end() ; ++it )
            (*it)->setRGB( rgb );

          return( true );
        }
      };

      /**
         @brief 点描画クラス ( RGB 成分用メンバ関数を利用した処理を行う )

         下地の色も考慮して描画される点が GPSet とは異なる。
         処理内容はコンストラクタ時の引数 op に依存する。
      **/
      class GRGBOp : public GPixelDraw
      {
        RGB_Op< RGB > op_; // RGB 成分への処理用関数オブジェクト

      public:

        /* コンストラクタ */

        /// @brief RGB 成分用メンバ関数とその引数を指定して構築
        ///
        /// 例えば、以下のコードは描画領域 draw の矩形 c0 - c1 の領域の RGB コードに対して透明度 50% の白を重ねあわせる
        /// @code
        /// GRGBOp op( &RGB::overlay, RGB( RGB::MAX, RGB::MAX, RGB::MAX, RGB::MAX / 2 );
        /// GBox::Fill( draw, op, c0, c1 );
        /// @endcode
        ///
        /// @param mf RGB 成分用メンバ関数 ( RGB 型の引数を一つ持つメンバ関数が対象 )
        /// @param rgb 引数
        GRGBOp( RGB& ( RGB::*mf )( RGB ), RGB rgb )
          : op_( mf, rgb ) {}

        /// @brief RGB 成分用メンバ関数のセット
        ///
        /// @param mf セットする RGB 成分用メンバ関数
        void setMemFun( RGB& ( RGB::*mf )( RGB ) )
        { op_.setMemFun( mf ); }

        // RGB 成分のセット
        virtual void setRGB( RGB rgb )
        { op_.setArg( rgb ); }

        // セットされた RGB 成分の取得
        virtual RGB getRGB() const
        { return( op_.getArg() ); }

        /// @brief 指定した座標のピクセルを RGB 成分用関数で処理する
        ///
        /// RGB 成分用メンバ関数 mf または描画領域 draw が未定義の場合は assert を実行する。
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c 操作するピクセルの座標
        /// @return 描画対象 draw のメンバ関数 canvas::DrawingArea_IF::point での RGB コード取得または
        ///         canvas::DrawingArea_IF::pset での描画に失敗した場合 false を返す
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c )
        {
          RGB bg;
          if ( ! draw.point( c, &bg ) )
            return( false );
          op_( &bg );

          return( draw.pset( c, bg ) );
        }
      };

      /**
         @brief ランダムな色の点描画クラス

         描画領域 canvas::DrawingArea_IF の指定した位置のピクセルにランダムな RGB コードを重ねあわせる。
         重ねあわせ方法は RGB 成分用メンバ関数 ( RGB::overlay など ) を通して指定することができる。
      **/
      class GRandomSet : public GPixelOp
      {
        RGB& ( RGB::*mf_ )( RGB ); // RGB 成分用メンバ関数へのポインタ
        RGB::color_type alpha_;    // 下地との重ね合わせの強さ(アルファチャンネル)

      public:

        /// @brief RGB 成分用メンバ関数と下地との重ね合わせの強さ ( アルファチャンネル ) を指定して構築
        ///
        /// @param mf RGB 成分用メンバ関数
        /// @param alpha 下地との重ね合わせの強さ ( アルファチャンネル )
        GRandomSet( RGB& ( RGB::*mf )( RGB ), RGB::primary_type alpha )
          : mf_( mf ), alpha_( alpha << 24 ) {}

        /// @brief RGB 成分用メンバ関数のセット
        ///
        /// @param mf セットする RGB 成分用メンバ関数
        void setMemFun( RGB& ( RGB::*mf )( RGB ) )
        { mf_ = mf; }

        /// @brief 下地との重ね合わせの強さ ( アルファチャンネル ) をセットする
        ///
        /// @param alpha 下地との重ね合わせの強さ ( アルファチャンネル )
        void setAlpha( RGB::primary_type alpha )
        { alpha_ = ( alpha << 24 ); }

        /// @brief セットした下地との重ね合わせの強さ ( アルファチャンネル ) を返す
        ///
        /// @return 下地との重ね合わせの強さ ( アルファチャンネル )
        RGB::primary_type getAlpha() const
        { return( alpha_ >> 24 ); }

        /// @brief ランダムな色で描画する
        ///
        /// 描画対象 draw または RGB 成分用メンバ関数 mf が未定義の場合は assert を実行する。
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c 操作するピクセルの座標
        /// @return 描画対象 draw のメンバ関数 canvas::DrawingArea_IF::pset の戻り値 ( draw が未定義なら必ず false になる )
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c )
        {
          RGB fg( alpha_ | ( static_cast< int >( static_cast< double >( rand() ) / ( static_cast< double >( RAND_MAX ) + 1 ) * 0x1000000 ) ) );
          GRGBOp gRPGOp( mf_, fg ); // mf_ が未定義の場合 assert が実行される

          return( gRPGOp( draw, c ) );
        }
      };

      /**
         @brief ライン・パターンに従った点描画クラス

         8 ビットのサイズのビット列 ( ライン・パターン ) を使い、ビットが立っている位置のピクセルだけを操作する。
         ライン描画などのように隣り合ったピクセルを連続して描画していることを前提としていることに注意。
         例えばピクセル操作に GPSet を指定し、ライン・パターンを 1010_1010 として GLine::DrawLine を利用すれば、点線を描画することができる。
      **/
      class GLinePatternSet : public GPixelOp
      {
        static const unsigned int PATTERN_SIZE = std::numeric_limits< unsigned char >::digits; // ライン・パターンのサイズ

        GPixelOp* pixelOp_;     // ピクセル操作用関数オブジェクトへのポインタ
        char index_;            // ライン・パターン上の現在位置

        unsigned char pattern_; // ライン・パターン

      public:

        /* コンストラクタ */

        /// @brief ピクセル操作用関数とライン・パターンを指定して構築
        ///
        /// @param pixelOp ピクセル操作用関数オブジェクト
        /// @param pattern ライン・パターン
        GLinePatternSet( GPixelOp& pixelOp, unsigned char pattern = 0xFF )
          : pixelOp_( &pixelOp ), index_( 0 ), pattern_( pattern ) {}

        /// @brief ライン・パターンをセットする
        ///
        /// @param pattern 新たなライン・パターン
        void setPattern( unsigned char pattern )
        { pattern_ = pattern; }

        /// @brief ピクセル操作用関数をセットする
        ///
        /// @param pixelOp 新たなピクセル操作用関数
        void setPixelOp( GPixelOp& pixelOp )
        { pixelOp_ = &pixelOp; }

        /// @brief ライン・パターン上の現在位置をリセットする
        ///
        /// 現在位置は GLinePatternSet::operator() が呼び出される度に更新されるため、いったんリセットしたい場合に用いる。
        void reset() { index_ = 0; }

        /// @brief ライン・パターンに従ってピクセル操作する
        ///
        /// 描画領域 draw またはピクセル操作用関数 pixelOp が未定義の場合は assert を実行する。
        ///
        /// @param draw ピクセル操作対象の canvas::DrawingArea_IF
        /// @param c 操作するピクセルの座標
        ///
        /// @return ライン・パターン上のビットがゼロでピクセル操作しなかった場合は false を返す
        /// @return ピクセル操作した場合は、その戻り値を返す
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c );
      };

      /**
         @brief 矩形パターンに従った点描画クラス

         あるパターンをコピーして描画に利用することができる。例えば GPaint::AreaPaint に
         適用すれば、パターンの繰り返しによる領域の塗りつぶしができる。
      **/
      class GBoxPatternSet : public GPixelOp
      {
        GRGBOp gRGBOp_;                 // 点描画関数オブジェクト
        const canvas::DrawingArea_IF& pattern_; // 矩形パターンへの参照

      public:

        /* コンストラクタ */

        /// @brief RGB 用メンバ関数へのポインタと矩形パターンを指定して構築
        ///
        /// @param mf RGB コード用メンバ関数へのポインタ
        /// @param pattern 対象の矩形パターンへの参照
        GBoxPatternSet( RGB& ( RGB::*mf )( RGB ), const canvas::DrawingArea_IF& pattern )
          : gRGBOp_( mf, RGB( 0, 0, 0 ) ), pattern_( pattern ) {}

        /// @brief RGB コード用メンバ関数のセット
        ///
        /// @param mf RGB コード用メンバ関数へのポインタ
        void setMemFun( RGB& ( RGB::*mf )( RGB ) )
        { gRGBOp_.setMemFun( mf ); }

        /// @brief 矩形パターンを描画する
        ///
        /// 描画対象 draw または RGB 用メンバ関数が未定義の場合は assert を実行する。
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c 描画するピクセルの座標
        ///
        /// @return 矩形パターンのサイズがゼロ以下の場合は何もせず false を返す
        /// @return 矩形パターンからの RGB コード取得時に false が返された場合は何もせず false を返す
        /// @return 点描画した場合は、その戻り値を返す
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c );
      };

      /**
         @brief 色反転描画クラス

         色を反転して描画する。
      **/
      class GNotSet : public GPixelOp
      {
      public:

        /// @brief 色反転描画処理
        ///
        /// アルファチャンネルは対象外となる ( 元の透明度のままで描画される )。
        ///
        /// 描画対象 draw が未定義なら assert を実行する。
        ///
        /// @param draw 描画対象の canvas::DrawingArea_IF
        /// @param c 描画するピクセルの座標
        ///
        /// @return 描画対象 draw のメンバ関数 canvas::DrawingArea_IF::point が false を返した場合、描画せずに false を返す
        /// @return 色反転描画した場合は、その戻り値を返す
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c );
      };

      /**
         @brief 描画領域のピクセルを別の描画領域にコピーする
      **/
      class GPixelCopy : public GPixelOp
      {
        canvas::DrawingArea_IF& dest_; // コピー先のパターン
        Coord< int > offset_;  // コピー元の原点を基準としたコピー先原点の位置

      public:

        /// @brief コピー先のパターンとコピー元との原点のオフセットを指定して構築
        ///
        /// オフセットはコピー元の原点を基準としたコピー先原点の位置として指定する。
        ///
        /// コピー先 dest が未定義だった場合は assert を実行する。
        ///
        /// @param dest コピー先のパターン
        /// @param offset コピー元の原点を基準としたコピー先原点の位置
        GPixelCopy( canvas::DrawingArea_IF& dest, Coord< int > offset )
          : dest_( dest ), offset_( offset )
        { assert( &dest_ != 0 ); }

        /// @brief 描画領域からピクセルをコピーする
        ///
        /// 位置はコピー元の原点を基準とする。
        ///
        /// コピー元 src が未定義だった場合は assert を実行する。
        ///
        /// @param src コピー元の描画領域
        /// @param c RGB 成分を取得する位置(コピー元の原点が基準)
        /// @return RGB 成分の取得・セットに成功したら true を返す
        virtual bool operator()( canvas::DrawingArea_IF& src, Coord< int > c );
      };

      /**
         @brief 十字カーソルの描画

         中央を描画色、左右上下の 4 ピクセルを背景色として十字の形のドットを描画する。
         但し、前回描画したドットと隣り合っていた場合は重なる部分の描画は行わない。
         線分描画のように連続して描画するような場合に利用する。

         <pre>
           ■        ■■■■■
         ■□■ → ■□□□□□■
           ■        ■■■■■
         </pre>
      **/
      class CrossDot : public GPixelDraw
      {
        Coord< int > pre_; // 前回描画した座標
        bool isFirst_;     // 初めての描画なら true

        RGB fg_; // 描画色
        RGB bg_; // 背景色

      public:

        /// @brief 描画色・背景色を指定して構築
        ///
        /// @param fg 描画色
        /// @param bg 背景色
        CrossDot( RGB fg, RGB bg )
          : isFirst_( true ), fg_( fg ), bg_( bg ) {}

        /// @brief 描画色 ( RGB 成分 ) のセット
        ///
        /// @param fg 描画色
        virtual void setRGB( RGB fg )
        { fg_ = fg; }

        /// @brief セットした描画色 ( RGB 成分 ) の取得
        ///
        /// @return セットした描画色
        virtual RGB getRGB() const
        { return( fg_ ); }

        /// @brief 背景色 ( RGB 成分 ) のセット
        ///
        /// @param bg 背景色
        void setBackground( RGB bg )
        { bg_ = bg; }

        /// @brief セットした背景色 ( RGB 成分 ) の取得
        ///
        /// @return セットした背景色
        RGB getBackground() const
        { return( bg_ ); }

        /// @brief 十字カーソルの描画
        ///
        /// 描画領域 draw が未定義だった場合は assert を実行する。
        ///
        /// @param draw 描画領域
        /// @param c 描画位置
        /// @return 常に true
        virtual bool operator()( canvas::DrawingArea_IF& draw, Coord< int > c );
      };

    } // namespace pen

  } // namespace gl

} // namespace tadah_fussy

#endif
