#include "gl_pixbuf.h"
#include "gl_g_pattern_op.h"

#include <assert.h>
#include <set>
#include <common/error_library.h>

namespace tadah_fussy
{
  namespace gl
  {
    using std::string;
    using std::vector;

    using error_lib::NNNum;
    using error_lib::PNum;

    /*****************************************************
       ヘルパ関数
    ****************************************************/

    namespace
    {
      /*
        RGB 成分 rgb から guint 型に変換する
      */
      guint RGBAtoGUInt( RGB rgb )
      {
        return( ( rgb.r << 24 ) | ( rgb.g << 16 ) | ( rgb.b << 8 ) | rgb.a );
      }

      /*
        guint型 color から RGB 成分に変換する
      */
      RGB GUIntToRGBA( guint color )
      {
        RGB rgb;

        rgb.r = ( color >> 24 ) & RGB::MAX;
        rgb.g = ( color >> 16 ) & RGB::MAX;
        rgb.b = ( color >> 8 ) & RGB::MAX;
        rgb.a = color & RGB::MAX;

        return( rgb );
      }

    } // namespace

    namespace canvas
    {
      /*****************************************************
          GDK_Pixbuf クラス
      ****************************************************/

      /*
        GDK_Pixbuf::overlay() : 全レイヤを重ねあわせた画像を返す

        メモリの確保ができなかった場合、std::bad_alloc 例外を投げる
        キャンバスのサイズがゼロの場合は NULL を返す
      */
      GdkPixbuf* GDK_Pixbuf::overlay()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::overlay" << std::endl;
#endif
        if ( ( ! changed_ ) && buff_ != 0 ) return( buff_ );

        // キャンバスサイズの異常
        assert( canvasSize().x >= 0 && canvasSize().y >= 0 );
        if ( canvasSize().x == 0 || canvasSize().y == 0 )
          return( 0 ); // サイズがゼロなら NULL を返す

        // キャンバスの初期化
        if ( buff_ != 0 )
          g_object_unref( buff_ );
        buff_ = gdk_pixbuf_new( GDK_COLORSPACE_RGB, canvasHasAlpha(), BITS_PER_SAMPLE, canvasSize().x, canvasSize().y );
        if ( buff_ == 0 )
          throw std::bad_alloc();
        gdk_pixbuf_fill( buff_, canvasBgColor_ );

        for ( vector< Layer >::const_iterator l = layers_.begin() ; l != layers_.end() ; ++l ) {
          if ( ! l->isValid() ) continue;
          if ( ! l->isVisible ) continue;
          Coord< int > dest( ( ( l->offset ).x < 0 ) ? 0 : ( l->offset ).x, ( ( l->offset ).y < 0 ) ? 0 : ( l->offset ).y );
          Coord< int > size( l->width(), l->height() );
          if ( ( l->offset ).x < 0 ) size.x += ( l->offset ).x;
          if ( ( l->offset ).y < 0 ) size.y += ( l->offset ).y;
          if ( size.x + dest.x > gdk_pixbuf_get_width( buff_ ) )
            size.x = gdk_pixbuf_get_width( buff_ ) - dest.x;
          if ( size.y + dest.y > gdk_pixbuf_get_height( buff_ ) )
            size.y = gdk_pixbuf_get_height( buff_ ) - dest.y;
#ifdef DEBUG
          std::cout << "dest = " << dest.x << "," << dest.y << std::endl;
          std::cout << "size = " << size.x << "," << size.y << std::endl;
          std::cout << "offset = " << ( l->offset ).x << "," << ( l->offset ).y << std::endl;
#endif
          gdk_pixbuf_composite( l->pixbuf, buff_, dest.x, dest.y, size.x, size.y,
                                ( l->offset ).x, ( l->offset ).y, 1, 1,
                                GDK_INTERP_NEAREST, l->alpha );
        }

        changed_ = false;

        return( buff_ );
      }

      /*
        GDK_Pixbuf::clearLayer : レイヤを全て消去する
      */
      void GDK_Pixbuf::clearLayer()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::clearLayer" << std::endl;
#endif
        for ( size_type i = 0 ; i < layers_.size() ; ++i ) {
          if ( layers_[i].pixbuf != 0 )
            g_object_unref( layers_[i].pixbuf );
        }
        layers_.clear();

        current_ = 0;
        changed_ = true;
      }

      /* GDK_Pixbufコンストラクタ */

      /*
        空の状態で構築 ( レイヤは存在しない )
      */
      GDK_Pixbuf::GDK_Pixbuf()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::GDK_Pixbuf()" << std::endl;
#endif
        current_ = 0;

        canvasBgColor_ = 0;
        canvasSize_ = Coord< int >( 0, 0 );
        canvasHasAlpha_ = true;

        changed_ = true;
        buff_ = 0;
      }

      /*
        キャンバスのサイズ sz、背景色 rgb、アルファ・チャンネルの有無 hasAlpha を指定して構築
        ( レイヤは存在しない )
      */
      GDK_Pixbuf::GDK_Pixbuf( Coord< int > sz, RGB rgb, bool hasAlpha )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::GDK_Pixbuf( sz, rgb, hasAlpha, alpha )" << std::endl;
#endif
        current_ = 0;

        canvasBgColor_ = RGBAtoGUInt( rgb );
        canvasSize_ = Coord< int >( NNNum( sz.x ), NNNum( sz.y ) );
        canvasHasAlpha_ = hasAlpha;

        changed_ = true;
        buff_ = 0;
      }

      /*
        ファイル名 fileName を指定して構築

        画像ファイルを読み込んだレイヤが一つ生成される。
        キャンバスのサイズ、背景色、アルファ・チャンネルの有無は生成したレイヤと等しくなる
      */
      GDK_Pixbuf::GDK_Pixbuf( const string& fileName )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::GDK_Pixbuf( fileName )" << std::endl;
#endif
        loadFromFile( fileName );

        changed_ = true;
        buff_ = 0;
      }

      /*
        デストラクタ
      */
      GDK_Pixbuf::~GDK_Pixbuf()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::~GDK_Pixbuf" << std::endl;
#endif
        clearLayer();

        if ( buff_ != 0 )
          g_object_unref( buff_ );
      }

      /*
        GDK_Pixbuf::addLayer : 空のレイヤを追加する
      */
      void GDK_Pixbuf::addLayer()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::addLayer" << std::endl;
#endif
        layers_.push_back( Layer() );

        changed_ = true;
      }

      /*
        GDK_Pixbuf::addLayer : size の大きさのレイヤを追加する

        offset はレイヤのキャンバスに対する相対位置、rgb は初期化用の RGB コード、
        hasAlpha はアルファ・チャンネルを持つか、alpha はアルファ・チャンネルを持つ場合の値を
        それぞれ表す。
      */
      void GDK_Pixbuf::addLayer( Coord< int > offset, Coord< int > size, RGB rgb, bool hasAlpha, RGB::primary_type alpha )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::addLayer" << std::endl;
#endif
        layers_.push_back( Layer() );
        layers_.back().pixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, hasAlpha, BITS_PER_SAMPLE, NNNum( size.x ), NNNum( size.y ) );
        if ( layers_.back().pixbuf == 0 )
          throw std::bad_alloc();
        layers_.back().bgColor = RGBAtoGUInt( rgb );
        layers_.back().offset = offset;
        layers_.back().alpha = alpha;
        gdk_pixbuf_fill( layers_.back().pixbuf, layers_.back().bgColor );

        changed_ = true;
      }

      /*
        GDK_Pixbuf::addFromFile : 画像ファイル fileName を新たなレイヤとして追加する
      */
      void GDK_Pixbuf::addFromFile( const std::string& fileName )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::addFromFile" << std::endl;
#endif
        layers_.push_back( Layer() );

        GError* error = 0; // エラー捕捉用

        layers_.back().pixbuf = gdk_pixbuf_new_from_file( fileName.c_str(), &error );

        // エラー発生
        if ( error != 0 ) {
          std::runtime_error ex( error->message );
          g_error_free( error );
          layers_.pop_back(); // 追加したレイヤは削除
          throw ex;
        }

        // error がセットされないのに pixbuf_ は NULL だった場合(念のため)
        if ( layers_.back().pixbuf == 0 ) {
          layers_.pop_back(); // 追加したレイヤは削除
          throw std::runtime_error( "Image buffer not created (unknown error) : fileName = " + fileName );
        }

        changed_ = true;
      }

      /*
        GDK_Pixbuf::eraseLayer : 現在のレイヤを消去する

        現在のレイヤ番号は一つ減る(消去したレイヤの下側のレイヤが対象になる)。
        但し、現在のレイヤが 0 番だった場合は変化しない。
      */
      void GDK_Pixbuf::eraseLayer()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::eraseLayer" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        std::vector< Layer >::iterator l = layers_.begin() + current_;
        if ( l->pixbuf != 0 )
          g_object_unref( l->pixbuf );
        layers_.erase( l );
        if ( current_ > 0 )
          --current_;

        changed_ = true;
      }

      /*
        現在のレイヤの位置を入れ替える

        入れ替え位置は指定した番号のレイヤの前とする ( 例えば 0 を指定した場合は入れ替えたレイヤが 0 番目になる。
        一番上に移動したい場合は最上位のレイヤの番号 + 1 を指定する。
        指定した添字が範囲外だった場合は例外 std::out_of_range を投げる
      */
      void GDK_Pixbuf::changeLayer( size_type upperLayer )
      {
        // 0 1 2 3 4 ... current_ = 2
        //     ^
        // upperLayer = 4 のとき
        // 0 1 2 3 4*5 ... current_ = 2
        //     ^
        // 0 1   2 3*4 ... current_ = 3
        //
        // upperLayer = 0 のとき
        // 0*1 2 3 4 5 ... current_ = 3
        //       ^
        // 0*1 2   3 4 ... current_ = 0
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::changeLayer" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        if ( upperLayer > layers_.size() )
          throw std::out_of_range( "GDK_Pixbuf::changeLayer invalid index ( out of range )" );
        if ( current_ == upperLayer || current_ + 1 == upperLayer ) return;

        if ( upperLayer == layers_.size() ) {
          layers_.push_back( layers_[current_] );
        } else {
          layers_.insert( layers_.begin() + upperLayer, layers_[current_] );
        }

        if ( upperLayer < current_ )
          ++current_;

        // current_ の参照先を代入しているので参照を一つ増やす
        g_object_ref( layers_[current_].pixbuf );
        eraseLayer();

        current_ = ( upperLayer < current_ ) ? upperLayer : upperLayer - 1;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::copyLayer : 現在のレイヤをコピーする
      */
      void GDK_Pixbuf::copyLayer()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::copyLayer" << std::endl;
#endif
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        layers_.push_back( Layer() );
        Layer& newLayer = layers_.back();
        newLayer.bgColor = layers_[current_].bgColor;
        newLayer.offset = layers_[current_].offset;
        newLayer.alpha = layers_[current_].alpha;
        newLayer.isVisible = layers_[current_].isVisible;

        newLayer.pixbuf = gdk_pixbuf_copy( layers_[current_].pixbuf );
        if ( newLayer.pixbuf == NULL ) {
          layers_.pop_back();
          throw std::bad_alloc();
        }

        changed_ = true;
      }

      /*
        GDK_Pixbuf::superimpose : 現在のレイヤを下側の画像に重ね合わせる

        現在のレイヤが不可視状態でも処理は行う。
        重ねあわせた後、現在のレイヤは消去される。
      */
      void GDK_Pixbuf::superimpose()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::superimpose" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        // 下地は対象外
        if ( current_ == 0 ) return;

        Layer& upper = layers_[current_];
        Layer& lower = layers_[current_ - 1];

        if ( upper.isValid() ) {
          if ( ! lower.isValid() ) {
            lower = upper;                // この時点で GdkPixbuf ポインタが二重になっている
            g_object_ref( lower.pixbuf ); // eraseLayer 内で reference count が一つ減らされるのであらかじめ加えておく
          } else {
            // 新たなオフセット(左上原点)・右隅 + 1・サイズ
            Coord< int > newOffset( std::min( upper.offset.x, lower.offset.x ),
                                    std::min( upper.offset.x, lower.offset.x ) );
            Coord< int > newEnd( std::max( upper.width() + upper.offset.x, lower.width() + lower.offset.x ),
                                 std::max( upper.height() + upper.offset.y, lower.height() + lower.offset.y ) );
            Coord< int > newSize( newEnd.x - newOffset.x, newEnd.y - newOffset.y );

            GdkPixbuf* buff = gdk_pixbuf_new( GDK_COLORSPACE_RGB, lower.hasAlpha(), BITS_PER_SAMPLE, newSize.x, newSize.y );
            gdk_pixbuf_fill( buff, lower.bgColor );

            gdk_pixbuf_composite( lower.pixbuf, buff, lower.offset.x - newOffset.x, lower.offset.y - newOffset.y,
                                  lower.width(), lower.height(),
                                  lower.offset.x - newOffset.x, lower.offset.y - newOffset.y, 1, 1,
                                  GDK_INTERP_NEAREST, lower.alpha );
            gdk_pixbuf_composite( upper.pixbuf, buff, upper.offset.x - newOffset.x, upper.offset.y - newOffset.y,
                                  upper.width(), upper.height(),
                                  upper.offset.x - newOffset.x, upper.offset.y - newOffset.y, 1, 1,
                                  GDK_INTERP_NEAREST, upper.alpha );
            g_object_unref( lower.pixbuf );
            lower.pixbuf = buff;
            lower.offset = newOffset;
          }
        }
        eraseLayer();

        changed_ = true;
      }

      /*
        GDK_Pixbuf::setCurrentLayer : 現在のレイヤ番号を変更する
      */
      void GDK_Pixbuf::setCurrentLayer( size_type index )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::setCurrentLayer" << std::endl;
#endif
        if ( index >= layers_.size() )
          throw std::out_of_range( "GDK_Pixbuf::setCurrentLayer : invalid index ( out of range )" );

        current_ = index;
      }

      /*
        GDK_Pixbuf::size : 現レイヤのサイズを返す
      */
      Coord< int > GDK_Pixbuf::size() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::size" << std::endl;
#endif
        // 空の状態なら ( 0, 0 ) を返す
        if ( empty() )
          return( Coord< int >( 0, 0 ) );

        assert( layers_.size() > current_ );

        const Layer& l = layers_[current_];
        return( Coord< int >( l.width(), l.height() ) );
      }

      /*
        GDK_Pixbuf::resize : 現レイヤのサイズを変更する
      */
      void GDK_Pixbuf::resize( Coord< int > size )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::resize" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        Layer& l = layers_[current_];
        GdkPixbuf* newPixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, l.hasAlpha(), BITS_PER_SAMPLE, NNNum( size.x ), NNNum( size.y ) );
        if ( newPixbuf == 0 )
          throw std::bad_alloc();
        gdk_pixbuf_fill( newPixbuf, l.bgColor );

        if ( l.isValid() ) {
          int width = std::min( l.width(), size.x );
          int height = std::min( l.height(), size.y );
          gdk_pixbuf_composite( l.pixbuf, newPixbuf, 0, 0, width, height,
                                0, 0, 1, 1, GDK_INTERP_NEAREST, RGB::MAX );
          g_object_unref( l.pixbuf );
        }

        l.pixbuf = newPixbuf;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::getBgColor : 現レイヤの背景色を返す
      */
      RGB GDK_Pixbuf::getBgColor() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::getBgColor" << std::endl;
#endif
        // 空の状態なら RGB( 0, 0, 0, 0 ) を返す
        if ( empty() ) return( RGB( 0, 0, 0, 0 ) );

        assert( layers_.size() > current_ );

        return( GUIntToRGBA( layers_[current_].bgColor ) );
      }
      /*
        GDK_Pixbuf::setBgColor : 現レイヤの背景色をセットする
      */
      void GDK_Pixbuf::setBgColor( RGB rgb )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::setBgColor" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        layers_[current_].bgColor = RGBAtoGUInt( rgb );

        changed_ = true;
      }

      /*
        GDK_Pixbuf::fill : 現レイヤを背景色で塗りつぶす
      */
      void GDK_Pixbuf::fill()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::fill" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        gdk_pixbuf_fill( layers_[current_].pixbuf, layers_[current_].bgColor );

        changed_ = true;
      }

      /*
        GDK_Pixbuf::getOffset : 現レイヤのオフセット値を返す
      */
      Coord< int > GDK_Pixbuf::getOffset() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::getOffset" << std::endl;
#endif
        // 空の状態なら ( 0, 0 ) を返す
        if ( empty() ) return( Coord< int >( 0, 0 ) );

        assert( layers_.size() > current_ );

        return( layers_[current_].offset );
      }
      /*
        GDK_Pixbuf::setOffset : 現レイヤのオフセット値をセットする
      */
      void GDK_Pixbuf::setOffset( Coord< int > offset )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::setOffset" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        layers_[current_].offset = offset;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::getCoord : 現レイヤのキャンバスに対する相対位置を求める
      */
      Coord< double > GDK_Pixbuf::getCoord( Coord< double > c ) const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::getCoord" << std::endl;
#endif
        // 空の状態なら c をそのまま返す
        if ( empty() ) return( c );

        assert( layers_.size() > current_ );

        const Layer& l = layers_[current_];

        c.x -= l.offset.x;
        c.y -= l.offset.y;

        return( c );
      }

      /*
        GDK_Pixbuf::getAlpha : 現レイヤのアルファ・チャンネルを返す
      */
      RGB::primary_type GDK_Pixbuf::getAlpha() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::getAlpha" << std::endl;
#endif
        // 空の状態ならゼロを返す
        if ( empty() ) return( 0 );

        assert( layers_.size() > current_ );

        return( layers_[current_].alpha );
      }
      /*
        GDK_Pixbuf::setAlpha : 現レイヤのアルファ・チャンネルをセットする
      */
      void GDK_Pixbuf::setAlpha( RGB::primary_type alpha )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::setAlpha" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        layers_[current_].alpha = alpha;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::hasAlpha : 現在のレイヤがのアルファチャンネルを持つ場合は true を返す
      */
      bool GDK_Pixbuf::hasAlpha() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::hasAlpha" << std::endl;
#endif
        // 空の状態なら false を返す
        if ( empty() ) return( false );

        assert( layers_.size() > current_ );

        return( layers_[current_].hasAlpha() );
      }

      /*
        GDK_Pixbuf::addAlpha : アルファ・チャンネルの追加

        substitudeRGB : 指定した RGB 成分 ( rgb ) と同じ色のピクセルを透明色に置き換えるなら true を指定する
        rgb : 透明色に置き換えるときの RGB 成分。substitudeRGB が false なら無視される。
      */
      void GDK_Pixbuf::addAlpha( bool substitudeRGB, RGB rgb )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::addAlpha" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        if ( ! layers_[current_].isValid() ) return;
        if ( layers_[current_].hasAlpha() ) return;

        GdkPixbuf* newPixbuf = gdk_pixbuf_add_alpha( layers_[current_].pixbuf, ( substitudeRGB ) ? TRUE : FALSE, rgb.r, rgb.g, rgb.b );
        if ( newPixbuf == nullptr )
          throw std::runtime_error( "GDK_Pixbuf::addAlpha : New pixbuf not created" );

        g_object_unref( layers_[current_].pixbuf );
        layers_[current_].pixbuf = newPixbuf;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::removeAlpha : アルファ・チャンネルの削除

        bgColor : 重ね合わせ時の背景色
      */
      void GDK_Pixbuf::removeAlpha( RGB bgColor )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::removeAlpha" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        Layer& l = layers_[current_];
        if ( ! l.isValid() ) return;
        if ( ! l.hasAlpha() ) return;

        GdkPixbuf* newPixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, FALSE, BITS_PER_SAMPLE, l.width(), l.height() );
        const RGB::primary_type* src = l.origin(); //+ c.y * l.rowstride() + c.x * l.nChannels();
        RGB::primary_type* dest = gdk_pixbuf_get_pixels( newPixbuf );
        RGB fg;
        int32_t srcNC = l.nChannels();
        int32_t destNC = gdk_pixbuf_get_n_channels( newPixbuf );
        for ( int32_t y = 0 ; y < l.height() ; ++y ) {
          for ( int32_t x = 0 ; x < l.width() ; ++x ) {
            for ( size_t i = RGB::RED ; i < RGB::RGBA_SIZE ; ++i )
              fg[i] = src[x * srcNC + i];
            RGB bg( bgColor );
            bg.overlay( fg );
            for ( size_t i = RGB::RED ; i < RGB::RGB_SIZE ; ++i )
              dest[x * destNC + i] = bg[i];
          }
          src += l.rowstride();
          dest += gdk_pixbuf_get_rowstride( newPixbuf );
        }

        g_object_unref( l.pixbuf );
        l.pixbuf = newPixbuf;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::isVisible : 現在のレイヤが可視状態なら true を返す
      */
      bool GDK_Pixbuf::isVisible() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::isVisible" << std::endl;
#endif
        // 空の状態なら false を返す
        if ( empty() ) return( false );

        assert( layers_.size() > current_ );

        return( layers_[current_].isVisible );
      }

      /*
        GDK_Pixbuf::visible : 現在のレイヤを可視状態にする
      */
      void GDK_Pixbuf::visible()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::visible" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        layers_[current_].isVisible = true;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::invisible : 現在のレイヤを不可視状態にする
      */
      void GDK_Pixbuf::invisible()
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::invisible" << std::endl;
#endif
        // 空の状態なら何もしない
        if ( empty() ) return;

        assert( layers_.size() > current_ );

        layers_[current_].isVisible = false;

        changed_ = true;
      }

      /*
        GDK_Pixbuf::point : 座標 c の RGB 成分を palet に取得する
      */
      bool GDK_Pixbuf::point( Coord< int > c, RGB* palet ) const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::point" << std::endl;
#endif
        assert( palet != nullptr );

        // 空の状態なら false を返す
        if ( empty() ) return( false );

        assert( layers_.size() > current_ );

        const Layer& l = layers_[current_];

        if ( ! l.isValid() ) return( false );

        if ( c.x < 0 || c.x >= l.width() ) return( false );
        if ( c.y < 0 || c.y >= l.height() ) return( false );

        const RGB::primary_type* pixel = l.origin() + c.y * l.rowstride() + c.x * l.nChannels();
        if ( l.nChannels() > 3 )
          palet->receiveFromParam( pixel[0], pixel[1], pixel[2], pixel[3] );
        else
          palet->receiveFromParam( pixel[0], pixel[1], pixel[2] );

        return( true );
      }

      /*
        GDK_Pixbuf::pset : 座標 c のピクセルを palet でセットする
      */
      bool GDK_Pixbuf::pset( Coord< int > c, RGB palet )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::pset" << std::endl;
#endif
        // 空の状態なら false を返す
        if ( empty() ) return( false );

        assert( layers_.size() > current_ );

        Layer& l = layers_[current_];

        if ( ! l.isValid() ) return( false );

        if ( c.x < 0 || c.x >= l.width() ) return( false );
        if ( c.y < 0 || c.y >= l.height() ) return( false );

        guchar* pixel = l.origin() + c.y * l.rowstride() + c.x * l.nChannels();
        if ( l.nChannels() > 3 ) {
          palet.sendToParam( pixel, pixel + 1, pixel + 2, pixel + 3 );
        } else {
          palet.sendToParam( pixel, pixel + 1, pixel + 2 );
        }

        changed_ = true;

        return( true );
      }

      /*
        GDK_Pixbuf::getCanvasBgColor : キャンバスの背景色を返す
      */
      RGB GDK_Pixbuf::getCanvasBgColor() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::getCanvasBgColor" << std::endl;
#endif
        return( GUIntToRGBA( canvasBgColor_ ) );
      }
      /*
        GDK_Pixbuf::setCanvasBgColor : キャンバスの背景色をセットする
      */
      void GDK_Pixbuf::setCanvasBgColor( RGB rgb )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::setCanvasBgColor" << std::endl;
#endif
        canvasBgColor_ = RGBAtoGUInt( rgb );

        changed_ = true;
      }

      /*
        GDK_Pixbuf::canvasSize : キャンバスのサイズを返す
      */
      Coord< int > GDK_Pixbuf::canvasSize() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::canvasSize" << std::endl;
#endif
        return( canvasSize_ );
      }

      /*
        GDK_Pixbuf::canvasResize : キャンバスのサイズを変更する
      */
      void GDK_Pixbuf::canvasResize( Coord< int > size )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::canvasResize" << std::endl;
#endif
        canvasSize_ = Coord< int >( NNNum( size.x ), NNNum( size.y ) );

        changed_ = true;
      }

      /*
        GDK_Pixbuf::canvasHasAlpha : キャンバスがアルファチャンネルを持つ場合は true を返す
      */
      bool GDK_Pixbuf::canvasHasAlpha() const
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::canvasHasAlpha" << std::endl;
#endif
        return( canvasHasAlpha_ );
      }

      /*
        GDK_Pixbuf::addCanvasAlpha : キャンバスにアルファ・チャンネルを追加する
      */
      void GDK_Pixbuf::setCanvasAlpha( bool flag )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::addCanvasAlpha" << std::endl;
#endif
        if ( canvasHasAlpha_ != flag ) {
          canvasHasAlpha_ = flag;
          changed_ = true;
        }
      }

      /*
        GDK_Pixbuf::loadFromFile : 画像ファイル fileName で初期化する
      */
      void GDK_Pixbuf::loadFromFile( const std::string& fileName )
      {
#ifdef DEBUG
        std::cout << "GDK_Pixbuf::loadFromFile" << std::endl;
#endif
        clearLayer();
        addFromFile( fileName );

        current_ = 0;

        // キャンバスのプロパティは画像ファイルのプロパティに合わせる
        canvasBgColor_ = layers_[current_].bgColor;
        canvasSize_ = size();
        canvasHasAlpha_ = hasAlpha();

        changed_ = true;
      }

      /*****************************************************
          GTK_DrawingArea クラス
      ****************************************************/

      /*
        GTK_DrawingArea::init : メンバ変数の初期化

        オフセットと倍率を初期値に戻す
      */
      void GTK_DrawingArea::init()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::init" << std::endl;
#endif
        offset_.x = offset_.y = 0;
        scale_.x = scale_.y = 1.0;

        recalcCoord();
        emitValChangeSignal();
      }

      /*
        GTK_DrawingArea::recalcCoordSub : 座標の再計算

        widget()のサイズが変わったときや、オフセット・スケール値が変更されたときに呼び出される

        pixbufSize : GDK_Pixbufオブジェクトの画像サイズ
        widgetSize : widget()のサイズ
        offset : GDK_Pixbuf オブジェクトを基準とした描画エリアの原点位置を代入する変数へのポインタ
        drawSize : 描画するサイズ ( 描画エリアと GDK_Pixbuf の小さい方 ) を代入する変数へのポインタ
      */
      void GTK_DrawingArea::recalcCoordSub( int pixbufSize, int widgetSize, double* offset, int* drawSize )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::recalcCoordSub" << std::endl;
#endif
        // widget() が pixbuf より大きい場合は、画像が中央になるように座標を調整
        if ( widgetSize >= pixbufSize ) {
          *offset = ( widgetSize - pixbufSize ) / 2;
          *drawSize = pixbufSize;
        } else {
          if ( *offset > 0 ) *offset = 0;
          // 右/下端の状態をチェック
          if ( pixbufSize + *offset < widgetSize )
            *offset = widgetSize - pixbufSize;
          *drawSize = widgetSize;
        }
      }

      /*
        GTK_DrawingArea::recalcCoord : 座標の再計算

        widget()のサイズが変わったときや、オフセット・スケール値が変更されたときに呼び出される
      */
      void GTK_DrawingArea::recalcCoord()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::recalcCoord" << std::endl;
#endif
        recalcCoordSub( round( pixbuf_.canvasSize().x * scale_.x ), gtk_widget_get_allocated_width( widget() ), &( offset_.x ), &( drawSize_.x ) );
        recalcCoordSub( round( pixbuf_.canvasSize().y * scale_.y ), gtk_widget_get_allocated_height( widget() ), &( offset_.y ), &( drawSize_.y ) );

#ifdef DEBUG
        std::cout << "offset = " << offset_.x << "," << offset_.y << std::endl;
        std::cout << "drawSize = " << drawSize_.x << "," << drawSize_.y << std::endl;
        std::cout << "scale = " << scale_.x << "," << scale_.y << std::endl;
#endif
      }

      /*
        GTK_DrawingArea::configure_event : サイズ変更時のイベント

        GtkWidget *widget : widget()オブジェクト
        GdkEventExpose *event : GdkEventExposeオブジェクト(未使用)
        gpointer draw : GTK_DrawingAreaオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea::configure_event( GtkWidget *widget, GdkEventConfigure *event, gpointer draw )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::configure_event : ( " << event->x << ", " << event->y << " ) - ( "
                  << event->width << ", " << event->height << " )" << std::endl;
#endif
        GTK_DrawingArea* d = (GTK_DrawingArea*)draw;

        d->recalcCoord();

        // ワークスペースの再作成
        // 描画されていた内容はリセットされる
        if ( d->workspace_ != 0 ) {
          d->clearWorkspace();
          d->createWorkspace();
        }

        d->emitValChangeSignal();

        return( FALSE );
      }

      /*
        GTK_DrawingArea::drawEvent : 再描画時のイベント

        widget : widget()オブジェクト (未使用)
        cr : Cairo コンテキスト ( widget() と接続 )
        draw : GTK_DrawingAreaオブジェクト ( 自己参照ポインタ )
      */
      gint GTK_DrawingArea::drawEvent( GtkWidget *widget, cairo_t* cr, gpointer draw )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::drawEvent" << std::endl;
#endif
        GTK_DrawingArea* d = (GTK_DrawingArea*)draw;

        d->draw( cr, GDK_INTERP_BILINEAR );

        return( FALSE );
      }

      /*
        GTK_DrawingArea::setEvent : イベントハンドラのセット
      */
      void GTK_DrawingArea::setEvent()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::setEvent" << std::endl;
#endif
        g_signal_connect( G_OBJECT( widget() ), "draw", G_CALLBACK( drawEvent ), this );
        g_signal_connect( G_OBJECT( widget() ), "configure_event", G_CALLBACK( configure_event ), this );

        gtk_widget_set_events( widget(),
                               GDK_EXPOSURE_MASK |
                               GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
                               GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
                               GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                               GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
                               GDK_SCROLL_MASK );

        valChangeSignal_ = g_signal_lookup( "redraw", G_TYPE_OBJECT );
        if ( valChangeSignal_ == 0 )
          valChangeSignal_ = g_signal_new( "redraw", G_TYPE_OBJECT, G_SIGNAL_RUN_FIRST, 0, 0, 0, 0, G_TYPE_NONE, 0 );
      }

      /*
        GTK_DrawingArea::draw : 画像を scale_ 倍した後 offset_ だけシフトして、drawSize_ の大きさだけ cairo コンテキスト cr を通して描画する
      */
      void GTK_DrawingArea::draw( cairo_t* cr, GdkInterpType interpType )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::draw( cairo_t*, GdkInterpType )" << std::endl;
        std::cout << "offset_ = ( " << offset_.x << ", " << offset_.y << " )" << std::endl;
        std::cout << "drawSize_ = ( " << drawSize_.x << ", " << drawSize_.y << " )" << std::endl;
        std::cout << "scale_ = ( " << scale_.x << ", " << scale_.y << " )" << std::endl;
#endif
        GdkRectangle rect;
        if ( ! gdk_cairo_get_clip_rectangle( cr, &rect ) ) return;

#ifdef DEBUG
        std::cout << "( x, y ) = ( " << rect.x << ", " << rect.y << " )" << std::endl;
        std::cout << "( w, h ) = ( " << rect.width << ", " << rect.height << " )" << std::endl;
#endif

        // 拡大・縮小用のバッファ ( アルファチャンネルを持つ場合にチェッカーボードと重ね合わされる )
        GdkPixbuf* buff = pixbuf_.overlay();
        if ( buff == nullptr ) return;

        // pixbuf を描画する開始位置
        Coord< double > drawOrigin( ( offset_.x > 0 ) ? offset_.x : 0, ( offset_.y > 0 ) ? offset_.y : 0 );

        if ( layer_ != nullptr )
          g_object_unref( layer_ );
        layer_ = gdk_pixbuf_new( GDK_COLORSPACE_RGB, FALSE, BITS_PER_SAMPLE, rect.width, rect.height );
        if ( layer_ == nullptr )
          throw std::bad_alloc();
        gdk_pixbuf_fill( layer_, RGBAtoGUInt( pixbuf_.getCanvasBgColor() ) );

        // 透明部分はチェッカーボードと重ね合わせる
#ifdef DEBUG
        std::cout << "dest = " << drawOrigin.x << "," << drawOrigin.y << std::endl;
        std::cout << "size = " << drawSize_.x << "," << drawSize_.y << std::endl;
        std::cout << "offset = " << offset_.x << "," << offset_.y << std::endl;
#endif
        gdk_pixbuf_composite_color( buff, layer_, drawOrigin.x, drawOrigin.y,
                                    std::min( drawSize_.x, rect.width ), std::min( drawSize_.y, rect.height ),
                                    offset_.x, offset_.y, scale_.x, scale_.y,
                                    interpType, RGB::MAX, drawOrigin.x, drawOrigin.y, 16, 0xaaaaaa, 0x555555 );

        // ワークスペースがある場合はさらに重ねあわせる
        GdkPixbuf* canvas = nullptr;
        if ( workspace_ != nullptr ) {
          canvas = gdk_pixbuf_copy( layer_ );
          gdk_pixbuf_composite( workspace_, canvas, 0, 0, rect.width, rect.height,
                                0, 0, 1, 1,
                                interpType, RGB::MAX );
          gdk_cairo_set_source_pixbuf( cr, canvas, 0, 0 );
        } else {
          gdk_cairo_set_source_pixbuf( cr, layer_, 0, 0 );
        }

        cairo_paint( cr );

        if ( canvas != 0 )
          g_object_unref( canvas );

#ifdef DEBUG
        std::cout << "GTK_DrawingArea::draw( cairo_t*, GdkInterpType ) End" << std::endl;
#endif
      }

      /*
        GTK_DrawingArea::draw : 画像を scale_ 倍した後 offset_ だけシフトして、drawSize_ の大きさだけ widget に描画する
      */
      void GTK_DrawingArea::draw( GtkWidget* widget, GdkInterpType interpType )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::draw( GtkWidget*, GdkInterpType )" << std::endl;
#endif
        assert( widget != nullptr );

        // GdkWidgetへのポインタを抽出
        GdkWindow* drawable = gtk_widget_get_window( widget );
        if ( drawable == nullptr ) return;

        cairo_region_t* cairoRegion = cairo_region_create();
        GdkDrawingContext* drawingContext = gdk_window_begin_draw_frame( drawable, cairoRegion );
        cairo_t* cr = gdk_drawing_context_get_cairo_context( drawingContext );

        draw( cr, interpType );

        gdk_window_end_draw_frame( drawable, drawingContext );
        cairo_region_destroy( cairoRegion );
      }

      void GTK_DrawingArea::draw( GdkInterpType interpType ) //const
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::draw( GdkInterpType )" << std::endl;
#endif
        recalcCoord();

        if ( gtk_widget_is_drawable( widget() ) )
          draw( widget(), interpType );
      }

      void GTK_DrawingArea::emitValChangeSignal() const
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::emitValChangeSignal" << std::endl;
#endif
        if ( valChangeSignal_ > 0 && GTK_IS_WIDGET( widget() ) )
          g_signal_emit( widget(), valChangeSignal_, 0 );
      }

      /* GTK_DrawingArea コンストラクタ */

      /*
        空のレイヤを構築
      */
      GTK_DrawingArea::GTK_DrawingArea()
        : pixbuf_(), drawingArea_( 0 ), workspace_( 0 ), layer_( 0 ), valChangeSignal_( 0 )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::GTK_DrawingArea()" << std::endl;
#endif
        drawingArea_ = gtk_drawing_area_new();

        setEvent();

        init();
      }

      /*
        キャンバスのサイズ sz、背景色 rgb、アルファ・チャンネルの有無 hasAlpha を指定して構築
      */
      GTK_DrawingArea::GTK_DrawingArea( Coord< int > sz, RGB rgb, bool hasAlpha )
        : pixbuf_( sz, rgb, hasAlpha ), drawingArea_( 0 ), workspace_( 0 ), layer_( 0 ), valChangeSignal_( 0 )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::GTK_DrawingArea( Coord< int >, RGB, bool, RGB::primary_type )" << std::endl;
#endif
        drawingArea_ = gtk_drawing_area_new();

        setEvent();

        init();
      }

      /*
        ファイル名 fileName を指定して構築
      */
      GTK_DrawingArea::GTK_DrawingArea( const string& fileName )
        : pixbuf_( fileName ), drawingArea_( 0 ), workspace_( 0 ), layer_( 0 ), valChangeSignal_( 0 )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::GTK_DrawingArea( const string& )" << std::endl;
#endif
        drawingArea_ = gtk_drawing_area_new();

        setEvent();

        init();
      }

      /*
        デストラクタ
      */
      GTK_DrawingArea::~GTK_DrawingArea()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::~GTK_DrawingArea" << std::endl;
#endif
        clearWorkspace();
        if ( layer_ != 0 )
          g_object_unref( layer_ );
        if ( GTK_IS_WIDGET( widget() ) )
          gtk_widget_destroy( widget() );
      }

      /*
        GTK_DrawingArea::getCoord : 指定した widget 上の座標 c を現レイヤの座標に変換する
      */
      Coord< double > GTK_DrawingArea::getCoord( Coord< double > c ) const
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::getCoord" << std::endl;
#endif
        return( pixbuf_.getCoord( Coord< double >
                                  ( ( c.x - offset_.x ) / scale_.x,
                                    ( c.y - offset_.y ) / scale_.y ) ) );
      }

      /*
        GTK_DrawingArea::setScale : 描画倍率を scale_ にセットする
      */
      void GTK_DrawingArea::setScale( Coord< double > scale, bool emitSignal )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::setScale" << std::endl;
#endif
        Coord< double > newScale = Coord< double >( PNum( scale.x ), PNum( scale.y ) );

        if ( newScale.x == scale_.x && newScale.y == scale_.y ) return;

        offset_.x *= newScale.x / scale_.x;
        offset_.y *= newScale.y / scale_.y;

        scale_ = newScale;

        recalcCoord();

        if ( emitSignal )
          emitValChangeSignal();

#ifdef DEBUG
        std::cout << "GTK_DrawingArea::setScale End" << std::endl;
#endif
      }

      /*
        GTK_DrawingArea::setOffset : 描画位置を offset_ にセットする
      */
      void GTK_DrawingArea::setOffset( Coord< double > c, bool emitSignal )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::setOffset" << std::endl;
#endif
        if ( offset_.x == c.x && offset_.y == c.y ) return;

        offset_.x = c.x;
        offset_.y = c.y;

        recalcCoord();

        if ( emitSignal )
          emitValChangeSignal();
      }

      /*
        GTK_DrawingArea::shiftOffset : 描画位置をシフトする
      */
      void GTK_DrawingArea::shiftOffset( Coord< double > c, bool emitSignal )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::shiftOffset" << std::endl;
#endif
        offset_.x += c.x;
        offset_.y += c.y;

        recalcCoord();

        if ( emitSignal )
          emitValChangeSignal();
      }

      /*
        GTK_DrawingArea::createWorkspace : ワークスペースの生成
      */
      void GTK_DrawingArea::createWorkspace()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::createWorkspace" << std::endl;
#endif
        if ( workspace_ != nullptr ) return;

        Coord< int > newSize( gtk_widget_get_allocated_width( widget() ), gtk_widget_get_allocated_height( widget() ) );

        workspace_ = gdk_pixbuf_new( GDK_COLORSPACE_RGB, TRUE, BITS_PER_SAMPLE, newSize.x, newSize.y );
        gdk_pixbuf_fill( workspace_, 0 );
      }

      /*
        GTK_DrawingArea::clearWorkspace : ワークスペースの消去
      */
      void GTK_DrawingArea::clearWorkspace()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::clearWorkspace" << std::endl;
#endif
        if ( workspace_ == nullptr ) return;

        g_object_unref( workspace_ );
        workspace_ = nullptr;
      }

      /*
        GDK_Pixbuf::size : ワークスペースのサイズを返す
      */
      Coord< int > GTK_DrawingArea::size() const
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::size" << std::endl;
#endif
        return( Coord< int >( ( workspace_ != 0 ) ? gdk_pixbuf_get_width( workspace_ ) : 0,
                              ( workspace_ != 0 ) ? gdk_pixbuf_get_height( workspace_ ) : 0 ) );
      }

      /*
        GDK_Pixbuf::resize : ワークスペースのサイズを変更する

        外部からリサイズすることはできないものとする。
      */
      void GTK_DrawingArea::resize( Coord< int > size )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::resize" << std::endl;
#endif
        if ( workspace_ == nullptr ) return;

        GdkPixbuf* newPixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, TRUE, BITS_PER_SAMPLE, NNNum( size.x ), NNNum( size.y ) );
        if ( newPixbuf == nullptr )
          throw std::bad_alloc();
        gdk_pixbuf_fill( newPixbuf, 0 );

        g_object_unref( workspace_ );

        workspace_ = newPixbuf;
      }

      /*
        GTK_DrawingArea::point : 座標 c の RGB 成分を palet に取得する
      */
      bool GTK_DrawingArea::point( Coord< int > c, RGB* palet ) const
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::point" << std::endl;
#endif
        assert( palet != nullptr );

        if ( layer_ == nullptr ) return( false );

        if ( c.x < 0 || c.x >= gdk_pixbuf_get_width( layer_ ) ) return( false );
        if ( c.y < 0 || c.y >= gdk_pixbuf_get_height( layer_ ) ) return( false );

        const RGB::primary_type* pixel = gdk_pixbuf_get_pixels( layer_ ) +
          c.y * gdk_pixbuf_get_rowstride( layer_ ) + c.x * gdk_pixbuf_get_n_channels( layer_ );
        if ( gdk_pixbuf_get_n_channels( layer_ ) > 3 )
          palet->receiveFromParam( pixel[0], pixel[1], pixel[2], pixel[3] );
        else
          palet->receiveFromParam( pixel[0], pixel[1], pixel[2] );

        return( true );
      }

      /*
        GTK_DrawingArea::pset : 座標 c のピクセルを palet でセットする
      */
      bool GTK_DrawingArea::pset( Coord< int > c, RGB palet )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea::pset" << std::endl;
#endif
        if ( workspace_ == nullptr ) return( false );

        if ( c.x < 0 || c.x >= gdk_pixbuf_get_width( workspace_ ) ) return( false );
        if ( c.y < 0 || c.y >= gdk_pixbuf_get_height( workspace_ ) ) return( false );

        guchar* pixel = gdk_pixbuf_get_pixels( workspace_ ) +
          c.y * gdk_pixbuf_get_rowstride( workspace_ ) + c.x * gdk_pixbuf_get_n_channels( workspace_ );
        if ( gdk_pixbuf_get_n_channels( workspace_ ) > 3 ) {
          palet.sendToParam( pixel, pixel + 1, pixel + 2, pixel + 3 );
        } else {
          palet.sendToParam( pixel, pixel + 1, pixel + 2 );
        }

        return( true );
      }

    } // namespace canvas

    namespace gtk_event
    {
      /*****************************************************
         GTK_DrawingArea_Eventクラス用 イベント処理基底クラス
      *****************************************************/

      /*
        GTK_DrawingArea_Event::connect : シグナル signal に対するコールバック関数 func の接続

        ident はシグナルに対する識別子
      */
      bool GTK_DrawingArea_Event::connect( const string& ident, const string& signal, GCallback func )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_Event::connect" << std::endl;
#endif
        if ( id_.find( ident ) != id_.end() ) {
          std::cerr << "ident [" << ident << "] already registered." << std::endl;
          return( false );
        }

        guint handleId = g_signal_connect( G_OBJECT( drawingArea_.widget() ), signal.c_str(), func, this );
        if ( handleId > 0 ) {
          id_[ident] = handleId;
          return( true );
        } else {
          return( false );
        }
      }

      /*
        GTK_DrawingArea_Event::block : シグナルに対するコールバック関数のブロック
      */
      void GTK_DrawingArea_Event::block( const std::string& ident )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_Event::block" << std::endl;
#endif
        std::map< std::string, guint >::const_iterator cit = id_.find( ident );
        if ( cit != id_.end() && G_IS_OBJECT( drawingArea_.widget() ) )
          g_signal_handler_block( G_OBJECT( drawingArea_.widget() ), cit->second );
      }

      /*
        GTK_DrawingArea_Event::block : シグナルに対するコールバック関数のブロック解除
      */
      void GTK_DrawingArea_Event::unblock( const std::string& ident )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_Event::unblock" << std::endl;
#endif
        std::map< std::string, guint >::const_iterator cit = id_.find( ident );
        if ( cit != id_.end() && G_IS_OBJECT( drawingArea_.widget() ) )
          g_signal_handler_unblock( G_OBJECT( drawingArea_.widget() ), cit->second );
      }

      /*
        GTK_DrawingArea_Event::block : 全シグナルに対するコールバック関数のブロック
      */
      void GTK_DrawingArea_Event::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_Event::block" << std::endl;
#endif
        for ( std::map< std::string, guint >::const_iterator cit = id_.begin() ;
              cit != id_.end() ; ++cit )
          if ( G_IS_OBJECT( drawingArea_.widget() ) )
            g_signal_handler_block( G_OBJECT( drawingArea_.widget() ), cit->second );
      }

      /*
        GTK_DrawingArea_Event::block : 全シグナルに対するコールバック関数のブロック解除
      */
      void GTK_DrawingArea_Event::unblock()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_Event::unblock" << std::endl;
#endif
        for ( std::map< std::string, guint >::const_iterator cit = id_.begin() ;
              cit != id_.end() ; ++cit )
          if ( G_IS_OBJECT( drawingArea_.widget() ) )
            g_signal_handler_unblock( G_OBJECT( drawingArea_.widget() ), cit->second );
      }

      /*
        GTK_DrawingArea_Event デストラクタ
      */
      GTK_DrawingArea_Event::~GTK_DrawingArea_Event()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_Event::~GTK_DrawingArea_Event" << std::endl;
#endif
        for ( std::map< std::string, guint >::const_iterator cit = id_.begin() ;
              cit != id_.end() ; ++cit )
          if ( G_IS_OBJECT( drawingArea_.widget() ) ) // drawingArea_.widget() が後で解体される保証はない
            if ( g_signal_handler_is_connected( G_OBJECT( drawingArea_.widget() ), cit->second ) )
              g_signal_handler_disconnect( G_OBJECT( drawingArea_.widget() ), cit->second );
      }

    } // namespace gtk_event

    /*****************************************************
         GTK_DrawingArea用 scale 操作
    *****************************************************/

    namespace
    {
      // デフォルト値
      double scale_[] = {
                         1.0 / 256, 1.0 / 180, 1.0 / 128, 1.0 / 90, 1.0 / 64, 1.0 / 45, 1.0 / 32, 1.0 / 23,
                         1.0 / 16, 1.0 / 11, 1.0 / 8, 1.0 / 5.5, 1.0 / 4, 1.0 / 3, 1.0 / 2, 1.0 / 1.5,
                         1, 1.5, 2, 3, 4, 5.5, 8, 11, 16, 23, 32, 45, 64, 90, 128, 180, 256,
      };

      // ひとつ前のスケールに戻る
      double decScale( double curScale )
      {
        double* p = std::lower_bound( scale_, scale_ + sizeof( scale_ ) / sizeof( scale_[0] ), curScale );
        if ( p == scale_ )
          return( curScale );

        return( *( p - 1 ) );
      }

      // 次のスケールに進む
      double incScale( double curScale )
      {
        double* p = std::upper_bound( scale_, scale_ + sizeof( scale_ ) / sizeof( scale_[0] ), curScale );
        if ( p == scale_ + sizeof( scale_ ) / sizeof( scale_[0] ) )
          return( curScale );

        return( *p );
      }

    } // namespace

    namespace gtk_event
    {
      /*****************************************************
         GTK_DrawingAreaクラス用 画像シフト機能

         マウスドラッグによって、画像の移動を行う処理
      *****************************************************/

      /*
        GTK_DrawingArea_ImageShiftEvent コンストラクタ

        描画対象の GTK_DrawingArea を指定して構築
      */
      GTK_DrawingArea_ImageShiftEvent::GTK_DrawingArea_ImageShiftEvent( canvas::GTK_DrawingArea& drawingArea )
        : GTK_DrawingArea_Event( drawingArea ), mouse( Coord<int>( -1, -1 ) ), scrollTime_( 0 )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ImageShiftEvent::GTK_DrawingArea_ImageShiftEvent" << std::endl;
#endif
        connect( "MotionNotify", "motion_notify_event", G_CALLBACK( event ) );
        connect( "Scroll", "scroll-event", G_CALLBACK( scrollEvent ) );
        connect( "Redraw", "redraw", G_CALLBACK( redrawEvent ) );
      }

      /*
        GTK_DrawingArea_ImageShiftEvent::event : マウスドラッグ時のイベント

        widget : 対象のwidget(GTK_DrawingArea内のwidgetでここでは未使用)
        event : GdkEventExposeオブジェクト
        gp : GTK_DrawingArea_ImageShiftEventオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ImageShiftEvent::event( GtkWidget *widget, GdkEventMotion *event, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ImageShiftEvent::event" << std::endl;
#endif
        GTK_DrawingArea_ImageShiftEvent* d = (GTK_DrawingArea_ImageShiftEvent*)gp;
        canvas::GTK_DrawingArea& drawingArea = d->drawingArea();
        Coord< int > c;
        GdkModifierType state;

        gdk_window_get_device_position( event->window, event->device, &( c.x ), &( c.y ), &state );
        if ( ( state & GDK_BUTTON1_MASK ) && ( d->mouse.x >= 0 && d->mouse.y >= 0 ) ) {
          Coord<int> buff( c );
          buff -= d->mouse;
          drawingArea.shiftOffset( Coord< double >( buff.x, buff.y ), true );
        }

        d->mouse = c;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_ImageShiftEvent::scrollEvent : マウスホイール使用時のイベント

        widget : GTK_DrawingArea内のウィジェット
        event : GdkEventScrollオブジェクト
        gp : GTK_DrawingArea_ImageShiftEventオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ImageShiftEvent::scrollEvent( GtkWidget* widget, GdkEventScroll* event, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ImageShiftEvent::scrollEvent" << std::endl;
#endif
        GTK_DrawingArea_ImageShiftEvent* d = (GTK_DrawingArea_ImageShiftEvent*)gp;
        canvas::GTK_DrawingArea& drawingArea = d->drawingArea();

        // 前回から 100msec 以内のイベントはスキップ
        if ( event->time < d->scrollTime_ + 100 && d->scrollTime_ != 0 ) return( FALSE );

        // モディファイア・キー ( Ctrl, Shift, Alt, Super, Hyper, Meta ) のみをONにするマスク
        GdkModifierType modifiers = gtk_accelerator_get_default_mod_mask();
        // シフトキーが押されているかをチェック
        bool shiftKeyPressed = ( ( event->state & modifiers ) == GDK_SHIFT_MASK );
        // Ctrlキーが押されているかをチェック
        bool ctrlKeyPressed = ( ( event->state & modifiers ) == GDK_CONTROL_MASK );

        Coord< int > c;
        switch ( event->direction ) {
        case GDK_SCROLL_UP:
          if ( shiftKeyPressed ) {
            drawingArea.shiftOffset( Coord< double >( gtk_widget_get_allocated_width( drawingArea.widget() ) / 2, 0 ),
                                     true );
          } else if ( ctrlKeyPressed ) {
            Coord< double > scale = drawingArea.getScale();
            Coord< double > newScale( incScale( scale.x ), incScale( scale.y ) );
            if ( scale.x < newScale.x || scale.y < newScale.y ) {
              drawingArea.setScale( newScale );
              drawingArea.shiftOffset( Coord< double >
                                       ( -event->x * ( newScale.x / scale.x - 1 ),
                                         -event->y * ( newScale.y / scale.y - 1 ) ),
                                       true );
            }
          } else {
            drawingArea.shiftOffset( Coord< double >( 0, gtk_widget_get_allocated_height( drawingArea.widget() ) / 2 ),
                                     true );
          }
          break;
        case GDK_SCROLL_DOWN:
          if ( shiftKeyPressed ) {
            drawingArea.shiftOffset( Coord< double >( -gtk_widget_get_allocated_width( drawingArea.widget() ) / 2, 0 ),
                                     true );
          } else if ( ctrlKeyPressed ) {
            Coord< double > scale = drawingArea.getScale();
            Coord< double > newScale( decScale( scale.x ), decScale( scale.y ) );
            if ( scale.x > newScale.x || scale.y > newScale.y ) {
              drawingArea.setScale( newScale );
              drawingArea.shiftOffset( Coord< double >
                                       ( -event->x * ( newScale.x / scale.x - 1 ),
                                         -event->y * ( newScale.y / scale.y - 1 ) ),
                                       true );
            }
          } else {
            drawingArea.shiftOffset( Coord< double >( 0, -gtk_widget_get_allocated_height( drawingArea.widget() ) / 2 ),
                                     true);
          }
          break;
        default:
          break;
        }

        d->scrollTime_ = event->time;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_ImageShiftEvent::redrawEvent : オフセット・スケールの変更による再描画

        widget : GTK_DrawingArea内のウィジェット
        gp : GTK_DrawingArea_ImageShiftEventオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ImageShiftEvent::redrawEvent( GtkWidget* widget, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ImageShiftEvent::redrawEvent" << std::endl;
#endif
        GTK_DrawingArea_ImageShiftEvent* d = (GTK_DrawingArea_ImageShiftEvent*)gp;
        canvas::GTK_DrawingArea& drawingArea = d->drawingArea();

        drawingArea.draw();

        return( FALSE );
      }

    } // namespace gtk_event

    namespace canvas
    {
      /*****************************************************
          GTK_DrawingArea用 Scrolled Windowクラス
      *****************************************************/

      /*
        GTK_DrawingArea_ScrolledWindow::hEvent, vEvent : 位置変更時のイベント

        adj : 対象の GtkAdjustment
        gp : GTK_DrawingArea_ScrolledWindowオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ScrolledWindow::hEvent( GtkAdjustment* adj, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::hEvent" << std::endl;
#endif
        GTK_DrawingArea_ScrolledWindow* gtkDA_SW = (GTK_DrawingArea_ScrolledWindow*)gp;
        GTK_DrawingArea& drawingArea = *( gtkDA_SW->drawingArea() );
        ( gtkDA_SW->value_ ).x = gtk_adjustment_get_value( adj );

        drawingArea.setOffset( Coord< double >( -( gtkDA_SW->value_ ).x, -( gtkDA_SW->value_ ).y ), true );

        return( FALSE );
      }
      gint GTK_DrawingArea_ScrolledWindow::vEvent( GtkAdjustment* adj, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::vEvent" << std::endl;
#endif
        GTK_DrawingArea_ScrolledWindow* gtkDA_SW = (GTK_DrawingArea_ScrolledWindow*)gp;
        GTK_DrawingArea& drawingArea = *( gtkDA_SW->drawingArea() );
        ( gtkDA_SW->value_ ).y = gtk_adjustment_get_value( adj );

        drawingArea.setOffset( Coord< double >( -( gtkDA_SW->value_ ).x, -( gtkDA_SW->value_ ).y ), true );

        return( FALSE );
      }

      /*
        GTK_DrawingArea_ScrolledWindow::scrollEvent : マウスホイール使用時のイベント

        widget : GTK_DrawingArea内のウィジェット
        event : GdkEventScrollオブジェクト
        gp : GTK_DrawingArea_ScrolledWindowオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ScrolledWindow::scrollEvent( GtkWidget* widget, GdkEventScroll* event, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::scrollEvent" << std::endl;
#endif
        GTK_DrawingArea_ScrolledWindow* gtkDA_SW = (GTK_DrawingArea_ScrolledWindow*)gp;
        GTK_DrawingArea& drawingArea = *( gtkDA_SW->drawingArea() );

        // 前回から 100msec 以内のイベントはスキップ
        if ( event->time < gtkDA_SW->scrollTime_ + 100 && gtkDA_SW->scrollTime_ != 0 ) return( FALSE );

        // モディファイア・キー ( Ctrl, Shift, Alt, Super, Hyper, Meta ) のみをONにするマスク
        GdkModifierType modifiers = gtk_accelerator_get_default_mod_mask();
        // シフトキーが押されているかをチェック
        bool shiftKeyPressed = ( ( event->state & modifiers ) == GDK_SHIFT_MASK );
        // Ctrl キーが押されているかをチェック
        bool ctrlKeyPressed = ( ( event->state & modifiers ) == GDK_CONTROL_MASK );

        switch ( event->direction ) {
        case GDK_SCROLL_UP:
          if ( shiftKeyPressed ) {
            ( gtkDA_SW->value_ ).x -= gtk_widget_get_allocated_width( drawingArea.widget() ) / 2;
            gtk_adjustment_set_value( gtkDA_SW->hAdj_, ( gtkDA_SW->value_ ).x );
            ( gtkDA_SW->value_ ).x = gtk_adjustment_get_value( gtkDA_SW->hAdj_ );
          } else if ( ctrlKeyPressed ) {
            gtkDA_SW->changeScale( true, Coord< double >( event->x, event->y ) );
          } else {
            ( gtkDA_SW->value_ ).y -= gtk_widget_get_allocated_height( drawingArea.widget() ) / 2;
            gtk_adjustment_set_value( gtkDA_SW->vAdj_, ( gtkDA_SW->value_ ).y );
            ( gtkDA_SW->value_ ).y = gtk_adjustment_get_value( gtkDA_SW->vAdj_ );
          }
          break;
        case GDK_SCROLL_DOWN:
          if ( shiftKeyPressed ) {
            ( gtkDA_SW->value_ ).x += gtk_widget_get_allocated_width( drawingArea.widget() ) / 2;
            gtk_adjustment_set_value( gtkDA_SW->hAdj_, ( gtkDA_SW->value_ ).x );
            ( gtkDA_SW->value_ ).x = gtk_adjustment_get_value( gtkDA_SW->hAdj_ );
          } else if ( ctrlKeyPressed ) {
            gtkDA_SW->changeScale( false, Coord< double >( event->x, event->y ) );
          } else {
            ( gtkDA_SW->value_ ).y += gtk_widget_get_allocated_height( drawingArea.widget() ) / 2;
            gtk_adjustment_set_value( gtkDA_SW->vAdj_, ( gtkDA_SW->value_ ).y );
            ( gtkDA_SW->value_ ).y = gtk_adjustment_get_value( gtkDA_SW->vAdj_ );
          }
          break;
        default:
          break;
        }

        gtkDA_SW->scrollTime_ = event->time;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_ScrolledWindow::sizeChangedEvent : サイズ変更時のイベント

        widget : GTK_DrawingArea内のウィジェット
        event : GdkEventConfigureオブジェクト
        gp : GTK_DrawingArea_ScrolledWindowオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ScrolledWindow::sizeChangedEvent( GtkWidget *widget, GdkEventConfigure *event, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::sizeChangeEvent" << std::endl;
        std::cout << "event ( " << event->x << ", " << event->y << " )" << std::endl;
        std::cout << "event ( " << event->width << ", " << event->height << " )" << std::endl;
#endif
        GTK_DrawingArea_ScrolledWindow* gtkDA_SW = (GTK_DrawingArea_ScrolledWindow*)gp;

        gtkDA_SW->setUpper();

        gtk_adjustment_set_page_size( gtkDA_SW->hAdj_, event->width );
        gtk_adjustment_set_page_size( gtkDA_SW->vAdj_, event->height );

        gtk_adjustment_set_page_increment( gtkDA_SW->hAdj_, event->width );
        gtk_adjustment_set_page_increment( gtkDA_SW->vAdj_, event->height );

        if ( gtk_adjustment_get_value( gtkDA_SW->hAdj_ ) < gtk_adjustment_get_lower( gtkDA_SW->hAdj_ ) )
          gtk_adjustment_set_value( gtkDA_SW->hAdj_, gtk_adjustment_get_lower( gtkDA_SW->hAdj_ ) );
        if ( gtk_adjustment_get_value( gtkDA_SW->hAdj_ ) > gtk_adjustment_get_upper( gtkDA_SW->hAdj_ ) - gtk_adjustment_get_page_size( gtkDA_SW->hAdj_ ) )
          gtk_adjustment_set_value( gtkDA_SW->hAdj_, gtk_adjustment_get_upper( gtkDA_SW->hAdj_ ) - gtk_adjustment_get_page_size( gtkDA_SW->hAdj_ ) );

        if ( gtk_adjustment_get_value( gtkDA_SW->vAdj_ ) < gtk_adjustment_get_lower( gtkDA_SW->vAdj_ ) )
          gtk_adjustment_set_value( gtkDA_SW->vAdj_, gtk_adjustment_get_lower( gtkDA_SW->vAdj_ ) );
        if ( gtk_adjustment_get_value( gtkDA_SW->vAdj_ ) > gtk_adjustment_get_upper( gtkDA_SW->vAdj_ ) - gtk_adjustment_get_page_size( gtkDA_SW->vAdj_ ) )
          gtk_adjustment_set_value( gtkDA_SW->vAdj_, gtk_adjustment_get_upper( gtkDA_SW->vAdj_ ) - gtk_adjustment_get_page_size( gtkDA_SW->vAdj_ ) );

        return( FALSE );
      }

      /*
        GTK_DrawingArea_ScrolledWindow::redrawEvent : オフセット・スケールの変更による再描画

        widget : GTK_DrawingArea内のウィジェット
        gp : GTK_DrawingArea_ScrolledWindowオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_ScrolledWindow::redrawEvent( GtkWidget* widget, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::redrawEvent" << std::endl;
#endif
        GTK_DrawingArea_ScrolledWindow* gtkDA_SW = (GTK_DrawingArea_ScrolledWindow*)gp;
        GTK_DrawingArea& drawingArea = *( gtkDA_SW->drawingArea() );

        drawingArea.draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_ScrolledWindow::connect : シグナルに対するコールバック関数の接続

        ident : 内部で利用する識別子
        eventName : signal の名称
        obj : コールバック対象のオブジェクト
        func : コールバック関数
      */
      void GTK_DrawingArea_ScrolledWindow::connect( const string& ident, const string& eventName, GObject* obj, GCallback func )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::connect" << std::endl;
#endif
        if ( eventInfo_.find( ident ) == eventInfo_.end() )
          eventInfo_.insert( std::pair< string, event >( ident, event( g_signal_connect( obj, eventName.c_str(), func, this ), obj ) ) );
      }

      /*
        GTK_DrawingArea_ScrolledWindow コンストラクタ

        スクロールバーを付ける対象の GTK_DrawingArea オブジェクトを指定して構築
      */
      GTK_DrawingArea_ScrolledWindow::GTK_DrawingArea_ScrolledWindow( GTK_DrawingArea& drawingArea )
        : drawingArea_( drawingArea ), value_( Coord<gdouble>( 0.0, 0.0 ) ), scrollTime_( 0 )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::GTK_DrawingArea_ScrolledWindow" << std::endl;
#endif
        /* ウィジェットの宣言 */

        gtk_widget_set_hexpand( GTK_WIDGET( drawingArea.widget() ), TRUE );
        gtk_widget_set_vexpand( GTK_WIDGET( drawingArea.widget() ), TRUE );

        // レイアウト用コンテナ
        table_ = gtk_grid_new();
        gtk_grid_set_row_homogeneous( GTK_GRID( table_ ), FALSE );
        gtk_grid_set_column_homogeneous( GTK_GRID( table_ ), FALSE );
        gtk_grid_set_row_spacing( GTK_GRID( table_ ), 0 );
        gtk_grid_set_column_spacing( GTK_GRID( table_ ), 0 );

        // 位置調節用 I/F
        hAdj_ = GTK_ADJUSTMENT( gtk_adjustment_new( 0.0, 0.0, 1.0, 1.0, 0, 0 ) );
        vAdj_ = GTK_ADJUSTMENT( gtk_adjustment_new( 0.0, 0.0, 1.0, 1.0, 0, 0 ) );

        // スクロールバー
        hBar_ = gtk_scrollbar_new( GTK_ORIENTATION_HORIZONTAL, hAdj_ );
        vBar_ = gtk_scrollbar_new( GTK_ORIENTATION_VERTICAL, vAdj_ );

        /* ウィジェットの配置 */
        gtk_grid_attach( GTK_GRID( table_ ), drawingArea_.widget(), 0, 0, 1, 1 );
        gtk_grid_attach( GTK_GRID( table_ ), GTK_WIDGET( vBar_ ), 1, 0, 1, 1 );
        gtk_grid_attach( GTK_GRID( table_ ), GTK_WIDGET( hBar_ ), 0, 1, 1, 1 );

        connect( "hAdj-value-changed", "value-changed", G_OBJECT( hAdj_ ), G_CALLBACK( hEvent ) );
        connect( "vAdj-value-changed", "value-changed", G_OBJECT( vAdj_ ), G_CALLBACK( vEvent ) );

        connect( "scroll", "scroll-event", G_OBJECT( drawingArea_.widget() ), G_CALLBACK( scrollEvent ) );
        connect( "configure-event", "configure-event", G_OBJECT( drawingArea_.widget() ), G_CALLBACK( sizeChangedEvent ) );
        connect( "redraw", "redraw", G_OBJECT( drawingArea_.widget() ), G_CALLBACK( redrawEvent ) );

        gtk_widget_show_all( table_ );
      }

      /*
        GTK_DrawingArea_ScrolledWindow デストラクタ

        シグナルハンドラの切断を行う
      */
      GTK_DrawingArea_ScrolledWindow::~GTK_DrawingArea_ScrolledWindow()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::~GTK_DrawingArea_ScrolledWindow" << std::endl;
#endif
        std::map<string, event>::iterator it = eventInfo_.begin();
        for ( ; it == eventInfo_.end() ; ++it )
          if ( G_IS_OBJECT( ( it->second ).second ) )
            if ( g_signal_handler_is_connected( ( it->second ).second, ( it->second ).first ) )
              g_signal_handler_disconnect( ( it->second ).second, ( it->second ).first );
      }

      /*
        GTK_DrawingArea_ScrolledWindow::initAdjustment : スクロール位置(画像表示位置)の初期化
      */
      void GTK_DrawingArea_ScrolledWindow::initAdjustment()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::initAdjustment" << std::endl;
#endif
        gtk_adjustment_set_value( hAdj_, 0.0 );
        gtk_adjustment_set_value( vAdj_, 0.0 );

        setUpper();
      }

      void GTK_DrawingArea_ScrolledWindow::setUpper()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::setUpper" << std::endl;
        std::cout << "width = " << drawingArea_.pixbuf().canvasSize().x << " * " << drawingArea_.getScale().x << " = " << drawingArea_.pixbuf().canvasSize().x * drawingArea_.getScale().x;
        std::cout << " height = " << drawingArea_.pixbuf().canvasSize().y << " * " << drawingArea_.getScale().y << " = " << drawingArea_.pixbuf().canvasSize().y * drawingArea_.getScale().y << std::endl;
#endif
        if ( drawingArea_.pixbuf().canvasSize().x == 0 || drawingArea_.pixbuf().canvasSize().y == 0 ) return;

        gtk_adjustment_set_upper( hAdj_, drawingArea_.pixbuf().canvasSize().x * drawingArea_.getScale().x );
        gtk_adjustment_set_upper( vAdj_, drawingArea_.pixbuf().canvasSize().y * drawingArea_.getScale().y );
      }

      void GTK_DrawingArea_ScrolledWindow::changeScale( bool incScale_, Coord< double > mouse )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::changeScale" << std::endl;
        std::cout << mouse.x << ", " << mouse.y << std::endl;
#endif
        Coord< double > scale = drawingArea_.getScale();
        Coord< double > newScale( ( incScale_ ) ? incScale( scale.x ) : decScale( scale.x ),
                                  ( incScale_ ) ? incScale( scale.y ) : decScale( scale.y ) );

#ifdef DEBUG
        std::cout << "scale = " << scale.x << "," << scale.y << std::endl;
        std::cout << "newScale = " << newScale.x << "," << newScale.y << std::endl;
#endif
        if ( scale.x == newScale.x && scale.y == newScale.y ) return;

        value_.x /= scale.x;
        value_.y /= scale.y;
        drawingArea_.setScale( newScale, true );
        setUpper();
        value_.x = value_.x * newScale.x + mouse.x * ( newScale.x / scale.x - 1 );
        value_.y = value_.y * newScale.y + mouse.y * ( newScale.y / scale.y - 1 );
        gtk_adjustment_set_value( hAdj_, value_.x );
        gtk_adjustment_set_value( vAdj_, value_.y );
        value_.x = gtk_adjustment_get_value( hAdj_ );
        value_.y = gtk_adjustment_get_value( vAdj_ );

#ifdef DEBUG
        std::cout << "GTK_DrawingArea_ScrolledWindow::changeScale End" << std::endl;
#endif
      }

    } // namespace canvas

    /*****************************************************
        描画用関数オブジェクト ヘルパ関数
    *****************************************************/

    namespace
    {
      // ワークスペースへの描画用ピクセル操作
      pen::GNotSet notSet_; // 反転描画
      pen::GPSet erase_( RGB( 0, 0, 0, 0 ) ); // ドットの消去

      // カーソル描画を行わない場合に NULL ポインタを指定するために利用
      //draw::DrawFuncIF_1p* drawNone_ = nullptr;

      /*
        現レイヤ上の座標から widget 上の座標に変換する

        c : 現レイヤ上の座標
        layerOffset : 現レイヤのキャンバスに対するオフセット値
        canvasOffset : キャンバスの widget に対するオフセット値
        scale : 拡大・縮小率
      */
      int GetInvCoord( double c, double layerOffset, double canvasOffset, double scale )
      { return( std::round( ( c + layerOffset ) * scale + canvasOffset ) ); }

      /*
        DrawHLine : 描画領域 draw に水平方向の線分を pixelOp で描画する

        c0 は始点(左側)の座標値、ex は終点(右側)の X 座標をそれぞれ表す
      */
      void DrawHLine( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, int ex )
      {
        for ( ; c0.x <= ex ; ++( c0.x ) )
          pixelOp( draw, c0 );
      }

      /*
        DrawVLine : 描画領域 draw に垂直方向の線分を pixelOp で描画する

        c0 は始点(上側)の座標値、ey は終点(下側)の Y 座標をそれぞれ表す
      */
      void DrawVLine( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, int ey )
      {
        for ( ; c0.y <= ey ; ++( c0.y ) )
          pixelOp( draw, c0 );
      }

    }

    namespace draw
    {
      /*****************************************************
        描画用関数オブジェクト 十字描画
      *****************************************************/

      /*
        DrawCross::operator() : 描画領域 draw 上に十字を pixelOp で描画する

        c は描画位置(十字の交差する点)を表す
      */
      void DrawCross::operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c )
      {
        Coord< int > drawSize = draw.size();

        DrawHLine( draw, pixelOp, Coord< int >( 0, c.y ), drawSize.x - 1 );
        DrawVLine( draw, pixelOp, Coord< int >( c.x, 0 ), drawSize.y - 1 );
      }

      /*****************************************************
         描画用関数オブジェクト 円形ブラシ形状の描画
      *****************************************************/

      /*
        DrawCircleBrush::operator() : 描画領域 draw 上に楕円ペン形状をを pixelOp で描画する

        o は楕円の中心 ( 焦点の中点 ) を表す
      */
      void DrawCircleBrush::operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o )
      {
        canvas::GTK_DrawingArea* gda = dynamic_cast< canvas::GTK_DrawingArea* >( &draw );

        int r = brush_.getR() * ( ( gda != nullptr ) ? gda->getScale().x : 1 );

        shape::circle::Stroke( draw, pixelOp, o, r );
        if ( r >= MIN_RADIUS_TO_DRAW_CROSS ) {
          shape::line::DrawH( draw, pixelOp, Coord< int >( o.x - MIN_RADIUS_TO_DRAW_CROSS, o.y ), o.x + MIN_RADIUS_TO_DRAW_CROSS );
          shape::line::DrawV( draw, pixelOp, Coord< int >( o.x, o.y - MIN_RADIUS_TO_DRAW_CROSS ), o.y + MIN_RADIUS_TO_DRAW_CROSS );
        }
      }

      /*****************************************************
         描画用関数オブジェクト 楕円ペン形状の描画
      *****************************************************/

      /*
        DrawEllipsePen::operator() : 描画領域 draw 上に楕円ペン形状をを pixelOp で描画する

        o は楕円の中心 ( 焦点の中点 ) を表す
      */
      void DrawEllipsePen::operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o )
      {
        canvas::GTK_DrawingArea* gda = dynamic_cast< canvas::GTK_DrawingArea* >( &draw );

        int r = pen_.getR() * ( ( gda != nullptr ) ? gda->getScale().x : 1 );
        Coord< int > k = pen_.getK();

        shape::ellipse::Stroke( draw, pixelOp, o, r, k );
        if ( r >= MIN_RADIUS_TO_DRAW_CROSS ) {
          shape::line::DrawH( draw, pixelOp, Coord< int >( o.x - MIN_RADIUS_TO_DRAW_CROSS, o.y ), o.x + MIN_RADIUS_TO_DRAW_CROSS );
          shape::line::DrawV( draw, pixelOp, Coord< int >( o.x, o.y - MIN_RADIUS_TO_DRAW_CROSS ), o.y + MIN_RADIUS_TO_DRAW_CROSS );
        }
      }

    } // namespace draw

    namespace gtk_event
    {
      /*******************************************************************
        GTK_DrawingArea_SelectionEvent : GTK_DrawingArea選択設定用イベント
      *******************************************************************/

      /*
        GTK_DrawingArea_SelectionEvent : コンストラクタ

        描画対象 drawingArea を指定して構築
        マウスボタンのクリック・リリース、キーボードのクリック・リリース、マウスの移動に対する
        コールバック関数への接続をする
      */
      GTK_DrawingArea_SelectionEvent::GTK_DrawingArea_SelectionEvent( canvas::GTK_DrawingArea& drawingArea )
        : GTK_DrawingArea_Event( drawingArea )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_SelectionEvent::GTK_DrawingArea_SelectionEvent" << std::endl;
#endif
        connect( "EnterNotify", "enter-notify-event", G_CALLBACK( enterNotifyEvent ) );
        connect( "LeaveNotify", "leave-notify-event", G_CALLBACK( leaveNotifyEvent ) );
        connect( "ButtonPress", "button-press-event", G_CALLBACK( buttonPressEvent ) );
        connect( "ButtonRelease", "button-release-event", G_CALLBACK( buttonReleaseEvent ) );
        connect( "KeyPress", "key-press-event", G_CALLBACK( keyPressEvent ) );
        connect( "KeyRelease", "key-release-event", G_CALLBACK( keyReleaseEvent ) );
        connect( "MotionNotify", "motion_notify_event", G_CALLBACK( motionNotifyEvent ) );
        connect( "Scroll", "scroll-event", G_CALLBACK( scrollEvent ) );
        connect( "Configure", "configure_event", G_CALLBACK( configureEvent ) );
        connect( "Redraw", "redraw", G_CALLBACK( redrawEvent ) );

        gtk_widget_set_can_focus( drawingArea.widget(), TRUE );
      }

      /*
        マウス移動時のコールバック関数

        widget : widget()オブジェクト
        event : GdkEventMotionイベント
        gp : GTK_DrawingArea_SelectionEventオブジェクト(自己参照ポインタ)
      */
      gint GTK_DrawingArea_SelectionEvent::motionNotifyEvent( GtkWidget *widget, GdkEventMotion *event, gpointer gp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_SelectionEvent::motionNotifyEvent" << std::endl;
#endif
        GTK_DrawingArea_SelectionEvent* daEvt = (GTK_DrawingArea_SelectionEvent*) gp;

        Coord< int > c;
        GdkModifierType state;
        getEventInfo( event, c, state );

        return( daEvt->motionNotifyAction( c, state ) );
      }

      /*******************************************************************************
        GTK_DrawingArea_DrawingEvent : GTK_DrawingArea選択範囲描画設定用イベント
      *******************************************************************************/

      /*
        イベントを定義する描画領域と点描画関数オブジェクトを指定して構築
      */
      GTK_DrawingArea_DrawingEvent::GTK_DrawingArea_DrawingEvent( canvas::GTK_DrawingArea& drawingArea, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_SelectionEvent( drawingArea )
      { setPixelOp( pixelOp ); }

      /*
        GTK_DrawingArea_DrawingEvent::setPixelOp : 点描画関数をセットする
      */
      void GTK_DrawingArea_DrawingEvent::setPixelOp( pen::GPixelOp* pixelOp )
      {
        assert( pixelOp != nullptr );

        pixelOp_ = pixelOp;
      }

      /*******************************************************************************
        GTK_DrawingArea_1PassSelectionEvent : GTK_DrawingArea選択設定用イベント(1パス)
      *******************************************************************************/

      /* GTK_DrawingArea_1PassSelectionEvent : コンストラクタ */

      // cursor0, draw を別々に定義
      GTK_DrawingArea_1PassSelectionEvent::GTK_DrawingArea_1PassSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea, draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_1p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( cursor0 ), draw_( draw ), isFirst_( true )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::GTK_DrawingArea_1PassSelectionEvent( GTK_DrawingArea&, DrawFuncIF_1p*&, DrawFuncIF_1p*&, GPixelOp*& )" << std::endl;
#endif
        assert( draw_ != nullptr );
        assert( drawingArea.wsIsValid() );
      }

      // cursor0 は描画なしの場合
      GTK_DrawingArea_1PassSelectionEvent::GTK_DrawingArea_1PassSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea, draw::DrawFuncIF_1p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( nullptr ), draw_( draw ), isFirst_( true )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::GTK_DrawingArea_1PassSelectionEvent( GTK_DrawingArea&, DrawFuncIF_1p*&, GPixelOp*& )" << std::endl;
#endif
        assert( draw_ != nullptr );
        assert( drawingArea.wsIsValid() );
      }

      // デストラクタ (カーソルの描画を消去して内容を初期化する)
      GTK_DrawingArea_1PassSelectionEvent::~GTK_DrawingArea_1PassSelectionEvent()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::~GTK_DrawingArea_1PassSelectionEvent" << std::endl;
        reset();
#endif
      }

      /*
        GTK_DrawingArea_1PassSelectionEvent::reset : 描画内容のクリア
      */
      void GTK_DrawingArea_1PassSelectionEvent::reset()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::reset" << std::endl;
#endif
        if ( ( ! isFirst_ ) && cursor0_ != nullptr ) {
          Coord< int > drawCoord( std::round( c_.x ), std::round( c_.y ) );
          (*cursor0_)( drawingArea(), erase_, drawCoord );
        }
      }

      /*
        GTK_DrawingArea_1PassSelectionEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_1PassSelectionEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::buttonPressAction" << std::endl;
#endif
        Coord< double > dc = drawingArea().getCoord( Coord< double >( event->x, event->y ) );
        Coord< int > c( std::round( dc.x ), std::round( dc.y ) );

        if ( event->button == 1 ) {
          reset();
          (*draw_)( drawingArea().pixbuf(), pixelOp(), c );
          isFirst_ = true;
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_1PassSelectionEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_1PassSelectionEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::motionNotifyAction" << std::endl;
#endif
        reset();
        c_ = Coord< double >( c.x, c.y );
        if ( cursor0_ != 0 ) {
          Coord< int > drawCoord( std::round( c_.x ), std::round( c_.y ) );
          (*cursor0_)( drawingArea(), notSet_, drawCoord );
          drawingArea().draw();
        }

        isFirst_ = false;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_1PassSelectionEvent::redrawAction : オフセット・スケール変更時のコールバック関数
      */
      gint GTK_DrawingArea_1PassSelectionEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::redrawAction" << std::endl;
#endif
        if ( cursor0_ != nullptr ) {
          reset();
          Coord< int > drawCoord( std::round( c_.x ), std::round( c_.y ) );
          (*cursor0_)( drawingArea(), notSet_, drawCoord );
          drawingArea().draw();
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_1PassSelectionEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去され、ステータスも初期状態になる
      */
      void GTK_DrawingArea_1PassSelectionEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_1PassSelectionEvent::block" << std::endl;
#endif
        reset();
        isFirst_ = true;
        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /*
        GTK_DrawingArea_1PassSelectionEvent::setDraw : 描画関数をセットする
      */
      void GTK_DrawingArea_1PassSelectionEvent::setDraw( draw::DrawFuncIF_1p* draw )
      {
        assert( draw != nullptr );

        draw_ = draw;
      }

      /*******************************************************************************
        GTK_DrawingArea_2PassSelectionEvent : GTK_DrawingArea選択設定用イベント(2パス)
      *******************************************************************************/

      /* GTK_DrawingArea_2PassSelectionEvent : コンストラクタ */

      // cursor0, cursor1, draw を別々に定義
      GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea,
        draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_2p* cursor1,
        draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( cursor0 ), cursor1_( cursor1 ),
          draw_( draw ), status_( NO_DRAW )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent( GTK_DrawingArea&, DrawFuncIF_1p*&, DrawFuncIF_2p*&, DrawFuncIF_2p*&, GPixelOp*& )" << std::endl;
#endif
        assert( cursor1_ != nullptr && draw_ != nullptr );
        assert( drawingArea.wsIsValid() );
      }

      // cursor1 と draw は同じ描画方式の場合
      GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea,
        draw::DrawFuncIF_1p* cursor0,
        draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( cursor0 ), cursor1_( draw ),
          draw_( draw ), status_( NO_DRAW )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent( GTK_DrawingArea&, DrawFuncIF_1p*&, DrawFuncIF_2p*&, GPixelOp*& )" << std::endl;
#endif
        assert( draw_ != nullptr );
        assert( drawingArea.wsIsValid() );
      }

      // cursor0 は描画なしの場合
      GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea,
        draw::DrawFuncIF_2p* cursor1,
        draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( 0 ), cursor1_( cursor1 ),
          draw_( draw ), status_( NO_DRAW )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent( GTK_DrawingArea&, DrawFuncIF_2p*&, DrawFuncIF_2p*&, GPixelOp*& )" << std::endl;
#endif
        assert( cursor1_ != nullptr && draw_ != nullptr );
        assert( drawingArea.wsIsValid() );
      }

      // cursor0 は描画なしで、cursor1 と draw は同じ描画方式の場合
      GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea,
        draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( 0 ), cursor1_( draw ),
          draw_( draw ), status_( NO_DRAW )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::GTK_DrawingArea_2PassSelectionEvent( GTK_DrawingArea&, DrawFuncIF_2p*&, GPixelOp*& )" << std::endl;
#endif
        assert( draw_ != nullptr );
        assert( drawingArea.wsIsValid() );
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_2PassSelectionEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::buttonPressAction" << std::endl;
#endif
        // 現レイヤ上の座標値
        Coord< double > c = drawingArea().getCoord( Coord< double >( event->x, event->y ) );

        // 左クリック
        if ( event->button == 1 ) {
          reset();
          // 2 点指定状態なら現レイヤに描画を行い描画なし状態に
          if ( status_ == CURSOR1 ) {
            // どちらも現レイヤ上の座標値
            Coord< int > ic0 = Coord< int >( std::round( l0_.x ), std::round( l0_.y ) );
            Coord< int > ic1( std::round( c.x ), std::round( c.y ) );
            (*draw_)( drawingArea().pixbuf(), pixelOp(), ic0, ic1 );
            status_ = NO_DRAW;
            // 描画なし・1 点指定状態なら 2 点指定状態に切り替え
          } else {
            // c0_ は現レイヤの座標、c1_ は widget 上の座標であることに注意
            l0_ = c;
            c0_ = c1_ = Coord< int >( std::round( event->x ), std::round( event->y ) );
            // 2 点指定カーソルを描画
            (*cursor1_)( drawingArea(), notSet_, c0_, c1_ );
            status_ = CURSOR1;
          }
          // 右クリック
        } else if ( event->button == 3 ) {
          reset();
          status_ = NO_DRAW;
        } else {
          return( FALSE );
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_2PassSelectionEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::motionNotifyAction" << std::endl;
#endif
        reset();

        // 2 点指定カーソルの場合
        if ( status_ == CURSOR1 ) {
          // c1_ は widget 上の座標であることに注意
          c1_ = c;
          // widget 上の座標に変換した座標を c0_ に代入
          c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          (*cursor1_)( drawingArea(), notSet_, c0_, c1_ );
          // 1 点指定カーソルの場合
        } else {
          // l0_ は現レイヤ上の座標であることに注意
          l0_ = drawingArea().getCoord( Coord< double >( c.x, c.y ) );
          c0_ = c;
          if ( cursor0_ != 0 ) {
            (*cursor0_)( drawingArea(), notSet_, c0_ );
          }
          status_ = CURSOR0;
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::redrawAction : カーソルの再描画
      */
      gint GTK_DrawingArea_2PassSelectionEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::redrawAction" << std::endl;
#endif
        reset();

        // 2 点指定カーソルの場合
        if ( status_ == CURSOR1 ) {
          // widget 上の座標に変換した座標を c0_ に代入
          c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          (*cursor1_)( drawingArea(), notSet_, c0_, c1_ );
          // 1 点指定カーソルの場合
        } else {
          if ( cursor0_ != nullptr ) {
            (*cursor0_)( drawingArea(), notSet_, c0_ );
          }
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::reset : 描画内容のクリア
      */
      void GTK_DrawingArea_2PassSelectionEvent::reset()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::reset" << std::endl;
#endif
        // 2 点指定カーソルの場合
        if ( status_ == CURSOR1 ) {
          (*cursor1_)( drawingArea(), erase_, c0_, c1_ );
          // 1 点指定カーソルの場合
        } else if ( status_ == CURSOR0 && cursor0_ != nullptr ) {
          (*cursor0_)( drawingArea(), erase_, c0_ );
        }
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去され、ステータスも初期状態になる
      */
      void GTK_DrawingArea_2PassSelectionEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_2PassSelectionEvent::block" << std::endl;
#endif
        reset();
        status_ = NO_DRAW;
        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::setCursor1 : 一点が決まった後のカーソルをセットする
      */
      void GTK_DrawingArea_2PassSelectionEvent::setCursor1( draw::DrawFuncIF_2p* cursor1 )
      {
        assert( cursor1 != nullptr );

        cursor1_ = cursor1;
      }

      /*
        GTK_DrawingArea_2PassSelectionEvent::setDraw : 描画関数をセットする
      */
      void GTK_DrawingArea_2PassSelectionEvent::setDraw( draw::DrawFuncIF_2p* draw )
      {
        assert( draw != nullptr );

        draw_ = draw;
      }

      /**********************************************************************************
        GTK_DrawingArea_3PassSelectionEvent : GTK_DrawingArea選択設定用イベント(3パス)
      **********************************************************************************/

      /* GTK_DrawingArea_3PassSelectionEvent : コンストラクタ */

      // cursor0, cursor1, cursor2, draw を別々に定義
      GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent
        ( canvas::GTK_DrawingArea& drawingArea,
          draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_2p* cursor1, draw::DrawFuncIF_3p* cursor2,
          draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
        cursor0_( cursor0 ), cursor1_( cursor1 ), cursor2_( cursor2 ),
        draw_( draw ), status_( 0 )
        {
#ifdef DEBUG
          std::cout << "GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea&, draw::DrawFuncIF_1p*&, draw::DrawFuncIF_2p*&, draw::DrawFuncIF_3p*&, draw::DrawFuncIF_3p*&, pen::GPixelOp* &pixelOp )" << std::endl;
#endif
          assert( cursor1_ != nullptr && cursor2_ != nullptr && draw_ != nullptr );
          assert( drawingArea.wsIsValid() );
        }

      // cursor2 と draw は同じ描画方式の場合
      GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent
        ( canvas::GTK_DrawingArea& drawingArea,
          draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_2p* cursor1,
          draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
        cursor0_( cursor0 ), cursor1_( cursor1 ), cursor2_( draw ),
        draw_( draw ), status_( 0 )
        {
#ifdef DEBUG
          std::cout << "GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea&, draw::DrawFuncIF_1p*&, draw::DrawFuncIF_2p*&, draw::DrawFuncIF_3p*&, pen::GPixelOp* &pixelOp )" << std::endl;
#endif
          assert( cursor1_ != nullptr && draw_ != nullptr );
          assert( drawingArea.wsIsValid() );
        }

      // cursor0 は描画なしの場合
      GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent
        ( canvas::GTK_DrawingArea& drawingArea,
          draw::DrawFuncIF_2p* cursor1, draw::DrawFuncIF_3p* cursor2,
          draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
        cursor0_( nullptr ), cursor1_( cursor1 ), cursor2_( cursor2 ),
        draw_( draw ), status_( 0 )
        {
#ifdef DEBUG
          std::cout << "GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea&, draw::DrawFuncIF_2p*&, draw::DrawFuncIF_3p*&, draw::DrawFuncIF_3p*&, pen::GPixelOp* &pixelOp )" << std::endl;
#endif
          assert( cursor1_ != nullptr && cursor2_ != nullptr && draw_ != nullptr );
          assert( drawingArea.wsIsValid() );
        }

      // cursor0 は描画なしで、cursor2 と draw は同じ描画方式の場合
      GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent
        ( canvas::GTK_DrawingArea& drawingArea,
          draw::DrawFuncIF_2p* cursor1,
          draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
        cursor0_( nullptr ), cursor1_( cursor1 ), cursor2_( draw ),
        draw_( draw ), status_( 0 )
        {
#ifdef DEBUG
          std::cout << "GTK_DrawingArea_3PassSelectionEvent::GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea&, draw::DrawFuncIF_2p*&, draw::DrawFuncIF_3p*&, pen::GPixelOp* &pixelOp )" << std::endl;
#endif
          assert( cursor1_ != nullptr && draw_ != nullptr );
          assert( drawingArea.wsIsValid() );
        }

      /*
        GTK_DrawingArea_3PassSelectionEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_3PassSelectionEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_3PassSelectionEvent::buttonPressAction" << std::endl;
#endif
        // 現レイヤ上の座標値
        Coord< double > c = drawingArea().getCoord( Coord< double >( event->x, event->y ) );

        if ( event->button == 1 ) {
          reset();
          switch( status_ ) {
          case CURSOR2:
            {
              Coord< int > ic0( std::round( l0_.x ), std::round( l0_.y ) );
              Coord< int > ic1( std::round( l1_.x ), std::round( l1_.y ) );
              Coord< int > ic2( std::round( c.x ), std::round( c.y ) );
              (*draw_)( drawingArea().pixbuf(), pixelOp(), ic0, ic1, ic2 );
            }
            status_ = NO_DRAW;
            break;
          case CURSOR1:
            l1_ = c;
            c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
            c1_ = c2_ = Coord< int >( std::round( event->x ), std::round( event->y ) );
            if ( cursor2_ != nullptr ) {
              (*cursor2_)( drawingArea(), notSet_, c0_, c1_, c2_ );
            }
            status_ = CURSOR2;
            break;
          default:
            l0_ = l1_ = c;
            c0_ = c1_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                      GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
            if ( cursor1_ != nullptr ) {
              (*cursor1_)( drawingArea(), notSet_, c0_, c1_ );
            }
            status_ = CURSOR1;
            break;
          }
        } else if ( event->button == 3 ) {
          reset();
          status_ = NO_DRAW;
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_3PassSelectionEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_3PassSelectionEvent::motionNotifyAction" << std::endl;
#endif
        reset();
        switch( status_ ) {
        case CURSOR2:
          c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          c1_ = Coord< int >( GetInvCoord( l1_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l1_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          c2_ = c;
          if ( cursor2_ != nullptr ) {
            (*cursor2_)( drawingArea(), notSet_, c0_, c1_, c2_ );
          }
          break;
        case CURSOR1:
          c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          c1_ = c;
          if ( cursor1_ != nullptr ) {
            (*cursor1_)( drawingArea(), notSet_, c0_, c1_ );
          }
          break;
        default:
          c0_ = c;
          if ( cursor0_ != nullptr ) {
            (*cursor0_)( drawingArea(), notSet_, c0_ );
          }
          status_ = CURSOR0;
          break;
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::redrawAction : カーソルの再描画
      */
      gint GTK_DrawingArea_3PassSelectionEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_3PassSelectionEvent::redrawAction" << std::endl;
#endif
        reset();

        switch( status_ ) {
        case CURSOR2:
          c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          c1_ = Coord< int >( GetInvCoord( l1_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l1_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          if ( cursor2_ != nullptr ) {
            (*cursor2_)( drawingArea(), notSet_, c0_, c1_, c2_ );
          }
          break;
        case CURSOR1:
          c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          if ( cursor1_ != nullptr ) {
            (*cursor1_)( drawingArea(), notSet_, c0_, c1_ );
          }
          break;
        default:
          if ( cursor0_ != nullptr ) {
            (*cursor0_)( drawingArea(), notSet_, c0_ );
          }
          break;
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::reset : 描画内容のクリア
      */
      void GTK_DrawingArea_3PassSelectionEvent::reset()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_3PassSelectionEvent::reset" << std::endl;
#endif
        switch( status_ ) {
        case CURSOR2:
          if ( cursor2_ != nullptr ) {
            (*cursor2_)( drawingArea(), erase_, c0_, c1_, c2_ );
          }
          break;
        case CURSOR1:
          if ( cursor1_ != nullptr ) {
            (*cursor1_)( drawingArea(), erase_, c0_, c1_ );
          }
          break;
        case CURSOR0:
          if ( cursor0_ != nullptr ) {
            (*cursor0_)( drawingArea(), erase_, c0_ );
          }
          break;
        }
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去され、ステータスも初期状態になる
      */
      void GTK_DrawingArea_3PassSelectionEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_3PassSelectionEvent::block" << std::endl;
#endif
        reset();
        status_ = NO_DRAW;
        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::setCursor1 : 一点が決まった後のカーソルをセットする
      */
      void GTK_DrawingArea_3PassSelectionEvent::setCursor1( draw::DrawFuncIF_2p* cursor1 )
      {
        assert( cursor1 != nullptr );

        cursor1_ = cursor1;
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::setCursor2 : 二点が決まった後のカーソルをセットする
      */
      void GTK_DrawingArea_3PassSelectionEvent::setCursor2( draw::DrawFuncIF_3p* cursor2 )
      {
        assert( cursor2 != nullptr );

        cursor2_ = cursor2;
      }

      /*
        GTK_DrawingArea_3PassSelectionEvent::setDraw : 描画関数をセットする
      */
      void GTK_DrawingArea_3PassSelectionEvent::setDraw( draw::DrawFuncIF_3p* draw )
      {
        assert( draw != nullptr );

        draw_ = draw;
      }

      /******************************************************************************************
         GTK_DrawingArea_MultiPassSelectionEvent 用ヘルパ関数
      ******************************************************************************************/

      namespace
      {
        // 頂点(指定した点)を移動する場合のドラッグの有効範囲
        const int VERTEX_CRITERIA = 4;

        /*
          GetVertex : クリック位置 c に等しい(ドラッグ有効範囲内にある)頂点の位置を v の中から返す

          c は現レイヤ上の座標値を表す。戻り値としては添字が返される。
          位置の重なった要素があった場合、末尾に近い側が必ず返される。
        */
        template< class T >
        typename vector< Coord< T > >::size_type
        GetVertex( const Coord< T >& c, const vector< Coord< T > >& v )
        {
#ifdef DEBUG
          std::cout << "GetVertex" << std::endl;
#endif
          typedef typename vector< Coord< T > >::size_type size_type;

          size_type sz = v.size();
          if ( sz == 0 ) return( 0 );

          for ( size_type i = sz ; i > 0 ; --i )
            if ( c.x >= v[i - 1].x - VERTEX_CRITERIA && c.x <= v[i - 1].x + VERTEX_CRITERIA &&
                 c.y >= v[i - 1].y - VERTEX_CRITERIA && c.y <= v[i - 1].y + VERTEX_CRITERIA )
              return( i - 1 );

          return( sz );
        }

        /*
          DrawLineWithBox : 端に Box の付いたラインの描画
        */
        void DrawLineWithBox( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const vector< Coord< int > >& v )
        {
#ifdef DEBUG
          std::cout << "DrawLineWithBox" << std::endl;
#endif
          vector< Coord< int > >::size_type sz = v.size();

          for ( std::vector< Coord< int > >::size_type i = 1 ; i < sz ; ++i ) {
            shape::box::Stroke( draw, pixelOp, Coord< int >( v[i - 1].x - VERTEX_CRITERIA, v[i - 1].y - VERTEX_CRITERIA ),
                          Coord<int>( v[i - 1].x + VERTEX_CRITERIA, v[i - 1].y + VERTEX_CRITERIA ) );
            shape::line::Draw( draw, pixelOp, v[i - 1], v[i] );
          }
          if ( sz > 0 )
            shape::box::Stroke( draw, pixelOp, Coord< int >( v[sz - 1].x - VERTEX_CRITERIA, v[sz - 1].y - VERTEX_CRITERIA ),
                          Coord<int>( v[sz - 1].x + VERTEX_CRITERIA, v[sz - 1].y + VERTEX_CRITERIA ) );
        }

      } // namespace

      /******************************************************************************************
        GTK_DrawingArea_MultiPassSelectionEvent : GTK_DrawingArea選択設定用イベント(任意の数のパス)
      ******************************************************************************************/

      /*
        GTK_DrawingArea_MultiPassSelectionEvent : コンストラクタ

        drawingArea : 描画領域
        drawMulti : 描画関数へのポインタの参照
        pixelOp : 点描画関数へのポインタの参照
      */
      GTK_DrawingArea_MultiPassSelectionEvent::GTK_DrawingArea_MultiPassSelectionEvent
        ( canvas::GTK_DrawingArea& drawingArea,
          draw::DrawFuncIF_Multiple* drawMulti, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ), drawMulti_( drawMulti )
        {
#ifdef DEBUG
          std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::GTK_DrawingArea_MultiPassSelectionEvent" << std::endl;
#endif
          assert( drawMulti_ != nullptr );
          assert( drawingArea.wsIsValid() );
        }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_MultiPassSelectionEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::buttonPressAction" << std::endl;
#endif
        // widget基準の座標値
        Coord< int > c( std::round( event->x ), std::round( event->y ) );
        // 現レイヤ基準の座標値
        Coord< double > l = drawingArea().getCoord( Coord< double >( event->x, event->y ) );

        if ( event->button == 1 ) {
          draggedVertex_ = GetVertex( l, vec_l_ );
          if ( draggedVertex_ >= vec_l_.size() ) {
            drawNewLine( l, c );
            draggedVertex_ = vec_l_.size() - 1;
          }
        } else if ( event->button == 3 ) {
          draggedVertex_ = GetVertex( l, vec_l_ );
          if ( draggedVertex_ < vec_l_.size() ) {
            DrawLineWithBox( drawingArea(), erase_, vec_c_ );
            vec_l_.erase( vec_l_.begin() + draggedVertex_ );
            vec_c_.erase( vec_c_.begin() + draggedVertex_ );
            DrawLineWithBox( drawingArea(), notSet_, vec_c_ );
          }
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::buttonReleaseAction : ボタンリリース時のコールバック関数

        リリース後に点のドラッグが確実に発生しないようにする

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_MultiPassSelectionEvent::buttonReleaseAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::buttonReleaseAction" << std::endl;
#endif
        // draggedVertex_ が範囲外なら modifyLine は作用しない
        if ( event->button == 1 )
          draggedVertex_ = vec_l_.size();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_MultiPassSelectionEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::motionNotifyAction" << std::endl;
#endif
        // 現レイヤ基準の座標値
        Coord< double > l = drawingArea().getCoord( Coord< double >( c.x, c.y ) );

        if ( state & GDK_BUTTON1_MASK ) {
          modifyLine( l, c );
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::redrawAction : オフセット・スケール変更時の再描画
      */
      gint GTK_DrawingArea_MultiPassSelectionEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::redrawAction" << std::endl;
#endif
        //reset();
        DrawLineWithBox( drawingArea(), erase_, vec_c_ );
        for ( vector< Coord< double > >::size_type i = 0 ; i < vec_l_.size() ; ++i ) {
          vec_c_[i] = Coord< int >( GetInvCoord( vec_l_[i].x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                    GetInvCoord( vec_l_[i].y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
        }
        DrawLineWithBox( drawingArea(), notSet_, vec_c_ );

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::modifyLine : ドラッグによる直線の変更

        c : 頂点の新たな位置
      */
      void GTK_DrawingArea_MultiPassSelectionEvent::modifyLine( const Coord< double >& l, const Coord< int >& c )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::modifyLine" << std::endl;
#endif
        if ( draggedVertex_ >= vec_l_.size() ) return;

        DrawLineWithBox( drawingArea(), erase_, vec_c_ );

        vec_l_[draggedVertex_] = l;
        vec_c_[draggedVertex_] = c;

        DrawLineWithBox( drawingArea(), notSet_, vec_c_ );
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::drawNewLine : 新たな線分の追加

        c : 新たな頂点の位置
      */
      void GTK_DrawingArea_MultiPassSelectionEvent::drawNewLine( const Coord< double >& l, const Coord< int >& c )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::drawNewLine" << std::endl;
#endif
        vec_l_.push_back( l );
        vec_c_.push_back( c );
        shape::box::Stroke( drawingArea(), notSet_, Coord< int >( c.x - VERTEX_CRITERIA, c.y - VERTEX_CRITERIA ),
                      Coord< int >( c.x + VERTEX_CRITERIA, c.y + VERTEX_CRITERIA ) );

        if ( vec_c_.size() > 1 ) {
          std::vector< Coord< int > >::size_type i = vec_c_.size() - 2;
          shape::line::Draw( drawingArea(), notSet_, vec_c_[i], vec_c_.back() );
        }
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::operator() : 描画関数オブジェクト drawMulti と点描画関数 pixelOp を外部から渡して描画する
      */
      void GTK_DrawingArea_MultiPassSelectionEvent::operator()( draw::DrawFuncIF_Multiple& drawMulti, pen::GPixelOp& pixelOp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::operator()( draw::DrawFuncIF_Multiple&, pen::GPixelOp& )" << std::endl;
#endif

        DrawLineWithBox( drawingArea(), erase_, vec_c_ );
        vector< Coord< int > > vec_l_buff;
        for ( vector< Coord< double > >::const_iterator i = vec_l_.begin() ; i != vec_l_.end() ; ++i )
          vec_l_buff.push_back( Coord< int >( std::round( i->x ), std::round( i->y ) ) );
        drawMulti( drawingArea().pixbuf(), pixelOp, vec_l_buff );
        vec_l_.clear();
        vec_c_.clear();
        drawingArea().draw();
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去され、点列もクリアされる
      */
      void GTK_DrawingArea_MultiPassSelectionEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_MultiPassSelectionEvent::block" << std::endl;
#endif

        DrawLineWithBox( drawingArea(), erase_, vec_c_ );
        vec_l_.clear();
        vec_c_.clear();
        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /*
        GTK_DrawingArea_MultiPassSelectionEvent::setDraw : 描画関数をセットする
      */
      void GTK_DrawingArea_MultiPassSelectionEvent::setDraw( draw::DrawFuncIF_Multiple* drawMulti )
      {
        assert( drawMulti != nullptr );

        drawMulti_ = drawMulti;
      }

      /*******************************************************************************
        GTK_DrawingArea_PenEvent : GTK_DrawingArea選択設定用イベント ( ペン描画 )
      *******************************************************************************/

      /* GTK_DrawingArea_PenEvent : コンストラクタ */

      GTK_DrawingArea_PenEvent::GTK_DrawingArea_PenEvent
      ( canvas::GTK_DrawingArea& drawingArea, draw::DrawFuncIF_1p* cursor0, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( cursor0 ),
          drawn_( false ), hasPre_( false ), clicked_( false ), shiftClicked_( false ),
          preScale_( drawingArea.getScale().x )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::GTK_DrawingArea_PenEvent( canvas::GTK_DrawingArea&, draw::DrawFuncIF_1p*&, pen::GPixelOp*& )" << std::endl;
#endif
        assert( drawingArea.wsIsValid() );
      }

      GTK_DrawingArea_PenEvent::GTK_DrawingArea_PenEvent
      ( canvas::GTK_DrawingArea& drawingArea, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp ),
          cursor0_( nullptr ),
          drawn_( false ), hasPre_( false ), clicked_( false ), shiftClicked_( false ),
          preScale_( drawingArea.getScale().x )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::GTK_DrawingArea_PenEvent( canvas::GTK_DrawingArea&, pen::GPixelOp*& )" << std::endl;
#endif
        assert( drawingArea.wsIsValid() );
      }

      // デストラクタ (カーソルの描画を消去して内容を初期化する)
      GTK_DrawingArea_PenEvent::~GTK_DrawingArea_PenEvent()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::~GTK_DrawingArea_PenEvent" << std::endl;
#endif
        reset();
      }

      /*
        GTK_DrawingArea_PenEvent::reset : 描画内容のクリア
      */
      void GTK_DrawingArea_PenEvent::drawLine( pen::GPixelOp& pixelOp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::drawLine : pre_ = " << pre_.x << "," << pre_.y << std::endl;
#endif
        if ( ! hasPre_ ) return;

        Coord< int > drawCoord0( std::round( pre_.x ), std::round( pre_.y ) );
        Coord< int > drawCoord1( std::round( c_.x ), std::round( c_.y ) );
        shape::line::DrawH( drawingArea(), pixelOp, Coord< int >( drawCoord0.x - 4, drawCoord0.y ), drawCoord0.x + 4 );
        shape::line::DrawV( drawingArea(), pixelOp, Coord< int >( drawCoord0.x, drawCoord0.y - 4 ), drawCoord0.y + 4 );
        shape::line::Draw( drawingArea(), pixelOp, drawCoord0, drawCoord1 );
      }

      /*
        GTK_DrawingArea_PenEvent::reset : カーソルのクリア
      */
      void GTK_DrawingArea_PenEvent::reset()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::reset" << std::endl;
#endif
        // カーソルを描画していない、またはクリック中の場合は何もしない
        if ( ( ! drawn_ ) || clicked_ ) return;

        if ( preScale_ != drawingArea().getScale().x ) {
          shape::box::Fill( drawingArea(), erase_, Coord< int >( 0, 0 ), Coord< int >( drawingArea().size().x - 1, drawingArea().size().y - 1 ) );
          preScale_ = drawingArea().getScale().x;
        } else {
          draw( erase_ );
        }

        drawn_ = false;
      }

      /*
        GTK_DrawingArea_PenEvent::draw : カーソルの描画
      */
      void GTK_DrawingArea_PenEvent::drawCursor()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::draw" << std::endl;
#endif
        // クリック中の場合は何もしない
        if ( clicked_ ) return;

        draw( notSet_ );

        if ( preScale_ != drawingArea().getScale().x )
          preScale_ = drawingArea().getScale().x;

        drawn_ = true;
      }

      /*
        GTK_DrawingArea_PenEvent::drawCursor : カーソルの描画
      */
      void GTK_DrawingArea_PenEvent::draw( pen::GPixelOp& pixelOp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::drawCursor" << std::endl;
#endif
        if ( cursor0_ != nullptr ) {
          Coord< int > drawCoord( std::round( c_.x ), std::round( c_.y ) );
          (*cursor0_)( drawingArea(), pixelOp, drawCoord );
          drawingArea().draw();
        }
        if ( shiftClicked_ ) {
          drawLine( pixelOp );
          drawingArea().draw();
        }
      }

      /*
        GTK_DrawingArea_PenEvent::drawPixbuf : Pixbuf への描画

        c は現在のマウス位置
      */
      void GTK_DrawingArea_PenEvent::drawPixbuf( Coord< double > c )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::drawPixbuf" << std::endl;
#endif
        Coord< double > dc1 = drawingArea().getCoord( c );
        Coord< int > c1( std::round( dc1.x ), std::round( dc1.y ) );

        if ( shiftClicked_ ) {
          Coord< int > c0( std::round( pre_l_.x ), std::round( pre_l_.y ) );
          shape::line::Draw( drawingArea().pixbuf(), pixelOp(), c0, c1 );
        } else {
          ( pixelOp() )( drawingArea().pixbuf(), c1 );
        }

        drawingArea().draw();
      }

      /*
        マウスがウィジェット内に入った時の動作

        フォーカスを ON にする
      */
      gint GTK_DrawingArea_PenEvent::enterNotifyAction( GdkEvent* event )
      {
        gtk_widget_grab_focus( drawingArea().widget() );

        return( FALSE );
      }

      /*
        マウスがウィジェット内から出た時の動作

        カーソルを消去する
      */
      gint GTK_DrawingArea_PenEvent::leaveNotifyAction( GdkEvent* event )
      {
        reset();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::keyPressAction : キークリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_PenEvent::keyPressAction( GdkEventKey* event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::keyPressAction" << std::endl;
#endif
        // シフトキーが押されているかをチェック
        bool shift = ( event->keyval == GDK_KEY_Shift_L ) || ( event->keyval == GDK_KEY_Shift_R );

        if ( shift != shiftClicked_ ) {
          reset();
          shiftClicked_ = shift;
          drawCursor();
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::keyReleaseAction : キーリリース時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_PenEvent::keyReleaseAction( GdkEventKey* event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::keyReleaseAction : keyval="; std::hex( std::cout ); std::cout << event->keyval << std::endl;
#endif
        // シフトキーがリリースされたのかをチェック
        bool shiftReleased = ( event->keyval == GDK_KEY_Shift_L ) || ( event->keyval == GDK_KEY_Shift_R );

        if ( shiftReleased && ( shiftReleased == shiftClicked_ ) ) {
          reset();
          shiftClicked_ = ! shiftReleased;
          drawCursor();
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_PenEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::buttonPressAction" << std::endl;
#endif
        Coord< double > dc = drawingArea().getCoord( Coord< double >( event->x, event->y ) );
        Coord< int > c( std::round( dc.x ), std::round( dc.y ) );

        // モディファイア・キー ( Ctrl, Shift, Alt, Super, Hyper, Meta ) のみをONにするマスク
        GdkModifierType modifiers = gtk_accelerator_get_default_mod_mask();
        // シフトキーが押されているかをチェック
        bool shiftKeyPressed = ( ( event->state & modifiers ) == GDK_SHIFT_MASK );
        // Ctrl キーが押されているかをチェック
        //bool ctrlKeyPressed = ( ( event->state & modifiers ) == GDK_CONTROL_MASK );

        if ( event->button == 1 ) {
          reset();
          shiftClicked_ = shiftKeyPressed;
          drawPixbuf( Coord< double >( event->x, event->y ) );
          pre_ = Coord< double >( event->x, event->y );
          pre_l_ = drawingArea().getCoord( pre_ );
          hasPre_ = true;
          clicked_ = true;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::buttonReleaseAction : ボタンリリース時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_PenEvent::buttonReleaseAction( GdkEventButton *event )
      {
        if ( event->button == 1 )
          clicked_ = false;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_PenEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::motionNotifyAction" << std::endl;
#endif
        reset();
        c_ = Coord< double >( c.x, c.y );
        drawCursor();

        if ( clicked_ ) {
          drawPixbuf( Coord< double >( c.x, c.y ) );
          pre_ = Coord< double >( c.x, c.y );
          pre_l_ = drawingArea().getCoord( pre_ );
          hasPre_ = true;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::redrawAction : オフセット・スケール変更時のコールバック関数
      */
      gint GTK_DrawingArea_PenEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::redrawAction" << std::endl;
#endif
        if ( hasPre_ ) {
          pre_ = Coord< double >( GetInvCoord( pre_l_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                  GetInvCoord( pre_l_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
        }
        if ( cursor0_ != 0 ) {
          reset();
          Coord< int > drawCoord( std::round( c_.x ), std::round( c_.y ) );
          (*cursor0_)( drawingArea(), notSet_, drawCoord );
          drawn_ = true;
          drawingArea().draw();
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_PenEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去され、ステータスも初期状態になる
      */
      void GTK_DrawingArea_PenEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_PenEvent::block" << std::endl;
#endif
        reset();
        shiftClicked_ = false;
        hasPre_ = false;
        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /******************************************************************************************
         GTK_DrawingArea_FergusonCoonsSelectionEvent 用ヘルパ関数
      ******************************************************************************************/

      namespace
      {
        // 微分係数を表す線分の長さの、微分係数の値に対する比率
        const double DIFF_SCALE = 5.0;

        /*
          GetDiffVertex : クリック位置 c に等しい(ドラッグ有効範囲内にある)頂点の位置を p + d の中から返す

          微分係数を表すカーソルが探索対象。
          c は widget 上の座標値を表す。戻り値としては添字が返される。
          位置の重なった要素があった場合、末尾に近い側が必ず返される。
        */
        vector< Coord< double > >::size_type
        GetDiffVertex( const Coord< int >& c, const vector< Coord< int > >& p, const vector< Coord< double > >& d )
        {
#ifdef DEBUG
          std::cout << "GetDiffVertex" << std::endl;
#endif
          typedef vector< Coord< double > >::size_type size_type;

          size_type sz = p.size();
          if ( sz == 0 ) return( 0 );

          for ( size_type i = sz ; i > 0 ; --i ) {
            Coord< int > o( std::round( p[i - 1].x + d[i - 1].x / DIFF_SCALE ), std::round( p[i - 1].y + d[i - 1].y / DIFF_SCALE ) );
            if ( c.x >= o.x - VERTEX_CRITERIA && c.x <= o.x + VERTEX_CRITERIA &&
                 c.y >= o.y - VERTEX_CRITERIA && c.y <= o.y + VERTEX_CRITERIA )
              return( i - 1 );
          }

          return( sz );
        }

      } // namespace

      /********************************************************************************************************************
        GTK_DrawingArea_FergusonCoonsSelectionEvent : GTK_DrawingArea選択設定用イベント ( Ferguson/Coons 曲線描画用 )
      *********************************************************************************************************************/

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent : コンストラクタ

        drawingArea : 描画領域
        pixelOp : 点描画関数へのポインタの参照
      */
      GTK_DrawingArea_FergusonCoonsSelectionEvent::GTK_DrawingArea_FergusonCoonsSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea, pen::GPixelOp* pixelOp )
        : GTK_DrawingArea_DrawingEvent( drawingArea, pixelOp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::GTK_DrawingArea_FergusonCoonsSelectionEvent" << std::endl;
#endif
        assert( drawingArea.wsIsValid() );

        currentScale_ = drawingArea.getScale();
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_FergusonCoonsSelectionEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::buttonPressAction" << std::endl;
#endif
        // widget基準の座標値
        Coord< int > c( std::round( event->x ), std::round( event->y ) );
        // 現レイヤ基準の座標値
        Coord< double > l = drawingArea().getCoord( Coord< double >( event->x, event->y ) );

        if ( event->button == 1 ) {
          draggedVertex_ = GetDiffVertex( c, vec_c_, vec_d_ );
          if ( draggedVertex_ < vec_l_.size() ) {
            isDiff_ = true;
          } else {
            isDiff_ = false;
            draggedVertex_ = GetVertex( l, vec_l_ );
            if ( draggedVertex_ >= vec_l_.size() ) {
              drawNewLine( l, c );
              draggedVertex_ = vec_l_.size() - 1;
            }
          }
        } else if ( event->button == 3 ) {
          draggedVertex_ = GetVertex( l, vec_l_ );
          if ( draggedVertex_ < vec_l_.size() ) {
            reset();
            vec_l_.erase( vec_l_.begin() + draggedVertex_ );
            vec_c_.erase( vec_c_.begin() + draggedVertex_ );
            vec_d_.erase( vec_d_.begin() + draggedVertex_ );
            draggedVertex_ = vec_l_.size();
            draw();
          }
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::buttonReleaseAction : ボタンリリース時のコールバック関数

        リリース後に点のドラッグが確実に発生しないようにする

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_FergusonCoonsSelectionEvent::buttonReleaseAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::buttonReleaseAction" << std::endl;
#endif
        // draggedVertex_ が範囲外なら modifyLine は作用しない
        if ( event->button == 1 )
          draggedVertex_ = vec_l_.size();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_FergusonCoonsSelectionEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::motionNotifyAction" << std::endl;
#endif
        // 現レイヤ基準の座標値
        Coord< double > l = drawingArea().getCoord( Coord< double >( c.x, c.y ) );

        if ( state & GDK_BUTTON1_MASK ) {
          if ( isDiff_ ) {
            modifyDiffLine( l, c );
          } else {
            modifyLine( l, c );
          }
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::redrawAction : オフセット・スケール変更時の再描画
      */
      gint GTK_DrawingArea_FergusonCoonsSelectionEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::redrawAction" << std::endl;
#endif
        reset();

        Coord< double > newScale = drawingArea().getScale();

        for ( vector< Coord< double > >::size_type i = 0 ; i < vec_l_.size() ; ++i ) {
          vec_c_[i] = Coord< int >( GetInvCoord( vec_l_[i].x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                    GetInvCoord( vec_l_[i].y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          vec_d_[i].x *= newScale.x / currentScale_.x;
          vec_d_[i].y *= newScale.y / currentScale_.y;
        }

        currentScale_ = newScale;

        draw();

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::modifyLine : ドラッグによる直線の変更

        c : 頂点の新たな位置
      */
      void GTK_DrawingArea_FergusonCoonsSelectionEvent::modifyLine( const Coord< double >& l, const Coord< int >& c )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::modifyLine" << std::endl;
#endif
        if ( draggedVertex_ >= vec_l_.size() ) return;

        reset();

        vec_l_[draggedVertex_] = l;
        vec_c_[draggedVertex_] = c;

        draw();
      }

      void GTK_DrawingArea_FergusonCoonsSelectionEvent::modifyDiffLine( const Coord< double >& l, const Coord< int >& c )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::modifyDiffLine" << std::endl;
#endif
        if ( draggedVertex_ >= vec_l_.size() ) return;

        reset();

        vec_d_[draggedVertex_] = Coord< double >( ( c.x - vec_c_[draggedVertex_].x ) * DIFF_SCALE, ( c.y - vec_c_[draggedVertex_].y ) * DIFF_SCALE );

        draw();
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::drawNewLine : 新たな線分の追加

        c : 新たな頂点の位置
      */
      void GTK_DrawingArea_FergusonCoonsSelectionEvent::drawNewLine( const Coord< double >& l, const Coord< int >& c )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::drawNewLine" << std::endl;
#endif
        reset();

        vec_l_.push_back( l );
        vec_c_.push_back( c );
        vec_d_.push_back( Coord< double >( 0, 0 ) );

        draw();
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::reset : 描画内容のクリア
      */
      void GTK_DrawingArea_FergusonCoonsSelectionEvent::reset()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::reset" << std::endl;
#endif

        for ( vector< double >::size_type i = 0 ; i < vec_d_.size() ; ++i ) {
          shape::box::Stroke( drawingArea(), erase_, Coord< int >( vec_c_[i].x - VERTEX_CRITERIA, vec_c_[i].y - VERTEX_CRITERIA ),
                        Coord< int >( vec_c_[i].x + VERTEX_CRITERIA, vec_c_[i].y + VERTEX_CRITERIA ) );
          Coord< int > d( std::round( vec_c_[i].x + vec_d_[i].x / DIFF_SCALE ), std::round( vec_c_[i].y + vec_d_[i].y / DIFF_SCALE ) );
          shape::circle::Fill( drawingArea(), erase_, d, VERTEX_CRITERIA );
          shape::line::Draw( drawingArea(), erase_, d, vec_c_[i] );
        }
        shape::curve::StrokeFergusonCoonsSpline( drawingArea(), erase_, vec_c_, vec_d_, 0.4 );
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::draw : 再描画
      */
      void GTK_DrawingArea_FergusonCoonsSelectionEvent::draw()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::draw" << std::endl;
#endif
        for ( vector< double >::size_type i = 0 ; i < vec_d_.size() ; ++i ) {
          shape::box::Stroke( drawingArea(), notSet_, Coord< int >( vec_c_[i].x - VERTEX_CRITERIA, vec_c_[i].y - VERTEX_CRITERIA ),
                        Coord< int >( vec_c_[i].x + VERTEX_CRITERIA, vec_c_[i].y + VERTEX_CRITERIA ) );
          Coord< int > d( std::round( vec_c_[i].x + vec_d_[i].x / DIFF_SCALE ), std::round( vec_c_[i].y + vec_d_[i].y / DIFF_SCALE ) );
          shape::circle::Fill( drawingArea(), notSet_, d, VERTEX_CRITERIA );
          shape::line::Draw( drawingArea(), notSet_, d, vec_c_[i] );
        }
        shape::curve::StrokeFergusonCoonsSpline( drawingArea(), notSet_, vec_c_, vec_d_, 0.4 );
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::operator() : 描画関数オブジェクト drawMulti と点描画関数 pixelOp を外部から渡して描画する
      */
      void GTK_DrawingArea_FergusonCoonsSelectionEvent::operator()( pen::GPixelOp& pixelOp )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::operator()( pen::GPixelOp& )" << std::endl;
#endif
        reset();

        vector< Coord< int > > vec_l_buff;
        for ( vector< Coord< double > >::const_iterator i = vec_l_.begin() ; i != vec_l_.end() ; ++i ) {
          vec_l_buff.push_back( Coord< int >( std::round( i->x ), std::round( i->y ) ) );
        }
        vector< Coord< double > > vec_d_buff;
        for ( vector< Coord< double > >::const_iterator i = vec_d_.begin() ; i != vec_d_.end() ; ++i ) {
          vec_d_buff.push_back( Coord< double >( i->x / currentScale_.x, i->y / currentScale_.y ) );
        }
        shape::curve::StrokeFergusonCoonsSpline( drawingArea().pixbuf(), pixelOp, vec_l_buff, vec_d_buff, 0.4 );
        vec_l_.clear();
        vec_c_.clear();
        vec_d_.clear();
        drawingArea().draw();
      }

      /*
        GTK_DrawingArea_FergusonCoonsSelectionEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去され、点列もクリアされる
      */
      void GTK_DrawingArea_FergusonCoonsSelectionEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_FergusonCoonsSelectionEvent::block" << std::endl;
#endif
        reset();

        vec_l_.clear();
        vec_c_.clear();
        vec_d_.clear();
        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

    } // namespace gtk_event

    /******************************************************************************************
          GTK_DrawingArea_RectangleSelectionEvent 用ヘルパ関数
    ******************************************************************************************/

    namespace
    {
      const int RECT_MAX_CRITERIA = 48;  // 矩形 4 隅のドラッグ・エリア最大値
      const int RECT_MIN_CRITERIA = 16;  // 矩形 4 隅のドラッグ・エリア最小値
      const int RECT_CRITERIA_RATIO = 3; // 矩形の長さに対する矩形 4 隅のドラッグ・エリアの長さの比率

      // 頂点の判定
      // iccvvhh
      // ^- 矩形内部ならビットを立てる
      //  ^- 矩形中央(水平・垂直)
      //    ^- 垂直方向の端付近(上・下)
      //      ^- 水平方向の端付近(右・左)
      enum Vertex { LEFT = 1, RIGHT = 2, UP = 4, DOWN = 8,
                    LU = 5, RU = 6, LD = 9, RD = 10, LRUD = 15,
                    HCENTER = 16, VCENTER = 32, CENTER = 48,
                    INNER = 64,
      };

      /*
        GetRectCursor : 矩形カーソルのドラッグ・エリアの大きさとその位置を求める

        lu, rd はそれぞれ矩形左上・右下の座標に固定されている(反転していることを想定していない)ことに注意

        lu, rd : 矩形カーソルの左上・右下の座標値
        lu_min, lu_max : ドラッグ・エリア左上側の座標値(最小・最大)
        rd_min, rd_max : ドラッグ・エリア右下側の座標値(最小・最大)
        criteriaSize : ドラッグエリアのサイズ
      */
      template< class T >
      void GetRectCursor( Coord< T > lu, Coord< T > rd, Coord< T >* lu_min, Coord< T >* lu_max, Coord< T >* rd_min, Coord< T >* rd_max, Coord< T >* criteriaSize )
      {
        *criteriaSize = Coord< T >( ( rd.x - lu.x + 1 ) / RECT_CRITERIA_RATIO, ( rd.y - lu.y + 1 ) / RECT_CRITERIA_RATIO );
        if ( criteriaSize->x > RECT_MAX_CRITERIA ) criteriaSize->x = RECT_MAX_CRITERIA;
        if ( criteriaSize->y > RECT_MAX_CRITERIA ) criteriaSize->y = RECT_MAX_CRITERIA;

        if ( criteriaSize->x < RECT_MIN_CRITERIA || criteriaSize->y < RECT_MIN_CRITERIA ) {
          criteriaSize->x = RECT_MIN_CRITERIA;
          criteriaSize->y = RECT_MIN_CRITERIA;
          lu_min->x = lu.x - RECT_MIN_CRITERIA;
          lu_max->x = lu.x;
          lu_min->y = lu.y - RECT_MIN_CRITERIA;
          lu_max->y = lu.y;
          rd_min->x = rd.x;
          rd_max->x = rd.x + RECT_MIN_CRITERIA;
          rd_min->y = rd.y;
          rd_max->y = rd.y + RECT_MIN_CRITERIA;
        } else {
          lu_min->x = lu.x;
          lu_max->x = lu.x + criteriaSize->x;
          lu_min->y = lu.y;
          lu_max->y = lu.y + criteriaSize->y;
          rd_min->x = rd.x - criteriaSize->x;
          rd_max->x = rd.x;
          rd_min->y = rd.y - criteriaSize->y;
          rd_max->y = rd.y;
        }
      }

      /*
        GetRectClickLocation : 矩形 lu - rd に対して選択位置 c を判定する

        lu, rd はそれぞれ矩形左上・右下の座標に固定されている(反転していることを想定していない)ことに注意

        01 ... 左側 ; 10 ... 右側 ; 100 ... 上側 ; 1000 ... 下側 ; 10000 ... 横方向中央 ; 100000 ... 縦方向中央
        0101 ... 左上 ; 0110 ... 右上 ; 1010 ... 右下 ; 1001 ... 左下
        1000000 ... 矩形の内部

        ; それ以外 ... 該当なし
      */
      template< class T >
      int GetRectClickLocation( Coord< T > c, Coord< T > lu, Coord< T > rd )
      {
        Coord< T > lu_min, lu_max;
        Coord< T > rd_min, rd_max;
        Coord< T > criteriaSize;
        GetRectCursor( lu, rd, &lu_min, &lu_max, &rd_min, &rd_max, &criteriaSize );

        int ans = 0;

        // 上下左右の端付近か
        if ( c.x >= lu_min.x && c.x <= lu_max.x )
          ans += LEFT;
        if ( c.x >= rd_min.x && c.x <= rd_max.x )
          ans += RIGHT;

        if ( c.y >= lu_min.y && c.y <= lu_max.y )
          ans += UP;
        if ( c.y >= rd_min.y && c.y <= rd_max.y )
          ans += DOWN;

        // 中央付近か
        Coord< double > center( static_cast< double >( lu.x + rd.x ) / 2,
                                static_cast< double >( lu.y + rd.y ) / 2 );
        if ( c.x >= center.x - criteriaSize.x && c.x <= center.x + criteriaSize.x )
          ans += HCENTER;
        if ( c.y >= center.y - criteriaSize.y && c.y <= center.y + criteriaSize.y )
          ans += VCENTER;

        // 内部にあるか
        if ( c.x >= lu.x && c.y >= lu.y && c.x <= rd.x && c.y <= rd.y )
          ans += INNER;

        return( ans );
      }

      /*
        DrawRectWithBox : 四隅に Box の付いた矩形の描画

        lu, rd はそれぞれ矩形左上・右下の座標に固定されている(反転していることを想定していない)ことに注意
      */
      template< class T >
      void DrawRectWith4Box( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< T > lu, Coord< T > rd )
      {
#ifdef DEBUG
        std::cout << "DrawRectWithBox" << std::endl;
#endif
        Coord< T > lu_min, lu_max;
        Coord< T > rd_min, rd_max;
        Coord< T > criteriaSize;
        GetRectCursor( lu, rd, &lu_min, &lu_max, &rd_min, &rd_max, &criteriaSize );

        shape::box::Stroke( draw, pixelOp, Coord< int >( std::round( lu.x ), std::round( lu.y ) ), Coord< int >( std::round( rd.x ), std::round( rd.y ) ) );
        shape::box::Stroke( draw, pixelOp, Coord< int >( std::round( lu_min.x ), std::round( lu_min.y ) ), Coord< int >( std::round( lu_max.x ), std::round( lu_max.y ) ) );
        shape::box::Stroke( draw, pixelOp, Coord< int >( std::round( lu_min.x ), std::round( rd_min.y ) ), Coord< int >( std::round( lu_max.x ), std::round( rd_max.y ) ) );
        shape::box::Stroke( draw, pixelOp, Coord< int >( std::round( rd_min.x ), std::round( lu_min.y ) ), Coord< int >( std::round( rd_max.x ), std::round( lu_max.y ) ) );
        shape::box::Stroke( draw, pixelOp, Coord< int >( std::round( rd_min.x ), std::round( rd_min.y ) ), Coord< int >( std::round( rd_max.x ), std::round( rd_max.y ) ) );
      }

    } // namespace

    namespace draw
    {

      /******************************************************************************************
           RectangleCopy : 矩形領域のコピー
      ******************************************************************************************/

      void RectangleCopy::operator()( canvas::DrawingArea_IF& draw, Coord< int > c0, Coord< int > c1 )
      {
        pattern_.copy( draw, c0, c1 );
      }

      /******************************************************************************************
           RectangleFill : 矩形領域の塗りつぶし
      ******************************************************************************************/

      void RectangleFill::operator()( canvas::DrawingArea_IF& draw, Coord< int > c0, Coord< int > c1 )
      {
        pen::GPSet pset( rgb );
        shape::box::Fill( draw, pset, c0, c1 );
      }

    } // namespace draw

    namespace gtk_event
    {
      /******************************************************************************************
           GTK_DrawingArea_RectangleSelectionEvent : GTK_DrawingArea選択設定用イベント(矩形選択)
      ******************************************************************************************/

      GTK_DrawingArea_RectangleSelectionEvent::GTK_DrawingArea_RectangleSelectionEvent
      ( canvas::GTK_DrawingArea& drawingArea )
        : GTK_DrawingArea_SelectionEvent( drawingArea ) {}

      /*
        GTK_DrawingArea_RectangleSelectionEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_RectangleSelectionEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::buttonPressAction" << std::endl;
#endif
        if ( &drawingArea() == nullptr ) return( FALSE );

        // widget基準の座標値
        Coord< int > c( std::round( event->x ), std::round( event->y ) );
        // 現レイヤ基準の座標値
        Coord< double > l = drawingArea().getCoord( Coord< double >( event->x, event->y ) );

        if ( event->button == 1 ) {
          // 範囲の微調整時
          if ( status_ == CURSOR1 ) {
            clickLocation_ = GetRectClickLocation( c, c0_, c1_ );
            pre_ = c;
            // 終点の選択時
          } else if ( status_ == CURSOR0 ) {
            shape::box::Stroke( drawingArea(), erase_, c0_, c1_ );
            c1_ = c; l1_ = l;
            if ( c0_.x > c1_.x ) std::swap( c0_.x, c1_.x );
            if ( c0_.y > c1_.y ) std::swap( c0_.y, c1_.y );
            DrawRectWith4Box( drawingArea(), notSet_, c0_, c1_ );
            status_ = CURSOR1;
          } else {
            c0_ = c1_ = c;
            l0_ = l1_ = l;
            shape::box::Stroke( drawingArea(), notSet_, c0_, c1_ );
            status_ = CURSOR0;
          }
        } else if ( event->button == 3 ) {
          if ( status_ == CURSOR0 ) {
            shape::box::Stroke( drawingArea(), erase_, c0_, c1_ );
            status_ = NO_DRAW;
          } else if ( status_ == CURSOR1 ) {
            DrawRectWith4Box( drawingArea(), erase_, c0_, c1_ );
            status_ = NO_DRAW;
          }
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::buttonReleaseAction : ボタンリリース時のコールバック関数

        リリース後にドラッグが確実に発生しないようにする

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_RectangleSelectionEvent::buttonReleaseAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::buttonReleaseAction" << std::endl;
#endif
        // clickLocation_ が範囲外ならドラッグは発生しない
        if ( event->button == 1 )
          clickLocation_ = 0;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_RectangleSelectionEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::motionNotifyAction" << std::endl;
#endif
        if ( &drawingArea() == nullptr ) return( FALSE );

        // 現レイヤ基準の座標値
        Coord< double > l = drawingArea().getCoord( Coord< double >( c.x, c.y ) );

        if ( status_ == CURSOR1 ) {
          if ( state & GDK_BUTTON1_MASK ) {
            modifyRect( l, c );
          }
        } else if ( status_ == CURSOR0 ) {
          shape::box::Stroke( drawingArea(), erase_, c0_, c1_ );
          c1_ = c;
          l1_ = l;
          shape::box::Stroke( drawingArea(), notSet_, c0_, c1_ );
        }
        pre_ = c;

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::redrawAction : オフセット・スケール変更時の再描画
      */
      gint GTK_DrawingArea_RectangleSelectionEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::redrawAction" << std::endl;
#endif
        if ( &drawingArea() == nullptr ) return( FALSE );

        if ( status_ == CURSOR0 ) {
          shape::box::Stroke( drawingArea(), erase_, c0_, c1_ );
        } else if ( status_ == CURSOR1 ) {
          DrawRectWith4Box( drawingArea(), erase_, c0_, c1_ );
        }

        c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                            GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
        c1_ = Coord< int >( GetInvCoord( l1_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                            GetInvCoord( l1_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );

        if ( status_ == CURSOR0 ) {
          shape::box::Stroke( drawingArea(), notSet_, c0_, c1_ );
        } else if ( status_ == CURSOR1 ) {
          DrawRectWith4Box( drawingArea(), notSet_, c0_, c1_ );
        }

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::modifyRect : クリック・ドラッグ位置に応じた矩形の再描画
      */
      void GTK_DrawingArea_RectangleSelectionEvent::modifyRect( Coord< double > l, Coord< int > c )
      {
        DrawRectWith4Box( drawingArea(), erase_, c0_, c1_ );

        if ( ( clickLocation_ & LEFT ) != 0 ) {
          c0_.x += c.x - pre_.x;
          l0_ = drawingArea().getCoord( Coord< double >( c0_.x, c0_.y ) );
          if ( c0_.x > c1_.x ) {
            std::swap( c0_.x, c1_.x );
            std::swap( l0_.x, l1_.x );
            clickLocation_ &= ( ~LEFT );
            clickLocation_ |= RIGHT;
          }
        }
        if ( ( clickLocation_ & RIGHT ) != 0 ) {
          c1_.x += c.x - pre_.x;
          l1_ = drawingArea().getCoord( Coord< double >( c1_.x, c1_.y ) );
          if ( c0_.x > c1_.x ) {
            std::swap( c0_.x, c1_.x );
            std::swap( l0_.x, l1_.x );
            clickLocation_ &= ( ~RIGHT );
            clickLocation_ |= LEFT;
          }
        }
        if ( ( clickLocation_ & UP ) != 0 ) {
          c0_.y += c.y - pre_.y;
          l0_ = drawingArea().getCoord( Coord< double >( c0_.x, c0_.y ) );
          if ( c0_.y > c1_.y ) {
            std::swap( c0_.y, c1_.y );
            std::swap( l0_.y, l1_.y );
            clickLocation_ &= ( ~UP );
            clickLocation_ |= DOWN;
          }
        }
        if ( ( clickLocation_ & DOWN ) != 0 ) {
          c1_.y += c.y - pre_.y;
          l1_ = drawingArea().getCoord( Coord< double >( c1_.x, c1_.y ) );
          if ( c0_.y > c1_.y ) {
            std::swap( c0_.y, c1_.y );
            std::swap( l0_.y, l1_.y );
            clickLocation_ &= ( ~DOWN );
            clickLocation_ |= UP;
          }
        }

#ifdef DEBUG
        std::cout << clickLocation_ << std::endl;
#endif
        if ( ( ( clickLocation_ & LRUD ) == 0 ) && ( ( clickLocation_ & INNER ) != 0 ) ) {
          c0_.x += c.x - pre_.x;
          c0_.y += c.y - pre_.y;
          c1_.x += c.x - pre_.x;
          c1_.y += c.y - pre_.y;
          l0_ = drawingArea().getCoord( Coord< double >( c0_.x, c0_.y ) );
          l1_ = drawingArea().getCoord( Coord< double >( c1_.x, c1_.y ) );
        }

        DrawRectWith4Box( drawingArea(), notSet_, c0_, c1_ );
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::operator() : 矩形への操作
      */
      void GTK_DrawingArea_RectangleSelectionEvent::operator()( draw::RectangleOperator& op )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::operator()" << std::endl;
#endif
        if ( &drawingArea() == 0 ) return;
        if ( status_ != CURSOR1 ) return;

        op( drawingArea().pixbuf(), Coord< int >( l0_.x, l0_.y ), Coord< int >( l1_.x, l1_.y ) );
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::reset : カーソルの消去とステータスの初期化
      */
      void GTK_DrawingArea_RectangleSelectionEvent::reset()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::reset" << std::endl;
#endif
        if ( &drawingArea() == nullptr ) return;

        if ( status_ == CURSOR0 ) {
          shape::box::Stroke( drawingArea(), erase_, c0_, c1_ );
        } else if ( status_ == CURSOR1 ) {
          DrawRectWith4Box( drawingArea(), erase_, c0_, c1_ );
        }

        status_ = NO_DRAW;
      }

      /*
        GTK_DrawingArea_RectangleSelectionEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去される
      */
      void GTK_DrawingArea_RectangleSelectionEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_RectangleSelectionEvent::block" << std::endl;
#endif
        reset();

        GTK_DrawingArea_Event::block();
      }

      /***********************************************************************************************
          GTK_DrawingArea_PasteEvent : GTK_DrawingArea選択設定用イベント(矩形パターンの貼り付け)
      ************************************************************************************************/

      GTK_DrawingArea_PasteEvent::GTK_DrawingArea_PasteEvent
      ( canvas::DrawingArea_IF& src, canvas::GTK_DrawingArea& drawingArea, pen::GPixelDraw* pixelDraw )
        : GTK_DrawingArea_SelectionEvent( drawingArea ), src_( src ), pixelDraw_( pixelDraw )
      {
        assert( &src_ != nullptr && pixelDraw_ != nullptr );
      }

      /*
        GTK_DrawingArea_PasteEvent::setPixelDraw : ピクセル描画用関数オブジェクトをセットする
      */
      void GTK_DrawingArea_PasteEvent::setPixelDraw( pen::GPixelDraw* pixelDraw )
      {
        assert( pixelDraw != nullptr );

        pixelDraw_ = pixelDraw;
      }

      /***********************************************************************************************
          GTK_DrawingArea_NormalPasteEvent : GTK_DrawingArea選択設定用イベント(矩形パターンの貼り付け)
      ************************************************************************************************/

      GTK_DrawingArea_NormalPasteEvent::GTK_DrawingArea_NormalPasteEvent
      ( canvas::DrawingArea_IF& src, canvas::GTK_DrawingArea& drawingArea, pen::GPixelDraw* pixelDraw )
        : GTK_DrawingArea_PasteEvent( src, drawingArea, pixelDraw ),
          c0_( -1, -1 ), c1_( -1, -1 ), l0_( -1, -1 ), l1_( -1, -1 ),
          clickLocation_( 0 ), isFirst_( true ),
          revX_( false ), revY_( false ), swapXY_( false ) {}

      /*
        GTK_DrawingArea_NormalPasteEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_NormalPasteEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_NormalPasteEvent::buttonPressAction" << std::endl;
#endif
        if ( event->button == 1 ) {
          if ( swapXY_ )
            clickLocation_ = GetRectClickLocation( Coord< int >( std::round( event->x ), std::round( event->y ) ), c0_, Coord< int >( c0_.x + std::abs( c1_.y - c0_.y ), c0_.y + std::abs( c1_.x - c0_.x ) ) );
          else
            clickLocation_ = GetRectClickLocation( Coord< int >( std::round( event->x ), std::round( event->y ) ), c0_, c1_ );
          pre_ = Coord< int >( std::round( event->x ), std::round( event->y ) );
        } else if ( event->button == 3 ) {
          operator()();
          isFirst_ = true;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::buttonReleaseAction : ボタンリリース時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_NormalPasteEvent::buttonReleaseAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_NormalPasteEvent::buttonReleaseAction" << std::endl;
#endif
        if ( event->button == 1 )
          clickLocation_ = 0;

        return( FALSE );
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_NormalPasteEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_NormalPasteEvent::motionNotifyAction" << std::endl;
#endif
        if ( src().size().x == 0 || src().size().y == 0 ) return( FALSE );

        if ( isFirst_ || ( clickLocation_ & INNER ) != 0 ) {

          // 前回の描画内容を消去
          cursorErase();

          // 左上側の座標値
          if ( isFirst_ ) {
            c0_ = c;
          } else {
            Coord< int > d( c.x - pre_.x, c.y - pre_.y );
            c0_ += d;
#ifdef DEBUG
            std::cout << c.x << "," << c.y << " ";
            std::cout << pre_.x << "," << pre_.y << " ";
            std::cout << c0_.x << "," << c0_.y << std::endl;
#endif
          }
          l0_ = drawingArea().getCoord( Coord< double >( c0_.x , c0_.y ) );

          // 右下側の座標値
          l1_ = Coord< double >( l0_.x + src().size().x - 1, l0_.y + src().size().y - 1 );
          c1_ = Coord< int >( GetInvCoord( l1_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                              GetInvCoord( l1_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );

          // 新たな座標値に描画
          cursorDraw();

          drawingArea().draw();

          isFirst_ = false;
          pre_ = c;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::redrawAction : オフセット・スケール変更時の再描画
      */
      gint GTK_DrawingArea_NormalPasteEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_NormalPasteEvent::redrawAction" << std::endl;
#endif
        if ( src().size().x == 0 || src().size().y == 0 ) return( FALSE );

        // 前回の描画内容を消去
        cursorErase();

        // widget上の座標値を再計算(スケールのみを考慮する)
        c0_ = Coord< int >( GetInvCoord( l0_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                            GetInvCoord( l0_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
        c1_ = Coord< int >( GetInvCoord( l1_.x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                            GetInvCoord( l1_.y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );

        // 再描画
        cursorDraw();

        drawingArea().draw();

        return( FALSE );
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::cursorDraw : カーソルの描画
      */
      void GTK_DrawingArea_NormalPasteEvent::cursorDraw()
      {
        pen::GPSet pset;
        draw::ResizePut( src(), drawingArea(), pset, c0_, c1_, revX_, revY_, swapXY_, interpolation::NearestNeighbor() );
        if ( swapXY_ )
          shape::box::Stroke( drawingArea(), notSet_, c0_, Coord< int >( c0_.x + std::abs( c1_.y - c0_.y ), c0_.y + std::abs( c1_.x - c0_.x ) ) );
        else
          shape::box::Stroke( drawingArea(), notSet_, c0_, c1_ );
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::cursorErase : カーソルのクリア
      */
      void GTK_DrawingArea_NormalPasteEvent::cursorErase()
      {
        if ( isFirst_ ) return;

        if ( swapXY_ )
          shape::box::Fill( drawingArea(), erase_, c0_, Coord< int >( c0_.x + std::abs( c1_.y - c0_.y ), c0_.y + std::abs( c1_.x - c0_.x ) ) );
        else
          shape::box::Fill( drawingArea(), erase_, c0_, c1_ );
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::operator() : パターンの描画
      */
      void GTK_DrawingArea_NormalPasteEvent::operator()()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_NormalPasteEvent::operator()" << std::endl;
#endif
        if ( src().size().x == 0 || src().size().y == 0 ) return;

        // 描画内容を消去
        cursorErase();

        // 現レイヤに描画
        draw::Put( src(), drawingArea().pixbuf(), pixelDraw(),
                   Coord< int >( std::round( l0_.x ), std::round( l0_.y ) ),
                   revX_, revY_, swapXY_ );

        drawingArea().draw();
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去される
      */
      void GTK_DrawingArea_NormalPasteEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_NormalPasteEvent::block" << std::endl;
#endif
        cursorErase();
        isFirst_ = true;

        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /*
        GTK_DrawingArea_NormalPasteEvent::setDirection : 貼り付け方向のセット
      */
      bool GTK_DrawingArea_NormalPasteEvent::setDirection( Direction direction )
      {
        cursorErase();

        switch ( direction ) {
        case normal:
          revX_ = false;
          revY_ = false;
          swapXY_ = false;
          break;
        case reverseH:
          revX_ = true;
          revY_ = false;
          swapXY_ = false;
          break;
        case reverseV:
          revX_ = false;
          revY_ = true;
          swapXY_ = false;
          break;
        case turnAround:
          revX_ = true;
          revY_ = true;
          swapXY_ = false;
          break;
        case anticlockwise:
          revX_ = true;
          revY_ = false;
          swapXY_ = true;
          break;
        case clockwise:
          revX_ = false;
          revY_ = true;
          swapXY_ = true;
          break;
        default:
          return( false );
          break;
        }

        // 再描画
        cursorDraw();

        drawingArea().draw();

        return( true );
      }

      /******************************************************************************************
           GTK_DrawingArea_TransPasteEvent 用ヘルパ関数
      ******************************************************************************************/

      namespace
      {
        /*
          IsInner : 点 c が vertex を頂点とする矩形の内側かを判定する
        */
        bool IsInner( const vector< Coord< int > >& vertex, Coord< int > c )
        {
          std::multiset< int > xCand;

          for ( vector< Coord< int > >::size_type i = 0 ; i < vertex.size() ; ++i ) {
            int y0 = std::min( vertex[i].y, vertex[( i + 1 ) % vertex.size()].y );
            int y1 = std::max( vertex[i].y, vertex[( i + 1 ) % vertex.size()].y );
            if ( c.y < y0 || c.y > y1 ) continue;
            if ( y0 == y1 ) return( true );
            xCand.insert( std::round
                          ( ( vertex[( i + 1 ) % vertex.size()].x - vertex[i].x ) * ( c.y - vertex[i].y ) /
                            ( vertex[( i + 1 ) % vertex.size()].y - vertex[i].y ) + vertex[i].x ) );
          }
          std::multiset< int >::const_iterator cit = xCand.lower_bound( c.x );
          if ( *cit == c.x ) return( true );

          return( std::distance( xCand.begin(), cit ) % 2 != 0 );
        }

        /*
          CalcCenter : v を頂点とする矩形の対角線の交点を求める
        */
        Coord< double > CalcCenter( const vector< Coord< double > >& v )
        {
          Coord< double > sum = math_lib::SumAll< Coord< double > >( v );

          return( Coord< double >( sum.x / v.size(), sum.y / v.size() ) );
        }

        /*
          Scaling : 頂点列 v の i 番目の頂点と cm の比率から各頂点を拡大・縮小する
        */
        void Scaling( vector< Coord< double > >* v, vector< Coord< double > >::size_type i, Coord< double > cm )
        {
          Coord< double > o = CalcCenter( *v );
          Coord< double > cr( (*v)[i].x, (*v)[i].y );
          double dr = coord_lib::Distance< double >( cr, o );
          double dm = coord_lib::Distance< double >( cm, o );

          for ( vector< Coord< double > >::iterator it = v->begin() ;
                it != v->end() ; ++it ) {
            it->x = ( it->x - o.x ) * dm / dr + o.x;
            it->y = ( it->y - o.y ) * dm / dr + o.y;
          }
        }

        /*
          Arctan : 逆正接 tan^-1( y / x ) の計算
        */
        double Arctan( double x, double y )
        {
          const double pi = math_lib::Pi< double >();

          if ( x == 0 )
            return( ( y >= 0 ) ? pi / 2 : pi * 3 / 2 );

          double theta = std::atan( y / x );
          if ( x < 0 ) {
            theta += pi;
          } else if ( y < 0 ) {
            theta += pi * 2;
          }

          return( theta );
        }

        /*
          Rotation : 頂点列 v の i 番目の頂点と cm の中心角の差から各頂点を回転する
        */
        void Rotation( vector< Coord< double > >* v, vector< Coord< double > >::size_type i, Coord< double > cm )
        {
          Coord< double > o = CalcCenter( *v );
          Coord< double > cr( (*v)[i].x, (*v)[i].y );
          double thetaR = Arctan( cr.x - o.x, cr.y - o.y );
          double thetaM = Arctan( cm.x - o.x, cm.y - o.y );
          double theta = thetaM - thetaR;

          for ( vector< Coord< double > >::iterator it = v->begin() ;
                it != v->end() ; ++it ) {
            Coord< double > buff( *it );
            it->x = ( buff.x - o.x ) * std::cos( theta ) - ( buff.y - o.y ) * std::sin( theta ) + o.x;
            it->y = ( buff.x - o.x ) * std::sin( theta ) + ( buff.y - o.y ) * std::cos( theta ) + o.y;
          }
        }

      } // namespace

        /***********************************************************************************************
           GTK_DrawingArea_TransPasteEvent : GTK_DrawingArea選択設定用イベント ( 矩形パターンの自由変形 )
        ************************************************************************************************/

      GTK_DrawingArea_TransPasteEvent::GTK_DrawingArea_TransPasteEvent
      ( canvas::DrawingArea_IF& src, canvas::GTK_DrawingArea& drawingArea,
        pen::GPixelDraw* pixelDraw, interpolation::Base* ip,
        bool priorFirst )
        : GTK_DrawingArea_PasteEvent( src, drawingArea, pixelDraw ), ip_( ip ),
          cv_( 4 ), lv_( 4 ), draggedVertex_( 4 ), isFirst_( true ), priorFirst_( priorFirst )
      {
        assert( ip_ != nullptr );
      }

      /*
        GTK_DrawingArea_TransPasteEvent::buttonPressAction : ボタンクリック時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_TransPasteEvent::buttonPressAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_TransPasteEvent::buttonPressAction" << std::endl;
#endif
        if ( event->button == 1 ) {
          Coord< int > c( std::round( event->x ), std::round( event->y ) );
          draggedVertex_ = GetVertex( c, cv_ );
          isInner_ = IsInner( cv_, c );
          pre_ = Coord< int >( std::round( event->x ), std::round( event->y ) );
        } else if ( event->button == 3 ) {
          operator()();
          isFirst_ = true;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_TransPasteEvent::buttonReleaseAction : ボタンリリース時のコールバック関数

        event : GdkEventButtonイベント
      */
      gint GTK_DrawingArea_TransPasteEvent::buttonReleaseAction( GdkEventButton *event )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_TransPasteEvent::buttonReleaseAction" << std::endl;
#endif
        if ( event->button == 1 ) {
          draggedVertex_ = cv_.size();
          isInner_ = false;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_TransPasteEvent::motionNotifyAction : マウス移動時のコールバック関数

        c : 現在のマウスの位置
        state : 現在のマウスの状態
      */
      gint GTK_DrawingArea_TransPasteEvent::motionNotifyAction( const Coord< int >& c, GdkModifierType& state )
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_TransPasteEvent::motionNotifyAction" << std::endl;
#endif
        if ( src().size().x == 0 || src().size().y == 0 ) return( FALSE );

        // モディファイア・キー ( Ctrl, Shift, Alt, Super, Hyper, Meta ) のみをONにするマスク
        GdkModifierType modifiers = gtk_accelerator_get_default_mod_mask();
        // シフトキーが押されているかをチェック
        bool shiftKeyPressed = ( ( state & modifiers ) == GDK_SHIFT_MASK );
        // Ctrl キーが押されているかをチェック
        bool ctrlKeyPressed = ( ( state & modifiers ) == GDK_CONTROL_MASK );

        if ( isFirst_ || draggedVertex_ < cv_.size() || isInner_ ) {

          // 前回の描画内容を消去
          if ( ! isFirst_ )
            resetCursor();

          // 左上側の座標値
          if ( isFirst_ ) {
            cv_[0] = c;
            lv_[0] = drawingArea().getCoord( Coord< double >( cv_[0].x , cv_[0].y ) );
            lv_[1] = Coord< double >( lv_[0].x + src().size().x - 1, lv_[0].y );
            lv_[2] = Coord< double >( lv_[0].x + src().size().x - 1, lv_[0].y + src().size().y - 1 );
            lv_[3] = Coord< double >( lv_[0].x, lv_[0].y + src().size().y - 1 );
            for ( vector< Coord< int > >::size_type i = 1 ; i < cv_.size() ; ++i )
              cv_[i] = Coord< int >( GetInvCoord( lv_[i].x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                     GetInvCoord( lv_[i].y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
          } else if ( draggedVertex_ < cv_.size() ) {
            if ( ctrlKeyPressed ) {
              Scaling( &lv_, draggedVertex_, drawingArea().getCoord( Coord< double >( c.x, c.y ) ) );
              for ( vector< Coord< double > >::size_type i = 0 ; i < lv_.size() ; ++i )
                cv_[i] = Coord< int >( GetInvCoord( lv_[i].x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                       GetInvCoord( lv_[i].y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
            } else if ( shiftKeyPressed ) {
              Rotation( &lv_, draggedVertex_, drawingArea().getCoord( Coord< double >( c.x, c.y ) ) );
              for ( vector< Coord< double > >::size_type i = 0 ; i < lv_.size() ; ++i )
                cv_[i] = Coord< int >( GetInvCoord( lv_[i].x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                       GetInvCoord( lv_[i].y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
            } else {
              Coord< int > d( c.x - pre_.x, c.y - pre_.y );
              cv_[draggedVertex_] += d;
              lv_[draggedVertex_] = drawingArea().getCoord( Coord< double >( cv_[draggedVertex_].x , cv_[draggedVertex_].y ) );
            }
#ifdef DEBUG
            std::cout << c.x << "," << c.y << " ";
            std::cout << pre_.x << "," << pre_.y << " ";
            std::cout << c0_.x << "," << c0_.y << std::endl;
#endif
          } else {
            Coord< int > d( c.x - pre_.x, c.y - pre_.y );
            for ( vector< Coord< int > >::size_type i = 0 ; i < cv_.size() ; ++i ) {
              cv_[i] += d;
              lv_[i] = drawingArea().getCoord( Coord< double >( cv_[i].x , cv_[i].y ) );
            }
          }

          // 新たな座標値に描画
          drawCursor();

          drawingArea().draw();

          isFirst_ = false;
          pre_ = c;
        }

        return( FALSE );
      }

      /*
        GTK_DrawingArea_TransPasteEvent::redrawAction : オフセット・スケール変更時の再描画
      */
      gint GTK_DrawingArea_TransPasteEvent::redrawAction()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_TransPasteEvent::redrawAction" << std::endl;
#endif
        if ( src().size().x == 0 || src().size().y == 0 ) return( FALSE );

        // 前回の描画内容を消去
        resetCursor();

        // widget上の座標値を再計算(スケールのみを考慮する)
        for ( vector< Coord< int > >::size_type i = 0 ; i < cv_.size() ; ++i ) {
          cv_[i] = Coord< int >( GetInvCoord( lv_[i].x, drawingArea().pixbuf().getOffset().x, drawingArea().getOffset().x, drawingArea().getScale().x ),
                                 GetInvCoord( lv_[i].y, drawingArea().pixbuf().getOffset().y, drawingArea().getOffset().y, drawingArea().getScale().y ) );
        }

        // 再描画
        drawCursor();

        drawingArea().draw();

        return( FALSE );
      }

      void GTK_DrawingArea_TransPasteEvent::resetCursor()
      {
        shape::box::Fill( drawingArea(), erase_, Coord< int >( 0, 0 ), Coord< int >( drawingArea().size().x - 1, drawingArea().size().y - 1 ) );
        DrawLineWithBox( drawingArea(), erase_, cv_ );
        shape::line::Draw( drawingArea(), erase_, cv_.back(), cv_.front() );
      }

      void GTK_DrawingArea_TransPasteEvent::drawCursor()
      {
        pen::GPSet pset;
        draw::TransPut( src(), drawingArea(), pset, cv_[0], cv_[1], cv_[2], cv_[3], interpolation::NearestNeighbor(), false );
        DrawLineWithBox( drawingArea(), notSet_, cv_ );
        shape::line::Draw( drawingArea(), notSet_, cv_.back(), cv_.front() );
      }

      /*
        GTK_DrawingArea_TransPasteEvent::operator() : パターンの描画
      */
      void GTK_DrawingArea_TransPasteEvent::operator()()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_TransPasteEvent::operator()" << std::endl;
#endif
        if ( src().size().x == 0 || src().size().y == 0 ) return;

        // 描画内容を消去
        resetCursor();

        // 現レイヤに描画
        vector< Coord< int > > lvBuff( 4 );
        for ( vector< Coord< int > >::size_type i = 0 ; i < lvBuff.size() ; ++i )
          lvBuff[i] = Coord< int >( std::round( lv_[i].x ), std::round( lv_[i].y ) );
        draw::TransPut( src(), drawingArea().pixbuf(), pixelDraw(),
                              lvBuff[0], lvBuff[1], lvBuff[2], lvBuff[3], *ip_, priorFirst_ );

        drawingArea().draw();
      }

      /*
        GTK_DrawingArea_TransPasteEvent::block : コールバック関数のブロック

        ブロック前に現在のカーソルは消去される
      */
      void GTK_DrawingArea_TransPasteEvent::block()
      {
#ifdef DEBUG
        std::cout << "GTK_DrawingArea_TransPasteEvent::block" << std::endl;
#endif
        resetCursor();
        isFirst_ = true;

        drawingArea().draw();

        GTK_DrawingArea_Event::block();
      }

      /*
        GTK_DrawingArea_TransPasteEvent::setIp : 補間用関数オブジェクトをセットする
      */
      void GTK_DrawingArea_TransPasteEvent::setIp( interpolation::Base* ip )
      {
        assert( ip != nullptr );

        ip_ = ip;
      }

    } // namespace gtk_event

  } // namespace gl

} // namespace tadah_fussy
