/**
   @file gl_pixbuf.h
   @brief gtk+ による DrawingArea_IF の実装

   @author tadah_fussy
   @date 2015/01/18 コメントを Doxygen 用に変更 
   @date 2018/03/08 アルファ・チャンネルを考慮して重ね合わせするように pset を変更
   @date 2018/04/20 GTK_DrawingArea_FergusonCoonsSelectionEvent の追加
   @date 2021/02/25 ワーニングの出力に対応
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GL_PIXBUF_20210402_H
#define GL_PIXBUF_20210402_H

#include <stdio.h>
#include <glib.h>
#include <math.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <png.h>

#include <limits>
#include <exception>
#include <iostream>

#include "gl_drawing_method.h"
#include "gl_g_interpolation.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace canvas
    {

      /**
         @brief GdkPixbuf からなるレイヤ集合

         下地となるキャンバスの情報 ( サイズ・背景色・アルファ・チャンネルの有無 ) と、その上に重ね合わされる複数の
         レイヤ ( GdkPixbufオブジェクト・背景色・キャンバスとのオフセット・全体にかかるアルファチャンネル・可視/
         不可視のフラグ ) から構成される。空の状態、キャンバス情報の指定、画像ファイルの指定により構築が可能。

         重ねあわせた結果は GDK_Pixbuf::overlay によって GdkPixbuf オブジェクトへのポインタとして得ることができる。
         重ね合わせはキャンバスのサイズの範囲内のみで行われることに注意。また、生成された GdkPixbuf オブジェクトは
         内部でキャッシュされるため、外部から g_object_unref により参照を外さないこと。

         @todo removeAlpha, copyLayer の実装
      **/
      class GDK_Pixbuf : public DrawingArea_IF
      {
        static const int BITS_PER_SAMPLE = 8; // RGB 成分のビット数

        // 各レイヤの定義
        struct Layer
        {
          GdkPixbuf* pixbuf;       // GdkPixbuf オブジェクト
          guint32 bgColor;         // 背景色
          Coord< int > offset;     // キャンバスとのオフセット
          RGB::primary_type alpha; // レイヤの画素全体に対するアルファ・チャンネル
          bool isVisible;          // 可視状態なら true

          // デフォルト・コンストラクタ
          Layer()
            : pixbuf( 0 ), bgColor( 0 ), offset( 0, 0 ), alpha( RGB::MAX ), isVisible( true ) {}

          /// 画像は有効か
          bool isValid() const { return( pixbuf != 0 ); }

          // origin : 画像データ開始位置を返す
          guchar* origin() const
          { return( ( isValid() ) ? gdk_pixbuf_get_pixels( pixbuf ) : 0 ); }

          // 画像の幅(横方向のサイズ)を返す
          int width() const { return( ( isValid() ) ? gdk_pixbuf_get_width( pixbuf ) : 0 ); }

          // 画像の高さ(縦方向のサイズ)を返す
          int height() const { return( ( isValid() ) ? gdk_pixbuf_get_height( pixbuf ) : 0 ); }

          // 画像の一行分のサイズを返す
          int rowstride() const { return( ( isValid() ) ? gdk_pixbuf_get_rowstride( pixbuf ) : 0 ); }

          // 1 ピクセル分のバイト数を返す
          int nChannels() const { return( ( isValid() ) ? gdk_pixbuf_get_n_channels( pixbuf ) : 0 ); }

          // アルファチャンネルを持つ場合は true を返す
          bool hasAlpha() const { return( ( isValid() ) ? ( gdk_pixbuf_get_has_alpha( pixbuf ) == TRUE ) : false ); }
        };

      public:

        /// レイヤの番号の型
        using size_type = std::vector< Layer >::size_type;

      private:

        // レイヤの集合(キャンバス)
        // 0 番目の画像サイズがキャンバス・サイズを表す
        std::vector< Layer > layers_;
        // 現在操作できるレイヤの番号
        size_type current_;

        // キャンバスの背景色
        guint32 canvasBgColor_;
        // キャンバスのサイズ
        Coord< int > canvasSize_;
        // キャンバスはアルファ・チャンネルを持つか
        bool canvasHasAlpha_;

        // 画像の更新があった場合は overlay を実行するまで true となる
        bool changed_;
        // 重ね合わせ後の画像用バッファ ( changed_ が true になったら更新する )
        GdkPixbuf* buff_;

        // コピーコンストラクタと代入演算子の多重定義は非公開
        explicit GDK_Pixbuf( const GDK_Pixbuf& d );
        GDK_Pixbuf& operator=( const GDK_Pixbuf& d );

      public:

        /* コンストラクタ・デストラクタ */

        /// @brief 空のレイヤを構築
        ///
        /// レイヤは全くない状態で初期化する。
        /// キャンバスのサイズはゼロ、背景色はゼロ ( 黒色、アルファ・チャンネル = 0 ) で、アルファ・チャンネルは持つ状態で初期化する。
        ///
        /// @note レイヤのサイズを変更しない限り、GDK_Pixbuf::addLayer でレイヤを追加しても重ねあわせた結果は得られないことに注意。
        GDK_Pixbuf();

        /// @brief キャンバスのサイズ・背景色・アルファ・チャンネルの有無を指定して構築
        ///
        /// レイヤは全くない状態で初期化する。
        /// 画像の大きさに負数を設定した場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param size キャンバスのサイズ
        /// @param rgb キャンバスの背景色
        /// @param hasAlpha キャンバスがアルファ・チャンネルを持つなら true
        GDK_Pixbuf( Coord< int > size, RGB rgb, bool hasAlpha );

        /// @brief ファイル名を指定して構築
        ///
        /// 画像ファイルを読み込んだレイヤが一つ生成される。
        /// キャンバスのサイズ、背景色、アルファ・チャンネルの有無は生成したレイヤと等しくなる。
        ///
        /// ファイルの読み込みに失敗した場合はレイヤは生成されず、例外 std::runtime_error を投げる。
        ///
        /// @param fileName 画像ファイル名
        explicit GDK_Pixbuf( const std::string& fileName );

        /// @brief デストラクタ
        ~GDK_Pixbuf();

        /* レイヤ操作・プロパティ */

        /// @brief 空のレイヤを追加する
        ///
        /// 背景色は透明(黒色)、キャンバスとのオフセットはゼロ、全体にかかるアルファチャンネルは RGB::MAX、
        /// 可視/不可視のフラグは true で初期化される。
        void addLayer();

        /// @brief 指定したサイズのレイヤを追加する
        ///
        /// 画像の大きさに負数を設定した場合は例外 ExceptionNegativeNumber を投げる。
        /// メモリが確保できなかった場合、例外 std::bad_alloc を投げる。
        ///
        /// @param offset レイヤの描画開始位置(左上隅)
        /// @param size 追加するレイヤのサイズ
        /// @param rgb 初期化時に塗りつぶす RGB 成分
        /// @param hasAlpha アルファ・チャンネルを持つ場合は true
        /// @param alpha 合成時のアルファ・チャンネル
        void addLayer( Coord< int > offset, Coord< int > size, RGB rgb, bool hasAlpha, RGB::primary_type alpha );

        /// @brief 指定したファイルの画像を新たなレイヤとして追加する
        ///
        /// ファイルの読み込みに失敗した場合はレイヤは生成されず、例外 std::runtime_error を投げる。
        ///
        /// @param fileName 追加する画像ファイル名
        void addFromFile( const std::string& fileName );

        /// @brief 現在のレイヤを消去する
        ///
        /// 現在のレイヤ番号は一つ減る(消去したレイヤの下側のレイヤが対象になる)。
        /// 但し、現在のレイヤが 0 番だった場合は変化しない。
        void eraseLayer();

        /// @brief レイヤを全て消去する
        void clearLayer();

        /// @brief 現在のレイヤの位置を入れ替える
        ///
        /// 入れ替え位置は指定した番号のレイヤの前とする ( 例えば 0 を指定した場合は 0 番目のレイヤの前に移動して 0 番目のレイヤになる )。
        /// 末尾に移動させたい場合は末尾の番号 + 1 を指定する。
        /// 指定した添字が範囲外だった場合は例外 std::out_of_range を投げる
        ///
        /// @param upperLayer 入れ替え位置の上側のレイヤ番号
        void changeLayer( size_type upperLayer );

        /// @brief 現在のレイヤをコピーする
        ///
        /// コピーしたレイヤは末尾に追加される。
        /// メモリが確保できなかった場合、例外 std::bad_alloc を投げる。
        void copyLayer();

        /// @brief 現在のレイヤを下側の画像に重ね合わせる
        ///
        /// 現在のレイヤが 0 番目なら何も処理しない。現在のレイヤが不可視状態でも処理は行う。
        /// 重ねあわせた後、現在のレイヤは消去される。消去後のレイヤ番号は重ね合わされた下側に変更される。
        void superimpose();

        /// @brief 現在のレイヤ番号を返す
        ///
        /// レイヤが存在しない場合はゼロを返す
        ///
        /// @return 現在のレイヤ番号
        size_type getCurrentLayer() const
        { return( current_ ); }
        /// @brief 現在のレイヤ番号を変更する
        ///
        /// 指定した添字が範囲外だった場合は例外 std::out_of_range を投げる
        /// 特に、レイヤが存在しない場合は必ず例外が投げられることに注意。
        ///
        /// @param index 新しいレイヤ番号
        void setCurrentLayer( size_type index );

        /// @brief レイヤ数を返す
        ///
        /// @return レイヤ数
        size_type layerCount() const
        { return( layers_.size() ); }

        /// @brief 現レイヤの画像サイズを返す
        ///
        /// レイヤが存在しない場合は ( 0, 0 ) を返す。
        ///
        /// @return 現レイヤの画像サイズ
        virtual Coord< int > size() const;

        /// @brief 現レイヤの大きさを変更する
        ///
        /// 元の画像は左上原点を重ねた状態で複製される。
        ///
        /// 画像の大きさに負数を設定した場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param sz 変更する画像の大きさ
        virtual void resize( Coord< int > sz );

        /// @brief 現レイヤの背景色を返す
        ///
        /// レイヤが存在しない場合は RGB( 0, 0, 0, 0 ) を返す。
        ///
        /// @return 現レイヤの背景色
        RGB getBgColor() const;
        /// @brief 現レイヤの背景色をセットする
        ///
        /// @param rgb 新しい背景色
        void setBgColor( RGB rgb );

        /// @brief 現レイヤを背景色で塗りつぶす
        void fill();

        /// @brief 現レイヤのオフセット値を返す
        ///
        /// レイヤが存在しない場合は ( 0, 0 ) を返す。
        ///
        /// @return 現レイヤのオフセット値
        Coord< int > getOffset() const;
        /// @brief 現レイヤのオフセット値をセットする
        ///
        /// @param offset 新たなオフセット値
        void setOffset( Coord< int > offset );

        /// @brief 現レイヤのキャンバスに対する相対位置を求める
        ///
        /// 例えばオフセット値が ( -10, 10 ) なら、c = ( 5, 5 ) のとき ( 15, -5 ) が返される
        ///
        /// @param c キャンバス上の位置
        /// @return 現レイヤにおいて c が指し示す位置
        Coord< double > getCoord( Coord< double > c ) const;

        /// @brief 現レイヤのアルファ・チャンネルを返す
        ///
        /// 全画素に共通に掛かるアルファ・チャンネルの値を返す。
        ///
        /// @return 現レイヤのアルファ・チャンネル
        RGB::primary_type getAlpha() const;
        /// @brief 現レイヤのアルファ・チャンネルをセットする
        ///
        /// 全画素に共通に掛かるアルファ・チャンネルの値をセットする。
        ///
        /// @param alpha アルファ・チャンネルの値
        void setAlpha( RGB::primary_type alpha );

        // @brief 現レイヤの RGB 成分のビット数を返す
        //int bps() const;
        // @brief 現レイヤの画像の一行分のサイズを返す
        //int rowstride() const;
        // @brief 現レイヤの 1 ピクセル分のバイト数を返す
        //int nChannels() const;

        /// @brief 現レイヤがアルファチャンネルを持つ場合は true を返す
        ///
        /// @return 現レイヤがアルファチャンネルを持つ場合は true
        bool hasAlpha() const;
        /// @brief 現レイヤにアルファ・チャンネルを追加する
        ///
        /// @param substitudeRGB 指定した RGB 成分 ( rgb ) と同じ色のピクセルを透明色に置き換えるなら true を指定する
        /// @param rgb 透明色に置き換えるときの RGB 成分。substitudeRGB が false なら無視される。
        void addAlpha( bool substitudeRGB, RGB rgb );

        /// @brief 現レイヤからアルファ・チャンネルを削除する
        ///
        /// 削除前に、各ピクセルのアルファ・チャンネルから色コードを計算 ( RGB::overlay を利用 ) して引数で指定した
        /// bgColor と重ね合わせる。
        ///
        /// @param bgColor 重ね合わせ時の背景色
        void removeAlpha( RGB bgColor );

        /// @brief 現在のレイヤが可視状態なら true を返す
        ///
        /// @return 現在のレイヤが可視状態なら true
        bool isVisible() const;

        /// @brief 現在のレイヤを可視状態にする
        void visible();

        /// @brief 現在のレイヤを不可視状態にする
        void invisible();

        /// @brief 全レイヤを重ねあわせた画像を返す
        ///
        /// メモリの確保ができなかった場合、std::bad_alloc 例外を投げる。
        /// キャンバスのサイズがゼロなら NULL が返される。
        ///
        /// @note 返り値の GdkPixbuf が指すオブジェクトは内部でバッファに保持しているため、
        /// 受け取ったポインタで g_object_unref を行って参照を外さないこと。
        ///
        /// @return 全レイヤを重ねあわせた画像(GdkPixbufへのポインタ)
        GdkPixbuf* overlay();

        /// @brief レイヤが一つも存在しない場合は true を返す
        ///
        /// @return レイヤが一つも存在しない場合は true
        bool empty() const
        { return( layers_.empty() ); }

        //GdkPixbuf* overlay( Coord< double > offset, Coord< double > scale, Coord< int > size, GdkInterpType interpType ) const;

        /* レイヤへの描画 */

        /// @brief 指定した座標の RGB 成分を取得する
        ///
        /// 取得対象は現レイヤとなる。また、指定した座標値は、現レイヤ上の座標を意味することに注意。
        /// キャンバス上の座標から現レイヤ上の座標に変換する場合は GDK_Pixbuf::getCoord を使う。
        ///
        /// @param c RGB 成分を取得する座標値
        /// @param palet 取得した RGB 成分を代入する変数へのポインタ
        /// @return 取得に成功した場合 true を返す
        virtual bool point( Coord< int > c, RGB* palet ) const;

        /// @brief 指定した RGB 成分で点描画する
        ///
        /// 描画対象は現レイヤとなる。また、指定した座標値は、現レイヤ上の座標を意味することに注意。
        /// キャンバス上の座標から現レイヤ上の座標に変換する場合は GDK_Pixbuf::getCoord を使う。
        ///
        /// @param c RGB 成分で描画する座標値
        /// @param palet 描画する RGB 成分
        /// @return 描画に成功した場合 true を返す
        virtual bool pset( Coord< int > c, RGB palet );

        /* キャンバスへの操作・プロパティ */

        /// @brief キャンバスの背景色を返す
        ///
        /// @return キャンバスの背景色
        RGB getCanvasBgColor() const;
        /// @brief キャンバスの背景色をセットする
        ///
        /// @param rgb セットする RGB 成分
        void setCanvasBgColor( RGB rgb );

        /// @brief キャンバスのサイズを返す
        ///
        /// @return キャンバスのサイズ
        Coord< int > canvasSize() const;

        /// @brief キャンバスのサイズを変更する
        ///
        /// 画像の大きさに負数を設定した場合は例外 ExceptionNegativeNumber を投げる。
        ///
        /// @param size 変更後のキャンバス・サイズ
        void canvasResize( Coord< int > size );

        /// @brief キャンバスがアルファチャンネルを持つ場合は true を返す
        ///
        /// @return キャンバスがアルファチャンネルを持つ場合は true
        bool canvasHasAlpha() const;
        /// @brief キャンバスのアルファ・チャンネルをセット・リセットする
        ///
        /// @param flag セットするなら true、リセットするなら false とする
        void setCanvasAlpha( bool flag );

        /* 初期化 */

        // @brief 指定したサイズの画像で初期化する
        //
        // 全てのレイヤはいったんクリアされる
        //
        // @param size 画像サイズ
        // @param rgb 初期化時に塗りつぶす RGB 成分
        // @param hasAlpha アルファ・チャンネルを持つ場合は true
        // @param alpha 合成時のアルファ・チャンネル
        //void initLayer( Coord< int > size, RGB rgb, bool hasAlpha, RGB::primary_type alpha );

        /// @brief 画像ファイルを読み込む
        ///
        /// ファイルの読み込みに失敗した場合はレイヤは生成されず、例外 std::runtime_error を投げる。
        ///
        /// @param fileName 読み込む画像ファイル
        void loadFromFile( const std::string& fileName );
      };

      /**
         @brief GtkDrawingArea を利用した描画領域

         実際の描画内容は GDK_Pixbuf が保持し、GTK_DrawingArea::pset を使いこのオブジェクトを通して描画された
         内容はワークスペース ( 最上位に表示されるオフセットやスケールに依存しない領域 ) に反映される。
         但し、GTK_DrawingArea::point を使って得たピクセル情報は GDK_Pixbuf の内容となることに注意。

         GTK_DrawingArea オブジェクト draw を使って GDK_Pixbuf に描画を行う場合は、draw.pset ではなく
         draw.pixbuf().pset とすることで実現できる。
      **/
      class GTK_DrawingArea : public DrawingArea_IF
      {
        static const int BITS_PER_SAMPLE = 8; // RGB 成分のビット数

        /* メンバ変数 */

        // GDK_Pixbuf オブジェクト
        GDK_Pixbuf pixbuf_;
        // GtkDrawingAreaオブジェクトへのポインタ
        GtkWidget* drawingArea_;

        // scale に依存しない最上位のレイヤ
        // 矩形指定などに利用する
        GdkPixbuf* workspace_;

        // 全レイヤを重ねあわせた結果
        // RGB 成分の取得 ( point ) で利用する
        GdkPixbuf* layer_;

        Coord< double > offset_; // 描画エリアを基準としたキャンバスの位置
        Coord< int > drawSize_;  // 描画する画像のサイズ
        Coord< double > scale_;  // 描画倍率

        guint valChangeSignal_;  // オフセット・スケール値変更時に発行するシグナル

        /* 非公開メンバ関数群*/

        // メンバ変数の初期化
        void init();

        // 座標の再計算
        void recalcCoordSub( int pixbufSize, int widgetSize, double* offset, int* drawSize );
        void recalcCoord();

        // コールバック関数
        static gint configure_event( GtkWidget *widget, GdkEventConfigure *event, gpointer draw );
        static gint drawEvent( GtkWidget *widget, cairo_t* cr, gpointer draw );

        // イベントハンドラのセット
        void setEvent();

        // 画像の描画・表示
        void draw( cairo_t* cr, GdkInterpType interpType );
        void draw( GtkWidget* widget, GdkInterpType interpType );

        // valChangeSignal_ の emit
        void emitValChangeSignal() const;

        // コピーコンストラクタと代入演算子の多重定義は非公開
        explicit GTK_DrawingArea( const GTK_DrawingArea& d );
        GTK_DrawingArea& operator=( const GTK_DrawingArea& d );

        // ワークスペースのリサイズは非公開
        virtual void resize( Coord< int > sz );

      public:

        /* コンストラクタ・デストラクタ */

        /// @brief 空の状態で初期化する
        ///
        /// レイヤは存在しない状態で初期化される。
        GTK_DrawingArea();

        /// @brief キャンバスのサイズ、背景色、アルファ・チャンネルの有無を指定して構築 ( レイヤはまだない状態となる )
        ///
        /// @param sz 画像サイズ
        /// @param rgb 背景色
        /// @param hasAlpha アルファ・チャンネルを持つなら true
        GTK_DrawingArea( Coord< int > sz, RGB rgb, bool hasAlpha );

        /// @brief 画像ファイルを指定して構築
        ///
        /// 画像ファイルを読み込んだレイヤが一つ生成される。
        ///
        /// @param fileName 画像ファイル名
        GTK_DrawingArea( const std::string& fileName );

        /// @brief デストラクタ
        ~GTK_DrawingArea();

        /* キャンバス操作・プロパティ */

        /// @brief 指定した widget 上の座標を現レイヤの座標に変換する
        ///
        /// マウスのクリック位置などに対して現レイヤ上の座標を求める時に利用する
        ///
        /// @param c 指定した widget 上の座標
        /// @return 画像上の相対座標
        Coord< double > getCoord( Coord< double > c ) const;

        /// @brief 描画倍率を取得する
        ///
        /// @return 描画倍率
        Coord< double > getScale() const
        { return( scale_ ); }

        /// @brief 描画倍率をセットする
        ///
        /// 値がゼロ以下だった場合、例外 ExceptionNotPositiveNumber を投げる。
        ///
        /// @note 値を変更した後で再描画のためのシグナル(redraw)を発行したい場合は emitSignal を true にする。
        /// デフォルトでは false となっていることに注意。"redraw" シグナルに対するイベントハンドラ内でこの関数
        /// を使用する場合、決して emitSignal を true にしてはならない。true にした場合、無限にシグナルが発行
        /// されてしまう。
        ///
        /// @param scale セットする描画倍率
        /// @param emitSignal "redraw" シグナルを発行する場合は true にする
        void setScale( Coord< double > scale, bool emitSignal = false );

        /// @brief キャンバスの表示位置を取得する
        ///
        /// 画像は、指定した座標値へシフトして widget 上に描画される。
        /// 座標が負値の場合、画像の端部分は表示されない状態になる。
        /// 描画倍率により拡大・縮小された画像に対する表示位置となることに注意
        ///
        /// @return キャンバスの表示位置
        Coord< double > getOffset() const
        { return( offset_ ); }

        /// @brief キャンバスの表示位置を変更する
        ///
        /// 画像は、指定した座標値へシフトして widget 上に描画される。
        /// 座標が負値の場合、画像の端部分は表示されない状態になる。
        /// 描画倍率により拡大・縮小された画像に対する表示位置となることに注意
        ///
        /// @note 値を変更した後で再描画のためのシグナル(redraw)を発行したい場合は emitSignal を true にする。
        /// デフォルトでは false となっていることに注意。"redraw" シグナルに対するイベントハンドラ内でこの関数
        /// を使用する場合、決して emitSignal を true にしてはならない。true にした場合、無限にシグナルが発行
        /// されてしまう。
        ///
        /// @param c 変更する表示位置
        /// @param emitSignal "redraw" シグナルを発行する場合は true にする
        void setOffset( Coord< double > c, bool emitSignal = false );

        /// @brief キャンバスの表示位置をシフトする
        ///
        /// widget 上に描画する画像の位置を指定した値だけシフトする。
        /// 指定値が正なら右・下、負なら左・上の方向へシフトする。
        /// 座標が正値になった場合、widget 上には余白が生じる状態になる。
        /// 描画倍率により拡大・縮小された画像に対する表示位置となることに注意 ( 倍率による補正は行わない )。
        ///
        /// @note 値を変更した後で再描画のためのシグナル(redraw)を発行したい場合は emitSignal を true にする。
        /// デフォルトでは false となっていることに注意。"redraw" シグナルに対するイベントハンドラ内でこの関数
        /// を使用する場合、決して emitSignal を true にしてはならない。true にした場合、無限にシグナルが発行
        /// されてしまう。
        ///
        /// @param c シフトする値
        /// @param emitSignal "redraw" シグナルを発行する場合は true にする
        void shiftOffset( Coord< double > c, bool emitSignal = false );

        // @brief 画像のリフレッシュ
        //
        // 画像を変更したり、表示座標を変更した場合はこの関数を使ってリフレッシュする必要がある。
        //void reflesh();

        /// @brief 画像を描画する
        ///
        /// @param interpType スケール変更時の補間方法
        void draw( GdkInterpType interpType = GDK_INTERP_BILINEAR );// const;

        /* ワークスペースの操作・プロパティ*/

        /// @brief ワークスペースの生成
        void createWorkspace();

        /// @brief ワークスペースの消去
        void clearWorkspace();

        /// @brief ワークスペースが有効かをチェックする
        ///
        /// @return ワークスペースが有効なら true を返す
        bool wsIsValid() const
        { return( workspace_ != 0 ); }

        /// @brief ワークスペースの画像サイズを返す ( widget のサイズに等しい )
        ///
        /// @return ワークスペースの画像サイズ
        virtual Coord< int > size() const;

        /* ワークスペースへの描画・パレット取得 */

        /// @brief 指定した座標の RGB 成分を取得する
        ///
        /// 取得できる RGB 成分は GDK_Pixbuf の全レイヤを重ねあわせた結果から得られることに注意。
        ///
        /// @param c RGB 成分を取得する座標値
        /// @param palet 取得した RGB 成分を代入する変数へのポインタ
        /// @return 取得に成功した場合 true を返す
        virtual bool point( Coord< int > c, RGB* palet ) const;

        /// @brief 指定した座標を指定した RGB 成分で描画する
        ///
        /// 描画対象はワークスペースであり、GDK_Pixbuf ではないことに注意。GDK_Pixbuf に描画する場合は
        /// GTK_DrawingArea::pixbuf() を通して行う。
        ///
        /// @param c 指定した RGB 成分で描画する座標値
        /// @param palet 描画に使う RGB 成分
        /// @return 描画に成功した場合 true を返す
        virtual bool pset( Coord< int > c, RGB palet );

        /* ウィジェット操作 */

        /// @brief GDK_Pixbuf への参照を返す
        ///
        /// @return 描画対象の widget
        GDK_Pixbuf& pixbuf() { return( pixbuf_ ); }

        /// @brief 描画対象の widget へのポインタを返す
        ///
        /// @return 描画対象の widget
        GtkWidget* widget() const { return( drawingArea_ ); }

        /// @brief 描画対象の widget を表示する
        void show() const { gtk_widget_show_all( widget() ); }
      };

    } // namespace canvas

    namespace gtk_event
    {
      /**
         @brief GTK_DrawingArea クラス用 Event 作成クラス
      **/
      class GTK_DrawingArea_Event
      {
        canvas::GTK_DrawingArea& drawingArea_;      // 描画領域
        std::map< std::string, guint > id_; // イベントID

        // コピー・コンストラクタと代入演算子は利用不可
        GTK_DrawingArea_Event( const GTK_DrawingArea_Event& event );
        GTK_DrawingArea_Event& operator=( const GTK_DrawingArea_Event& event );

      protected:

        /// @brief シグナルに対するコールバック関数の接続
        ///
        /// @param ident シグナルに対する識別子
        /// @param signal シグナルの名称
        /// @param func コールバック関数
        /// @return 接続に成功したら true を返す
        bool connect( const std::string& ident, const std::string& signal, GCallback func );

        /// @brief 登録された GTK_DrawingArea への参照を返す
        ///
        /// @return 登録された GTK_DrawingArea への参照
        canvas::GTK_DrawingArea& drawingArea() { return( drawingArea_ ); }

      public:

        /* コンストラクタ */

        /// @brief 対象の GTK_DrawingArea を指定して構築
        ///
        /// @param drawingArea 対象の GTK_DrawingArea
        GTK_DrawingArea_Event( canvas::GTK_DrawingArea& drawingArea )
          : drawingArea_( drawingArea ) {}

        /// @brief デストラクタ
        ///
        /// 登録された全てのイベントを切断する
        virtual ~GTK_DrawingArea_Event();

        /// @brief シグナルに対するコールバック関数のブロック
        ///
        /// @param ident シグナルの識別子
        virtual void block( const std::string& ident );

        /// @brief シグナルに対するコールバック関数のブロック解除
        ///
        /// @param ident シグナルの識別子
        virtual void unblock( const std::string& ident );

        /// @brief 全シグナルに対するコールバック関数のブロック
        virtual void block();

        /// @brief 全シグナルに対するコールバック関数のブロック解除
        virtual void unblock();
      };

      /**
         @brief GTK_DrawingAreaクラス用 画像シフト機能

         マウス・ドラッグによる画像のシフトの他に、マウスホイールで上下の移動、Shift キーを併用することで左右の移動ができる。
      **/
      class GTK_DrawingArea_ImageShiftEvent : public GTK_DrawingArea_Event
      {
        Coord< int > mouse; // 前回にマウスドラッグした時のリリース位置

        guint32 scrollTime_; // スクロールイベントが発行された時の時間(msec) イベントの多発を抑制するために利用

        // マウスドラッグによって、画像の移動を行う処理
        static gint event( GtkWidget *widget, GdkEventMotion *event, gpointer gp );

        // マウスドラッグによって、画像の移動を行う処理
        static gint scrollEvent( GtkWidget* widget, GdkEventScroll* event, gpointer gp );

        // オフセット・スケールの変更による画像の再描画
        static gint redrawEvent( GtkWidget* widget, gpointer gp );

      public:

        /* コンストラクタ */

        /// @brief 対象の GTK_DrawingArea を指定して構築
        ///
        /// @param drawingArea 対象の GTK_DrawingArea
        GTK_DrawingArea_ImageShiftEvent( canvas::GTK_DrawingArea& drawingArea );
      };

    } // namespace gtk_event

    namespace canvas
    {
      /**
         @brief GTK_DrawingArea 用 Scrolled Window クラス

         GTK_DrawingArea にスクロールバーを追加した widget

         スクロールバーによる画像のシフトの他に、マウスホイールで上下の移動、Shift キーを併用することで左右の移動ができる。
      **/
      class GTK_DrawingArea_ScrolledWindow
      {
        using event = std::pair< guint, GObject* >; // イベントID, 対象オブジェクト格納用コンテナ型

        GTK_DrawingArea& drawingArea_; // 描画領域
        Coord< gdouble > value_;       // スクロールバーが示す現在位置
        GtkWidget* table_;             // レイアウト用コンテナ
        GtkAdjustment* hAdj_; // 位置調節用 I/F (H)
        GtkAdjustment* vAdj_; // 位置調節用 I/F (V)
        GtkWidget* hBar_;
        GtkWidget* vBar_;
        std::map<std::string, event> eventInfo_; // イベント情報( ident, Event )

        guint32 scrollTime_; // スクロールイベントが発行された時の時間(msec) イベントの多発を抑制するために利用

        // コピー・コンストラクタと代入演算子は利用不可
        GTK_DrawingArea_ScrolledWindow( const GTK_DrawingArea_ScrolledWindow& scrolledWin );
        GTK_DrawingArea_ScrolledWindow& operator=( const GTK_DrawingArea_ScrolledWindow& scrolledWin );

        // 位置変更時のイベント
        static gint hEvent( GtkAdjustment* adj, gpointer gp ); // 水平スクロールバー
        static gint vEvent( GtkAdjustment* adj, gpointer gp ); // 垂直スクロールバー

        // マウスホイール使用時(スクロール)のイベント
        static gint scrollEvent( GtkWidget* widget, GdkEventScroll* event, gpointer gp );

        // ウインドウサイズ変更時のイベント
        static gint sizeChangedEvent( GtkWidget *widget, GdkEventConfigure *event, gpointer gp );

        // オフセット・スケールの変更による画像の再描画
        static gint redrawEvent( GtkWidget* widget, gpointer gp );

        // シグナルに対するコールバック関数の接続
        void connect( const std::string& ident, const std::string& event, GObject* adj, GCallback func );

        // スクロールバーの上限を変更する
        void setUpper();

        // スケールの変更
        void changeScale( bool incScale, Coord< double > mouse );

      public:

        /* コンストラクタ */

        /// @brief スクロールバーを追加する GTK_DrawingArea オブジェクトを指定して構築
        ///
        /// @param drawingArea 対象の GTK_DrawingArea オブジェクト
        explicit GTK_DrawingArea_ScrolledWindow( GTK_DrawingArea& drawingArea );

        /// @brief デストラクタ
        ~GTK_DrawingArea_ScrolledWindow();

        /// @brief スクロール位置 ( 画像表示位置 ) を初期化する
        void initAdjustment();

        /// @brief 内部に保持している画像領域 ( GTK_DrawingArea ) へのポインタを返す
        ///
        /// @return 画像領域 ( GTK_DrawingArea ) へのポインタ
        GTK_DrawingArea* drawingArea() const { return( &drawingArea_ ); }

        /// @brief 内部に保持している水平位置調節用インターフェースへのポインタを返す
        ///
        /// @return 水平位置調節用インターフェース ( GtkAdjustment ) へのポインタ
        GtkAdjustment* hAdjustment() const { return( hAdj_ ); }
        /// @brief 内部に保持している垂直位置調節用インターフェースへのポインタを返す
        ///
        /// @return 垂直位置調節用インターフェース ( GtkAdjustment ) へのポインタ
        GtkAdjustment* vAdjustment() const { return( vAdj_ ); }

        /// @brief GTK_DrawingArea オブジェクトとスクロールバーを乗せたコンテナ・ウィジェット ( GtkTable ) へのポインタを返す
        ///
        /// @return コンテナ・ウィジェット ( GtkTable ) へのポインタ
        GtkWidget* widget() const { return( table_ ); }
      };

    } // namespace canvas

    namespace draw
    {
      /**
         @brief 描画用関数オブジェクト I/F

         全ての描画用関数オブジェクトの基底
      **/
      struct DrawFuncIF
      {
        /// @brief 仮想デストラクタ(何もしない)
        virtual ~DrawFuncIF() {}
      };

      /**
         @brief 描画用関数オブジェクト 一座標指定 I/F

         画像上の点を一つだけ指定して描画を行う関数オブジェクト
      **/
      struct DrawFuncIF_1p : public DrawFuncIF
      {
        /// @brief 描画用関数オブジェクト
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c 描画座標
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c ) = 0;
      };

      /**
         @brief 描画用関数オブジェクト 二座標指定 I/F

         画像上の点を二つ指定して描画を行う関数オブジェクト
      **/
      struct DrawFuncIF_2p : public DrawFuncIF
      {
        /// @brief 描画用関数オブジェクト
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1 ) = 0;
      };

      /**
         @brief 描画用関数オブジェクト 三座標指定 I/F

         画像上の点を三つ指定して描画を行う関数オブジェクト
      **/
      struct DrawFuncIF_3p : public DrawFuncIF
      {
        /// @brief 描画用関数オブジェクト
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        /// @param c2 描画座標 2
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1, Coord< int > c2 ) = 0;
      };

      /**
         @brief 描画用関数オブジェクト 任意の数の座標指定 I/F

         画像上の点を任意の数だけ指定して描画を行う関数オブジェクト
      **/
      struct DrawFuncIF_Multiple : public DrawFuncIF
      {
        /// @brief 描画用関数オブジェクト(何も処理しない)
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 描画座標の配列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c ) = 0;
      };

      /**
         @brief 描画用関数オブジェクト 一座標指定 アダプタ

         DrawFuncIF_1p に任意の描画関数を適用するためのアダプタ
      **/
      class DrawFuncAdapter_1p : public DrawFuncIF_1p
      {
        // 組み込む関数へのポインタの型
        using ptr_func = void ( * )( canvas::DrawingArea_IF&, pen::GPixelOp&, Coord< int > );

        // 組み込む関数へのポインタ
        ptr_func func_;

      public:

        /* コンストラクタ */

        /// @brief 呼び出す描画関数を指定して構築
        ///
        /// @param func 呼び出す関数へのポインタ
        DrawFuncAdapter_1p( ptr_func func )
          : func_( func ) {}

        /// @brief 指定した関数を呼び出して描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c 描画座標
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c )
        {
          if ( func_ != 0 )
            (*func_)( draw, pixelOp, c );
        }
      };

      /**
         @brief 描画用関数オブジェクト 二座標指定 アダプタ

         DrawFuncIF_2p に任意の描画関数を適用するためのアダプタ
      **/
      class DrawFuncAdapter_2p : public DrawFuncIF_2p
      {
        // 組み込む関数へのポインタの型
        using ptr_func = void ( * )( canvas::DrawingArea_IF&, pen::GPixelOp&, Coord< int >, Coord< int > );

        // 組み込む関数へのポインタ
        ptr_func func_;

      public:

        /* コンストラクタ */

        /// @brief 呼び出す描画関数を指定して構築
        ///
        /// @param func 組み込む関数へのポインタ
        DrawFuncAdapter_2p( ptr_func func )
          : func_( func ) {}

        /// @brief 指定した関数を呼び出して描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1 )
        {
          if ( func_ != 0 )
            (*func_)( draw, pixelOp, c0, c1 );
        }
      };

      /**
         @brief 描画用関数オブジェクト 三座標指定 アダプタ

         DrawFuncIF_3p に任意の描画関数を適用するためのアダプタ
      **/
      class DrawFuncAdapter_3p : public DrawFuncIF_3p
      {
        // 組み込む関数へのポインタの型
        using ptr_func = void ( * )( canvas::DrawingArea_IF&, pen::GPixelOp&, Coord< int >, Coord< int >, Coord< int > );

        // 組み込む関数へのポインタ
        ptr_func func_;

      public:

        /* コンストラクタ */

        /// @brief 呼び出す描画関数を指定して構築
        ///
        /// @param func 組み込む関数へのポインタ
        DrawFuncAdapter_3p( ptr_func func )
          : func_( func ) {}

        /// @brief 指定した関数を呼び出して描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        /// @param c2 描画座標 2
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1, Coord< int > c2 )
        {
          if ( func_ != 0 )
            (*func_)( draw, pixelOp, c0, c1, c2 );
        }
      };

      /**
         @brief 描画用関数オブジェクト 任意の数の座標指定 アダプタ

         DrawFuncIF_Multiple に任意の描画関数を適用するためのアダプタ
      **/
      class DrawFuncAdapter_Multiple : public DrawFuncIF_Multiple
      {
        // 組み込む関数へのポインタの型
        using ptr_func = void ( * )( canvas::DrawingArea_IF&, pen::GPixelOp&, const std::vector< Coord< int > >& );

        // 組み込む関数へのポインタ
        ptr_func func_;

      public:

        /* コンストラクタ */

        /// @brief 呼び出す描画関数を指定して構築
        ///
        /// @param func 組み込む関数へのポインタ
        DrawFuncAdapter_Multiple( ptr_func func )
          : func_( func ) {}

        /// @brief 指定した関数を呼び出して描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 描画座標の配列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c )
        {
          if ( func_ != 0 )
            (*func_)( draw, pixelOp, vec_c );
        }
      };

      /**
         @brief 描画用関数オブジェクト 線分描画
      **/
      class DrawLine : public DrawFuncIF_2p
      {
      public:

        bool connect4; // Trueの場合は 4連結で描画する

        /* コンストラクタ */

        /// @brief 線分の連結方法を指定して構築
        ///
        /// @param isConnect4 True の場合は 4連結で描画する
        DrawLine( bool isConnect4 = false )
          : connect4( isConnect4 ) {}

        /// @brief 線分 c0 - c1 を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1 )
        { shape::line::Draw( draw, pixelOp, c0, c1, connect4 ); }
      };

      /**
         @brief 描画用関数オブジェクト 矩形描画
      **/
      class DrawBox : public DrawFuncIF_2p
      {
      public:

        bool filled; // Trueの場合は塗りつぶし

        /* コンストラクタ */

        /// @brief 矩形の内部を塗りつぶすかどうかを指定して構築
        ///
        /// @param isFilled True の場合は塗りつぶし
        DrawBox( bool isFilled = false )
          : filled( isFilled ) {}

        /// @brief c0,c1 を対角とする矩形を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1 )
        { shape::box::Draw( draw, pixelOp, c0, c1, filled ); }
      };

      /**
         @brief 描画用関数オブジェクト 十字描画
      **/
      class DrawCross : public DrawFuncIF_1p
      {
      public:

        /// @brief c0,c1 を対角とする矩形を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c 描画位置(十字の交差する点)
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c );
      };

      /**
         @brief 描画用関数オブジェクト 楕円描画
      **/
      class DrawEllipse : public DrawFuncIF_2p
      {
      public:

        bool filled; // Trueの場合は塗りつぶし

        /* コンストラクタ */

        /// @brief 楕円の内部を塗りつぶすかどうかを指定して構築
        ///
        /// @param isFilled True の場合は塗りつぶし
        DrawEllipse( bool isFilled = false )
          : filled( isFilled ) {}

        /// @brief 二点 c0,c1 を通る楕円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, Coord< int > c1 )
        { shape::ellipse::Draw_2p( draw, pixelOp, c0, c1, filled ); }
      };

      /**
         @brief 描画用関数オブジェクト 真円描画
      **/
      class DrawCircle : public DrawFuncIF_2p
      {
      public:

        bool filled; // Trueの場合は塗りつぶし

        /* コンストラクタ */

        /// @brief 円の内部を塗りつぶすかどうかを指定して構築
        ///
        /// @param isFilled True の場合は塗りつぶし
        DrawCircle( bool isFilled = false )
          : filled( isFilled ) {}

        /// @brief o を中心とし c を通る円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param o 円の中心
        /// @param c 円周を通る 1 点
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o, Coord< int > c )
        { shape::circle::Draw( draw, pixelOp, o, coord_lib::Distance< int >( o, c ), filled ); }
      };

      /**
         @brief 描画用関数オブジェクト 真円描画

         二点を直径の両端として円を決定し、描画する
      **/
      class DrawCircle_2p : public DrawFuncIF_2p
      {
      public:

        bool filled; // Trueの場合は塗りつぶし

        /* コンストラクタ */

        /// @brief 真円の内部を塗りつぶすかどうかを指定して構築
        ///
        /// @param isFilled True の場合は塗りつぶし
        DrawCircle_2p( bool isFilled = false )
          : filled( isFilled ) {}

        /// @brief 二点 c0,c1 を直径とする真円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, Coord< int > c1 )
        { shape::circle::Draw_2p( draw, pixelOp, c0, c1, filled ); }
      };

      /**
         @brief 描画用関数オブジェクト 三角形描画
      **/
      class DrawTriangle : public DrawFuncIF_3p
      {
      public:

        /// @brief 三点 c0,c1,c2 を頂点とする三角形を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        /// @param c2 描画座標 2
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1, Coord< int > c2 )
        {
          shape::line::Draw( draw, pixelOp, c0, c1 );
          shape::line::Draw( draw, pixelOp, c1, c2 );
          shape::line::Draw( draw, pixelOp, c2, c0 );
        }
      };

      /**
         @brief 描画用関数オブジェクト 真円描画

         円周上の三点から中心と半径を決定して描画する
      **/
      class DrawCircle_3p : public DrawFuncIF_3p
      {
      public:

        bool filled; // Trueの場合は塗りつぶし

        /* コンストラクタ */

        /// @brief 真円の内部を塗りつぶすかどうかを指定して構築
        ///
        /// @param isFilled True の場合は塗りつぶし
        DrawCircle_3p( bool isFilled = false )
          : filled( isFilled ) {}

        /// @brief 三点 c0,c1,c2 を通る真円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        /// @param c2 描画座標 2
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1, Coord< int > c2 )
        { shape::circle::Draw_3p( draw, pixelOp, c0, c1, c2, filled ); }
      };

      /**
         @brief 描画用関数オブジェクト 線分・真円の描画

         二点を通る線分と、それを直径とする真円を同時に描画する
      **/
      class DrawCircleWithLine : public DrawCircle_2p
      {
      public:

        /// @brief 二点 c0,c1 を結んだ線分と、それを直径とする真円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > c0, Coord< int > c1 )
        {
          shape::line::Draw( draw, pixelOp, c0, c1 );
          shape::circle::Stroke_2p( draw, pixelOp, c0, c1 );
        }
      };

      /**
         @brief 描画用関数オブジェクト 三角形・真円描画

         三点を頂点とする三角形と、それに外接する真円を描画する
      **/
      class DrawCircleWithTriangle : public DrawCircle_3p
      {
      public:

        /// @brief 三点 c0,c1,c2 を頂点とする三角形と、それに外接する真円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param c0 描画座標 0
        /// @param c1 描画座標 1
        /// @param c2 描画座標 2
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > c0, Coord< int > c1, Coord< int > c2 )
        {
          DrawTriangle()( draw, pixelOp, c0, c1, c2 );
          shape::circle::Stroke_3p( draw, pixelOp, c0, c1, c2 );
        }
      };

      /**
         @brief 描画用関数オブジェクト 円弧描画

         最初の一点を中心とし、他の二点を端点とする円弧を描画する
      **/
      class DrawArc : public DrawCircle_3p
      {
        bool filled_; // true の場合は塗りつぶし

      public:

        DrawArc( bool isFilled = false )
          : filled_( isFilled ) {}

        /// @brief o を中心とし、二点 c0, c1 を端点とする円弧を描画する
        ///
        /// 円弧は、c0 から c1 へ時計回りに描画される。
        ///
        /// 円弧の半径は o から c0 までの距離で決定する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param o 円弧の中心座標
        /// @param c0 円弧の両端座標 1
        /// @param c1 円弧の両端座標 2
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 Coord< int > o, Coord< int > c0, Coord< int > c1 )
        {
          int r = std::round( coord_lib::Distance< int >( o, c0 ) );
          shape::arc::Draw( draw, pixelOp, o, r, c0, c1, false, filled_ );
        }
      };

      /**
         @brief 描画用関数オブジェクト 複数の直線描画
      **/
      class DrawMultipleLine : public DrawFuncIF_Multiple
      {
      public:

        bool closed; ///< 末尾と開始の点を連結する場合は true とする

        /* コンストラクタ */

        /// @brief 末尾と開始の点を連結するか(閉多角形とするか)を指定して構築
        ///
        /// @param closed 末尾と開始の点を連結するなら true を指定する
        DrawMultipleLine( bool closed )
          : closed( closed ) {}

        /// @brief 座標の配列を端点とする線分を描画する
        ///
        /// 線分は、添字を一つずつたどりながら描画するため、描画結果は各線分が連結された形になる。
        /// 但し、配列の末尾と開始は連結しない。
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 描画座標の配列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, const std::vector< Coord< int > >& vec_c )
        {
          shape::line::DrawMultiple( draw, pixelOp, vec_c );
          if ( closed && vec_c.size() > 2 )
            shape::line::Draw( draw, pixelOp, vec_c.back(), vec_c.front() );
        }
      };

      /**
         @brief 描画用関数オブジェクト ポリゴン描画
      **/
      class DrawPolygon : public DrawFuncIF_Multiple
      {
        shape::polygon::METHOD method_; // 領域の判定方法

      public:

        /* コンストラクタ */

        /// @brief 領域の判定方法を指定して構築
        ///
        /// @param method 領域の判定方法(CROSSING_NUMBER,WINDING_NUMBER)
        DrawPolygon( shape::polygon::METHOD method )
          : method_( method ) {}

        /// @brief 座標の配列を頂点とするポリゴンを描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c ポリゴンの頂点の配列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c )
        { shape::polygon::Fill( draw, pixelOp, vec_c, method_ ); }
      };

      /**
         @brief 描画用関数オブジェクト Catmull-Rom 曲線描画
      **/
      struct DrawCatmullRomSpline : public DrawFuncIF_Multiple
      {
        double d; ///< 微分係数計算時の補正係数

        /* コンストラクタ */

        /// @brief 微分係数計算時の補正係数を指定して構築
        ///
        /// @param d 微分係数計算時の補正係数
        DrawCatmullRomSpline( double d )
          : d( d ) {}

        /// @brief 座標の配列を点列とする曲線を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 点列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c )
        { shape::curve::StrokeCatmullRomSpline( draw, pixelOp, vec_c, d, 0.4 ); }
      };

      /**
         @brief 描画用関数オブジェクト ベジェ曲線描画
      **/
      class DrawBezierSpline : public DrawFuncIF_Multiple
      {
        size_t n_; // 次数

      public:

        /* コンストラクタ */

        /// @brief 描画に使う Bernstein 多項式の次数を指定して構築
        ///
        /// @param n 次数
        DrawBezierSpline( size_t n ) : n_( n ) {}

        /// @brief 座標の配列を点列とする曲線を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 点列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c )
        { shape::curve::StrokeBezierSpline( draw, pixelOp, vec_c, n_, 0.4 ); }
      };

      /**
         @brief 描画用関数オブジェクト 一様スプライン曲線描画
      **/
      class DrawUniformBSpline : public DrawFuncIF_Multiple
      {
        size_t n_; // 次数

      public:

        /* コンストラクタ */

        /// @brief 描画に使う B-スプライン基底関数の次数を指定して構築
        ///
        /// @param n 次数
        DrawUniformBSpline( size_t n ) : n_( n ) {}

        /// @brief 座標の配列を点列とする曲線を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 点列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c )
        { shape::curve::StrokeUniformBSpline( draw, pixelOp, vec_c, n_, 0.4 ); }
      };

      /**
         @brief 描画用関数オブジェクト 開一様スプライン曲線描画
      **/
      class DrawOpenUniformBSpline : public DrawFuncIF_Multiple
      {
        size_t n_; // 次数

      public:

        /* コンストラクタ */

        /// @brief 描画に使う B-スプライン基底関数の次数を指定して構築
        ///
        /// @param n 次数
        DrawOpenUniformBSpline( size_t n ) : n_( n ) {}

        /// @brief 座標の配列を点列とする曲線を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param vec_c 点列
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp,
                                 const std::vector< Coord< int > >& vec_c )
        { shape::curve::StrokeOpenUniformBSpline( draw, pixelOp, vec_c, n_, 0.4 ); }
      };

      /**
         @brief 描画用関数オブジェクト 円形ブラシ形状の描画

         pen::GCircleBrush のペン形状のカーソル描画に利用する
      **/
      class DrawCircleBrush : public DrawFuncIF_1p
      {
        static const int MIN_RADIUS_TO_DRAW_CROSS = 4;

        pen::GCircleBrush& brush_; // 対象のブラシ

      public:

        /* コンストラクタ */

        /// @brief 描画対象のブラシを指定して構築
        ///
        /// @param brush 描画対象のブラシ
        DrawCircleBrush( pen::GCircleBrush& brush )
          : brush_( brush ) {}

        /// @brief o を中心にペンサイズの円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param o 円の中心
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o );
      };

      /**
         @brief 描画用関数オブジェクト 楕円ペン形状の描画

         GEllipsePen のペン形状のカーソル描画に利用する
      **/
      class DrawEllipsePen : public DrawFuncIF_1p
      {
        static const int MIN_RADIUS_TO_DRAW_CROSS = 4;

        pen::GEllipsePen& pen_; // 対象のペン

      public:

        /* コンストラクタ */

        /// @brief 描画対象のペンを指定して構築
        ///
        /// @param pen 描画対象のペン
        DrawEllipsePen( pen::GEllipsePen& pen )
          : pen_( pen ) {}

        /// @brief o を中心にペンサイズの円を描画する
        ///
        /// @param draw 描画領域への参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        /// @param o 円の中心
        virtual void operator()( canvas::DrawingArea_IF& draw, pen::GPixelOp& pixelOp, Coord< int > o );
      };

    } // namespace draw

    namespace gtk_event
    {
      /**
         @brief GTK_DrawingArea への選択機能設定用イベント
      **/
      class GTK_DrawingArea_SelectionEvent : public GTK_DrawingArea_Event
      {
        // 対象の描画領域
        //GTK_DrawingArea& drawingArea_;

        // スクロール・バーの値変更に対するイベントのハンドルID
        //gulong hId_;
        //gulong vId_;

        // マウスイベント情報の取得
        static void getEventInfo( GdkEventMotion* event, Coord< int >& c, GdkModifierType& state )
        { gdk_window_get_device_position( event->window, event->device, &( c.x ), &( c.y ), &state ); }

        // マウスがウィジェット内に入った時のコールバック関数
        static gint enterNotifyEvent( GtkWidget* widget, GdkEvent* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->enterNotifyAction( event ) ); }

        // マウスがウィジェットから出た時のコールバック関数
        static gint leaveNotifyEvent( GtkWidget* widget, GdkEvent* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->leaveNotifyAction( event ) ); }

        // マウスボタンクリック時のコールバック関数
        static gint buttonPressEvent( GtkWidget* widget, GdkEventButton* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->buttonPressAction( event ) ); }

        // マウスボタンリリース時のコールバック関数
        static gint buttonReleaseEvent( GtkWidget* widget, GdkEventButton* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->buttonReleaseAction( event ) ); }

        // キークリック時のコールバック関数
        static gint keyPressEvent( GtkWidget* widget, GdkEventKey* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->keyPressAction( event ) ); }

        // キーリリース時のコールバック関数
        static gint keyReleaseEvent( GtkWidget* widget, GdkEventKey* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->keyReleaseAction( event ) ); }

        // マウス移動時のコールバック関数
        static gint motionNotifyEvent( GtkWidget* widget, GdkEventMotion *event, gpointer gp );

        // マウススクロール時のコールバック関数
        static gint scrollEvent( GtkWidget* widget, GdkEventScroll *event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->scrollAction( event ) ); }

        // サイズ更新時のコールバック関数
        static gint configureEvent( GtkWidget* widget, GdkEvent* event, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->configureAction( event ) ); }

        // オフセット・スケール変更時のコールバック関数
        static gint redrawEvent( GtkWidget* widget, gpointer gp )
        { return( ( (GTK_DrawingArea_SelectionEvent*) gp )->redrawAction() ); }

      protected:

        /// @brief マウスがウィジェット内に入った時の動作 (何もしない)
        ///
        /// @param event イベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint enterNotifyAction( GdkEvent* event ) { return( FALSE ); }

        /// @brief マウスがウィジェットから出た時の動作 (何もしない)
        ///
        /// @param event イベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint leaveNotifyAction( GdkEvent* event ) { return( FALSE ); }

        /// @brief マウスボタンクリック時の動作 (何もしない)
        ///
        /// @param event マウスボタン操作時のイベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint buttonPressAction( GdkEventButton* event ) { return( FALSE ); }

        /// @brief マウスボタンリリース時の動作 (何もしない)
        ///
        /// @param event マウスボタン操作時のイベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint buttonReleaseAction( GdkEventButton* event ) { return( FALSE ); }

        /// @brief キークリック時の動作 (何もしない)
        ///
        /// @param event キー操作時のイベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint keyPressAction( GdkEventKey* event ) { return( FALSE ); }

        /// @brief キーリリース時の動作 (何もしない)
        ///
        /// @param event キー操作時のイベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint keyReleaseAction( GdkEventKey* event ) { return( FALSE ); }

        /// @brief マウス移動時の動作 (何もしない)
        ///
        /// @param c マウスの位置
        /// @param state 修飾キーやマウスボタンの状態を表すビット列
        /// @return 常に FALSE
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state ) { return( FALSE ); }

        /// @brief マウススクロール時の動作 (何もしない)
        ///
        /// @param event イベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint scrollAction( GdkEventScroll* event ) { return( FALSE ); }

        /// @brief サイズ更新時の動作 (何もしない)
        ///
        /// @param event イベント構造体へのポインタ
        /// @return 常に FALSE
        virtual gint configureAction( GdkEvent *event ) { return( FALSE ); }

        /// @brief オフセット・スケール変更時の動作 (何もしない)
        ///
        /// @return 常に FALSE
        virtual gint redrawAction() { return( FALSE ); }

      public:

        /* コンストラクタ */

        /// @brief イベントを定義する描画領域を指定して実行
        ///
        /// 描画領域 drawingArea が未定義の場合は assert を実行する
        ///
        /// @param drawingArea 対象の描画領域への参照
        GTK_DrawingArea_SelectionEvent( canvas::GTK_DrawingArea& drawingArea );
      };

      /**
         @brief GTK_DrawingArea選択範囲描画用イベント

         マウス操作で描画を行うコールバック関数用の関数オブジェクト
      **/
      class GTK_DrawingArea_DrawingEvent : public GTK_DrawingArea_SelectionEvent
      {
        pen::GPixelOp* pixelOp_;      // 点描画関数へのポインタ

      protected:

        /// @brief 点描画関数への参照を返す
        ///
        /// @return 点描画関数への参照
        pen::GPixelOp& pixelOp() const
        { return( *pixelOp_ ); }

      public:

        /* コンストラクタ */

        /// @brief イベントを定義する描画領域と点描画関数オブジェクトを指定して構築
        ///
        /// @param drawingArea 描画対象
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_DrawingEvent( canvas::GTK_DrawingArea& drawingArea, pen::GPixelOp* pixelOp );

        /// @brief 点描画関数をセットする
        ///
        /// @param pixelOp 点描画関数へのポインタ
        //virtual void setPixelOp( pen::GPixelOp* pixelOp );
        void setPixelOp( pen::GPixelOp* pixelOp );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント ( 1 パス )

         1 パスの操作で描画を行うコールバック関数用の関数オブジェクト
         描画前のカーソルと実際の処理内容を指定して任意の位置への処理ができるようにする
      **/
      class GTK_DrawingArea_1PassSelectionEvent : public GTK_DrawingArea_DrawingEvent
      {
        draw::DrawFuncIF_1p* cursor0_; // 初期状態のカーソルへのポインタ
        draw::DrawFuncIF_1p* draw_;    // 描画関数へのポインタ

        bool isFirst_;           // 初期状態か？
        Coord< double > c_;      // 現在の座標値

        // マウスボタンクリック時の動作
        //
        // 初期状態のカーソルを消去し描画を行う
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウス移動時の操作
        //
        // マウス位置を再計算して初期状態のカーソルを再描画する
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更時の操作
        //
        // カーソルを再描画する
        virtual gint redrawAction();

        // カーソルの描画を消去して内容を初期化する
        void reset();

      public:

        /* コンストラクタ */

        /// @brief 初期状態のカーソルと描画内容を個々に定義して構築
        ///
        /// @param drawingArea 描画対象
        /// @param cursor0 処理前のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_1PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_1p* cursor0,
                                             draw::DrawFuncIF_1p* draw, pen::GPixelOp* pixelOp );

        /// @brief 描画内容のみを定義して構築 ( 初期状態のカーソルは非表示 )
        ///
        /// @param drawingArea 描画対象
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_1PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_1p* draw, pen::GPixelOp* pixelOp );

        /* デストラクタ */

        /// @brief デストラクタ (カーソルの描画を消去して内容を初期化する)
        ~GTK_DrawingArea_1PassSelectionEvent();

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        /// @brief 初期状態のカーソルをセットする
        ///
        /// @param cursor0 初期状態のカーソルへのポインタ
        void setCursor0( draw::DrawFuncIF_1p* cursor0 )
        { cursor0_ = cursor0; }

        /// @brief 描画関数をセットする
        ///
        /// @param draw 描画関数へのポインタ
        void setDraw( draw::DrawFuncIF_1p* draw );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント ( 2 パス )

         2 パスの操作で描画を行うコールバック関数用の関数オブジェクト
         描画前の二つのカーソルと実際の処理内容を指定して任意の位置への処理ができるようにする
      **/
      class GTK_DrawingArea_2PassSelectionEvent : public GTK_DrawingArea_DrawingEvent
      {
        // 描画状態
        // NO_DRAW ... 描画前(カーソル表示なし)
        // CURSOR0 ... 一つめのカーソル ( 1 点での描画 )
        // CURSOR1 ... 二つめのカーソル ( 2 点での描画 )
        enum Status { NO_DRAW, CURSOR0, CURSOR1, };

        draw::DrawFuncIF_1p* cursor0_; // 初期状態のカーソル
        draw::DrawFuncIF_2p* cursor1_; // 一点が決まった後のカーソル
        draw::DrawFuncIF_2p* draw_;    // 描画関数

        int status_;             // 現在の描画状態

        // 描画対象の座標値
        // c0_, c1_ はどちらも widget 上の座標を表す
        Coord< int > c0_, c1_;
        // 現レイヤに対する始点の座標値
        Coord< double > l0_;

        // マウスボタンクリック時の動作
        //
        // 左クリックで現在のカーソルを消去しステータスを次に移行させる
        // 描画前ステータスだった場合は描画を行い初期状態に戻す
        // 右クリックで現在のカーソルを消去しステータスを一つ戻す
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウス移動時の操作
        //
        // マウス位置を再計算してカーソルを再描画する
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更時の操作
        //
        // カーソルを再描画する
        virtual gint redrawAction();

        // カーソルの描画を消去して内容を初期化する
        void reset();

      public:

        /* コンストラクタ */

        /// @brief 二つのカーソルと描画内容を個々に定義して構築
        ///
        /// @param drawingArea 描画対象
        /// @param cursor0 クリック前のカーソルを描画する関数オブジェクトへのポインタ
        /// @param cursor1 一回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_2PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_2p* cursor1,
                                             draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp );

        /// @brief 一回クリックした後のカーソルと描画内容が同じである場合の定義
        ///
        /// @param drawingArea 描画対象
        /// @param cursor0 クリック前のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_2PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_1p* cursor0,
                                             draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp );

        /// @brief クリック前のカーソルは描画しない場合の定義
        ///
        /// @param drawingArea 描画対象
        /// @param cursor1 一回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_2PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_2p* cursor1,
                                             draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp );

        /// @brief クリック前のカーソルは描画せず、一回クリックした後のカーソルと描画内容が同じである場合の定義
        ///
        /// @param drawingArea 描画対象
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_2PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_2p* draw, pen::GPixelOp* pixelOp );

        /// @brief デストラクタ (カーソルの描画を消去して内容を初期化する)
        ~GTK_DrawingArea_2PassSelectionEvent()
        { reset(); status_ = NO_DRAW; }

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        /// @brief 初期状態のカーソルをセットする
        ///
        /// @param cursor0 初期状態のカーソルへのポインタ
        void setCursor0( draw::DrawFuncIF_1p* cursor0 )
        { cursor0_ = cursor0; }

        /// @brief 一点が決まった後のカーソルをセットする
        ///
        /// @param cursor1 一点が決まった後のカーソルへのポインタ
        void setCursor1( draw::DrawFuncIF_2p* cursor1 );

        /// @brief 描画関数をセットする
        ///
        /// @param draw 描画関数へのポインタ
        void setDraw( draw::DrawFuncIF_2p* draw );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント ( 3 パス )

         3 パスの操作で描画を行うコールバック関数用の関数オブジェクト
         描画前の三つのカーソルと実際の処理内容を指定して任意の位置への処理ができるようにする
      **/
      class GTK_DrawingArea_3PassSelectionEvent : public GTK_DrawingArea_DrawingEvent
      {
        // 描画状態
        // NO_DRAW ... 描画前(カーソル表示なし)
        // CURSOR0 ... 一つめのカーソル ( 1 点での描画 )
        // CURSOR1 ... 二つめのカーソル ( 2 点での描画 )
        // CURSOR2 ... 三つめのカーソル ( 3 点での描画 )
        enum Status { NO_DRAW, CURSOR0, CURSOR1, CURSOR2, };

        draw::DrawFuncIF_1p* cursor0_;      // 初期状態のカーソル
        draw::DrawFuncIF_2p* cursor1_;      // 一点が決まった後のカーソル
        draw::DrawFuncIF_3p* cursor2_;      // 二点が決まった後のカーソル
        draw::DrawFuncIF_3p* draw_;         // 描画関数

        int status_;                  // 現在の描画状態

        Coord< int > c0_, c1_, c2_; // 描画対象の座標値 ( widget に対する座標値 )
        Coord< double > l0_, l1_;   // 描画対象の座標値 ( 現レイヤに対する座標値 )

        // マウスボタンクリック時の操作
        //
        // 左クリックで現在のカーソルを消去しステータスを次に移行させる
        // 描画前ステータスだった場合は描画を行い初期状態に戻す
        // 右クリックで現在のカーソルを消去しステータスを一つ戻す
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウス移動時の操作
        //
        // マウス位置を再計算してカーソルを再描画する
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更時の操作
        //
        // カーソルを再描画する
        virtual gint redrawAction();

        // カーソルの描画を消去して内容を初期化する
        void reset();

      public:

        /* コンストラクタ */

        /// @brief 三つのカーソルと描画内容を個々に定義して構築
        ///
        /// @param drawingArea 描画対象
        /// @param cursor0 クリック前のカーソルを描画する関数オブジェクトへのポインタ
        /// @param cursor1 一回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param cursor2 二回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_2p* cursor1, draw::DrawFuncIF_3p* cursor2,
                                             draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp );

        /// @brief 二回クリックした後のカーソルと描画内容が同じである場合の定義
        ///
        /// @param drawingArea 描画対象
        /// @param cursor0 クリック前のカーソルを描画する関数オブジェクトへのポインタ
        /// @param cursor1 一回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_1p* cursor0, draw::DrawFuncIF_2p* cursor1,
                                             draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp );

        /// @brief クリック前のカーソルは描画しない場合の定義
        ///
        /// @param drawingArea 描画対象
        /// @param cursor1 一回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param cursor2 二回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_2p* cursor1, draw::DrawFuncIF_3p* cursor2,
                                             draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp );

        /// @brief クリック前のカーソルは描画せず、二回クリックした後のカーソルと描画内容が同じである場合の定義
        ///
        /// @param drawingArea 描画対象
        /// @param cursor1 一回クリックした後のカーソルを描画する関数オブジェクトへのポインタ
        /// @param draw 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_3PassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                             draw::DrawFuncIF_2p* cursor1,
                                             draw::DrawFuncIF_3p* draw, pen::GPixelOp* pixelOp );

        /// @brief デストラクタ (カーソルの描画を消去して内容を初期化する)
        ~GTK_DrawingArea_3PassSelectionEvent()
        { reset(); status_ = NO_DRAW; }

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        /// @brief 初期状態のカーソルをセットする
        ///
        /// @param cursor0 初期状態のカーソルへのポインタ
        void setCursor0( draw::DrawFuncIF_1p* cursor0 )
        { cursor0_ = cursor0; }

        /// @brief 一点が決まった後のカーソルをセットする
        ///
        /// @param cursor1 一点が決まった後のカーソルへのポインタ
        void setCursor1( draw::DrawFuncIF_2p* cursor1 );

        /// @brief 二点が決まった後のカーソルをセットする
        ///
        /// @param cursor2 二点が決まった後のカーソルへのポインタ
        void setCursor2( draw::DrawFuncIF_3p* cursor2 );

        /// @brief 描画関数をセットする
        ///
        /// @param draw 描画関数へのポインタ
        void setDraw( draw::DrawFuncIF_3p* draw );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント(複数のパス)

         複数の点を指定して描画を行うコールバック関数用の関数オブジェクト
      **/
      class GTK_DrawingArea_MultiPassSelectionEvent : public GTK_DrawingArea_DrawingEvent
      {
        draw::DrawFuncIF_Multiple* drawMulti_; // 描画関数へのポインタ

        std::vector< Coord< int > > vec_c_;    // 選択された点の配列(widget基準)
        std::vector< Coord< double > > vec_l_; // 選択された点の配列(現レイヤ基準)
        std::vector< Coord< double > >::size_type
        draggedVertex_;                      // ドラッグ対象の頂点の番号

        // マウスボタンクリック時の操作
        //
        // すでに選択された点の近くだった場合はその点を移動するモードに
        // そうでなければ新たな点を追加する
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウスボタンリリース時の操作
        //
        // 点の移動モードが動作しないよう変数を変更する
        virtual gint buttonReleaseAction( GdkEventButton *event );

        // マウス移動時の操作
        //
        // 左クリックが押されたままだった場合は点を移動する
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更による再描画
        virtual gint redrawAction();

        // ドラッグによる直線の変更
        void modifyLine( const Coord< double >& l, const Coord< int >& c );

        // 新しい線分の描画
        void drawNewLine( const Coord< double >& l, const Coord< int >& c );

      public:

        /* コンストラクタ */

        /// @brief 複数の点を引数に描画する関数を指定して構築
        ///
        /// @param drawingArea 描画対象
        /// @param drawMulti 処理関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_MultiPassSelectionEvent( canvas::GTK_DrawingArea& drawingArea,
                                                 draw::DrawFuncIF_Multiple* drawMulti, pen::GPixelOp* pixelOp );

        /// @brief 外部から渡した描画関数オブジェクト・点描画関数オブジェクトで処理する
        ///
        /// @param drawMulti 描画関数オブジェクトへの参照
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        void operator()( draw::DrawFuncIF_Multiple& drawMulti, pen::GPixelOp& pixelOp );

        /// @brief 内部に定義した描画関数・点描画関数で描画する
        void operator()() { (*this)( *drawMulti_, pixelOp() ); }

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        /// @brief 描画関数をセットする
        ///
        /// @param drawMulti 描画関数へのポインタ
        void setDraw( draw::DrawFuncIF_Multiple* drawMulti );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント ( ペン描画 )
      **/
      class GTK_DrawingArea_PenEvent : public GTK_DrawingArea_DrawingEvent
      {
        draw::DrawFuncIF_1p* cursor0_; // 初期状態のカーソルへのポインタ

        bool drawn_;              // カーソルが描画された状態なら true
        bool hasPre_;             // 前回クリックされた場所があれば true
        bool clicked_;            // クリックされた状態なら true
        bool shiftClicked_;       // シフトキーが押された状態なら true

        Coord< double > pre_;     // 前回クリックされた場所(widget基準)
        Coord< double > pre_l_;   // 前回クリックされた場所(現レイヤ基準)
        Coord< double > c_;       // 現在の座標値
        double preScale_;         // 前回描画した時のスケール ( X/Y は等しいことを前提 )

        // マウスがウィジェット内に入った時の動作
        virtual gint enterNotifyAction( GdkEvent* event );

        // マウスがウィジェット内から出た時の動作
        virtual gint leaveNotifyAction( GdkEvent* event );

        // マウスボタンクリック時の動作
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウスボタンリリース時の動作
        virtual gint buttonReleaseAction( GdkEventButton *event );

        // キークリック時の動作
        virtual gint keyPressAction( GdkEventKey* event );

        // キーリリース時の動作
        virtual gint keyReleaseAction( GdkEventKey* event );

        // マウス移動時の操作
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更時の操作
        virtual gint redrawAction();

        // シフトキークリック時の線分描画
        void drawLine( pen::GPixelOp& pixelOp );

        // pixelOp によるカーソルの描画
        void draw( pen::GPixelOp& pixelOp );

        // カーソル描画
        void drawCursor();

        // 描画領域への描画
        void drawPixbuf( Coord< double > c );

        // カーソルの描画を消去して内容を初期化する
        void reset();

      public:

        /* コンストラクタ */

        /// @brief 初期状態のカーソルと点描画関数を指定して構築
        ///
        /// @param drawingArea 描画対象
        /// @param cursor0 処理前のカーソルを描画する関数オブジェクトへのポインタ
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_PenEvent( canvas::GTK_DrawingArea& drawingArea,
                                  draw::DrawFuncIF_1p* cursor0,
                                  pen::GPixelOp* pixelOp );

        /// @brief 点描画関数を指定して構築
        ///
        /// カーソルは描画されない
        ///
        /// @param drawingArea 描画対象
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_PenEvent( canvas::GTK_DrawingArea& drawingArea,
                                  pen::GPixelOp* pixelOp );

        /* デストラクタ */

        /// @brief デストラクタ (カーソルの描画を消去して内容を初期化する)
        ~GTK_DrawingArea_PenEvent();

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        /// @brief 初期状態のカーソルをセットする
        ///
        /// @param cursor0 初期状態のカーソルへのポインタ
        void setCursor0( draw::DrawFuncIF_1p* cursor0 )
        { cursor0_ = cursor0; }
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント ( Ferguson/Coons 曲線描画用 )

         Ferguson/Coons 曲線描画のためのコールバック関数用の関数オブジェクト
      **/
      class GTK_DrawingArea_FergusonCoonsSelectionEvent : public GTK_DrawingArea_DrawingEvent
      {
        std::vector< Coord< int > > vec_c_;    // 選択された点の配列(widget基準)
        std::vector< Coord< double > > vec_l_; // 選択された点の配列(現レイヤ基準)
        std::vector< Coord< double > > vec_d_; // 選択された各点の微分係数(vec_c_基準)

        std::vector< Coord< double > >::size_type
        draggedVertex_;                      // ドラッグ対象の頂点の番号
        bool isDiff_;                          // ドラッグ対象が微分係数なら true
        Coord< double > currentScale_;         // 現在のスケール

        // マウスボタンクリック時の操作
        //
        // すでに選択された点の近くだった場合はその点を移動するモードに
        // そうでなければ新たな点を追加する
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウスボタンリリース時の操作
        //
        // 点の移動モードが動作しないよう変数を変更する
        virtual gint buttonReleaseAction( GdkEventButton *event );

        // マウス移動時の操作
        //
        // 左クリックが押されたままだった場合は点を移動する
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更による再描画
        virtual gint redrawAction();

        // ドラッグによる点の変更
        void modifyLine( const Coord< double >& l, const Coord< int >& c );

        // ドラッグによる微分係数の変更
        void modifyDiffLine( const Coord< double >& l, const Coord< int >& c );

        // 新しい点の描画
        void drawNewLine( const Coord< double >& l, const Coord< int >& c );

        // 描画内容のクリア
        void reset();

        // 再描画
        void draw();

      public:

        /* コンストラクタ */

        /// @brief 点描画関数を指定して構築
        ///
        /// @param drawingArea 描画対象
        /// @param pixelOp 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_FergusonCoonsSelectionEvent( canvas::GTK_DrawingArea& drawingArea, pen::GPixelOp* pixelOp );

        /// @brief 外部から点描画関数を指定して実行する
        ///
        /// @param pixelOp 点描画用関数オブジェクトへの参照
        void operator()( pen::GPixelOp& pixelOp );

        /// @brief 内部に定義した点描画関数で描画する
        void operator()() { (*this)( pixelOp() ); }

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();
      };

    } // namespace gtk_event

    namespace draw
    {
      /**
         @brief 矩形領域への操作用関数オブジェクト

         矩形の始点(左上)と終点(右下)を引数に処理を行う関数オブジェクトの基底クラス
      **/
      struct RectangleOperator : public DrawFuncIF
      {
        /// @brief 矩形領域への操作
        ///
        /// @param draw 描画領域への参照
        /// @param c0 矩形領域の始点(左上)
        /// @param c1 矩形領域の終点(右下)
        virtual void operator()( canvas::DrawingArea_IF& draw, Coord< int > c0, Coord< int > c1 ) = 0;
      };

      /**
         @brief 矩形領域のコピー
      **/
      class RectangleCopy : public RectangleOperator
      {
        canvas::GPattern& pattern_; // コピー先

      public:

        /// @brief コピー先を指定して構築
        ///
        /// @param pattern コピー先
        RectangleCopy( canvas::GPattern& pattern )
          : pattern_( pattern )
        { assert( &pattern_ != 0 ); }

        /// @brief 矩形領域のコピー
        ///
        /// @param draw 描画領域への参照
        /// @param c0 矩形領域の始点(左上)
        /// @param c1 矩形領域の終点(右下)
        virtual void operator()( canvas::DrawingArea_IF& draw, Coord< int > c0, Coord< int > c1 );
      };

      /**
         @brief 矩形領域の塗りつぶし
      **/
      struct RectangleFill : public RectangleOperator
      {
        RGB rgb; /// 塗りつぶすときの RGB 成分

        /// @brief 塗りつぶすときの RGB 成分を指定して構築
        ///
        /// @param rgb 塗りつぶすときの RGB 成分
        RectangleFill( RGB rgb )
          : rgb( rgb ) {}

        /// @brief 矩形領域の塗りつぶし
        ///
        /// @param draw 描画領域への参照
        /// @param c0 矩形領域の始点(左上)
        /// @param c1 矩形領域の終点(右下)
        virtual void operator()( canvas::DrawingArea_IF& draw, Coord< int > c0, Coord< int > c1 );
      };

    } // namespace draw

    namespace gtk_event
    {
      /**
         @brief GTK_DrawingArea選択設定用イベント(矩形選択)

         矩形の始点(左上)と終点(右下)を指定し、さらに位置や大きさの微調整を後で行うことができる。
      **/
      class GTK_DrawingArea_RectangleSelectionEvent : public GTK_DrawingArea_SelectionEvent
      {
        // 描画状態
        // NO_DRAW ... 描画前 ( 始点選択 )
        // CURSOR0 ... 1 つめのカーソル ( 終点選択 )
        // CURSOR1 ... 2 つめのカーソル ( 範囲の微調整 )
        enum Status { NO_DRAW, CURSOR0, CURSOR1, };

        int status_;  // 現在のステータス
        int clickLocation_; // クリックした位置の判定結果
        Coord< int > pre_; // 前回クリックした位置
        Coord< int > c0_, c1_; // 選択範囲の左上・右下座標(widget基準)
        Coord< double > l0_, l1_; // 選択範囲の左上・右下座標(現レイヤ基準)

        // マウスボタンクリック時の操作
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウスボタンリリース時の操作
        //
        // 選択範囲の移動モードが動作しないよう変数を変更する
        virtual gint buttonReleaseAction( GdkEventButton *event );

        // マウス移動時の操作
        //
        // 左クリックが押されたままだった場合、頂点が選択されていたらその頂点を、
        // 矩形範囲内が選択されていたら矩形そのものを移動する
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更による再描画
        virtual gint redrawAction();

        // クリック・ドラッグ位置に応じた矩形の再描画
        void modifyRect( Coord< double > l, Coord< int > c );

      public:

        /* コンストラクタ */

        /// @brief 選択対象の描画領域を指定して構築
        ///
        /// @param drawingArea 選択対象の描画領域
        GTK_DrawingArea_RectangleSelectionEvent( canvas::GTK_DrawingArea& drawingArea );

        /// @brief 選択範囲に対する任意の処理を行う
        ///
        /// @param op 矩形領域に対する処理用関数オブジェクトへの参照
        void operator()( draw::RectangleOperator& op );

        /// @brief カーソルの消去とステータスの初期化
        void reset();

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();
      };

      /**
         @brief GTK_DrawingArea選択範囲描画用イベント

         マウス操作で描画を行うコールバック関数用の関数オブジェクト
      **/
      class GTK_DrawingArea_PasteEvent : public GTK_DrawingArea_SelectionEvent
      {
        canvas::DrawingArea_IF& src_;   // 描画するパターンへの参照
        pen::GPixelDraw* pixelDraw_; // ピクセル描画用関数オブジェクトへのポインタ

      protected:

        /// @brief 描画パターンへの参照を返す
        ///
        /// @return 描画パターンへの参照
        canvas::DrawingArea_IF& src() const
        { return( src_ ); }

        /// @brief 点描画関数への参照を返す
        ///
        /// @return 点描画関数への参照
        pen::GPixelDraw& pixelDraw() const
        { return( *pixelDraw_ ); }

      public:

        /* コンストラクタ */

        /// @brief イベントを定義する描画領域と点描画関数オブジェクトを指定して構築
        ///
        /// @param src 描画するパターンへの参照
        /// @param drawingArea 描画対象
        /// @param pixelDraw 点描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_PasteEvent( canvas::DrawingArea_IF& src, canvas::GTK_DrawingArea& drawingArea, pen::GPixelDraw* pixelDraw );

        /// @brief 点描画関数をセットする
        ///
        /// @param pixelDraw 点描画関数へのポインタ
        void setPixelDraw( pen::GPixelDraw* pixelDraw );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント(矩形パターンの貼り付け)

         選択した範囲の矩形パターンの描画位置指定
      **/
      class GTK_DrawingArea_NormalPasteEvent : public GTK_DrawingArea_PasteEvent
      {
      public:

        enum Direction
          {
           normal,        // 通常の貼り付け
           reverseH,      // 水平方向の反転
           reverseV,      // 垂直方向の反転
           turnAround,    // 180度回転
           anticlockwise, // 90度回転(半時計回り)
           clockwise,     // 90度回転(時計回り)
          };

      private:

        //canvas::DrawingArea_IF& src_;     // 描画するパターンへの参照
        //GPixelDraw* pixelDraw_;   // ピクセル描画用関数オブジェクトへのポインタ

        Coord< int > pre_;        // 前回クリックした位置
        Coord< int > c0_, c1_;    // 現在の矩形範囲の左上(widget上)
        Coord< double > l0_, l1_; // 現在の矩形範囲の左上(現レイヤ上)

        int clickLocation_;       // クリック位置の判定結果
        bool isFirst_;            // 一回もカーソルを描画していなければ true

        bool revX_, revY_, swapXY_; // 貼り付け方法(水平反転・垂直反転・X/Yの入れ替え)

        // マウスボタンクリック時の操作
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウスボタンリリース時の操作
        virtual gint buttonReleaseAction( GdkEventButton *event );

        // マウス移動時の操作
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更による再描画
        virtual gint redrawAction();

        // カーソルの描画
        void cursorDraw();

        // カーソルのクリア
        void cursorErase();

      public:

        /* コンストラクタ */

        /// @brief 対象パターンとパターン描画領域を指定して構築
        ///
        /// @param src 描画するパターンへの参照
        /// @param drawingArea 描画対象領域への参照
        /// @param pixelDraw ピクセル描画用関数オブジェクトへのポインタ
        GTK_DrawingArea_NormalPasteEvent( canvas::DrawingArea_IF& src, canvas::GTK_DrawingArea& drawingArea, pen::GPixelDraw* pixelDraw );

        /// @brief パターンを描画領域に描画する
        void operator()();

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        /// @brief 貼り付け方向をセットする
        ///
        /// @param direction 貼り付け方向
        /// @return direction が対象外の値なら false を返す
        bool setDirection( Direction direction );

        /// @brief ピクセル描画用関数オブジェクトをセットする
        ///
        /// @note PixelDraw オブジェクトへのポインタにダウンキャストできない場合は assert を実行する。
        ///
        /// @param pixelOp セットするピクセル描画用関数オブジェクトへのポインタ
        //virtual void setPixelOp( pen::GPixelOp* pixelOp );
      };

      /**
         @brief GTK_DrawingArea選択設定用イベント ( 矩形パターンの自由変形 )
      **/
      class GTK_DrawingArea_TransPasteEvent : public GTK_DrawingArea_PasteEvent
      {
        //canvas::DrawingArea_IF& src_;   // 描画するパターンへの参照
        //GPixelDraw* pixelDraw_; // ピクセル描画用関数オブジェクトへのポインタ
        interpolation::Base* ip_; // 補間用関数オブジェクトへのポインタ

        Coord< int > pre_;             // 前回クリックした位置
        std::vector< Coord< int > > cv_;    // 現在の矩形範囲の 4 隅(widget上)
        std::vector< Coord< double > > lv_; // 現在の矩形範囲の 4 隅(現レイヤ上)

        std::vector< Coord< int > >::size_type draggedVertex_; // ドラッグ中の頂点
        bool isFirst_; // 一回もカーソルを描画していなければ true
        bool isInner_; // 矩形内部をクリックしている場合は true

        bool priorFirst_;         // 先に描画されたピクセルを優先する(Transput専用オプション)

        // マウスボタンクリック時の操作
        virtual gint buttonPressAction( GdkEventButton *event );

        // マウスボタンリリース時の操作
        virtual gint buttonReleaseAction( GdkEventButton *event );

        // マウス移動時の操作
        virtual gint motionNotifyAction( const Coord< int >& c, GdkModifierType& state );

        // オフセット・スケール変更による再描画
        virtual gint redrawAction();

        // カーソルの描画
        void drawCursor();

        // カーソルのリセット
        void resetCursor();

      public:

        // コンストラクタ

        /// @brief 対象パターンとパターン描画領域を指定して構築
        ///
        /// @param src 描画するパターンへの参照
        /// @param drawingArea 描画対象領域への参照
        /// @param pixelDraw ピクセル描画用関数オブジェクトへのポインタ
        /// @param ip 補間用関数オブジェクトへのポインタ
        /// @param priorFirst 最初に描画されたピクセルを優先するなら true (自由変形専用オプション)
        GTK_DrawingArea_TransPasteEvent( canvas::DrawingArea_IF& src, canvas::GTK_DrawingArea& drawingArea,
                                         pen::GPixelDraw* pixelDraw, interpolation::Base* ip,
                                         bool priorFirst );

        /// @brief パターンを描画領域に描画する
        void operator()();

        /// @brief イベントのブロック
        ///
        /// ブロック時にはカーソルの描画を消去して内容を初期化する
        virtual void block();

        // @brief ピクセル描画用関数オブジェクトをセットする
        //
        // @note PixelDraw オブジェクトへのポインタにダウンキャストできない場合は assert を実行する。
        //
        // @param pixelOp セットするピクセル描画用関数オブジェクトへのポインタ
        //virtual void setPixelOp( pen::GPixelOp* pixelOp );

        /// @brief 補間用関数オブジェクトをセットする
        ///
        /// @param ip セットする補間用関数オブジェクトへのポインタ
        void setIp( interpolation::Base* ip );
      };

    } // namespace gtk_event

  } // namespace gl

} // namespace tadah_fussy
#endif


