#include "gl_super_sampling_pattern.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace canvas
    {

      using error_lib::NNNum;

      /**************************************************************
        SuperSamplingPattern : Supersampling用パターン
      **************************************************************/

      /*
        SuperSamplingPatternコンストラクタ

        sz を実パターンのサイズ、mag を疑似サイズとの比率(倍率)として構築
      */
      SuperSamplingPattern::SuperSamplingPattern( const Coord< int >& sz, int mag )
        : patSize_( sz ), mag_( ( mag >= 0 ) ? mag : 0 )
      {
        if ( realSize().x < 0 ) patSize_.x = 0;
        if ( realSize().y < 0 ) patSize_.y = 0;

        if ( realSize().x > 0 && realSize().y > 0 )
          pattern_.resize( realSize().x * realSize().y );
      }

      /*
        SuperSamplingPattern::getPixel : パターンの rc 上の Pixel 値を pixel に取得する

        座標がパターンの範囲外なら何もせず falseを返す
      */
      bool SuperSamplingPattern::getPixel( Coord< int > rc, Pixel* pixel ) const
      {
        if ( rc.x < 0 || rc.x >= realSize().x || rc.y < 0 || rc.y >= realSize().y )
          return( false );

        *pixel = pattern_[rc.y * realSize().x + rc.x];

        return( true );
      }

      /*
        SuperSamplingPattern::point : パターンの rc 上の色コードを col に取得する

        指定した座標がパターン外部の場合は何もせず false を返す。
        パターン内に色コードが格納されていない場合は何もせず false を返す。
      */
      bool SuperSamplingPattern::point( Coord< int > rc, RGB* col ) const
      {
        Pixel pixel;
        if ( ! getPixel( rc, &pixel ) ) return( false );
        if ( pixel.cnt == 0 ) return( false );

        col->r = RGB::Check_RGBPrimary( std::round( static_cast< double >( pixel.r ) / static_cast< double >( pixel.cnt ) ) );
        col->g = RGB::Check_RGBPrimary( std::round( static_cast< double >( pixel.g ) / static_cast< double >( pixel.cnt ) ) );
        col->b = RGB::Check_RGBPrimary( std::round( static_cast< double >( pixel.b ) / static_cast< double >( pixel.cnt ) ) );

        return( true );
      }

      /*
        SuperSamplingPattern::pset : 擬似座標 vc へ色コード col で描画する(加算する)

        倍率がゼロ、座標がパターンの範囲外ならば何もせず false を返す
      */
      bool SuperSamplingPattern::pset( Coord<int> vc, RGB col )
      {
        if ( mag_ == 0 ) return( false );

        Coord<int> rc( vc.x / mag_, vc.y / mag_ ); // 実サイズ上の座標

        if ( rc.x < 0 || rc.x >= realSize().x || rc.y < 0 || rc.y >= realSize().y )
          return( false );

        pattern_[rc.y * realSize().x + rc.x].r += col.r;
        pattern_[rc.y * realSize().x + rc.x].g += col.g;
        pattern_[rc.y * realSize().x + rc.x].b += col.b;
        ++( pattern_[rc.y * realSize().x + rc.x].cnt );

        return( true );
      }

      /*
        SuperSamplingPattern::add : 実座標 rc へパターンへ Pixel 値を加算する

        描画 pset とは異なり、加算回数はカウントされない。誤差項を追加する場合等に利用
        座標がパターンの範囲外ならば何もせず falseを返す
      */
      bool SuperSamplingPattern::add( const Coord<int>& rc, const Pixel& pixel ) const
      {
        if ( rc.x < 0 || rc.x >= realSize().x || rc.y < 0 || rc.y >= realSize().y )
          return( false );

        pattern_[rc.y * realSize().x + rc.x].r += pixel.r * pattern_[rc.y * realSize().x + rc.x].cnt / pixel.cnt;
        pattern_[rc.y * realSize().x + rc.x].g += pixel.g * pattern_[rc.y * realSize().x + rc.x].cnt / pixel.cnt;
        pattern_[rc.y * realSize().x + rc.x].b += pixel.b * pattern_[rc.y * realSize().x + rc.x].cnt / pixel.cnt;

        return( true );
      }

      /*
        SuperSamplingPattern::put : 描画領域 draw の s の位置からパターンを描画する
      */
      void SuperSamplingPattern::put( DrawingArea_IF& draw, Coord<int> s )
      {
        if ( &draw == 0 ) return;

        // 描画オブジェクトの大きさを取得
        Coord<int> dSize = draw.size();

        // 描画開始位置はクリッピングエリア範囲内か
        if ( s.x >= dSize.x || s.y >= dSize.y ) return;

        // パターンサイズを読み込んで右下の座標を決定
        Coord<int> e( s.x + realSize().x - 1, s.y + realSize().y - 1 );

        // パターンはクリッピングエリア範囲内か
        if ( e.x < 0 || e.y < 0 ) return;

        // パターンからの読み込み位置
        Coord<int> p( 0, 0 );

        // クリッピング処理
        if ( s.x < 0 ) {
          p.x -= s.x;
          s.x = 0;
        }
        if ( e.x >= dSize.x )
          e.x = dSize.x - 1;
        if ( s.y < 0 ) {
          p.y -= s.y;
          s.y = 0;
        }
        if ( e.y >= dSize.y )
          e.y = dSize.y - 1;

        // パターンの描画
        int pxBuff = p.x;
        RGB rgb;

        for ( ; s.y <= e.y ; ++( s.y ) ) {
          for ( int x = s.x ; x <= e.x ; ++x ) {
            if ( point( p, &rgb ) ) {
              draw.pset( Coord<int>( x, s.y ), rgb );
            }
            ++( p.x );
          }
          p.x = pxBuff;
          ++( p.y );
        }
      }

      /****************************************************************
        SSP_ErrorDiffusion : Supersampling用パターン(誤差拡散法適用版)
      ****************************************************************/

      /*
        SSP_ErrorDiffusion::point : パターンの rc 上の色コードを col に取得する

        指定した座標がパターン外部の場合は何もせず false を返す。
        パターン内に色コードが格納されていない場合は何もせず false を返す。

        剰余を周囲に分散して加算する。比率は以下の通り。

        画素 1/2
        1/8 残り 1/4
      */
      bool SSP_ErrorDiffusion::point( Coord<int> rc, RGB* col ) const
      {
        Pixel pixel;
        if ( ! SuperSamplingPattern::getPixel( rc, &pixel ) ) return( false );
        if ( pixel.cnt == 0 ) return( false );

        // 色コードを取得 ( 範囲外であった場合に補正 )
        *col = RGB( RGB::Check_RGBPrimary( std::round( static_cast<double>( pixel.r ) / static_cast<double>( pixel.cnt ) ) ),
                    RGB::Check_RGBPrimary( std::round( static_cast<double>( pixel.g ) / static_cast<double>( pixel.cnt ) ) ),
                    RGB::Check_RGBPrimary( std::round( static_cast<double>( pixel.b ) / static_cast<double>( pixel.cnt ) ) ) );

        // 端数を取得
        pixel.r %= pixel.cnt;
        pixel.g %= pixel.cnt;
        pixel.b %= pixel.cnt;

        SuperSamplingPattern::Pixel buff( pixel ); // 端数計算用 Pixel 構造体

        // 右側に端数の半分を加算
        buff.r >>= 1;
        buff.g >>= 1;
        buff.b >>= 1;
        add( Coord<int>( rc.x + 1, rc.y ), buff );
        pixel.r -= buff.r;
        pixel.g -= buff.g;
        pixel.b -= buff.b;

        // 右下側に端数の1/4を加算
        buff.r >>= 1;
        buff.g >>= 1;
        buff.b >>= 1;
        add( Coord<int>( rc.x + 1, rc.y + 1 ), buff );
        pixel.r -= buff.r;
        pixel.g -= buff.g;
        pixel.b -= buff.b;

        // 左下側に端数の1/8を加算
        buff.r >>= 1;
        buff.g >>= 1;
        buff.b >>= 1;
        add( Coord<int>( rc.x - 1, rc.y + 1 ), buff );
        pixel.r -= buff.r;
        pixel.g -= buff.g;
        pixel.b -= buff.b;

        // 下側に端数の残りを加算
        add( Coord<int>( rc.x, rc.y + 1 ), pixel );

        return( true );
      }

      /***********************************************************************************
        SSP_FloydSteinberg : Supersampling用パターン(Floyd & Steinberg型誤差拡散法適用版)
      ***********************************************************************************/

      /*
        SSP_FloydSteinberg::point : パターンの rc 上の色コードを col に取得する

        指定した座標がパターン外部の場合は何もせず false を返す。
        パターン内に色コードが格納されていない場合は何もせず false を返す。

        剰余を周囲に分散して加算する。比率は以下の通り。

        画素 7/16
        3/16 5/16 1/16
      */
      bool SSP_FloydSteinberg::point( Coord<int> rc, RGB* col ) const
      {
        Pixel pixel;
        if ( ! SuperSamplingPattern::getPixel( rc, &pixel ) ) return( false );
        if ( pixel.cnt == 0 ) return( false );

        // 色コードを取得 ( 範囲外であった場合に補正 )
        *col = RGB( RGB::Check_RGBPrimary<int>( std::round( static_cast<double>( pixel.r ) / static_cast<double>( pixel.cnt ) ) ),
                    RGB::Check_RGBPrimary<int>( std::round( static_cast<double>( pixel.g ) / static_cast<double>( pixel.cnt ) ) ),
                    RGB::Check_RGBPrimary<int>( std::round( static_cast<double>( pixel.b ) / static_cast<double>( pixel.cnt ) ) ) );

        // 端数を取得
        pixel.r %= pixel.cnt;
        pixel.g %= pixel.cnt;
        pixel.b %= pixel.cnt;

        Pixel buff; // 端数計算用 Pixel 構造体

        // 右側に端数の7/16を加算
        buff.r = ( static_cast<int>( pixel.r ) * 7 ) >> 4;
        buff.g = ( static_cast<int>( pixel.g ) * 7 ) >> 4;
        buff.b = ( static_cast<int>( pixel.b ) * 7 ) >> 4;
        buff.cnt = pixel.cnt;
        add( Coord<int>( rc.x + 1, rc.y ), buff );

        // 左下側に端数の3/16を加算
        buff.r = ( static_cast<int>( pixel.r ) * 3 ) >> 4;
        buff.g = ( static_cast<int>( pixel.g ) * 3 ) >> 4;
        buff.b = ( static_cast<int>( pixel.b ) * 3 ) >> 4;
        add( Coord<int>( rc.x - 1, rc.y + 1 ), buff );

        // 下側に端数の5/16を加算
        buff.r = ( static_cast<int>( pixel.r ) * 5 ) >> 4;
        buff.g = ( static_cast<int>( pixel.g ) * 5 ) >> 4;
        buff.b = ( static_cast<int>( pixel.b ) * 5 ) >> 4;
        add( Coord<int>( rc.x - 1, rc.y + 1 ), buff );

        // 右下側に端数の1/16を加算
        pixel.r >>= 4;
        pixel.g >>= 4;
        pixel.b >>= 4;
        add( Coord<int>( rc.x + 1, rc.y + 1 ), pixel );

        return( true );
      }

      /*********************************************************************************
        SSP_JJN : Supersampling用パターン(Jarvis, Judice & Ninke型誤差拡散法適用版)
      *********************************************************************************/

      /*
        SSP_JJN::point : パターンの rc 上の色コードを col に取得する

        指定した座標がパターン外部の場合は何もせず false を返す。
        パターン内に色コードが格納されていない場合は何もせず false を返す。

        剰余を周囲に分散して加算する。比率は以下の通り。

        画素 7/48 5/48
        3/48 5/48 7/48 5/48 3/48
        1/48 3/48 5/48 3/48 1/48
      */
      bool SSP_JJN::point( Coord<int> rc, RGB* col ) const
      {
        Pixel pixel;
        if ( ! SuperSamplingPattern::getPixel( rc, &pixel ) ) return( false );
        if ( pixel.cnt == 0 ) return( false );

        // 色コードを取得 ( 範囲外であった場合に補正 )
        *col = RGB( RGB::Check_RGBPrimary<int>( std::round( static_cast<double>( pixel.r ) / static_cast<double>( pixel.cnt ) ) ),
                    RGB::Check_RGBPrimary<int>( std::round( static_cast<double>( pixel.g ) / static_cast<double>( pixel.cnt ) ) ),
                    RGB::Check_RGBPrimary<int>( std::round( static_cast<double>( pixel.b ) / static_cast<double>( pixel.cnt ) ) ) );

        // 端数を取得
        pixel.r %= pixel.cnt;
        pixel.g %= pixel.cnt;
        pixel.b %= pixel.cnt;

        SuperSamplingPattern::Pixel buff; // 端数計算用 Pixel 構造体

        // 端数の7/48を加算
        buff.r = std::round( static_cast<double>( pixel.r * 7 ) / 48.0 );
        buff.g = std::round( static_cast<double>( pixel.g * 7 ) / 48.0 );
        buff.b = std::round( static_cast<double>( pixel.b * 7 ) / 48.0 );
        buff.cnt = pixel.cnt;
        add( Coord<int>( rc.x + 1, rc.y ), buff );
        add( Coord<int>( rc.x, rc.y + 1 ), buff );

        // 端数の5/48を加算
        buff.r = std::round( static_cast<double>( pixel.r * 5 ) / 48.0 );
        buff.g = std::round( static_cast<double>( pixel.g * 5 ) / 48.0 );
        buff.b = std::round( static_cast<double>( pixel.b * 5 ) / 48.0 );
        add( Coord<int>( rc.x + 2, rc.y ), buff );
        add( Coord<int>( rc.x - 1, rc.y + 1 ), buff );
        add( Coord<int>( rc.x + 1, rc.y + 1 ), buff );
        add( Coord<int>( rc.x, rc.y + 2 ), buff );

        // 端数の3/48を加算
        buff.r = std::round( static_cast<double>( pixel.r * 3 ) / 48.0 );
        buff.g = std::round( static_cast<double>( pixel.g * 3 ) / 48.0 );
        buff.b = std::round( static_cast<double>( pixel.b * 3 ) / 48.0 );
        add( Coord<int>( rc.x - 2, rc.y + 1 ), buff );
        add( Coord<int>( rc.x + 2, rc.y + 1 ), buff );
        add( Coord<int>( rc.x - 1, rc.y + 2 ), buff );
        add( Coord<int>( rc.x + 1, rc.y + 2 ), buff );

        // 端数の1/48を加算
        buff.r = std::round( static_cast<double>( pixel.r ) / 48.0 );
        buff.g = std::round( static_cast<double>( pixel.g ) / 48.0 );
        buff.b = std::round( static_cast<double>( pixel.b ) / 48.0 );
        add( Coord<int>( rc.x - 2, rc.y + 2 ), buff );
        add( Coord<int>( rc.x + 2, rc.y + 2 ), buff );

        return( true );
      }

      /**************************************************************************
        SSP_OrderedDithering : Supersampling用パターン(オーダードディザ法適用版)
      **************************************************************************/

      /*
        SSP_OrderedDithering コンストラクタ

        SuperSamplingPattern 構築のため、実パターンのサイズ sz、疑似サイズとの比率(倍率) mag を指定する。
        ディザマトリックスの作成には、サイズ size ( 2 の指数表現 ; 2^size ) と
        ディザマトリックスの要素を加算する比率 dmi ( 値が大きいほど影響度が小さくなる ) を指定する。
      */
      SSP_OrderedDithering::SSP_OrderedDithering( const Coord<int>& sz, int mag, unsigned int size, unsigned int dmi )
        : SuperSamplingPattern( sz, mag )
      {
        dmi_ = ( dmi > 0 ) ? dmi : 1;
        matSize_ = ( size > 0 ) ? ( 1 << size ) : 2;
        matrix_.resize( matSize_ * matSize_ );

        // 2 x 2 サイズで初期化
        matrix_[0] = 0;
        matrix_[1] = 2;
        matrix_[matSize_] = 3;
        matrix_[matSize_ + 1] = 1;

        // ひとつ前のサイズの行列の要素を元に構築
        for ( unsigned int cnt = 2 ; cnt < matSize_ ; cnt *= 2 ) {
          for ( unsigned int r = 0 ; r < cnt ; ++r ) {
            for ( unsigned int c = 0 ; c < cnt ; ++c ) {
              matrix_[r * matSize_ + c] *= 4;
              matrix_[r * matSize_ + cnt + c] = matrix_[r * matSize_ + c] + 2;
              matrix_[( r + cnt ) * matSize_ + c] = matrix_[r * matSize_ + c] + 3;
              matrix_[( r + cnt ) * matSize_ + cnt + c] = matrix_[r * matSize_ + c] + 1;
            }
          }
        }

        // 中央値で減算する(平均値をゼロにする)
        for ( unsigned int r = 0 ; r < matSize_ ; ++r )
          for ( unsigned int c = 0 ; c < matSize_ ; ++c )
            matrix_[r * matSize_ + c] -= matSize_ * matSize_ / 2;
      }

      /*
        SSP_OrderedDithering::point : Ordered Dither Matrix から得た誤差を加算した rc 上の色コードを col に代入する
        rc は実サイズ上の座標を表す
      */
      bool SSP_OrderedDithering::point( Coord<int> rc, RGB* col ) const
      {
        Pixel pixel;
        if ( ! SuperSamplingPattern::getPixel( rc, &pixel ) ) return( false );
        if ( pixel.cnt == 0 ) return( false );

        double error = matrix_[( rc.y % matSize_ ) * matSize_ + ( rc.x % matSize_ )] / dmi_;
        col->r = RGB::Check_RGBPrimary<int>( std::round( ( static_cast<double>( pixel.r ) + error ) / static_cast<double>( pixel.cnt ) ) );
        col->g = RGB::Check_RGBPrimary<int>( std::round( ( static_cast<double>( pixel.g ) + error ) / static_cast<double>( pixel.cnt ) ) );
        col->b = RGB::Check_RGBPrimary<int>( std::round( ( static_cast<double>( pixel.b ) + error ) / static_cast<double>( pixel.cnt ) ) );

        return( true );
      }

    } // namespace canvas

  } // namespace gl

} // namespace tadah_fussy
