/**
   @file gl_super_sampling_pattern.h
   @brief スーパー・サンプリング用クラス定義ヘッダ

   @author tadah_fussy
   @date 2018/02/26 DrawingMethod.h から分離
   @date 2021/04/02 文字コードを UTF-8 に変更
**/
#ifndef GRAPHIC_LIBRARY_SUPER_SAMPLING_PATTERN_20210402_H
#define GRAPHIC_LIBRARY_SUPER_SAMPLING_PATTERN_20210402_H

#include <vector>

#include "gl_drawing_area_if.h"

namespace tadah_fussy
{
  namespace gl
  {
    namespace canvas
    {
      /**
         @brief Supersampling 用パターン

         パターンを擬似的に拡大して、その中で描画を行う。実画面にはパターンを縮小して再現する。
         実際に拡大した分の要素を用意するのではなく、要素数は実画面と同数としておき、一つの要素が
         複数の RGB 成分を持つものとみなす。従って、各要素は RGB 各成分の和とその個数を保持する。

         以下はその利用例である。
         @code
         Coord< int > size( 1024, 768 ); // 画像サイズ

         // 描画領域。引数として画像サイズを渡す。
         GPattern pattern( size );
         // 画像サイズの 1 / 4 の描画領域を用意。
         GPattern quarter( Coord< int >( size.x / 2, size.y / 2 ) );

         // ここで quarter に対して何らかの画像を描画

         SuperSamplingPattern ssp( size, 5 ); // 疑似解像度用パターン(5倍の解像度)
         GInterpolation_NN nn;                // 補間処理は最近傍法

         // quarter の画像を 2 倍 ( 擬似的には 10 倍 ) に拡大処理
         ResizePut( quarter, ssp, Coord< int >( 0, 0 ), Coord< int >( ssp.size().x - 1, ssp.size().y - 1 ), nn );

         // pattern に処理結果を出力
         ssp.put( pattern, Coord< int >( 0, 0 ) );
         @endcode

         DrawingArea_IF で用意されている仮想関数 DrawingArea_IF::point は非公開としているため、
         SuperSamplingPattern インスタンスを利用して Put 等により他の描画領域に描画を行うことは
         できない。代わりに、専用のメンバ関数 put が用意されておりこれを利用する。
      **/
      class SuperSamplingPattern : public DrawingArea_IF
      {
      protected:

        /// @brief パターンの要素の型
        struct Pixel {
          unsigned int r;   ///< 赤成分
          unsigned int g;   ///< 緑成分
          unsigned int b;   ///< 青成分
          unsigned int cnt; ///< 加算回数

          /// @brief デフォルト・コンストラクタ
          Pixel() : r( 0 ), g( 0 ), b( 0 ), cnt( 0 ) {}
        };

        /// @brief パターンのPixel値を取得する
        ///
        /// 指定した座標がパターン外部の場合、もっとも近いピクセルのPixel値を取得する
        ///
        /// @param rc 取得する位置(実サイズ上)
        /// @param pixel 取得したPixel値
        /// @return 座標がパターンの範囲外か、またはパターン内に色コードが格納されていなかった場合は falseを返す
        bool getPixel( Coord< int > rc, Pixel* pixel ) const;

        /// @brief パターンへのPixel値の加算
        ///
        /// 描画とは異なり、加算回数はカウントされない。誤差項を追加する場合等に利用する。
        /// 指定した座標がパターン外部の場合は処理しない。
        ///
        /// @param rc 描画する位置(実サイズ上)
        /// @param pixel 加算するPixel値
        /// @return 座標がパターンの範囲外ならば falseを返す
        bool add( const Coord< int >& rc, const Pixel& pixel ) const;

      private:

        mutable std::vector< Pixel > pattern_; // 矩形パターン
        Coord< int > patSize_; // 矩形パターンのサイズ
        int mag_;              // 実サイズに対する疑似解像度の倍率

        // リサイズ関数は使用しない
        virtual void resize( Coord< int > size ) {}

        /// @brief 実パターン・サイズの取得
        ///
        /// @return 実パターンのサイズ
        Coord< int > realSize() const
        {
          return( patSize_ );
        }

        /// @brief パターンの色コードを取得する
        ///
        /// 実サイズ上の色コードを取得するため、各 RGB 成分の和を加算回数で割る処理を行い、その結果を返す。
        /// 指定した座標がパターン外部の場合、もっとも近いピクセルの色コードを取得する。
        /// この仮想関数は外部からは利用できない。
        ///
        /// @param rc 取得する位置(実サイズ上)
        /// @param col 取得した色コードを代入する変数への参照
        /// @return 座標がパターンの範囲外か、またはパターン内に色コードが格納されていなかった場合は falseを返す
        virtual bool point( Coord< int > rc, RGB* col ) const;

      public:

        /* コンストラクタ */

        /// @brief 実際の大きさと倍率を指定して構築
        ///
        /// 各要素は Pixel のデフォルト値 ( 全要素がゼロ ) で初期化される。
        ///
        /// パターンの大きさ sz の X, Y 座標それぞれ負数の場合はゼロで初期化される。
        /// 倍率 mag が負数の場合はゼロで初期化される。
        ///
        /// @param sz パターンの実際の大きさ
        /// @param mag 倍率(疑似解像度との比率)
        SuperSamplingPattern( const Coord< int >& sz, int mag );

        /// @brief 要素を消去して初期化する
        ///
        /// 各要素は Pixel のデフォルト値 ( 全要素がゼロ ) で初期化される。
        void init() { pattern_.assign( pattern_.size(), Pixel() ); }

        /// @brief 要素を消去して初期化する(同時に倍率を変更する)
        ///
        /// 各要素は Pixel のデフォルト値 ( 全要素がゼロ ) で初期化される。
        /// 倍率 mag が負数の場合は無視される。
        ///
        /// @param mag 変更する倍率
        void init( int mag ) { init(); if ( mag >= 0 ) mag_ = mag; }

        /// @brief パターンへの描画
        ///
        /// 擬似サイズ上の座標を指定することに注意。実サイズ上の座標に変換された上で、Pixel の要素に値が加算される。
        /// 倍率 mag がゼロ、指定した座標がパターン外部の場合は処理しない
        ///
        /// @param vc 描画する位置(疑似サイズ上)
        /// @param col 描画する色コード
        /// @return 座標がパターンの範囲外ならば falseを返す
        virtual bool pset( Coord< int > vc, RGB col );

        /// @brief 疑似パターン・サイズの取得
        ///
        /// @return 擬似パターンのサイズ
        virtual Coord< int > size() const
        {
          return( Coord< int >( patSize_.x * mag_, patSize_.y * mag_ ) );
        }

        /// @brief パターンの描画
        ///
        /// 倍率 mag がゼロの場合は何も描画されないことに注意。
        ///
        /// @param draw 描画領域
        /// @param s パターンの描画開始位置
        void put( DrawingArea_IF& draw, Coord< int > s );
      };

      /**
         @brief 誤差拡散法を利用した Supersampling 用パターン

         誤差拡散法により剰余を周囲に分散して加算する。分散の仕方は以下の通り。

         <pre>
         画素 1/2
         1/8 残り 1/4
         </pre>
      **/
      class SSP_ErrorDiffusion : public SuperSamplingPattern
      {
        /// @brief パターンの色コードを取得する
        ///
        /// 指定した座標がパターン外部の場合、もっとも近いピクセルの色コードを取得する。
        /// この仮想関数は外部からは利用できない。
        ///
        /// @param rc 取得する位置(実サイズ上)
        /// @param col 取得した色コードを代入する変数への参照
        /// @return 座標がパターンの範囲外か、またはパターン内に色コードが格納されていなかった場合は falseを返す
        virtual bool point( Coord< int > rc, RGB* col ) const;

      public:

        /* コンストラクタ */

        /// @brief 実際の大きさと倍率を指定して構築
        ///
        /// パターンの大きさ sz の X, Y 座標それぞれ負数の場合はゼロで初期化される。
        /// 倍率 mag が負数の場合はゼロで初期化される。
        ///
        /// @param sz パターンの実際の大きさ
        /// @param mag 倍率(疑似解像度との比率)
        SSP_ErrorDiffusion( const Coord< int >& sz, int mag )
          : SuperSamplingPattern( sz, mag ) {}
      };

      /**
         @brief FloydSteinberg 型誤差拡散法を利用した Supersampling 用パターン

         FloydSteinberg 型誤差拡散法により剰余を周囲に分散して加算する。分散の仕方は以下の通り。

         <pre>
         画素 7/16
         3/16 5/16 1/16
         </pre>
      **/
      class SSP_FloydSteinberg : public SuperSamplingPattern
      {
        /// @brief パターンの色コードを取得する
        ///
        /// 指定した座標がパターン外部の場合、もっとも近いピクセルの色コードを取得する。
        /// この仮想関数は外部からは利用できない。
        ///
        /// @param rc 取得する位置(実サイズ上)
        /// @param col 取得した色コードを代入する変数への参照
        /// @return 座標がパターンの範囲外か、またはパターン内に色コードが格納されていなかった場合は falseを返す
        virtual bool point( Coord< int > rc, RGB* col ) const;

      public:

        /* コンストラクタ */

        /// @brief 実際の大きさと倍率を指定して構築
        ///
        /// パターンの大きさ sz の X, Y 座標それぞれ負数の場合はゼロで初期化される。
        /// 倍率 mag が負数の場合はゼロで初期化される。
        ///
        /// @param sz パターンの実際の大きさ
        /// @param mag 倍率(疑似解像度との比率)
        SSP_FloydSteinberg( const Coord< int >& sz, int mag )
          : SuperSamplingPattern( sz, mag ) {}
      };

      /**
         @brief Jarvis, Judice & Ninke 型誤差拡散法を利用した Supersampling 用パターン

         Jarvis, Judice & Ninke 型誤差拡散法により剰余を周囲に分散して加算する。分散の仕方は以下の通り。

         <pre>
         画素 7/48 5/48
         3/48 5/48 7/48 5/48 3/48
         1/48 3/48 5/48 3/48 1/48
         </pre>
      **/
      class SSP_JJN : public SuperSamplingPattern
      {
        /// @brief パターンの色コードを取得する
        ///
        /// 指定した座標がパターン外部の場合、もっとも近いピクセルの色コードを取得する。
        /// この仮想関数は外部からは利用できない。
        ///
        /// @param rc 取得する位置(実サイズ上)
        /// @param col 取得した色コードを代入する変数への参照
        /// @return 座標がパターンの範囲外か、またはパターン内に色コードが格納されていなかった場合は falseを返す
        virtual bool point( Coord< int > rc, RGB* col ) const;

      public:

        /* コンストラクタ */

        /// @brief 実際の大きさと倍率を指定して構築
        ///
        /// パターンの大きさ sz の X, Y 座標それぞれ負数の場合はゼロで初期化される。
        /// 倍率 mag が負数の場合はゼロで初期化される。
        ///
        /// @param sz パターンの実際の大きさ
        /// @param mag 倍率(疑似解像度との比率)
        SSP_JJN( const Coord< int >& sz, int mag )
          : SuperSamplingPattern( sz, mag ) {}
      };

      /**
         @brief オーダードディザ法を利用した Supersampling 用パターン
      **/
      class SSP_OrderedDithering : public SuperSamplingPattern
      {
        std::vector< int > matrix_;   // ディザマトリックス
        unsigned int matSize_; // matrixの行列サイズ
        int dmi_;              // ディザマトリックスの要素を加算する比率(値が大きいほど影響度が小さくなる)

        /// @brief パターンの色コードを取得する
        ///
        /// 指定した座標がパターン外部の場合、もっとも近いピクセルの色コードを取得する
        /// この仮想関数は外部からは利用できない
        ///
        /// @param rc 取得する位置(実サイズ上)
        /// @param col 取得した色コードを代入する変数への参照
        /// @return 座標がパターンの範囲外か、またはパターン内に色コードが格納されていなかった場合は falseを返す
        bool point( Coord< int > rc, RGB* col ) const;

      public:

        /* コンストラクタ */

        /// @brief 実際の大きさと倍率、ディザマトリックスのパラメータを指定して構築
        ///
        /// パターンの大きさ sz の X, Y 座標それぞれ負数の場合はゼロで初期化される。
        /// 倍率 mag が負数の場合はゼロで初期化される。
        /// ディザマトリックスのサイズ ditherSize がゼロの場合、1 に変換される。
        /// ディザマトリックスの要素を加算する比率 dmi がゼロの場合 1 に変換される。
        ///
        /// @param sz パターンの実際の大きさ
        /// @param mag 倍率(疑似解像度との比率)
        /// @param ditherSize ディザマトリックスのサイズ ( 2 の指数表現 ; 2^size )
        /// @param dmi ディザマトリックスの要素を加算する比率
        SSP_OrderedDithering( const Coord< int >& sz, int mag, unsigned int ditherSize = 3, unsigned int dmi = 4 );
      };

    } // namespace canvas

  } // namespace gl

} // namespace tadah_fussy

#endif
